--[[----------------------------------------------------------------------------

LmModelForwarder.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Lr.frameworks.substrate
local AgSystemInfo = require 'AgSystemInfo'
local AgConfiguration = require 'AgConfiguration'

	-- Lr.toolkits.wf_sync
local AgLrOzEndpoints = require 'AgLrOzEndpoints'

	-- WF.core.os
local WFLocalizationUtils = require 'WFLocalizationUtils'

	-- WF.oz_client.ims
local WFImsAppIdentity = require 'WFImsAppIdentity'

	-- WF.oz_client.managed_catalog
local WFCatModelForwarder = require 'WFCatModelForwarder' 

local logger = require 'LrLogger'( 'LmModelForwarder' )

--============================================================================--

-- @mark LmModelForwarder

--------------------------------------------------------------------------------

local LmModelForwarder = {}

local sRedirectUri = 'http://stophere.adobe.com'

--------------------------------------------------------------------------------

local function logAndRaiseError( errFmtStr, ... )

	local errStr = select( '#', ... ) > 0 and
					string.format( errFmtStr, ... ) or errFmtStr

	logger:error( errStr )

	error( errStr )

end

--------------------------------------------------------------------------------

local runModelCommand

--------------------------------------------------------------------------------

function LmModelForwarder.setRunModelCommandFunc( runModelCommandFunc )

	if type( runModelCommandFunc ) ~= "function" then

		logAndRaiseError( "invalid runModelCommandFunc type: %s", 
						type( runModelCommandFunc ) )

	end

	if runModelCommand then

		logAndRaiseError( "runModelCommand function already has been set on this thread" )

	end

	runModelCommand = runModelCommandFunc

end

--------------------------------------------------------------------------------

local function maybeRunModelCommand( params )

	if not runModelCommand then

		logAndRaiseError( "Tried to run model command before runModelCommand was set" )

	end

	return runModelCommand( params )

end

--------------------------------------------------------------------------------

-- [bzibble 2014-06-13]: fix #3772741: when the app language is set to automatic,
-- we need to translate the language code that we get from Mac OS to make it match
-- the format expected by IMS.

local function sanitizeLanguageCode( code )

	local replacements = {
		[ "zh-Hans" ] = "zh_cn", -- Chinese simplified
		[ "zh-Hant" ] = "zh_tw", -- Chinese traditional
	}
	
	return replacements[ code ] or code

end

--------------------------------------------------------------------------------

function LmModelForwarder.forwardBootstrapRootModel( syncAndGeneralBaseUrl, 
														rootCatalogDir, 
														processId )

	require 'AgWFSyncConfig'.maybeInstallSomething()
		-- Make sure we have IMS client ID and secret available.

	local clientId, clientSecret = WFImsAppIdentity.getClientIdAndSecret()

	local languageCode = sanitizeLanguageCode( WFLocalizationUtils.getLanguageCode() )

	local regionCode = WFLocalizationUtils.getRegionCode()
		-- FIXME [scouten 2016-02-03]: Not sure but these may be wrong. Hard to
		-- get correct language and region codes from non-main thread in Lr.

	local imsOverrides = {
		imsClientId = clientId,
		imsClientSecret = clientSecret,
		baseServerUrl = AgLrOzEndpoints.syncAndGeneralBaseUrl,
		systemUserName = AgSystemInfo.getUsername(),
		macAddress = AgSystemInfo.getMACAddress(),
		machineName = AgSystemInfo.getMachineName(),
		languageCode = languageCode,
		regionCode = regionCode,
		queryParamOverrides = {
			response_type = 'device',
			client_type = 'desktop',
			scope = 'openid,AdobeID',
			idp_flow = 'login',
			redirect_uri = sRedirectUri,
			ctx_id = 'lr5_c1',
		}
	}


	local maxSaves = 10	
	if AgConfiguration.DaemonOverrides.maxAsyncSaves ~= nil then
		maxSaves = AgConfiguration.DaemonOverrides.maxAsyncSaves
	end

	-- empirical testing has soon that setting maxRenders to the number of logical processors yields
	-- the best results, except during times of extreme stress
	
	local maxRenders = AgSystemInfo.numCPUs()

	if AgConfiguration.DaemonOverrides.maxAsyncRenders ~= nil then
		maxRenders = AgConfiguration.DaemonOverrides.maxAsyncRenders
	end
	
	local pixelProviderSettings = {maxSaves = maxSaves, maxRenders = maxRenders}
	
	-- SPAM [jandrew 2018-09-19]: The App locale is needed to set the preferred 
	-- Oz communication language on app startup. This is then leveraged by the
	-- search index to provide localized autocomplete suggestions.
	
	local langCode = string.lower( require 'AgZString'.getISOLanguageCode() )

	local locales = require 'AppAccountUtils'.getAppLocales()
	local currentLocale = locales[ langCode ]
	
	local rootModel =
			WFCatModelForwarder.makeForwardedModel(
				maybeRunModelCommand,
					'NmbUserModel', 'makeUserModel',
						syncAndGeneralBaseUrl, rootCatalogDir, processId,
						imsOverrides, pixelProviderSettings, currentLocale )

	return rootModel

end

--------------------------------------------------------------------------------

return LmModelForwarder
