--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- Rx library
local Rx = require 'Rx'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local WFUnknown = require 'WFUnknown'
local WFCatUtils = require 'WFCatUtils'
local WFTableMerge = require 'WFTableMerge'
local WFPlatformInfo = require 'WFPlatformInfo'


	-- WF.core.string
local WFSortOrderStrings = require 'WFSortOrderStrings'

	-- WF.lrmobile
local LrmCatAlbum = require 'LrmCatAlbum'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.rxLua
local WFRxPromiseBridge = require 'WFRxPromiseBridge'

	--WF.managed_catalog
local WFCatSharedWithYouAlbums = require 'WFCatSharedWithYouAlbums'

--------------------------------------------------------------------------------

local asKSPromise = WFRxPromiseBridge
local logger = WFLogger 'LrmAlbumAssetUtils'
	-- :setLevel 'DEBUG'
--==============================================================================

local LrmAlbumAssetUtils = {}

--==============================================================================

local function getHighestOrderForAlbumP( indexStore, albumId )

	return indexStore:queryP(
		[[select assetId, sortOrder from album_asset where albumId == ? order by sortOrder desc limit 1]],
		albumId )
	:mapDataP( function( _, row )
	
		local highestUserOrder = nil
		if row and WFUnknown.isKnown( row ) and type( row ) == 'table' and #row >= 1 then
			
			highestUserOrder = row[ 1 ][ 2 ] or nil
		end
		
		return 'data', highestUserOrder
	end )
end

--------------------------------------------------------------------------------

local function calculateNewHighOrderForAssets( assetIds, currentHighOrder )

	local orders = {}
	
	if currentHighOrder == nil then
		return KSPromises.resolvedPromiseP( 'data', nil )
	else
		local formerOrder = currentHighOrder

		for i = 1, #assetIds do
			
			local assetId = assetIds[ i ]
			local newOrder = WFSortOrderStrings.insertAfter_lex64( formerOrder, 'low' )
			orders[ assetId ] = newOrder
			formerOrder = newOrder
		end
		return KSPromises.resolvedPromiseP( 'data', orders )
	end
end

local calculateNewHighOrderForAssetsP = KSPromises.liftP( calculateNewHighOrderForAssets )

--------------------------------------------------------------------------------

local function calculateNewInsertOrderForAssets( assetIds, insertOrder, location )

	local orders = {}
	
	local formerOrder

	for i = 1, #assetIds do
			
		local assetId = assetIds[ i ]
		
		local newOrder 
		
		if formerOrder == nil then
		
			if location == "before" then
				
				newOrder = WFSortOrderStrings.insertBefore_lex64( insertOrder, 'high' )
				
			else
			
				newOrder = WFSortOrderStrings.insertAfter_lex64( insertOrder, 'low' )
			
			end
		
		else
		
			if location == 'before' then
			
				newOrder = WFSortOrderStrings.insertBetween_lex64( formerOrder, insertOrder )
				
			else
			
				newOrder = WFSortOrderStrings.insertBetween_lex64( insertOrder, formerOrder )
				
			end
		
		end
		
		
		
		orders[ assetId ] = newOrder
		formerOrder = newOrder
	end
		
	return KSPromises.resolvedPromiseP( 'data', orders )
		
end

local calculateNewInsertOrderForAssetsP = KSPromises.liftP( calculateNewInsertOrderForAssets )

--------------------------------------------------------------------------------

local function addAssetsToOwnAlbumP( catalog, --[[string]] albumId,
					--[[string]] assetIds, --[[table]] orders )
	local undoRedoAssetIds = {}
	
	local assetsPresent = {}
	
	local function checkForExistingAssetsP()
		
		local arguments = table.concat( assetIds, '\',\'' )
		arguments = string.format( '\'%s\'', arguments )
		
		local sqlQuery = [[ SELECT assetId FROM album_asset WHERE albumId == ? AND assetId IN ( ]] .. arguments .. [[ ) ]]
		return catalog:indexStore():queryP( sqlQuery, albumId )
		:mapDataP( function( _, idRows )
			for i = 1, #idRows do
				assetsPresent[ idRows[ i ][ 1 ] ] = true
			end
			return 'data', assetsPresent
		end )

	end
	
	local function createAlbumAssetDocument( assetId )
		
		if assetsPresent[ assetId ] then
			return 'data', true
		else
			local baseDocValue = WFCatUtils.createBaseOverlayWithUserCreatedFields( KSIdGenerator.uuidInLowercaseHex() )
			
			local albumAssetOverlay = {
				type = 'album_asset',
				album = {
					id = albumId,
				},
				asset = {
					id = assetId,
				},
			}
			
			if orders then
				local order = orders[ assetId ]
				albumAssetOverlay[ "order" ] = order
			end
			
			WFTableMerge.applyOverlay( albumAssetOverlay, baseDocValue )

			local innerResultP = catalog:docStore():applyOverlayP( albumAssetOverlay )

			table.insert( undoRedoAssetIds, assetId )
		
			return innerResultP
			
		end
					
	end
	
	local resultP = catalog:indexStore():queryP( [[SELECT docId FROM albums WHERE docId == ? ]], albumId )
		:mapP( function ( status, data )

			if status == 'data' and data and data [ 1 ] and data[ 1 ][ 1 ] then
				return checkForExistingAssetsP()
				:mapDataP( function ()
					return KSPromises.allP( createAlbumAssetDocument, assetIds )
				end )
			else
				return 'error', 'invalidAlbum'
			end
			
		end )
		:mapDataP( function( _, assetIds )
			return LrmCatAlbum.touchAlbumP( catalog, albumId )
				:mapDataP( function()
					return 'data', assetIds
				end )
		end )

	return resultP:mapDataP( function( _, assetIds )
		return 'data', assetIds, undoRedoAssetIds
	end )			
end

--------------------------------------------------------------------------------

--[[!
	\brief Adds assets to an album in this catalog.
	
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (array of strings) An array of unique IDs for the assets to add.
	\param orders (map of order per assetid) Map containing order for each asset id.
	\param maybeRedactionOptions (map) Map containing the values of redaction options.
			Valid only if the albumId belongs to another owner's space.
	
	TODO: The return results from this function are weird.  It should consistently
	return a boolean, or a list of ids, not a mix of the two as it does currently.
	
	\return <code>P< true | array ></code>
		A Promise object that resolves to \c true if the assets were already in the album,
		or else to the list of asset IDs that were actually added to the album.
		
]]--

function LrmAlbumAssetUtils.addAssetsToAlbumP( catalogService, --[[string]] albumId,
					--[[string]] assetIds, --[[table]] orders, --[[table]] maybeRedactionOptions )

	logger:debugf( "LrmAlbumAssetUtils.addAssetsToAlbumP: maybeRedactionOptions=%s", maybeRedactionOptions )
	
	local copyRenditionsFn = nil

	if WFPlatformInfo.isAndroid then

		copyRenditionsFn  = function( renditionInfo, docId )

			local md5 = renditionInfo and renditionInfo.renditionInfo and renditionInfo.renditionInfo.md5

			if not md5 or not docId or not WFPlatformInfo.isAndroid then

				return Rx.Observable.just( nil )

			end

			return Rx.Observable.create(function( observer )

				local function emitRenditionInfo()

					observer:onNext( renditionInfo )

					observer:onCompleted()

				end

				local putP = catalogService:getDataCacheP()

				:mapP ( function( status, dataCache )

					if status ~= 'data' then

						emitRenditionInfo()

						return

					end

					assert( dataCache )

					dataCache:getP( md5 )

					:mapP( function( status, data )

						if status ~= 'data' then

							emitRenditionInfo()

							return

						end

						local original_md5 = data:digest( 'MD5' )

						local new_md5 = original_md5 .. "_" .. docId

						dataCache:putP( new_md5, data )

						:mapP( function ( status )

							if status ~= 'data' then

								emitRenditionInfo()

								return

							end

							renditionInfo.renditionInfo.md5 = new_md5

							emitRenditionInfo()

						end)

						:runUntilDone()

					end )

					:runUntilDone()

				end)

				:runUntilDone()

				return Rx.Disposable.create(
					function()

						putP:close()

						putP = nil

					end )

				end)

		end

	end

	return catalogService:getWfCatalog()
	
		:call( asKSPromise )
		
		:mapDataP( function( _, wfCatalog )
		
			return WFCatSharedWithYouAlbums.getSpaceIdOfAlbumId( wfCatalog, albumId )
	
				:call( asKSPromise )
					
				:mapDataP( function( _, spaceId )
				
					logger:debugf( "LrmAlbumAssetUtils.addAssetsToAlbumP: spaceId=%s", spaceId )
				
					if spaceId then
					
						return WFCatSharedWithYouAlbums.addAssetsToAlbum( wfCatalog, spaceId, albumId, assetIds, maybeRedactionOptions, copyRenditionsFn )
							:call( asKSPromise )
							
					end
					
					return addAssetsToOwnAlbumP( catalogService, albumId, assetIds, orders )
				
				end )
		
		end )

	
end

--------------------------------------------------------------------------------

function LrmAlbumAssetUtils.addAssetsToAlbumWithOrderP( catalog, albumId, assetIds, maybeRedactionOptions )

	local currentHighOrderP = getHighestOrderForAlbumP( catalog:indexStore(), albumId )
	
	return calculateNewHighOrderForAssetsP( assetIds, currentHighOrderP )
	:mapDataP( function( status, orders )
	
		return LrmAlbumAssetUtils.addAssetsToAlbumP( catalog, albumId, assetIds, orders, maybeRedactionOptions )
	end )
end

--------------------------------------------------------------------------------

function LrmAlbumAssetUtils.addAssetsToAlbumWithCustomOrderP( catalog, albumId, assetIds, insertOrder, location, maybeRedactionOptions )
	
	return calculateNewInsertOrderForAssetsP( assetIds, insertOrder, location )
	:mapDataP( function( status, orders )
	
		return LrmAlbumAssetUtils.addAssetsToAlbumP( catalog, albumId, assetIds, orders, maybeRedactionOptions )
	end )
end



return LrmAlbumAssetUtils
