--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.reactive
local KSPromises = require 'KSPromises'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'
local KSDate = require 'KSDate'
local LrmCatAlbum = require 'LrmCatAlbum'

--==============================================================================

local LrmCatAlbumAsset = {}

--------------------------------------------------------------------------------
--[[!
	\brief Deletes assets from an album in this catalog.
	
	\param albumId (string) The unique ID of the album to modify.
	\param assetIds (array of strings) An array of unique IDs for the assets to delete.
	
	\return <code>P< true ></code> A Promise object that resolves to \c true
		if the request succeeded.
]]--

function LrmCatAlbumAsset.removeAssetsFromAlbumP( catalog, albumId, assetIds )

	WFArguments.check( 'OzCatalogService', 'string', 'table' )
		
	local albumAssetsPresent = {}

	local function getAlbumAssetIdsP()
		
		-- TODO [arahn 2013-11-05]: This is a SQL injection vulnerability.
		local arguments = table.concat( assetIds, '\',\'' )
		arguments = string.format( '\'%s\'', arguments )
		
		local sqlQuery = [[ SELECT docId, assetId FROM album_asset WHERE albumId == ? AND assetId IN ( ]] .. arguments .. [[ ) ]]
		return catalog:indexStore():queryP( sqlQuery, albumId )
		:mapDataP( function( _, idRows )
			for i = 1, #idRows do
				albumAssetsPresent[ idRows[ i ][ 2 ] ] = idRows[ i ][ 1 ]
			end
			return 'data', albumAssetsPresent
		end )

	end
	
	local function deleteAlbumAssetDocument( assetId )
	
		local albumAssetId = albumAssetsPresent[ assetId ]
		
		if albumAssetId then
		
			local localOnly = {
				album = {
					id = albumId,
				},
				asset = {
					id = assetId,
				},
			}
					
			local innerResultP = catalog:docStore():applyOverlayToDocumentP {
				docType = 'album_asset',
				docIds = albumAssetId,
				overlay = {
					_deleted = true,
					_localOnly = localOnly,
				},
			}
						
			return innerResultP

		else
			return 'data', true
		end

	end
	
	local resultP = catalog:indexStore():queryP( [[SELECT docId FROM albums WHERE docId == ? ]], albumId )
		:mapP( function ( status, data )

			if status == 'data' and data and data [ 1 ] and data[ 1 ][ 1 ] then
				return getAlbumAssetIdsP()
				:mapDataP( function ()
					return KSPromises.allP( deleteAlbumAssetDocument, assetIds )
				end )
			else
				return 'error', 'invalidAlbum'
			end
		
		end )
		:mapP( function( status, data )
			return catalog:docStore():docValueOnceP( 'album', albumId )
		end )
		:mapP( function( status, album )
			local coverId = KSTableUtils.derefOpt( album, 'cover', 'id' )
			
			if coverId and albumAssetsPresent[ coverId ] then
				return catalog:docStore():applyOverlayToDocumentP {
					docType = 'album',
					docId = albumId,
					overlay = {
						cover = {
							_delete = true,
						},
					},
				}
			end
		end )
		:mapP( function()
			return LrmCatAlbum.touchAlbumP( catalog, albumId )
		end )

		-- jrechtie, 2017-08-10, LRMI-4185408
		-- The album "updated" field is (mis-)used as a kind of "last import date" field in LrM and Nimbus.
		-- Do not touch the local timestamp on deletion of album_asset. This seems to be consistent with what Oz does.
	
	return resultP
	
end

--------------------------------------------------------------------------------

local function createIndex( array )
	local index = { }
	for i, asset in ipairs( array ) do
		index[ asset ] = i
	end
	
	return index
end

--------------------------------------------------------------------------------

local function createMarkedAssetArray( assets, touched )
	local markedAssets = { }
	
	for i, assetId in ipairs( assets ) do
		markedAssets[ i ] = {
			assetId = assetId,
			touched = not not touched[ assetId ] -- toboolean
		}
	end
	
	return markedAssets
end

--------------------------------------------------------------------------------

--[[!
	\brief Apply an ordering to a set of assets in the album.
	
	All assets whose ids are in the list will have a persisted ordering when this
	is complete. Any members of the album not in the list will be untouched.  The
	optional set of "touched" ids will be used as a hint to avoid unnecessary
	changes to the order of other items.
	
	Assets in the list that turn out to not actually be in the album for any
	reason (either they aren't really in there or have been deleted since the list
	was generated) will be ignored 	rather than interfere with the persistence of
	other ordering adjustments.
	
	If the assets are already in the specified order, no changes be made even if
	they appear in the touched id set (i.e., superfluous touched id entries should
	not trigger unnecessary changes).
	
	If changes are required, this routine will prefer to modify the order attribute
	of items that appear in the touched id set.
	
	\param albumId (string) The unique ID of the album.
	\param assetArray (table) An array of asset id strings in the desired order.
	\param optTouchedIds (table) a set of "touched" assets (asset id string keys)
	
	\return <code>P< true ></code> An indication that the order was saved.
 ]]--

function LrmCatAlbumAsset.updateAssetOrderingP( catalog, albumId, assetArray, optTouchedIds )

	return catalog:indexStore():queryP( [[select assetId, docId, sortOrder
			from album_asset where albumId = ?]], albumId )
		:mapDataP( function( status, results )
		
			local assetIdToIndex = createIndex( assetArray )
			local markedAssetArray = createMarkedAssetArray( assetArray, optTouchedIds or { } )
		
			for i, result in ipairs( results ) do
				local index = assetIdToIndex[ result[ 1 ] ]
				if index then
					markedAssetArray[ index ].albumAssetId = result[ 2 ]
					markedAssetArray[ index ].order = result[ 3 ]
				end
			end
			
			local WFAlbumOrdering = require 'WFAlbumOrdering'

			local updates = WFAlbumOrdering.adjustOrderFields( markedAssetArray )
			
			local docIds, overlays = { }, { }
			
			for i, update in pairs( updates ) do
				local docId = markedAssetArray[ i ].albumAssetId
				
				if docId then
					docIds[ #docIds + 1 ] = docId
					overlays[ #overlays + 1 ] = {
						order = update,
						userUpdated = KSDate.timeInIso8601Format()
					}
				end
			end

			if #docIds > 0 then
				return catalog:docStore():applyOverlayToDocumentP {
					docType = 'album_asset',
					docIds = docIds,
					overlays = overlays
				}
			else
				return KSPromises.resolvedPromiseP( 'data', true )
			end
		end ):runUntilDoneP()

end

--------------------------------------------------------------------------------

return LrmCatAlbumAsset
