--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.reactive
local KSPromises = require 'KSPromises'
local KSReactive = require 'KSReactive'
local KSReactivePromiseBridge = require 'KSReactivePromiseBridge'

	-- WF.core.utils
local WFUnknown = require 'WFUnknown'
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.core.rxlua
local WFRxPromiseBridge = require 'WFRxPromiseBridge'

	-- WF.core.os
local KSDate = require 'KSDate'

--==============================================================================

-- @mark LrmCatCollaboration namespace

--------------------------------------------------------------------------------

local LrmCatCollaboration = {}

local kIsCatalogComment = {}

local asKSPromise = WFRxPromiseBridge

--------------------------------------------------------------------------------

local kFavoriteFields = [[ created, fu.firstName, fu.lastName, fu.fullName,
		author_id, spacefavorite.assetId, spacefavorite.spaceId ]]

--------------------------------------------------------------------------------

local kCommentFields = [[ body, created, fu.firstName, fu.lastName, fu.fullName,
		author_id, comment.assetId, comment.spaceId ]]

--------------------------------------------------------------------------------

local nextIndex
do
	local index = 0
	nextIndex = function() index = index + 1 return index end
end

local kDocIdIndex				= nextIndex()
local _ --[[ kSourceIndex ]]	= nextIndex()
local _ --[[ kAlbumIdIndex ]]	= nextIndex()
local _ --[[ kAlbumNameIndex ]]	= nextIndex()
local _ --[[ kUpdateTimeStampIndex ]]	= nextIndex()
local kBodyIndex				= nextIndex()
local kCreatedIndex				= nextIndex()
local kAuthorFirstNameIndex		= nextIndex()
local kAuthorLastNameIndex		= nextIndex()
local kAuthorFullNameIndex		= nextIndex()
local kAuthorIdIndex			= nextIndex()
local kAssetIdIndex				= nextIndex()
local kSpaceIdIndex				= nextIndex()

--------------------------------------------------------------------------------

local kCommentOnSpaceByAlbumClause = [[
		FROM
			comment
		LEFT JOIN
			feedback_user fu
		ON
			fu.userId == comment.author_id
		WHERE
			comment.assetId == ? AND comment.spaceId
		IN (
			SELECT
				spaceId
			FROM
				space_album
			WHERE
				space_album.albumId == ?
		)
		ORDER BY
			created]]
			
--------------------------------------------------------------------------------

local function formatComment( idRow, optIndexOfFirstCommentField )
	
	local comment = {}
	
	local offset = 0
	
	if optIndexOfFirstCommentField then
		offset = optIndexOfFirstCommentField - kBodyIndex
	end
	
	-- doc id is always first, regardless of offset
	comment.id = idRow[ kDocIdIndex ]
	
	comment.body = idRow[ kBodyIndex + offset ]
	comment.created = idRow[ kCreatedIndex + offset ]
	comment.spaceId = idRow[ kSpaceIdIndex + offset ]
	comment.assetId = idRow[ kAssetIdIndex + offset ]
	comment.type = 'comment'
	
	comment.author = {
		firstName = idRow[ kAuthorFirstNameIndex + offset ],
		fullName = idRow[ kAuthorFullNameIndex + offset ],
		lastName = idRow[ kAuthorLastNameIndex + offset ],
		id = idRow[ kAuthorIdIndex + offset ],
	}
	
	return comment

end

--------------------------------------------------------------------------------

local function formatFavorite( idRow )
	
	local favorite = {}
	
	-- doc id is always first, regardless of offset
	favorite.id = idRow[ 1 ]
	
	favorite.created = idRow[ 6 ]
	favorite.assetId = idRow[ 11 ]
	favorite.spaceId = idRow[ 12 ]
	favorite.type = 'favorite'
	
	favorite.author = {
		firstName = idRow[ 7 ],
		fullName = idRow[ 9 ],
		lastName = idRow[ 8 ],
		id = idRow[ 10 ],
	}
	
	return favorite

end

--------------------------------------------------------------------------------

local function getSpaceForAlbumP( catalog, albumId )
	
	return catalog:indexStore():queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
		:mapDataP( function( _, idRows )
		
			-- If there is a space for this album, return the ID for that space.
			local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
			if spaceId then
				return 'data', spaceId
			end
			
			return 'error', "Album is not shared"
		end )

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.hasCommentsInAlbumB( catalog, albumId )

	return catalog:indexStore():queryB( [[
		SELECT
			comment.docId
		FROM
			comment
		WHERE
			spaceId
		IN (
			SELECT
				spaceId
			FROM
				space_album
			WHERE
				space_album.albumId == ? )
		AND 
			comment.assetId
		NOT IN (
			SELECT
				docId
			FROM 
				trashAssets )
		LIMIT 1 ]], albumId )
		
	:closingMapB( function( idRows )
	
		if idRows == WFUnknown.value() then
			return idRows
		end
		
		return idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
		
	end )

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.hasSpaceFavoritesInAlbumB( catalog, albumId )

	return catalog:indexStore():queryB( [[
		SELECT
			spacefavorite.docId
		FROM
			spacefavorite
		WHERE
			spaceId IN (SELECT spaceId FROM space_album WHERE albumId == ?)
		AND
			spacefavorite.assetId NOT IN (SELECT docId FROM trashAssets)
		LIMIT 1
			]], albumId )
		
	:closingMapB( function( idRows )
	
		if idRows == WFUnknown.value() then
			return idRows
		end
		
		return idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
		
	end )
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.addCommentP( catalog, albumId, assetId, comment )

	if not ( #comment > 0 ) then
		error( "OzCatalogService:addCommentP: The comment body must be a non-empty string", 2 )
	end
	
	local resultP
	if albumId then
		resultP = getSpaceForAlbumP( catalog, albumId )
	else
		resultP = KSPromises.resolvedPromiseP( "data", kIsCatalogComment )
	end
	
	local commentId
	
	return resultP:mapDataP( function( _, spaceId )
	
		assert( spaceId, "No sharing space" )
		
		if spaceId == kIsCatalogComment then
			spaceId = nil
		end
		
		commentId = KSIdGenerator.uuidInLowercaseHex()
		local created = KSDate.timeInIso8601Format( nil, false )
		
		-- fill in the other values that we expect the server will provide

		return catalog:ozUserP()
			:mapDataP_new( function( user )
			
				return user:userInfoCache():docValue( user:userId() )
					:call( asKSPromise )
					
			end )
			:mapDataP( function( _, userInfo )
			
				local author = {
					first_name = KSTableUtils.derefOpt( userInfo, 'first_name' ),
					last_name = KSTableUtils.derefOpt( userInfo, 'last_name' ),
					full_name = KSTableUtils.derefOpt( userInfo, 'full_name' ),
					id = userInfo._id,
				}
		
				local resultP = catalog:docStore():applyOverlayToDocumentP {
					docType = 'comment',
					docId = commentId,
					overlay = {
						subtype = 'user',
						body = comment,
						space = { id = spaceId },
						asset = { id = assetId },
						_localOnly = {
							created = created,
							author = author
						},
					},
				}
				
				return resultP
				
			end )
	end )
	:mapDataP( function( status,_ )
		return status, commentId
	end )
	:runUntilDoneP()
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.deleteCommentP( catalog, commentId )

	return catalog:docStore():applyOverlayToDocumentP {
		docType = 'comment',
		docIds = commentId,
		overlay = {
			_deleted = true,
		},
	}
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.deleteSpaceFavoriteP( catalog, spacefavoriteId )

	return catalog:docStore():applyOverlayToDocumentP {
		docType = 'favorite',
		docIds = spacefavoriteId,
		overlay = {
			_deleted = true,
		},
	}
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.toggleMySpaceFavoriteP( catalog, assetId, albumId )

	local function getSpaceForAlbumP( catalog, albumId )
		
			return catalog:indexStore():queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
				:mapDataP( function( _, idRows )
				
					-- If there is a space for this album, return the ID for that space.
					local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
					if spaceId then
						return 'data', spaceId
					end
					
					return 'error', "Album is not shared"
				end )
		
		end
		
		local userInfo
		local ozUserId 
		
		local toggleP = catalog:ozUserP()
			:mapDataP_new( function( user )
			
				ozUserId = user:userId()
				
				return user:userInfoCache():docValue( user:userId() )
					:call( asKSPromise )
				
			end )
			
			:mapDataP_new( function( localUserInfo )
			
				userInfo = localUserInfo
				
				return getSpaceForAlbumP( catalog, albumId )
				
			end )
			
			:mapDataP( function( _, spaceId )
			
				return catalog:indexStore():queryP( [[SELECT docId FROM spacefavorite WHERE assetId == ? AND
										spaceId == ? AND author_id == ? ]], assetId, spaceId, ozUserId )
										
					:mapDataP( function( _, idRows )
					
						local spaceFavoriteId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
						
						if not spaceFavoriteId then
						
							local author = {
								first_name = KSTableUtils.derefOpt( userInfo, 'first_name' ),
								last_name = KSTableUtils.derefOpt( userInfo, 'last_name' ),
								full_name = KSTableUtils.derefOpt( userInfo, 'full_name' ),
								id = userInfo._id,
							}

							spaceFavoriteId = KSIdGenerator.uuidInLowercaseHex()
					
							local overlay = {
								space = {
									id = spaceId,
								},
								asset = {
									id = assetId,
								},
								subtype = 'user',
								_localOnly = {
									created = KSDate.timeInIso8601Format( nil, false ),
									author = author,
								}
							}
							
							local resP = catalog:docStore():applyOverlayToDocumentP {
								docType = 'favorite',
								docId = spaceFavoriteId,
								overlay = overlay,
							}
							
							return resP
									
						else
						
							local resP = catalog:docStore():applyOverlayToDocumentP {
								docType = 'favorite',
								docId = spaceFavoriteId,
								overlay = { _deleted = true },
							}
							
							return resP
							
						end
						
						
							
					end )
					
			end )
			
		return toggleP
			
end

--------------------------------------------------------------------------------

local function getSyncTerms( docType )

	if docType == 'asset' then
		return {
			fetchSql = [[SELECT docId, spaceId FROM space_asset WHERE assetId == ?]],
			isShared = [[SELECT
							space_asset.docId, spaceId FROM space_asset
						INNER JOIN
							space
						ON
							space_asset.spaceId == space.docId
						WHERE
							assetId == ? AND space.private == 0]],
			docId = 'assetId',
			space_doc_id = 'space_asset_id',
			space_doc = 'space_asset',
			
		}
	elseif docType == 'album' then
		return {
			fetchSql = [[SELECT docId, spaceId FROM space_album WHERE albumId == ?]],
			isShared = [[SELECT
							space_album.docId, spaceId FROM space_album
						INNER JOIN
							space
						ON
							space_album.spaceId == space.docId
						WHERE
							albumId == ? AND space.private == 0]],
			docId = 'albumId',
			space_doc_id = 'space_album_id',
			space_doc = 'space_album',
		}
	else
		error "docType must be 'asset' or 'album'";
	end
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.shareAssetOrAlbum( catalog, docType, docId, spaceSubtype, allowDownload )

	-- Revel defaults the ability to download shared albums to true, while Lr wants it to be false
	-- adding a param to control this and, if not set, the Lr default will be used
	
	if allowDownload == nil then
		 allowDownload = false
	end
	
	
	-- The current sharing model for albums calls for a 1-1 mapping from an 'album' or 'asset' document to
	-- a 'space'. In order to share the document, a 'space' needs to be created and the document
	-- added to it.

	-- This function creates a space and adds the provided document to it by creating a
	-- 'space_album' or 'space_asset' document as appropriate. If a space already exists for this document,
	-- it returns the ID of that 'space'.
	
	local syncTerms = getSyncTerms( docType )
		
	local resultP = catalog:indexStore():queryP( syncTerms.fetchSql, docId )
		:mapDataP( function( _, idRows )
		
			-- If there is a space for this album, return the ID for that space.
			local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 2 ]
			if spaceId then
				
				local innerResultP = catalog:docStore():applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						private = false,
						
					}
				}
				
				return innerResultP:mapDataP( function()
					return 'data', spaceId
				end )
				
			end
			
			spaceId = KSIdGenerator.uuidInLowercaseHex()
			
			local space_doc_id = KSIdGenerator.uuidInLowercaseHex()

			-- Save the album information in a localOnly section in the space itself. Oz does
			-- not have a separate Add Album to Space API for 2.0. We need to send a single
			-- request to create a space and add the album to it.  When this API is present, we
			-- should sync the space_album document separately as is done for album_asset
			-- documents.
			
			local spaceResultP = catalog:docStore():applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						_localOnly = {
							[ docType ] = {
								[ syncTerms.docId ] = docId,
								[ syncTerms.space_doc_id ] = space_doc_id,
							},
						},
						subtype = 'shared',
						download = allowDownload,
						private = false,
						metadata = false,
						location = false,
					},
				}
		
			return spaceResultP:mapDataP( function()
			
				local innerResultP = catalog:docStore():applyOverlayToDocumentP {
					docType = syncTerms.space_doc,
					docIds = space_doc_id,
					overlay = {
						space = {
							id = spaceId,
						},
						[ docType ] = {
							id = docId,
						}
					},
				}
				
				return innerResultP:mapDataP( function()
					return 'data', spaceId
				end )
			
			end )
			
		end )
		
	return resultP

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.unshareAssetOrAlbumP( catalog, docType, docId )

	local syncTerms = getSyncTerms( docType )
		
	local resultP = catalog:indexStore():queryP( syncTerms.fetchSql, docId )
		:mapDataP( function( _, idRows )
		
			local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 2 ]
			
			if not idRows[ 1 ] then
				return 'error', docType .. " is not shared"
			end
			
			local spaceResultP = catalog:docStore():applyOverlayToDocumentP {
					docType = 'space',
					docIds = spaceId,
					overlay = {
						private = true,
					},
				}
		
			return spaceResultP:mapDataP( function ()
				return 'data', true
			end )
			
		end )
		
	return resultP
		
end

--------------------------------------------------------------------------------

local function sharedAlbumUrlB_imp( docStore, spaceId )

	return docStore:docValueCheckTypeB( 'space', spaceId )
		:closingMapB( function ( docValue )
		
			return KSTableUtils.derefOpt( docValue, 'links', '/rels/web_space', 'href' )
				or KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata',
													'links', '/rels/web_space', 'href' )
			
		end )

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.sharedAssetOrAlbumUrlB( catalog, docType, docId, optSpaceId )

	local syncTerms = getSyncTerms( docType )
		
	if optSpaceId then
		return sharedAlbumUrlB_imp( catalog:docStore(), optSpaceId )
	end
	
	local resultB = catalog:indexStore():queryB( syncTerms.fetchSql, docId )
		:closingMapSwitchB( function( idRow )
		
			if WFUnknown.isUnknown( idRow ) then
				return KSReactive.constantB( idRow )
			end
		
			local spaceId = idRow and idRow[ 1 ] and idRow[ 1 ][ 2 ]
			if spaceId then
				return sharedAlbumUrlB_imp( catalog:docStore(), spaceId )
			else
				return KSReactive.constantB( nil )
			end

		end )
	
	WFCloseSupport.mustClose( resultB )

	return resultB

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.sharedAssetOrAlbumUrlP( catalog, docType, docId, nonNilResult )

	local syncTerms = getSyncTerms( docType )

	if nonNilResult then

		local sharedUrlB = LrmCatCollaboration.sharedAssetOrAlbumUrlB( catalog, docType, docId )
		
		return KSReactivePromiseBridge.closingBehaviorToPromise( sharedUrlB,
			function ( data )
				return WFUnknown.isKnown( data ) and data ~= nil
			end )

	else

		return catalog:indexStore():queryP( syncTerms.fetchSql, docId )
		:mapDataP( function ( status, idRow )
		
			local spaceId = idRow and idRow[ 1 ] and idRow[ 1 ][ 2 ]
			if spaceId then
				return catalog:docStore():docValueOnceP( 'space', spaceId )
					:mapDataP( function( _, docValue )
				
					local url = KSTableUtils.derefOpt( docValue, 'links', '/rels/web_space', 'href' )
					return 'data', url
				end )
	
			else
				return 'data', nil
			end
			
		end )
		
	end

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.resolveAlbumFromSpaceP( catalog, spaceId )

	return catalog:indexStore():queryP( [[SELECT albumId FROM space_album WHERE spaceId == ?]],
							spaceId )
						
		:mapDataP( function( _, row )
		
			if row and row[ 1 ] and row[ 1 ][ 1 ] then
				return 'data', row[ 1 ][ 1 ]
			end
			
			return 'data', nil
			
		end )
		
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.shareAlbumP( catalog, albumId, spaceSubtype )
	
	return LrmCatCollaboration.shareAssetOrAlbum( catalog, 'album', albumId, spaceSubtype )
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.unshareAlbumP( catalog, albumId )

	return LrmCatCollaboration.unshareAssetOrAlbumP( catalog, 'album', albumId )
		
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.sharedAlbumUrlP( catalog, albumId, nonNilResult )

	return LrmCatCollaboration.sharedAssetOrAlbumUrlP( catalog, 'album', albumId, nonNilResult )

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.getCommentsB( catalog )

	return catalog:indexStore():queryB( [[
		SELECT DISTINCT comment.docId, "space", 0, "catalog", 0, ]] .. kCommentFields .. [[
		FROM
			comment
		INNER JOIN
			album_asset
		ON
			comment.assetId == album_asset.assetId
		INNER JOIN
			space_album
		ON
			space_album.albumId == album_asset.albumId
		LEFT JOIN
			feedback_user fu
		ON
			fu.userId == comment.author_id
		WHERE
			comment.spaceId == space_album.spaceId
		]] )
		
	:closingMapB( function( idRows )
	
		if idRows == WFUnknown.value() then
			return idRows
		end
		
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatComment( idRow )
		end
		
		return result
		
	end )
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.getCommentsForAlbumB( catalog, albumId )

	return catalog:indexStore():queryB( [[
		SELECT
			DISTINCT comment.docId, "space", 0, "catalog", 0, ]] .. kCommentFields .. [[
		FROM
			comment
		LEFT JOIN
			album_asset
		ON
			album_asset.assetId == comment.assetId
		LEFT JOIN
			feedback_user fu
		ON
			fu.userId == comment.author_id
		WHERE
			album_asset.albumId == ?
		AND
			comment.assetId
		NOT IN (
			SELECT 
				docId
			FROM 
				trashAssets )
		AND
			comment.spaceId
		IN (
			SELECT
				spaceId
			FROM
				space_album
			WHERE
				space_album.albumId == ? ) ]], albumId, albumId )
	:closingMapB( function( idRows )
	
		if idRows == WFUnknown.value() then
			return idRows
		end
		
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatComment( idRow )
		end
		
		return result
		
	end )
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.getCommentsOnAlbumAssetB( catalog, albumId, assetId )

	return catalog:indexStore():queryB( [[
		SELECT
			comment.docId, "space", 0, "catalog", 0, ]] .. kCommentFields .. [[
		]] .. kCommentOnSpaceByAlbumClause, assetId, albumId )
	:closingMapB( function( idRows )
	    
	    if idRows == WFUnknown.value() then
			return idRows
		end
		
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatComment( idRow )
		end
		return result
	end )

end

--------------------------------------------------------------------------------

function LrmCatCollaboration.getSpaceFavoritesOnAlbumB( catalog, albumId )

	return catalog:indexStore():queryB( [[
		SELECT
			spacefavorite.docId, "space", 0, "catalog", 0, ]] .. kFavoriteFields .. [[
		FROM
			spacefavorite
		LEFT JOIN
			album_asset
		ON
			album_asset.assetId == spacefavorite.assetId
		LEFT JOIN
			feedback_user fu
		ON
			fu.userId == spacefavorite.author_id
		WHERE
			album_asset.albumId == ?
		AND
			spacefavorite.assetId NOT IN (SELECT docId FROM trashAssets)
		AND
			spacefavorite.spaceId IN (SELECT spaceId FROM space_album WHERE albumId == ?)]], albumId, albumId )
		
	:closingMapB( function( idRows )
	
		if idRows == WFUnknown.value() then
			return idRows
		end
		
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatFavorite( idRow )
		end
		
		return result
		
	end )
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.getSpaceFavoritesB( catalog )

	return catalog:indexStore():queryB( [[
		SELECT DISTINCT spacefavorite.docId, "space", 0, "catalog", 0, ]] .. kFavoriteFields .. [[
		FROM
			spacefavorite
		INNER JOIN
			album_asset
		ON
			spacefavorite.assetId == album_asset.assetId
		INNER JOIN
			space_album
		ON
			space_album.albumId == album_asset.albumId
		LEFT JOIN
			feedback_user fu
		ON
			fu.userId == spacefavorite.author_id
		WHERE
			spacefavorite.spaceId == space_album.spaceId
		]] )
		
	:closingMapB( function( idRows )
	
		if idRows == WFUnknown.value() then
			return idRows
		end
		
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatFavorite( idRow )
		end
		
		return result
		
	end )
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.getSpaceFavoritesOnAlbumAssetB( catalog, albumId, assetId )

	return catalog:indexStore():queryB( [[
		SELECT
			spacefavorite.docId, "space", 0, "catalog", 0, ]] .. kFavoriteFields .. [[
		FROM
			spacefavorite
		LEFT JOIN
			feedback_user fu
		ON
			fu.userId == spacefavorite.author_id
		WHERE
			spacefavorite.assetId == ? AND spacefavorite.spaceId
		IN (
			SELECT
				spaceId
			FROM
				space_album
			WHERE
				space_album.albumId == ?
		)
		ORDER BY
			created]],
		assetId, albumId )
		
	:closingMapB( function( idRows )
	
	    if idRows == WFUnknown.value() then
			return idRows
		end
		
		local result = {}
		for i, idRow in ipairs( idRows ) do
			result[ #result + 1 ] = formatFavorite( idRow )
		end
		return result
		
	end )
	
end

--------------------------------------------------------------------------------

function LrmCatCollaboration.isMySpaceFavoriteB( catalog, assetId, albumId, userId )

	local function getSpaceForAlbumP( indexStore, albumId )
	
		return indexStore:queryP( [[SELECT spaceId FROM space_album WHERE albumId == ?]], albumId )
			:mapP( function( _, idRows )
			
				-- If there is a space for this album, return the ID for that space.
				local spaceId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
				if spaceId then
					return 'data', spaceId
				end
				
				return 'data', nil
			end )
	
	end
	
	local isMyFavoriteB = getSpaceForAlbumP( catalog:indexStore(), albumId )
	
		:untilResolvedB( 'data', nil )
		:closingMapSwitchB( function( spaceId )
		
			if not spaceId then
				return KSReactive.constantB( nil )
			end
			
			return catalog:indexStore():queryB( [[SELECT docId FROM spacefavorite WHERE assetId == ? AND
											spaceId == ? AND author_id == ? ]], assetId, spaceId, userId )

				:closingFilterB( function( result ) return result ~= WFUnknown.value() end, nil )
				:closingMapB( function( idRows )
				
					local spaceFavoriteId = idRows and idRows[ 1 ] and idRows[ 1 ][ 1 ]
					
					local isMyFavorite = spaceFavoriteId and true or false
					
					return isMyFavorite
					
				end )
				
		end )
		
	return isMyFavoriteB
	
end

--------------------------------------------------------------------------------

return LrmCatCollaboration
