--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local Rx = require 'Rx'

	-- WF.core.utils
local KSStringUtils = require 'KSStringUtils'
local KSTableUtils = require 'KSTableUtils'

local WFArguments = require 'WFArguments'
local WFPlatformInfo = require 'WFPlatformInfo'
--------------------------------------------------------------------------------

local LrmLibraryDatabaseModelImp = {}

--==============================================================================

function LrmLibraryDatabaseModelImp.docValue( parentModel, docId, docType )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )
	
	if docType == nil then
		return {
			value = docStore:docValue( docId )
		}
	
	else
	
		return {
			value = docStore:docValueOnceP( docType, docId )
		}
		
	end
		
end

--------------------------------------------------------------------------------

function LrmLibraryDatabaseModelImp.docValueWithFieldSpec( parentModel, docId, optFieldSpec )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )

	return {
		value = docStore:docValue( docId, nil, optFieldSpec )
	}

end

--------------------------------------------------------------------------------

function LrmLibraryDatabaseModelImp.ongoingDocValue( parentModel, docId, docType )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )
	
	if docType == nil or WFPlatformInfo.isAndroid then
	
		return {
			value = docStore:ongoingDocValue( docId )
		}
		
	else
	
		return {
			value = docStore:docValueB( docType, docId )
		}
	
	end
	
	
end

--------------------------------------------------------------------------------

function LrmLibraryDatabaseModelImp.bulkDocValue( parentModel, assetIds )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )
		
	return {
		value = docStore:bulkDocValue( assetIds )
	}
	
end

--------------------------------------------------------------------------------

if WFPlatformInfo.isiOS then

	function LrmLibraryDatabaseModelImp.applyOverlay( parentModel, overlay )

		local catalog = assert( parentModel.catalog )
		local docStore = assert( catalog:docStore() )

		return {
			value = docStore:applyOverlay( overlay )
		}

	end

else

	function LrmLibraryDatabaseModelImp.applyOverlay( parentModel, docId, docType, overlay )

		local catalog = assert( parentModel.catalog )
		local docStore = assert( catalog:docStore() )

		return {
			value = docStore:applyOverlayToDocumentP {
				docType = docType,
				docId = docId,
				overlay = overlay,
			}
		}

	end

end
--------------------------------------------------------------------------------

function LrmLibraryDatabaseModelImp.queryIndexStore( parentModel, args, ... )

	WFArguments.check( 'table', 'table|string' )

	local catalog = assert( parentModel.catalog )
	local indexStore = assert( catalog:indexStore() )

	if WFPlatformInfo.isAndroid then
		args = {sql = args}
	else
		local n = select( '#', ... )

		if n > 0 then

			local queryArguments = { ... }

			for i = 1, n do
				table.insert( args, queryArguments[ i ] )
			end

		end
	end

	if ( args.oneShot == true ) then

		return {
			value = indexStore:queryOnce( args )
				:map( function( rows )

					return rows

				end )
		}

	else

		return {
			value = indexStore:query( args )
				:map( function( rows )

					return rows

				end )
		}

	end

end

--------------------------------------------------------------------------------


function LrmLibraryDatabaseModelImp.makeAssetInfoModel( parentModel, assetIds, keyPaths, separator )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )

	local sep = separator or "/"
	local WFPlatformInfo = require 'WFPlatformInfo'

	local docValues = docStore:bulkDocValue( assetIds )
		:map( function( docValueArray )
		
			local trimmedDocs = {}
		
			for _, docValue in ipairs( docValueArray ) do
			
				local doc = {}

				for _, value in ipairs( keyPaths ) do

					if WFPlatformInfo.isAndroid then

						local infoPath = KSStringUtils.split( value, sep )
						doc [value] = KSTableUtils.derefOpt( docValue, unpack( infoPath ) )
						-- KSTableUtils.setNested( doc, v, unpack( infoPath ) )
						-- LrA still rely on the path based mechanism instead of nested way of doing it.
					else
						local v = KSTableUtils.derefOpt( docValue, unpack( value ) )
						KSTableUtils.setNested( doc, v, unpack( value ) )
					end
				end

				trimmedDocs[ docValue._id or docValue._docId ] = doc
			end
		if WFPlatformInfo.isAndroid then
			trimmedDocs[ 'state' ] = true
		end
		return trimmedDocs

	end )

	return {
		value = docValues,
	}

end

--------------------------------------------------------------------------------

function LrmLibraryDatabaseModelImp.setAssetCustomData( parentModel, assetId, attribute, data, isValidForCurrentRevision )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )

	local localOnly = {}
	local customFields = {}
	local customData = {
		isValidForCurrentRevision = isValidForCurrentRevision,
		currentRevision = '',
		data = data,
	}
	
	customFields[attribute] = customData
	localOnly['lrm_customData'] = customFields
	
	local result = docStore:docValue( assetId )
	:flatMapLatest( function( docValue )

		if ( docValue._empty or docValue._deleted ) then
			return Rx.Observable.just( { state = false, error = 'document empty or deleted' } )
		end

		customData.currentRevision = docValue._revId

		return docStore:applyOverlay( {
			_id = assetId,
			type = 'asset',
			_localOnly = localOnly,
		} )
		:map( function()
			return { state = true }
		end )
		:catch( function( error )
			return Rx.Observable.just( { state = false, error = error } )
		end )

	end )
	:catch( function( error )
		return Rx.Observable.just( { state = false, error = error } )
	end ) 

	return {
		value = result
	}
end

--------------------------------------------------------------------------------

function LrmLibraryDatabaseModelImp.compact( parentModel, optVacuum )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )
	
	return {
		value = docStore:compact( optVacuum )
	}

end

--------------------------------------------------------------------------------

function LrmLibraryDatabaseModelImp.makeBulkDocValueModel( parentModel, assetIds, keyPaths )

	local catalog = assert( parentModel.catalog )
	local docStore = assert( catalog:docStore() )

	local fieldSpec = {}

	-- the fieldSpec format is different from the C++ keyPaths format
	-- (which was chosen because it can be written as C++ literals)
	for _, infoPath in ipairs( keyPaths ) do
		KSTableUtils.setNested( fieldSpec, true, unpack( infoPath ) )
	end

	fieldSpec._id = true

	local docValues = docStore:bulkDocValue( assetIds, fieldSpec )
		:map( function( strippedDocValues )

			local result = {}

			for _, docValue in ipairs( strippedDocValues ) do
				result[ docValue._id ] = docValue
			end

			return result

		end )

	return {
		value = docValues
	}

end

--------------------------------------------------------------------------------

return LrmLibraryDatabaseModelImp
