--[[----------------------------------------------------------------------------

LrmOzPurgeAgent.lua


--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2019 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This class monitors the content of a single catalog's data cache and periodically
purges it.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local ReplaySubject = Rx.ReplaySubject
local BehaviorSubject = Rx.BehaviorSubject

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local WFFileUtils = require 'WFFileUtils'
local KSPathUtils = require 'KSPathUtils'

	-- WF.core.rxlua
local WFArguments = require 'WFArguments'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFCloseSupport = require 'WFCloseSupport'
local WFDiagnosticLog = require 'WFDiagnosticLog'
local WFCapabilities = require 'WFCapabilities'
local WFOzSyncConstants = require 'WFOzSyncConstants'

	-- WF.os.prefs
local WFPlatformInfo = require 'WFPlatformInfo'

local OzCatalogUploadStatus = require 'OzCatalogUploadStatus'

	-- WF.oz.purge_agent
local WFOzPurgeSettings = require 'WFOzPurgeSettings'

	-- WF.oz_client.sync_engine
-- local WFOzBinaryUploadUtils = require 'WFOzBinaryUploadUtils'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.lrmobile.modules.source
local LrmConfig = require 'LrmConfig'
local WFCatAssetUtils = require 'WFCatAssetUtils'
local WFCatSyncAutoDeleter = require 'WFCatSyncAutoDeleter'

	-- WF.oz_client.utils
local OzFilePaths = require 'OzFilePaths'

--------------------------------------------------------------------------------

LrmConfig.registerBinaryStrategyCheck()

local check = assert( WFArguments.check )

local derefUniversal = assert( WFCatAssetUtils.derefUniversal )
local universalPath = assert( WFCatAssetUtils.universalPath )

local runUntilDone = WFRxRunUntilDone

local defaultCachePollingInterval = 40	-- seconds; feel free to tune

local minimumFreeSpace = WFOzPurgeSettings.minimumFreeSpace()

local logger = WFLogger 'LrmOzPurgeAgent'
	-- :setLevel 'DEBUG'

local derefOpt = KSTableUtils.derefOpt

local batchSize = 100

--------------------------------------------------------------------------------

local getDiskSpaceNeeded = function( self )

	local path

	if WFPlatformInfo.isAndroid then
		path = KSPathUtils.parent( self._dataCachePath )
	elseif WFPlatformInfo.isiOS then
		path = OzFilePaths.getCatalogBasePath( self._catalogId )
	end

	local fileSystemFreeSpace, _ -- fileSystemSize
		= WFFileUtils.freeSpaceAtPath( path )

	local requiredSpace = math.max(0, minimumFreeSpace - fileSystemFreeSpace)
	return requiredSpace

end

--============================================================================--
--[[!
	\brief A \c LrmOzPurgeAgent monitors the content of a single catalog's data
	cache and periodically purges it to make additional disk space available.
	
	It should be used only by LrmSyncClientImp.	
--]]

-- @mark LrmOzPurgeAgent

local purgeAgentMethods = {
	
	_cachePollingInterval = defaultCachePollingInterval,
		-- used to calm queries - configurable for tests

	_docStore = false,
		-- (OzDocStore)
	
	_indexStore = false,
		-- (OzIndexStore)

	_dataCache = false,
		-- (KSDataCache) data cache for this doc store (Android only)
	
	_dataCachePath = false,
		-- (string) path to cache (Android only)
	
	_purgeUserUpdated = false,
		-- An IndexStore query listing purgeable assets by userUpdated/epoch
	
	_purgeAll = false,
		-- An IndexStore query listing all purgeable	
	
	_purgeOnDeleteChain = false,
		-- the reactive chain that is executed on purgeOnDelete (exposed for testing)
	
	_purgeOnDelete = false,
		-- Subscription to the docStore's document deleter
	
	_checkCacheKeysFrom = false,
		-- (false or string) where do we start cross-checking cache keys?

	_warnedMissingEntries = false,
		-- (table) entries that we've warned about (so we don't keep bleating about them)

	_binaryStrategy = false,
		-- (string) LrmConfig
	
	_purgeSub = false,
		-- A subscription to _purgeUserUpdated used to continously purge unused renditions.

	_catalogId = false,
		-- (string) CatalogId

	_deleteFeedCallback = false,
		-- A subscription to send callback when a profile was deleted successfully (Android only)

	_canPurge = false,
	-- A subscription to send callback when a profile was deleted successfully (Android only)

	_initializedSubject = false,
		-- An Observable that sends completed once initializaton is done

	_wfCatalogStream = false,
	-- wfCatalogStream observable

	_autoDeleter = false,
		-- (WFCatSyncAutoDeleter)

	private = {

		-- The following slots are internal, but are exposed for testing.
		
		_getDiskSpaceNeeded = getDiskSpaceNeeded,
			-- a function that returns the needed diskspace

		_purgeContinuously = false
			-- An observable that emits the progress of the continuous purge process
	},

	WFCloseSupport.addCloseSupportMethods,

}

-- platform dependent forward declarations

local cacheReferencesPathKey

local userUpdatedKey

local removeCacheEntriesForDocument

local absolutePath

local purgeBinariesDirectly

if WFPlatformInfo.isiOS then

	cacheReferencesPathKey = 'path'

	userUpdatedKey = 'userUpdated'

	purgeBinariesDirectly = { 
		original = true, 
	 	proxy = true,
		thumbnail2x = true,
		preview = true,
		xmp_develop = true,
		presetProfile = true
	}

	--------------------------------------------------------------------------------

	absolutePath = function( self, relativePath )
	
		if relativePath[ 1 ] == '/' then
			return relativePath
		else
			local basePath = OzFilePaths.getCatalogBasePath( self._catalogId )
			
			return KSPathUtils.child( basePath, relativePath )
			
		end
	
	end
	
	--------------------------------------------------------------------------------

	removeCacheEntriesForDocument = function( self, docValue )

		return Rx.Observable.just( docValue )
	
	end
	
elseif WFPlatformInfo.isAndroid then

	cacheReferencesPathKey = 'md5'

	userUpdatedKey = 'epoch'

	purgeBinariesDirectly = { 
		original = true, 
	 	proxy = true,
		xmp_develop = true,
		presetProfile = true
	}

	--------------------------------------------------------------------------------

	local function cacheEntriesForDocument( docValue )

		local renditionNames = { 'thumbnail', 'thumbnail2x', 'preview' }
		local keys = {}
		
		for _, renditionName in ipairs( renditionNames ) do
			
			local md5 = derefOpt( docValue, '_localOnly', renditionName, 'md5' )
		
			if md5 then
				table.insert( keys, md5 )
			end
						
		end

		return keys

	end

	--------------------------------------------------------------------------------

	removeCacheEntriesForDocument = function( self, docValue )

		local cacheEntries = cacheEntriesForDocument( docValue )

		if ( next( cacheEntries ) ) then

			-- remove the cache entries
			return self._dataCache:removeBulk( cacheEntries )
				:defaultIfEmpty( docValue )
				:map( function()
					return docValue
				end )
		
		end

		return Rx.Observable.just( docValue )
	
	end

	--------------------------------------------------------------------------------

	absolutePath = function( self, relativePath )
	
		return OzCatalogUploadStatus.checkAndCorrectLocalPath( relativePath )
		
	end

end

assert( cacheReferencesPathKey )

assert( removeCacheEntriesForDocument )

assert( absolutePath )

assert( purgeBinariesDirectly )

--------------------------------------------------------------------------------

function purgeAgentMethods:sendDiagnosticLogData()

	return {

		name = "Lrm Purge Agent",
		id = 'LrmOzPurgeAgent',
		weight = 521,

		body = {
			binaryStrategy = self._binaryStrategy,	
		}
	}
	
end

--------------------------------------------------------------------------------

local function purgeableAlbumAssetListSql()

	return string.format( [[
		SELECT docId, renditionName, %s, size 
		FROM cacheReferences 
		WHERE 
			cacheReferences.hasUploaded == 1
			AND cacheReferences.renditionName != 'thumbnail2x'
			AND docId IN ( 
				SELECT 
					album_asset.assetId 
				FROM 
					album_asset
				INNER JOIN 
					albums 
				ON 
					album_asset.albumId == albums.docId 
				WHERE 
					album_asset.albumId == ?)
			AND (docId NOT IN ( 
					SELECT 
					   album_asset.assetId 
					FROM 
					   album_asset 
					INNER JOIN 
					   albums 
					ON 
						   album_asset.albumId == albums.docId 
					WHERE 
						   albums.useOffline == 1 
					)
				)
		   LIMIT %d ]],
		   cacheReferencesPathKey, batchSize)
		   
end
 
 --------------------------------------------------------------------------------

 local function purgeUserUpdatedSql()
 
	 return string.format( [[ 
		SELECT docId, renditionName, %s, size 
		FROM cacheReferences
			WHERE %s < ( SELECT max( %s ) FROM cacheReferences )
				AND cacheReferences.hasUploaded == 1
				AND cacheReferences.renditionName != 'thumbnail2x'
				AND (docId NOT IN (
					SELECT 
						album_asset.assetId 
					FROM 
						album_asset 
					INNER JOIN 
						albums 
					ON 
						album_asset.albumId == albums.docId 
					WHERE 
						albums.useOffline == 1 
				)
			)
			ORDER BY %s ASC, size DESC LIMIT %d ]],
			cacheReferencesPathKey, userUpdatedKey, userUpdatedKey, userUpdatedKey, batchSize )
end

 --------------------------------------------------------------------------------

local function purgeAllSql()

	return string.format( [[ 
		SELECT docId, renditionName, %s, size 
		FROM cacheReferences 
			WHERE cacheReferences.hasUploaded == 1
				AND cacheReferences.renditionName != 'thumbnail2x'
				AND (docId NOT IN (
					SELECT 
						album_asset.assetId 
					FROM 
						album_asset 
					INNER JOIN 
						albums 
					ON 
						album_asset.albumId == albums.docId 
					WHERE 
						albums.useOffline == 1 
				)
			)
			LIMIT %d ]],
			cacheReferencesPathKey, batchSize )

end

--------------------------------------------------------------------------------

local function setBinaryStrategy( self, binaryStrategy )

	if self._binaryStrategy ~= binaryStrategy then

		self._binaryStrategy = binaryStrategy

		self._purgeUserUpdated = self._indexStore:query {
			sql = purgeUserUpdatedSql(),
			quiesceUntilUpToDate = true,
			queryId = 'purgeUserUpdated',
			calm = self._cachePollingInterval,
		}

		self._purgeAll = self._indexStore:query {
			sql = purgeAllSql(),
			quiesceUntilUpToDate = true,
			queryId = 'purgeAll',
			calm = self._cachePollingInterval
		}

	end

end

--------------------------------------------------------------------------------

local function setupPurgeOnDelete( self )

	local purgeAgentChangeCounterLocalDocId = 'LrmOzPurgeAgent-changeCounter'
	
	self._purgeOnDeleteChain = self._docStore:deletedDocsSinceNamedChangeCounter( purgeAgentChangeCounterLocalDocId )

		:flatMap( function( docValues, continueFrom )

			return Rx.Observable.from( docValues )
			
				:flatMap( function( docValue )

					local result = Rx.Observable.just( nil )

					if docValue.type == 'asset' then

						local hasTrashOrRestoreAsset = derefOpt( docValue, '_newId' ) 
						local deletedFromOzWithNewId = derefOpt( docValue, '_serverProvidedMetadata', 'new_id' )

						if hasTrashOrRestoreAsset then
						
							result = Rx.Observable.just( docValue._id )
								-- dont purge as it goes to trash
						
						elseif deletedFromOzWithNewId then
							
							local localOnly = {}
							
							if WFPlatformInfo.isiOS then
							
								localOnly.files = {
									master = derefOpt( docValue, '_localOnly', 'files', 'master' ),
									proxy = derefOpt( docValue, '_localOnly', 'files', 'proxy' ),
								}
						
							else
								
								localOnly = {
									originalFilePath = derefOpt( docValue, '_localOnly', 'originalFilePath' ),
									latestOriginal = derefOpt( docValue, '_localOnly', 'latestOriginal' ),
									original = derefOpt( docValue, '_localOnly', 'original' ),
									proxy = derefOpt( docValue, '_localOnly', 'proxy' ),
									files = derefOpt( docValue, '_localOnly', 'files' ),
								}
								
							end
						
							-- FIXME [jandrew 2019-11-07]: We should batch together these overlays
							-- and use `bulkApplyOverlay`.
							
							result = self._docStore:applyOverlay {
									_id = deletedFromOzWithNewId,
									_localOnly = localOnly
								}
								
								:map( function()
									return docValue._id
								end )
							
						else
							
							-- if we are here, this is a final delete of the asset
							
							self:purgeDeletedDocument( docValue )

							result = removeCacheEntriesForDocument( self, docValue )
							
								:map( function()
									return docValue._id
								end )
							
						end

					end
					
					return result
			
				end )
			
				:toArray()
				
				:flatMap( function( docIds )
			
					return self._docStore:updateDeletedDocsSinceNamedChangeCounter(
							purgeAgentChangeCounterLocalDocId, continueFrom )
							
							:map( function()
							
								return docIds
							
							end )
				
				end )

		end )
		
		:filter( function( docIds )
			return #docIds > 0
		end )
		
		:shareReplay( 1 )

	self._purgeOnDelete = self._purgeOnDeleteChain:subscribe()
	
end 

--------------------------------------------------------------------------------

--[[!
	\brief Initialize a purge agent for a specific catalog.
	
	\param catalogService (OzCatalogServiceViaWFCatalog) catalog service
	\param binaryStrategy (LrmConfig) prefer proxy or master
--]]

function purgeAgentMethods:__init( catalogService, binaryStrategy, cachePollingInterval, optArgs )

	check( 'LrmOzPurgeAgent', 'OzCatalogService', '?binaryStrategy', '?number', '?table' )
	
	if cachePollingInterval then
		self._cachePollingInterval = cachePollingInterval
	end
	
	optArgs = optArgs or {}
	
	self._wfCatalogStream = catalogService:getWfCatalog()
	self._canPurge = BehaviorSubject( false )

	self._docStore = catalogService:docStore()
	self._indexStore = catalogService:indexStore()

	self._warnedMissingEntries = {}

	self._initializedSubject = ReplaySubject( 1 )
	
	-- Setup the Auto Deleter for Cascade Deletions
	
	self._wfCatalogStream
		:take( 1 )
		:tap( function( wfCatalog )
			self._autoDeleter = WFCatSyncAutoDeleter( wfCatalog, optArgs )
			self:addCloseTarget( self._autoDeleter )
		end )
		:call( runUntilDone )
		
	WFDiagnosticLog.addLogProvider( self )

	if WFPlatformInfo.isAndroid then

		self._deleteFeedCallback = ReplaySubject( 1 )

		Rx.Observable.combineLatest( function( wfCatalog, dataCache )
	
				self._catalogId = wfCatalog:ozCatalogSpec():catalogId()
				
				self._dataCache = dataCache
				self:addCloseTarget( self._dataCache )
		
				local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'
	
				local catalogDir = KSPathUtils.child( documentsDir, self._catalogId )
				catalogDir = KSPathUtils.child( catalogDir, "Catalog.ozcat" )
	
				WFFileUtils.mkdir( catalogDir )
	
				self._dataCachePath = KSPathUtils.child( catalogDir, "previewCache" )

				if binaryStrategy then
					setBinaryStrategy( self, binaryStrategy )
				end
				
				setupPurgeOnDelete( self )

				self._initializedSubject:onNext( self._catalogId )
				self._initializedSubject:onCompleted()

				return wfCatalog
				
			end,
			catalogService._wfCatalogStream,
			catalogService:getDataCache() )
			:call( runUntilDone )

	elseif WFPlatformInfo.isiOS then
		
		catalogService._wfCatalogStream
			:map( function( wfCatalog )

				self._catalogId = wfCatalog:ozCatalogSpec():catalogId()

				if binaryStrategy then
					setBinaryStrategy( self, binaryStrategy )
				end

				setupPurgeOnDelete( self )

				self._initializedSubject:onNext( self._catalogId )
				self._initializedSubject:onCompleted()

				return wfCatalog
				
			end )
			:call( runUntilDone )
		
	end
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:getDeleteFeedCallback( )

    return self._deleteFeedCallback

end

--------------------------------------------------------------------------------

function purgeAgentMethods:mayBeStartPurging( )

	self._canPurge:onNext( true )

end

--------------------------------------------------------------------------------

function purgeAgentMethods:mayBeStopPurging( )

	self._canPurge:onNext( false )

end

--------------------------------------------------------------------------------

function purgeAgentMethods:close()

	local purgeOnDelete = self._purgeOnDelete
	self._purgeOnDelete = nil	
	
	if purgeOnDelete then
		purgeOnDelete:dispose()
	end

	self:stopContinuousPurge()	

end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeDeletedDocument( docValue )

	local hasTrashOrRestoreAsset = derefOpt( docValue, '_newId' ) or derefOpt (docValue, '_serverProvidedMetadata', 'new_id')

	if hasTrashOrRestoreAsset then
		return docValue
	end

	for renditionName, _ in pairs( purgeBinariesDirectly ) do

		local renditionPath

		if renditionName == 'thumbnail' or renditionName == 'thumbnail2x' then
			renditionPath = derefUniversal( docValue, 'files', 'renditions', renditionName, 'relativePath' )
		elseif renditionName == 'preview' then
			renditionPath = derefUniversal( docValue, 'files', 'renditions', '2048', 'relativePath' )
		elseif renditionName == 'proxy' then
			renditionPath = derefUniversal( docValue, 'files', 'proxy', '2560', 'relativePath' )
		elseif renditionName == 'original' then
			renditionPath = derefUniversal( docValue, 'files', 'original', 'master', 'relativePath' )
		elseif renditionName == 'latestOriginal' then
			renditionPath = derefUniversal( docValue, 'files', 'original', 'latestMaster', 'relativePath' )
		elseif renditionName == 'xmp_develop' then

			if WFPlatformInfo.isiOS then
				renditionPath = derefUniversal( docValue, 'files', renditionName, 'relativePath' )
			elseif WFPlatformInfo.isAndroid then

				local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'

				local xmpSha256 = derefUniversal( docValue, 'files', 'xmp_develop', 'contentId' )

				if xmpSha256 then
					local xmpPath = OzFilePaths.getRelativeXmpDevelopPath( self._catalogId, xmpSha256 )

					renditionPath = KSPathUtils.child( documentsDir, xmpPath )
				end
			end
		elseif renditionName == 'presetProfile' then
			renditionPath = derefUniversal( docValue, 'files', 'presetProfile', 'relativePath' )
		end

		if renditionPath then

			WFFileUtils.delete( absolutePath( self, renditionPath ) )

			if renditionName == 'presetProfile' and self._deleteFeedCallback then

				self._deleteFeedCallback:onNext( true )

			end

		end

	end

	return docValue

end

--------------------------------------------------------------------------------

--[[

\brief prepare purging of documents

\param victims as an array of { docId, renditionName, relativePath, size }
	   (the output of an index store query)
\param spaceNeeded: spaceNeeded to satisfy minimumFreeSpace threshold or
		-1 for no purge limit

\returns a table with
	- docs:  changed documents
	- removeFiles: files that can be removed
	- removeCache: cache entries to purge
	- spaceReclaimed: space reclaimed
	- purged: number of purged documents
]]--

local function preparePurge( self, victims, spaceNeeded )

	logger:infof( "preparePurge\nvictims = %s\nspaceNeeded = %s",
				   victims, spaceNeeded )

	return self._wfCatalogStream
	:flatMap( function( wfCatalog )

		if not WFCapabilities.getCapability( "LrmOzPurgeAgent.isPurgeEnabledForResidualOrOfflineUsers" ) then
			return Rx.Observable.just ( false )
		end

		local user = wfCatalog:ozUser()
		local isValidUser = user and user:userId() ~= WFOzSyncConstants.kGlobalDocumentId

		return Rx.Observable.combineLatest(

			function( canPurge, isSubscriptionExpired )

				local dontPerformPurge = false

				if isValidUser then

					-- Don't Purge assets in offline or residual mode
					if isSubscriptionExpired or not canPurge then
						dontPerformPurge = true
					end

				end

				return dontPerformPurge

			end, self._canPurge, user:isSubscriptionExpired()
		)

	end )
	:take( 1 )
	:map(function( dontPerformPurge )

		local result = {
			docs = {},
			removeFiles = {},
			removeCache = {},
			spaceReclaimed = 0,
			purged = 0
		}

		if dontPerformPurge then
			return result
		end

		for i, victim in ipairs( victims ) do

			local docId = victim[ 1 ]

			logger:debugf('purging %s', docId)

			local renditionName = victim[ 2 ]
			-- Hard-coded indices are based on indexStore queries purgeAllSql, purgeUserUpdatedSql 
			-- and purgeableAlbumAssetListSql

			local overlay = result.docs[ docId ] or {
				_id = docId
			}

			if renditionName == 'original' then

				KSTableUtils.setNested( overlay, { _delete = true },
					universalPath( 'files', 'original', 'master' ) )

				-- MEH [immisch 2019-03-13]: deal with the irregular layout in the old _localOnly/originalFilePath
				KSTableUtils.setNested( overlay, { _delete = true },
					unpack { '_localOnly', 'originalFilePath' } )

			elseif renditionName == 'latestOriginal' then

				KSTableUtils.setNested( overlay, { _delete = true },
					universalPath( 'files', 'original', 'latestMaster' ) )

				-- MEH [immisch 2019-03-13]: deal with the irregular layout in the old _localOnly/originalFilePath
				KSTableUtils.setNested( overlay, { _delete = true },
					unpack { '_localOnly', 'latestOriginal', 'filePath' } )


			elseif renditionName == 'proxy' then
				KSTableUtils.setNested( overlay, { _delete = true },
					universalPath( 'files', 'proxy', '2560' ) )
			elseif renditionName == 'thumbnail2x' or renditionName == 'thumbnail' then
				KSTableUtils.setNested( overlay, { _delete = true },
					universalPath( 'files', 'renditions', renditionName ) )
			elseif renditionName == 'preview' then
				KSTableUtils.setNested( overlay, { _delete = true },
					universalPath( 'files', 'renditions', '2048' ) )
			elseif renditionName == 'xmp_develop' then
				KSTableUtils.setNested( overlay, { _delete = true },
					universalPath( 'files', 'xmp_develop' ) )
			end

			if purgeBinariesDirectly[ renditionName ] then

				if renditionName == 'xmp_develop' then
					local documentsDir = KSPathUtils.getStandardFilePath 'carouselDocuments'

					result.removeFiles[ #result.removeFiles + 1 ] = KSPathUtils.child( documentsDir, victim[ 3 ] )

				else
					result.removeFiles[ #result.removeFiles + 1 ] = victim[ 3 ]
				end

			else
				result.removeCache[ #result.removeCache + 1 ] = victim[ 3 ]
			end

			result.docs[ docId ] = overlay

			result.spaceReclaimed = result.spaceReclaimed + victim[ 4 ]
			result.purged = result.purged + 1

			if ( spaceNeeded ~= -1 ) and ( result.spaceReclaimed >= spaceNeeded ) then
				return result
			end

		end

		return result

	end)

end

--------------------------------------------------------------------------------

local function purgeVictims(self, victims, spaceNeeded )

	local purged = 0

	logger:debugf('victims:%s, spaceNeeded:%s', victims, spaceNeeded)

	if not next( victims ) or spaceNeeded == 0 then
		return Rx.Observable.just( 0, spaceNeeded )
	end

	return preparePurge( self, victims, spaceNeeded )
	:flatMap(function( batch )

		if not next ( batch.docs ) then
			return Rx.Observable.just( 0, spaceNeeded )
		end
		
		local overlays = KSTableUtils.extractValues( batch.docs )

		return self._docStore:bulkApplyOverlay( overlays )
		:flatMap( function()
		
			purged = purged + batch.purged
			if spaceNeeded ~= -1 then
				spaceNeeded = math.max(0, spaceNeeded - batch.spaceReclaimed)
			end
		
			-- remove the files
			for i, fileName in ipairs(batch.removeFiles) do
				WFFileUtils.delete( absolutePath( self, fileName ) )
			end

			if not next( batch.removeCache) or WFPlatformInfo.isiOS then
				-- no cache entries to purge, everything was in the filesystem
				return Rx.Observable.just( purged, spaceNeeded )
			end
					
			-- remove the cache entries
			return self._dataCache:removeBulk( batch.removeCache )
			:map( function()
				return purged, spaceNeeded
			end )

		end )

	end)

end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeUntilDone( query, spaceNeeded, queryId )
    
    local sum = 0
    
    local purgeObservable = Rx.ReplaySubject( 1 )
    
    local querySubscription 
    
    querySubscription = query
    
        :flatMap( function( victims )

            self._indexStore:requestFromQuery( queryId, 100 )
            
            return purgeVictims( self, victims, spaceNeeded )
            
                :map( function( purged, spaceNeeded )
                
                    return purged, spaceNeeded, #victims
                    
                end )
            
        end )
        
        :map( function ( _purged, _spaceNeeded, victimCount )
            
            spaceNeeded = _spaceNeeded
    
            sum = sum + _purged
            
            return victimCount
            
        end )
        
        :filter( function( victimCount )
            return victimCount < 100
        end )
        
        :take( 1 )
        
        :subscribe( nil, nil, function()
            
            purgeObservable:onNext( sum )
            purgeObservable:onCompleted()
            
            querySubscription:dispose()
            
        end ) 
    
    return purgeObservable
    
end

--------------------------------------------------------------------------------

--[[!
	\brief Triggers an immediate purge cycle.
]]--

function purgeAgentMethods:purgeImmediately( binaryStrategy, noPurgeLimit )

	check( 'LrmOzPurgeAgent', '?binaryStrategy', '?boolean' )

	if binaryStrategy then
		setBinaryStrategy( self, binaryStrategy )
	end

	local purgeAll = self._indexStore:query {
		sql = purgeAllSql(),
		quiesceUntilUpToDate = true,
		queryId = 'purgeAll',
		calm = 1
	}

	return self._initializedSubject
		:flatMap( function()

			-- wait until initialized before calling _getDiskSpaceNeeded

			local spaceRequired = self.private._getDiskSpaceNeeded( self )
			if noPurgeLimit then
				spaceRequired = -1
			end

			return self:purgeUntilDone( purgeAll, spaceRequired, 'purgeAll' )
		end )

end

--------------------------------------------------------------------------------

function purgeAgentMethods:purgeForAlbum( albumId, binaryStrategy, noPurgeLimit )
	
	check( 'LrmOzPurgeAgent', 'uuid', '?binaryStrategy', '?boolean' )

	local indexStore = self._indexStore
	local docStore = self._docStore
	
	local albumDoc = docStore:docValue( albumId )
	
	local spaceRequired = minimumFreeSpace
	if noPurgeLimit then
		spaceRequired = -1
	end	

	-- IMPORTANT!!
	--
	-- Verify offline use is still disabled for the album otherwise we are gonna ddos Oz
	-- as the PurgeAgent will try to remove files locally whereas the BinaryDownloader 
	-- will download again

	local result = albumDoc
		:flatMap( function( docValue )
	
			local useOffline = derefOpt( docValue, '_localOnly', 'useOffline' )
			
			if useOffline == false then
			
				return indexStore:query { 
						sql = purgeableAlbumAssetListSql(),
						quiesceUntilUpToDate = true,
						queryId = 'purgeAlbum',
						albumId
					}
				
			else
			
				return Rx.Observable.just( {} )
				
			end
			
		end )

	
	return self._initializedSubject
		:flatMap( function()
			return self:purgeUntilDone( result, spaceRequired, 'purgeAlbum' )
		end )
		
end

--------------------------------------------------------------------------------

function purgeAgentMethods:stopContinuousPurge()

	check( 'LrmOzPurgeAgent' )

	if self._purgeSub then
		self._purgeSub:dispose()
		self._purgeSub = nil
	end
	
end

--------------------------------------------------------------------------------

function purgeAgentMethods:startContinuousPurge( binaryStrategy )

	check( 'LrmOzPurgeAgent', 'binaryStrategy' )

	if self._purgeSub and binaryStrategy == self._binaryStrategy then
		-- if we're already running and the binary strategy hasn't changed, we're done
		return
	end

	if self._purgeSub then
		self:stopContinuousPurge()
	end

	setBinaryStrategy( self, binaryStrategy )

	self.private._purgeContinuously = self._purgeUserUpdated
		:flatMap( function( potentialVictims )
		
			local spaceNeeded = self.private._getDiskSpaceNeeded( self )

			local purged = purgeVictims( self, potentialVictims, spaceNeeded )
			
			self._indexStore:requestFromQuery( 'purgeUserUpdated', batchSize )

			return purged

		end )
		:replay( nil, 1 )
		:refCount()


	self._purgeSub = self.private._purgeContinuously:subscribe()
	
	return self.private._purgeContinuously
		
end

--------------------------------------------------------------------------------


return Rx.Class 'LrmOzPurgeAgent'( purgeAgentMethods )
