--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- WF.core.reactive
local KSReactive = require 'KSReactive'

	-- WF.core.utils
local WFUnknown = require 'WFUnknown'
local KSTableUtils = require 'KSTableUtils'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.lrmobile
local LrmCatCollaboration = require 'LrmCatCollaboration'
local LrmAlbumUtils = require 'LrmAlbumUtils'

--==============================================================================

local LrmSocialActivityStatusModel = {}

--------------------------------------------------------------------------------

local function isAlbumSharedB( catalog, albumId )

	return LrmAlbumUtils.isAlbumSharedB( catalog:indexStore(), albumId )
	
end

--------------------------------------------------------------------------------

local function updateLastReadTimestamp( self )

	local now = KSDate.timeInIso8601Format( nil, false )
	
	self._catalog:docStore()
		:setLocalStateP( 'lastReadSocialActivityTimestamp', now )
		:runUntilDone()

end

--------------------------------------------------------------------------------

local function hasNewSocialActivity( newComments, newFavorites )
	
	if WFUnknown.isKnown( newComments ) or WFUnknown.isKnown( newFavorites ) then
	
		if ( newComments and #newComments > 0 ) or ( newFavorites and #newFavorites > 0 ) then
			return true
		end
		
	end
	
	return false
	
end

--------------------------------------------------------------------------------

local function getFavoritesB( catalog, lastReadTimestamp, context, albumId, assetId )

	local indexStore = catalog:indexStore()
	
	if context == 'catalog' then
	
		return indexStore:queryB(
			[[SELECT docId FROM spacefavorite
				WHERE datetime( created ) > datetime( ? ) LIMIT 1]],
			lastReadTimestamp )
			
	elseif context == 'album' then
		
		return indexStore:queryB(
			[[SELECT
				spacefavorite.docId
			FROM
				spacefavorite
			LEFT JOIN
				 album_asset
			WHERE
				spacefavorite.assetId == album_asset.assetId AND
				datetime( created ) > datetime( ? ) AND
				album_asset.albumId == ? LIMIT 1]],
			lastReadTimestamp, albumId )
			
	else
	
		return indexStore:queryB(
			[[SELECT
				docId
			FROM
				spacefavorite
			WHERE
				datetime( created ) > datetime( ? ) AND
				spacefavorite.assetId == ? LIMIT 1]],
			lastReadTimestamp, assetId )
			
	end
		
end

--------------------------------------------------------------------------------

local function getCommentsB( catalog, lastReadTimestamp, context, albumId, assetId )

	local indexStore = catalog:indexStore()
	
	if context == 'catalog' then
	
		return indexStore:queryB(
			[[SELECT docId FROM comment
				WHERE datetime( created ) > datetime( ? ) LIMIT 1]],
			lastReadTimestamp )
			
	elseif context == 'album' then
	
		return indexStore:queryB(
			[[SELECT
				comment.docId
			FROM
				comment
			LEFT JOIN
				 album_asset
			WHERE
				comment.assetId == album_asset.assetId AND
				datetime( created ) > datetime( ? ) AND
				album_asset.albumId == ? LIMIT 1]],
			lastReadTimestamp, albumId )
			
	else
	
		return indexStore:queryB(
			[[SELECT
				docId
			FROM
				comment
			WHERE
				datetime( created ) > datetime( ? ) AND
				comment.assetId == ? LIMIT 1]],
			lastReadTimestamp, assetId )
			
	end
	
end

--------------------------------------------------------------------------------

local function hasNewActivityB( catalog, context, albumId, assetId )

	local hasNewActivityB = catalog:docStore():localStateB( 'lastReadSocialActivityTimestamp' )
		:closingKnownMapB( function( activityTimestamp )
		
			if activityTimestamp._empty then
				return WFUnknown.value()
			end
			
			local now = KSDate.timeInIso8601Format( nil, false )
			
			return KSTableUtils.derefOpt( activityTimestamp, 'value' ) or now
			
		end )
	
		:closingMapSwitchB( function( lastReadTimestamp )
			
			if WFUnknown.isKnown( lastReadTimestamp ) then
			
				local hasNewCommentsB = getCommentsB( catalog, lastReadTimestamp,
														context, albumId, assetId )
					
				local hasNewFavoritesB = getFavoritesB( catalog, lastReadTimestamp,
														context, albumId, assetId )
					
				return KSReactive.closingDataCallB( hasNewSocialActivity, hasNewCommentsB, hasNewFavoritesB )
				
			end
			
			return KSReactive.constantB( WFUnknown.value() )
			
		end )
		
	return hasNewActivityB

end

--------------------------------------------------------------------------------

local function supportsSocialActivity( hasComments, hasFavorites )

	if WFUnknown.isUnknown( hasComments ) then
		hasComments = false
	end
	
	if WFUnknown.isUnknown( hasFavorites ) then
		hasFavorites = false
	end
	
	if hasComments or hasFavorites then
		return true
	end
	
	return false
	
end

--------------------------------------------------------------------------------

local function supportsActivityB( catalog, context, albumId, assetId )

	if context == 'asset' then
	
		return isAlbumSharedB( catalog, albumId )
		
	elseif context == 'album' then
	
		return KSReactive.closingDataCallB( supportsSocialActivity,
											LrmCatCollaboration.hasCommentsInAlbumB( catalog, albumId ),
											LrmCatCollaboration.hasSpaceFavoritesInAlbumB( catalog, albumId ) )
	elseif context == 'catalog' then
	
		return LrmAlbumUtils.hasAnySharedAlbumsB( catalog:indexStore() )
		
	end
	
end
	
--------------------------------------------------------------------------------

function LrmSocialActivityStatusModel.makeSocialActivityStatusModel( catalog, context, albumId, assetId  )

			
	return {
		_catalog = catalog,
		
		value = {
		
			hasNewActivity = hasNewActivityB( catalog, context, albumId, assetId )
								:closingFilterB( function( val ) return val ~= nil end, WFUnknown.value() ),
								
			supportsActivity = supportsActivityB( catalog, context, albumId, assetId )
								:closingFilterB( function( val ) return val ~= nil end, WFUnknown.value() ),
			
		},
		
		events = {
		
			updateLastReadTimestamp = updateLastReadTimestamp
			
		}
	}
	
end

--------------------------------------------------------------------------------

return LrmSocialActivityStatusModel
