--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.core.lua
local WFWorkerThread = require 'WFWorkerThread'

	-- WF.core.table
local WFTableArchiving = require 'WFTableArchiving'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

local WFCapabilities = require "WFCapabilities"

local LrmConfig = require "LrmConfig"

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local runUntilDone = WFRxRunUntilDone

LrmConfig.registerBinaryStrategyCheck()

--==============================================================================

-- @mark LrmSyncClient namespace

-- Manages the reference to Sync MessageProcessor where our Sync infrastructure
-- is running on. See LrmSyncClientImp for the actual implementation

--------------------------------------------------------------------------------

local syncClientMethods = {

	_syncClientImp = false,
		-- (WFWorkerThread) Reference to the SyncClient WorkerThread instance
	
	WFCloseSupport.addCloseSupportMethods
	
}

--------------------------------------------------------------------------------

function syncClientMethods:__init( wfCatalog, optIndexSpecScript, optSyncClientImpName )
	
	check( 'LrmSyncClient', 'WFCatalog', '+string|nil', '+string|nil' )
	
	optIndexSpecScript = optIndexSpecScript or 'WFCatIndexDefinitions.lua'
	optSyncClientImpName = optSyncClientImpName or 'LrmSyncClientImp'
	
	local options = {
		syncClientImpName = optSyncClientImpName,
		serverBase = wfCatalog:ozUser():serverBaseUrl(),
		userId = wfCatalog:ozUserId(),
		userInfoCacheFolder = wfCatalog:ozUser():userInfoCacheFolder(),
		catalogId = wfCatalog:ozCatalogId(),
		indexSpecScript = optIndexSpecScript,
		rootCatalogPath = wfCatalog:fileSpec():rootPath(),
		docStorePath = wfCatalog:fileSpec():docStorePath(),
		enableCloudTrash = WFCapabilities.getCapability( "WFCatDelete.enableCloudTrash" ) and true or false,
	}
	
	local optionsString = WFTableArchiving.tableToString( options )
	
	self:setOwned {
		_syncClientImp = WFWorkerThread( 'LrmSyncClientMessageProcessor.lua', optionsString )
	}
	
	WFCloseSupport.mustClose( self )
	
end

--------------------------------------------------------------------------------

function syncClientMethods:getDeleteFeedCallback( )

	check( 'LrmSyncClient' )

	return self._syncClientImp:request( 'getDeleteFeedCallback' )

end

function syncClientMethods:startUploading()

	check( 'LrmSyncClient' )
	return self._syncClientImp:request( 'startUploading' )
	:call( runUntilDone )

end

function syncClientMethods:stopUploading()

	check( 'LrmSyncClient' )
	return self._syncClientImp:request( 'stopUploading' )
	:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:startSync()

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'startSync' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:startSyncWithoutUploader()

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'startSyncWithoutUploader' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:startAccountsSync()

	check( 'LrmSyncClient' )

	return self._syncClientImp:request( 'startAccountsSync' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:stopSync( )

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'stopSync' )
		:call( runUntilDone )
		
end

--------------------------------------------------------------------------------

function syncClientMethods:stopAccountsSync()

	check( 'LrmSyncClient' )

	return self._syncClientImp:request( 'stopAccountsSync' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:startSyncInResidualMode()

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'startSyncInResidualMode' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:stopSyncInResidualMode( )

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'stopSyncInResidualMode' )
		:call( runUntilDone )
		
end

--------------------------------------------------------------------------------

function syncClientMethods:startSpaceSyncWithoutOzCatalog()

	check( 'LrmSyncClient' )

	return self._syncClientImp:request( 'startSpaceSyncWithoutOzCatalog' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:stopSpaceSyncWithoutOzCatalog()

	check( 'LrmSyncClient' )

	return self._syncClientImp:request( 'stopSpaceSyncWithoutOzCatalog' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:startContinuousPurge( binaryStrategy )

	check( 'LrmSyncClient' , 'binaryStrategy' )
	
	return self._syncClientImp:request( 'startContinuousPurge', binaryStrategy )
		:call( runUntilDone )
		
end


--------------------------------------------------------------------------------

function syncClientMethods:purgeImmediately( binaryStrategy, noPurgeLimit )

	check( 'LrmSyncClient' , '?binaryStrategy', '?boolean' )
	
	return self._syncClientImp:request( 'purgeImmediately', binaryStrategy, noPurgeLimit )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:purgeForAlbum( albumId, binaryStrategy, noPurgeLimit )

	check( 'LrmSyncClient' , 'uuid', '?binaryStrategy', '?boolean' )
	
	return self._syncClientImp:request( 'purgeForAlbum', albumId, binaryStrategy, noPurgeLimit )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:isPushingMetadata()

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'isPushingMetadata' )
	
end

--------------------------------------------------------------------------------

function syncClientMethods:isPullingMetadata()
	
	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'isPullingMetadata' )
	
end

--------------------------------------------------------------------------------

function syncClientMethods:getRequiredUpdateStream()

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'getRequiredUpdateStream' )
		
end

--------------------------------------------------------------------------------

function syncClientMethods:isOzCatalogDeleted()

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'isOzCatalogDeleted' )
		
end

--------------------------------------------------------------------------------

function syncClientMethods:pausePusherReturningCounter()

	-- Used for synchronizing trash operations with the pusher.

	check( 'LrmSyncClient' )

	return self._syncClientImp:request( 'pausePusherReturningCounter' )

end

--------------------------------------------------------------------------------

function syncClientMethods:resumePusher()

	-- Used for synchronizing trash operations with the pusher.

	check( 'LrmSyncClient' )

	return self._syncClientImp:request( 'resumePusher' )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

function syncClientMethods:getDetailedUploadProgress()

	check( 'LrmSyncClient' )
	
	return self._syncClientImp:request( 'getDetailedUploadProgress' )
	
end

-----------------------------------------------------------------------------------

return Rx.Class( 'LrmSyncClient', syncClientMethods )
