--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local WFArguments = require 'WFArguments'
local WFPlatformInfo = require 'WFPlatformInfo'
local KSTableUtils = require 'KSTableUtils'
local WFDeviceInfo = require "WFDeviceInfo"
local WFAnalytics = require "WFAnalytics"
local WFProgressReporter = require 'WFProgressReporter'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.core.lua.threading
local WFRequestDispatcher = require "WFRequestDispatcher"

	-- WF.oz_client.managed_catalog
local WFCatConflictResolver = require 'WFCatConflictResolver'
local WFCatErrorHandler = require 'WFCatErrorHandler'
local WFCatUserSynchronizer = require 'WFCatUserSynchronizer'
local WFCatStorage = require 'WFCatStorage'
local WFCatSyncClientBase = require 'WFCatSyncClientBase'

	-- WF.oz_client.lrmobile
local WFCatBinaryUploader = require 'WFCatBinaryUploader'
local WFOzSyncConstants = require 'WFOzSyncConstants'
local OzSyncMetadataMapper = require 'OzSyncMetadataMapper'
local LrmCatSyncStatus = require 'LrmCatSyncStatus'

	-- WF.oz_client.doc_store

local OzCatalogServiceViaWFCatalog = require "OzCatalogServiceViaWFCatalog"

	-- WF trash
local WFCatTrash = require 'WFCatTrash'

local LrmConfig = require "LrmConfig"

local LrmOzPurgeAgent = require "LrmOzPurgeAgent"

local LrmCatAlbum = require "LrmCatAlbum"

local WFLogger = require 'WFLogger'

local logger = WFLogger( 'LrmSyncClientImp' )

--------------------------------------------------------------------------------

LrmConfig.registerBinaryStrategyCheck()

local runUntilDone = assert( WFRxRunUntilDone )
local check = assert( WFArguments.check)
local derefOpt = assert( KSTableUtils.derefOpt )

local Observable = Rx.Observable

--------------------------------------------------------------------------------

local function trackSha256MasterUploadError( puller, docValue )

	if docValue then
	
		local docType = derefOpt( docValue, 'type' )
		
		if docType == 'asset' then
		
			local hasSha256Error = derefOpt( docValue, 'importSource', 'status', 'error', 'all', 'reason', 'sha256' ) == "does not match"
			
			if hasSha256Error then
			
				local uniqueDeviceId = derefOpt( docValue, 'importSource', 'uniqueDeviceId' )
			
				if uniqueDeviceId and ( uniqueDeviceId == WFDeviceInfo.uniqueDeviceId() ) then
			
					local errorEvent = {}
				
					errorEvent['event.type'] = 'error'
					errorEvent['event.workflow'] = 'SYNC'
					errorEvent['event.subcategory'] = 'error_handler'
					errorEvent['event.error_type'] = 'Sha256MismatchError'
					
					errorEvent['content.id'] = docValue._id
					errorEvent['content.type'] = derefOpt( docValue, 'subtype' )
					errorEvent['content.mimetype'] = derefOpt( docValue, 'importSource', 'contentType' )
					errorEvent['content.size'] = derefOpt( docValue, 'importSource', 'fileSize' )
					errorEvent['event.error_desc'] = derefOpt( docValue, 'importSource', 'sha256' )
					
					WFAnalytics.trackEvent( errorEvent )
					
				end
				
			end
			
		end
		
	end
	
	
end

--------------------------------------------------------------------------------

local function albumAssetsForFilteredAssetsQueryFn( catalog, assetIds )

	return catalog:indexStore():queryOnce {
				sql = [[SELECT album_asset.docId, album_asset.assetId FROM TEMP_ids JOIN album_asset ON 
 							TEMP_ids.docId = album_asset.assetId]],
				tempIds = assetIds,
		}
		
end

--------------------------------------------------------------------------------

local function deleteAssets( catalog, assetIds )

	local overlays = {}
	
	for _, assetId in ipairs( assetIds ) do
	
		overlays[ #overlays + 1 ] = {
			docType = 'asset',
			_id = assetId,
			_deleted = true
		}

	end
	
	return catalog:docStore():bulkApplyOverlay( overlays )
		:call( runUntilDone )

end
--==============================================================================

-- @mark LrmSyncClientImp namespace

--------------------------------------------------------------------------------

local syncClientImpMethods = {
		
	_syncClientBase = false,
		-- (WFCatSyncClientBase)
	
	_userId = false,
		--(string)
	
	_storageMonitor = false,
		-- (WFCatStorage)
	
	_binaryUploader = false,
		-- (WFCatBinaryUploader)

	_progressReporter = false,
		-- (WFProgressReporter) Reports back processed and total bytes of files
		-- currently being uploaded.

	_purgeAgent = false,
		-- (LrmOzPurgeAgent)
		
	_errorHandler = false,
		-- (WFCatErrorHandler)
		
	_userSynchronizer = false,
		-- (WFCatUserSynchronizer)
	
	_conflictResolver = false,
		-- (WFCatConflictResolver)

	_catalogServiceViaWFCatalog = false,
		-- (OzCatalogServiceViaWFCatalog)
		
	_checkAutoPurgeSubscription = false,
		-- (Rx Disposable)
	
	_cloudTrashAutoPurge = false,
		-- (Rx Disposable)
		
	attachMessageHandlers = false,
	
	WFCloseSupport.addCloseSupportMethods,
	
}

--------------------------------------------------------------------------------

function syncClientImpMethods:__init( options )

	check( 'LrmSyncClientImp', 'table' )
	
	-- Additional options for WFCatSyncClientBase
	
	options.noFileAccessDb = true
	options.noCatalogSearch = true
	options.pullerPretfetch = true
	options.pusherDelayBeforeSendingUpdates = 3
	options.albumAssetsForFilteredAssetsQueryFn = albumAssetsForFilteredAssetsQueryFn
	
	local syncClientBase = WFCatSyncClientBase( options )
	
	-- ConflictResolver
	
	local catalog = syncClientBase:getCatalog()
	
	local conflictResolver = WFCatConflictResolver( catalog:docStore(), catalog:indexStore() )
	
	-- Error Handler
	
	local errorHandler = WFCatErrorHandler( catalog, {
			deleteAssets = deleteAssets,
			deleteIndividualAlbum = LrmCatAlbum.deleteAlbum,
			getSyncStatusOfAssetsOfAlbumAssets = LrmCatSyncStatus.getSyncStatusOfAssetsOfAlbumAssets,
			getSyncStatusOfAlbumsOfAlbumAssets = LrmCatSyncStatus.getSyncStatusOfAlbumsOfAlbumAssets,
			getSyncStatusOfParentCollectionOfAlbums = LrmCatSyncStatus.getSyncStatusOfParentCollectionOfAlbums,
		} )

	-- Custom doc handling
	
	local catalogPuller = syncClientBase:getCatalogPuller()
	
	catalogPuller:addAssetTypeHandler( 'asset', trackSha256MasterUploadError )
	
	local catalogPusher = syncClientBase:getCatalogPusher()
	
	self._userId = assert( options.userId )
	
	local function remapAssetDocForUpdate( _, localDoc, payload )

		OzSyncMetadataMapper.remapGlobalDocumentId( self._userId, payload )
		
	end
	
	if catalogPusher then
		catalogPusher:addUpdateDocRemapFunction( 'asset', remapAssetDocForUpdate )
	end
	
	-- Cloud Trash Auto-Purge

	local isAuthenticatedUser = self._userId ~= WFOzSyncConstants.kGlobalDocumentId
	
	local catalogId = assert( options.catalogId )
	
	local user = catalog:ozUser()
	
	local hasValidConnectionToOz = (isAuthenticatedUser and
									catalogId ~= WFOzSyncConstants.kGlobalDocumentId) and	-- CatalogId check is needed for Created users as they have valid connection to Oz but no catalog on Oz.
									user:hasValidConnectionToOz() or
									Rx.Observable.just( false )

	self._checkAutoPurgeSubscription = Rx.Observable.combineLatest(

			function( isSubscriptionExpired, hasValidConnectionToOz )

				if isSubscriptionExpired or not hasValidConnectionToOz then
					return true
				end

				return false

			end,
			user:isSubscriptionExpired(), hasValidConnectionToOz
		)
		:distinctUntilChanged()
		:subscribe( function( shouldAutoPurge )

			if self._cloudTrashAutoPurge then
				self._cloudTrashAutoPurge:dispose()
				self._cloudTrashAutoPurge = nil
			end

			if shouldAutoPurge then
				self._cloudTrashAutoPurge = WFCatTrash.autoPurge( catalog ):subscribe()
			end

		end )

	-- User Synchronizer
	
	local userSynchronizer = WFCatUserSynchronizer( user )
	
	-- Binary Uploader

	local ozCatalog = OzCatalogServiceViaWFCatalog( Rx.Observable.just( catalog ) )

	local progressReporter = WFProgressReporter( "Binary Upload Progress Report" )
	self._progressReporter = progressReporter

	local uploadOptions = {
		deriveCatalogBasePath = true,
		progressReporter = progressReporter,
	}

	local catSpec = syncClientBase:getCatSpec()

	-- Binary Uploader
	local binaryUploader = WFCatBinaryUploader( catalog, nil, nil, catSpec, uploadOptions )
	
	-- Quota Monitoring
	
	local storageMonitor = WFCatStorage( catalog )
	
	if catalogPusher then 
	
		-- monitor quota
		
		local quotaMonitorSubscription = storageMonitor:isOverStorage()
			:distinctUntilChanged()
			:subscribe( function( isOverStorage )
				catalogPusher:setUserIsOverQuota( isOverStorage )
			end )
		
		-- monitor expiration
		
		local userExpiredSubscription = user:isSubscriptionExpired()
			:distinctUntilChanged()
			:subscribe( function( isExpired )
				catalogPusher:setUserIsExpired( isExpired )
			end )
		
		ozCatalog:addCloseAction( function()
		
			quotaMonitorSubscription:dispose()
			userExpiredSubscription:dispose()
			
		end )
	
	end
	
	local purgeAgent = LrmOzPurgeAgent( ozCatalog )
	
	self:setOwned {
		_syncClientBase = syncClientBase,
		_storageMonitor = storageMonitor,
		_binaryUploader = binaryUploader,
		_purgeAgent = purgeAgent,
		_errorHandler = errorHandler,
		_userSynchronizer = userSynchronizer,
		_conflictResolver = conflictResolver,
		_catalogServiceViaWFCatalog = ozCatalog
	}
	
	-- WFCatSyncClientMessageProcessor.lua uses rawget, so we have to stash 
	-- attachMessageHandlers here
	
	self.attachMessageHandlers = function( self )

		WFRequestDispatcher.onMessageDo {
			message = 'purgeImmediately',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

		WFRequestDispatcher.onMessageDo {
			message = 'purgeForAlbum',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

		WFRequestDispatcher.onMessageDo {
			message = 'startContinuousPurge',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

		WFRequestDispatcher.onMessageDo {
			message = 'stopContinuousPurge',
			target = self._purgeAgent,
			onlyIfCallbackIsStillValid = true,
		}

	end
	
	-- INFO [pkamin 2017-01-27]: we always run the user synchronizer for 
	-- authenticated user to refresh the account document even if sync is stopped
	-- ( e.g. trial expired )
	
	-- [pkamin 2017-09-11]: We only manage the UserSynchronizer on this message
	-- processor for Android. See LrmCatUserModel:userSynchronizer for iOS
	
	if isAuthenticatedUser and not WFPlatformInfo.isiOS then
		self._userSynchronizer:startContinuousUpdates()
	end
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:close()

	check( 'LrmSyncClientImp' )
	
	if self._checkAutoPurgeSubscription then
	
		self._checkAutoPurgeSubscription:dispose()
		self._checkAutoPurgeSubscription = nil
	
	end
	
	if self._cloudTrashAutoPurge then
	
		self._cloudTrashAutoPurge:dispose()
		self._cloudTrashAutoPurge = nil
	
	end
			
end

--------------------------------------------------------------------------------

function syncClientImpMethods:getDeleteFeedCallback( )

    check( 'LrmSyncClientImp' )

    if self._purgeAgent then
        return self._purgeAgent:getDeleteFeedCallback()
    end

    return Observable.empty()

end

--------------------------------------------------------------------------------

local uploadStateLogger = WFLogger 'UploadStateLogger'
	:setLevel 'debug'

function syncClientImpMethods:startUploading( )

	check( 'LrmSyncClientImp' )

	uploadStateLogger:debugf('Starting the Uploading')

	self._binaryUploader:startContinuousUploads()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:stopUploading( )

	check( 'LrmSyncClientImp' )

	uploadStateLogger:debugf('Stopping the Uploading')

	self._binaryUploader:stopContinuousUploads()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:startSync()

	check( 'LrmSyncClientImp' )

	logger:debugf( 'starting sync' )

	self._binaryUploader:startContinuousUploads()

	self._purgeAgent:mayBeStartPurging()
	
	return self._syncClientBase:startSync()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:startSyncWithoutUploader()

	check( 'LrmSyncClientImp' )

	logger:debugf( 'starting sync without uploader' )
	
	self._purgeAgent:mayBeStartPurging()
	
	return self._syncClientBase:startSync()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:startAccountsSync()

	check( 'LrmSyncClientImp' )

	logger:debugf('starting accounts sync')

	self._userSynchronizer:startContinuousUpdates()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:stopSync()

	check( 'LrmSyncClientImp' )
	
	logger:debugf( 'stopping sync' )

	self._binaryUploader:stopContinuousUploads()

	self._purgeAgent:mayBeStopPurging()

	return self._syncClientBase:stopSync()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:stopAccountsSync()

	check( 'LrmSyncClientImp' )

	logger:debugf('stopping accounts sync')

	self._userSynchronizer:stopContinuousUpdates()

end

------------------------------------------------------------------------

function syncClientImpMethods:startSyncInResidualMode()

	check( 'LrmSyncClientImp' )
	
	self._binaryUploader:startContinuousUploadOfOnlySpaceAssets()
	
	self._purgeAgent:mayBeStartPurging()
	
	return self._syncClientBase:startSyncInResidualMode()
	
end

------------------------------------------------------------------------

function syncClientImpMethods:stopSyncInResidualMode()

	check( 'LrmSyncClientImp' )

	self._binaryUploader:stopContinuousUploads()

	self._purgeAgent:mayBeStopPurging()

	return self._syncClientBase:stopSyncInResidualMode()

end

--------------------------------------------------------------------------------
function syncClientImpMethods:startSpaceSyncWithoutOzCatalog()

	check( 'LrmSyncClientImp' )

	self._binaryUploader:startContinuousUploadOfOnlySpaceAssets()

	self._purgeAgent:mayBeStartPurging()

	return self._syncClientBase:startSpaceMetadataSyncInFreemiumMode()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:stopSpaceSyncWithoutOzCatalog()

	check( 'LrmSyncClientImp' )

	self._binaryUploader:stopContinuousUploads()

	self._purgeAgent:mayBeStopPurging()

	return self._syncClientBase:stopSpaceMetadataSyncInFreemiumMode()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:isPushingMetadata()

	check( 'LrmSyncClientImp' )
	
	return self._syncClientBase:isUploadingMetadata()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:isPullingMetadata()

	check( 'LrmSyncClientImp' )
	
	return self._syncClientBase:isDownloadingMetadata()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:getRequiredUpdateStream()

	check( 'LrmSyncClientImp' )
	
	return self._syncClientBase:getRequiredUpdateStream()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:isOzCatalogDeleted()

	check( 'LrmSyncClientImp' )
	
	return self._syncClientBase:isOzCatalogDeleted()
	
end

--------------------------------------------------------------------------------

function syncClientImpMethods:pausePusherReturningCounter()

	check( 'LrmSyncClientImp' )

	return self._syncClientBase:pausePusherReturningCounter()

end

------------------------------------------------------------------------

function syncClientImpMethods:resumePusher()

	check( 'LrmSyncClientImp' )

	return self._syncClientBase:resumePusher()

end

--------------------------------------------------------------------------------

function syncClientImpMethods:getDetailedUploadProgress()

	check( 'LrmSyncClientImp' )
	
	return self._progressReporter:status()
	
end

--------------------------------------------------------------------------------

return Rx.Class( 'LrmSyncClientImp', syncClientImpMethods )
