--[[----------------------------------------------------------------------------

OzFilePaths.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2012 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local KSPathUtils = require "KSPathUtils"
local KSStringUtils = require "KSStringUtils"
local KSTableUtils = require "KSTableUtils"

local WFArguments = require "WFArguments"
local WFFileUtils = require "WFFileUtils"
local WFLogger = require "WFLogger"
local WFPlatformInfo = require "WFPlatformInfo"

local WFArguments_check = WFArguments.check
local KSPathUtils_child = KSPathUtils.child
local KSPathUtils_getStandardFilePath = KSPathUtils.getStandardFilePath

	-- WF.oz_client.sync_engine
local WFOzSyncConstants = require 'WFOzSyncConstants'
local logger = WFLogger "OzFilePaths"

--==============================================================================

local OzFilePaths = {}


--------------------------------------------------------------------------------

--[[!
	\brief construct a file path for the standard originalFolderDocuments path, based
	on a catalog ID. This is the standard base for relative paths.
	
	\param catalogId the GUID id for the catalog
	\return (string) the file path
--]]
function OzFilePaths.getCatalogBasePath( catalogId )

	WFArguments_check( "string" )
	
	local documentsDir = KSPathUtils_getStandardFilePath 'carouselDocuments'
	
	return KSPathUtils_child( documentsDir, catalogId )
		
end

--------------------------------------------------------------------------------

--[[!
	\brief Construct an absolute path from a relative path within the Catalog Path
	
	\param catalogId (string) the GUID id for the catalog
	\param path (string) the relative path within the Catalog Path
	\return (string) the absolute file path
--]]
function OzFilePaths.getAbsolutePathInCatalog( catalogId, path )

	WFArguments_check( 'string', 'string' )
	
	local catalogPath = OzFilePaths.getCatalogBasePath( catalogId )
	
	return KSPathUtils_child( catalogPath, path )
		
end

--------------------------------------------------------------------------------

--[[!
	\brief construct a file path for the standard originalFolderDocuments path, based
	on a catalog ID
	
	\param catalogId the GUID id for the catalog
	\return (string) the file path
--]]
function OzFilePaths.getStandardOriginalFolderDocumentsPath( catalogId )

	WFArguments_check( "string" )

	if not WFPlatformInfo.isAndroid then
		return KSPathUtils.child( OzFilePaths.getCatalogBasePath( catalogId ), 'Originals' )
	end

	local documentsDir = KSPathUtils_getStandardFilePath 'carouselDocuments'

	local originalsFolderDocuments = documentsDir

	if not WFOzSyncConstants.isGlobalDocumentId( catalogId ) then

		originalsFolderDocuments = KSPathUtils_child( documentsDir, catalogId )

	end

	originalsFolderDocuments = KSPathUtils_child( originalsFolderDocuments, "Originals" )

	return originalsFolderDocuments

end

--------------------------------------------------------------------------------

--[[!
	\brief construct a file path for the standard originalFolderCaches path, based
	on a catalog ID
	
	\param catalogId the GUID id for the catalog
	\return (string) the file path
--]]
function OzFilePaths.getStandardOriginalFolderCachesPath( catalogId )

	WFArguments_check( "string" )
	
	local cachesDir = KSPathUtils.getStandardFilePath 'carouselCaches'
	
	local originalsFolderCaches = KSPathUtils.child( cachesDir, catalogId )
	originalsFolderCaches = KSPathUtils.child( originalsFolderCaches, "Originals" )
	
	return originalsFolderCaches
	
end

--[[!
	\brief construct a file path for the standard originalFolderCaches path, based
	on a catalog ID

	\param docId the document id
	\param catalogIdForCreation if set, the intermediate directory should be created
	\return (string) the file path
--]]
function OzFilePaths.getRelativeBlobPath( docId, catalogIdForPathCreation )

	WFArguments_check( "string", "?string" )

	local path = KSPathUtils.child( 'blobs', string.sub( docId, 1, 2 ) )
	path = KSPathUtils.child( path, docId )

	if catalogIdForPathCreation then
		local absPath = KSPathUtils.child( OzFilePaths.getCatalogBasePath( catalogIdForPathCreation ), path )

		if not pcall( WFFileUtils.mkdir, absPath ) then
			logger:errorf( "could not create asset directory %s for %s", absPath, docId )

			return nil
		end
	end

	return path
end

--------------------------------------------------------------------------------
if WFPlatformInfo.isAndroid then
--[[!
	\brief construct a relative file path the external xmp develop document, based
	on a catalog ID and the sha256 of the contents.

	\param catalogId the GUID id for the catalog
	\param sha256 the sha256 of the develop settings
	\param createDir set to true if the intermediate directory should be created
	\return (string) the relative file name of the external develop settings
--]]
function OzFilePaths.getRelativeXmpDevelopPath( catalogId, sha256, createDir )

	WFArguments_check( "string", "string" )

	local xmpDevelopDir = ''

	if not WFOzSyncConstants.isGlobalDocumentId( catalogId ) then
		xmpDevelopDir = catalogId
    end

    xmpDevelopDir = KSPathUtils_child( xmpDevelopDir, "xmp_develop" )

	if createDir then

		local documentsPath = KSPathUtils.getStandardFilePath( 'carouselDocuments' )

		local absPath = KSPathUtils.child( documentsPath, xmpDevelopDir )

        WFFileUtils.mkdir( absPath )

	end

    return  KSPathUtils.child( xmpDevelopDir, sha256 .. '.rdf' )

end

--[[!
	\brief convenience function to construct a absolute file path for the external xmp develop document
	based on the docValue.

	\param docValue the docValue
	\return (string) the absolute file path to the external develop settings
--]]

function OzFilePaths.getAbsoluteXmpDevelopPath( docValue )

	WFArguments_check( "table" )

	local documentsPath = KSPathUtils.getStandardFilePath( 'carouselDocuments' )
	local xmpDevelopFilePath = KSTableUtils.derefOpt( docValue, '_localOnly', 'files', 'xmp_develop', 'path' )

	if documentsPath and xmpDevelopFilePath then
		return KSPathUtils.child( documentsPath, xmpDevelopFilePath )
	end

	return nil

end

end

--------------------------------------------------------------------------------

--[[!
	\brief construct a file path for an original based on a originalsFolderDocuments
	path and a capture date in format like this: 2012-10-08T14:15:44
	
	\param originalsFolderDocuments Absolute path to the originalFolderDocuments folder
	\param captureDate Date of the original asset that you're trying to find a home for
]]--

function OzFilePaths.originalsFolderForDate( originalsFolderDocuments, captureDate )

	WFArguments_check( "string", "string" )

	captureDate = captureDate:sub( 1, 10 )

	local folder = KSPathUtils_child( originalsFolderDocuments, captureDate:sub( 1, 4 ) ) -- yyyy
	folder = KSPathUtils_child( folder, captureDate:sub( 1, 7 ) ) -- yyyy-mm
	folder = KSPathUtils_child( folder, captureDate ) -- yyyy-mm-dd
	
	return folder

end

--------------------------------------------------------------------------------

--[[!
	\brief construct a relative file from an absolute path and an anchor point
	
	\param absPath Absolute path 
	\param anchor base directory used as an anchor point 9for example 'Originals'.
]]--

function OzFilePaths.makeRelativeFrom( absPath, anchor )

	local components = KSStringUtils.split( absPath, WFPlatformInfo.folderSeparator )

	for i = #components, 1, -1 do

		if components[ i ] == anchor then

			components = KSTableUtils.slice( components, i )

			return table.concat( components, WFPlatformInfo.folderSeparator )

		end

	end

	logger:errorf( "could not find anchor %s in absolute path %s", anchor, absPath )

	return nil

end

--==============================================================================

return OzFilePaths
