--[[----------------------------------------------------------------------------

WFAcrOperations.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.rxlua
local WFRxNativeObservable = require 'WFRxNativeObservable'

	-- WF.acr
local WFAcrOperationsImp = require 'WFAcrOperationsImp'

--------------------------------------------------------------------------------

assert( WFPlatformInfo.hasACR, "ACR integration not enabled" )

--------------------------------------------------------------------------------

local logger = require "WFLogger"( "WFAcrOperations" )

--============================================================================--

-- @mark WFAcrOperations class

--[[!
	\brief WFCrContext provides Lua access to ACR api,
	 which doesn't require cr_context
]]--

local WFAcrOperations = {}

--------------------------------------------------------------------------------

local function logAndWrapInErrorObservable( errStr, ... )

	local errorString = select( '#', ... ) == 0 and 
							errStr or string.format( errStr, ... )

	logger:error( errorString )

	return Rx.Observable.throw( errorString )

end


--------------------------------------------------------------------------------
--[[!
	\brief Get fingerprint, id, name, and camera model from a custom camera
	profile.

	\param path (string) path of the profile
	
	\return table containing the following info
		fingerprint (string) the camera profile's fingerprint
		uniqueId (string) the camera profile's id
		name (string) the camera profile's name
		model (string) the camera profile's camera model
]]--

local WFAcrOperationsImp_getCustomCameraProfileInfo = WFAcrOperationsImp.getCustomCameraProfileInfo

--------------------------------------------------------------------------------

function WFAcrOperations.getCustomCameraProfileInfo( profilePath )

	-- TODO: The existence check should happen on the worker thread. Not critical
	-- in this case, since we know the files in question will always reside on
	-- a the boot drive; always online.
	
	if type( profilePath ) ~= "string" or profilePath == "" or
		not WFFileUtils.exists( profilePath ) then

		return logAndWrapInErrorObservable( 
					"Invalid path passed to getCustomCameraProfileInfo()" )

	end

	return WFRxNativeObservable( 
				WFAcrOperationsImp_getCustomCameraProfileInfo, profilePath )

end

--------------------------------------------------------------------------------
--[[!
	\brief Get style favorite info.

	\return table containing the following info
		presetFavorites (table) A table containing fingerprint and boolean value
			for xml preset favorites
		profileFavorites (table) A table containing fingerprint and boolean value
			for xml profile favorites
]]--

local WFAcrOperationsImp_getStyleFavorites = WFAcrOperationsImp.getStyleFavorites

--------------------------------------------------------------------------------

function WFAcrOperations.getStyleFavorites()

	return WFRxNativeObservable( 
				WFAcrOperationsImp_getStyleFavorites )

end

--------------------------------------------------------------------------------

local WFAcrOperationsImp_refreshACRDatabases = WFAcrOperationsImp.refreshACRDatabases

--------------------------------------------------------------------------------

function WFAcrOperations.refreshACRDatabases()

	return WFRxNativeObservable( 
				WFAcrOperationsImp_refreshACRDatabases )

end

--------------------------------------------------------------------------------
--[[!
	\brief Retrurns the path info for all the standard profiles.

	\return (table) Map of modelname vs profile info list.
		key - (string) model name
		value - (table) array of tables containing the following info
			fileName (string) the file name of the profile file
			relativePath (string) the relative path of the profile file
]]--

local WFAcrOperationsImp_getCoreCameraProfilesMap = WFAcrOperationsImp.getCoreCameraProfilesMap

--------------------------------------------------------------------------------

function WFAcrOperations.getCoreCameraProfilesMap()

	return WFRxNativeObservable( 
				WFAcrOperationsImp_getCoreCameraProfilesMap )

end

--------------------------------------------------------------------------------

return WFAcrOperations
