--[[----------------------------------------------------------------------------

WFCatAssetInfoListUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2019 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Adobe and its suppliers, if any. The intellectual
and technical concepts contained herein are proprietary to Adobe
and its suppliers and are protected by all applicable intellectual
property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Adobe.

----------------------------------------------------------------------------]]--

local WFConfiguration = require 'WFConfiguration'

--==============================================================================

local WFCatAssetInfoListUtils = {}

--------------------------------------------------------------------------------

local SHOW_GRID_LABELS = WFConfiguration.Shared.showGridLabels
local UNPACKED_ASSET_INFO = WFConfiguration.Shared.transportAssetInfoUnpacked
local STRINGIFY_ASSET_INFO = WFConfiguration.Shared.transportAssetInfoAsStrings
local RAW_QUERYRESULT_ASSET_INFO = WFConfiguration.Shared.transportAssetInfoAsRawQueryResult

if not UNPACKED_ASSET_INFO and not STRINGIFY_ASSET_INFO then

	-- For now default to transporting as a raw query result. It has more risk
	-- on whether the client accidentally uses the asset info array element
	-- without calling the unpacking routine but has less risk in the decoding
	-- of the string.  Both the stringified query and transporting the raw query
	-- results have about the same performance, as the time spent encoding
	-- the extra array elements is offset by the time spent reg-ex-ing the
	-- and tonumber-ing the string values.  This was measured wholistically by
	-- timing switching from an album to a large (200k) all photos view.

	RAW_QUERYRESULT_ASSET_INFO = true

elseif RAW_QUERYRESULT_ASSET_INFO then

	STRINGIFY_ASSET_INFO = false
	UNPACKED_ASSET_INFO = false

elseif STRINGIFY_ASSET_INFO then

	RAW_QUERYRESULT_ASSET_INFO = false
	UNPACKED_ASSET_INFO = false

else

	UNPACKED_ASSET_INFO = true

end

local unpackedAssetRequiredFields = "docId, width, height, subtype, orientation, captureDate, duration, fileName"

--==============================================================================

if UNPACKED_ASSET_INFO then

	-- Unpacked asset info as originally transited from daemon in versions 2.3 and earlier

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.getRequiredFields()

		return unpackedAssetRequiredFields

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.getPackedAssetInfoDetails( currentAssetInfoFromQuery, additionalDataForAsset )

		-- LrDesktop originally transited the unpacked data so even though this says
		-- packed all the data is sent unpacked.

		local width = currentAssetInfoFromQuery[ 2 ]
		local height = currentAssetInfoFromQuery[ 3 ]

		return {
			id = currentAssetInfoFromQuery[ 1 ],
			width = width,
			height = height,
			subtype = currentAssetInfoFromQuery[ 4 ],
			orientation = currentAssetInfoFromQuery[ 5 ],
			captureDate = currentAssetInfoFromQuery[ 6 ],
			duration = currentAssetInfoFromQuery[ 7 ],
			fileName = currentAssetInfoFromQuery[ 8 ],
			aspectRatio = width / height,
			additionalAssetInfo = additionalDataForAsset,
		}

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.getPackedAssetInfoDetailsForArray( queryResult )

		local assetInfos = {}

		for i, row in ipairs( queryResult ) do

			local width = row[ 2 ]
			local height = row[ 3 ]

			assetInfos[ i ] = {
				id = row[ 1 ],
				width = width,
				height = height,
				subtype = row[ 4 ],
				orientation = row[ 5 ],
				captureDate = row[ 6 ],
				duration = row[ 7 ],
				fileName = row[ 8 ],
				aspectRatio = width / height,
			}

		end

		return assetInfos

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.processPackedAssetInfo( assetInfoOrString )

		-- table is transmitted unpacked just return the values

		return assetInfoOrString.id, assetInfoOrString.subtype, assetInfoOrString

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.unpackAssetIdAndSubtype( assetInfo )

		return assetInfo.id, assetInfo.subtype

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.unpackAssetInfoDetailsToTable( assetInfo )

		return assetInfo

	end

end

--==============================================================================

if STRINGIFY_ASSET_INFO then

	-- An array of strings are transited from daemon to client.

	-- Order of stringified asset info:
	-- id, subtype, width, height, orientation, captureDate, duration, fileName, additionalAssetInfo

	-- Note duration, fileName, and additionalAssetInfo can be nil.

	-- Note if fileName has the ; delimiter in it it won't matter as it is last field
	-- the other fields don't have any ';'.

	local kPackedAssetIdAndSubtypePattern = '([^;]+);([^;]+)'

	local kPackedInfoPattern = SHOW_GRID_LABELS
		and '([^;]+);([^;]+);([^;]+);([^;]+);([^;]+);([^;]+);([^;]*);([^;]*);(.*)'
		or  '([^;]+);([^;]+);([^;]+);([^;]+);([^;]+);([^;]+);([^;]*);(.*)'

	-- Note that search expects docId as first entry in query result rows.
	local packedAssetRequiredFields = "docId, docId || ';' || subtype || ';' || width || ';' || height || ';' || orientation || ';' || captureDate || ';' || ifnull(duration,'') || ';' || ifnull(fileName,'')"

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.getRequiredFields()

		return packedAssetRequiredFields

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.processPackedAssetInfo( assetInfo )

		-- It is too risky at the moment to leave the assetInfo in packed form
		-- for the grid to use as other things like share and export
		-- refer to assetInfos. So unpack everything during processPackedAssetInfo
		-- to reduce risk.

		-- If the assetInfo is in packed form this will expand the stringified
		-- version into a table, otherwise it will return the table as is.

		local packedAssetInfo = type( assetInfo ) == 'string' and assetInfo or assetInfo.packedAssetInfo

		if packedAssetInfo then

			-- Note additionalAssetInfo will only be present if SHOW_GRID_LABELS,
			-- config flag is on.

			local id, subtype, width, height, orientation, captureDate, duration, fileName, additionalAssetInfo
				= string.match( packedAssetInfo, kPackedInfoPattern )

			assetInfo = {
				id = tonumber( id ),
				subtype = subtype,
				width = tonumber( width ),
				height = tonumber( height ),
				orientation = tonumber( orientation ),
				captureDate = captureDate,
				duration = tonumber( duration ),
				fileName = fileName,
				aspectRatio = width / height,
				additionalAssetInfo = additionalAssetInfo,
			}

			-- Use assetInfo.id as it is converted to a number.
			return assetInfo.id, subtype, assetInfo

		end

		return assetInfo.id, assetInfo.subtype, assetInfo

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.getPackedAssetInfoDetails( currentAssetInfoFromQuery, additionalDataForAsset )

		-- WFCatAsset.getAssetInfoDetails converts a row from asset query into the
		-- assetInfo entry as expected by the client app.

		-- Currently this converts an table with docId, and stringified values to a
		-- single string containing the stringified values.

		return currentAssetInfoFromQuery[ 2 ]

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.getPackedAssetInfoDetailsForArray( queryResult )

		-- This function duplicates the implementation of getPackedAssetInfoDetails
		-- to eliminate the need for the function call.

		local assetInfos = {}

		for i, row in ipairs( queryResult ) do
			assetInfos[ i ] = row[ 2 ]
		end

		return assetInfos

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.unpackAssetIdAndSubtype( assetInfo )

		local packedAssetInfo = type( assetInfo ) == 'string' and assetInfo or assetInfo.packedAssetInfo

		if packedAssetInfo then
			local assetId, subtype = string.match( packedAssetInfo, kPackedAssetIdAndSubtypePattern )
			return tonumber( assetId ), subtype
		end

		return assetInfo.id, assetInfo.subtype

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.unpackAssetInfoDetailsToTable( assetInfo )

		-- For now table was already unpacked in processPackedAssetInfo.
		return assetInfo

	end

end

--------------------------------------------------------------------------------

if STRINGIFY_ASSET_INFO and SHOW_GRID_LABELS then

	WFCatAssetInfoListUtils.getPackedAssetInfoDetails = function( currentAssetInfoFromQuery, additionalDataForAsset )

		-- WFCatAsset.getAssetInfoDetails converts a row from asset query into the
		-- assetInfo entry as expected by the client app.

		-- Currently this converts an table with docId, and stringified values to a
		-- single string containing the stringified values.

		return currentAssetInfoFromQuery[ 2 ] .. ';' .. ( additionalDataForAsset and tostring( additionalDataForAsset ) or "" )

	end

	----------------------------------------------------------------------------

	WFCatAssetInfoListUtils.getPackedAssetInfoDetailsForArray = function( queryResult )

		-- This function duplicates the implementation of getPackedAssetInfoDetails
		-- to eliminate the need for the function call.

		-- Search passes relevance info for additionalDataForAsset however this
		-- function is called when there was none involved so add another ; if
		-- SHOW_GRID_LABELS is on

		local assetInfos = {}

		for i, row in ipairs( queryResult ) do
			assetInfos[ i ] = row[ 2 ] .. ';'
		end

		return assetInfos

	end

end

--==============================================================================

if RAW_QUERYRESULT_ASSET_INFO then

	-- The raw query result is transited in this case (array of arrays)

	function WFCatAssetInfoListUtils.getRequiredFields()

		return unpackedAssetRequiredFields

	end

	----------------------------------------------------------------------------

	WFCatAssetInfoListUtils.getPackedAssetInfoDetails = function( currentAssetInfoFromQuery, additionalDataForAsset )

		currentAssetInfoFromQuery[ 9 ] = additionalDataForAsset
		return currentAssetInfoFromQuery

	end

	----------------------------------------------------------------------------

	WFCatAssetInfoListUtils.getPackedAssetInfoDetailsForArray = function( queryResult )

		return queryResult

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.processPackedAssetInfo( packedAssetInfo )

		-- It is too risky at the moment to leave the assetInfo in packed form
		-- for the grid to use as other things like share and export
		-- refer to assetInfos. So unpack everything during processPackedAssetInfo
		-- to reduce risk.

		if packedAssetInfo[ 1 ] then

			local assetInfo = {
				id = packedAssetInfo[ 1 ],
				width = packedAssetInfo[ 2 ],
				height = packedAssetInfo[ 3 ],
				subtype = packedAssetInfo[ 4 ],
				orientation = packedAssetInfo[ 5 ],
				captureDate = packedAssetInfo[ 6 ],
				duration = packedAssetInfo[ 7 ],
				fileName = packedAssetInfo[ 8 ],
				aspectRatio = packedAssetInfo[ 2 ] / packedAssetInfo[ 3 ],
				additionalAssetInfo = packedAssetInfo[ 9 ],
			}

			return assetInfo.id, assetInfo.subtype, assetInfo

		end

		return packedAssetInfo.id, packedAssetInfo.subtype, packedAssetInfo

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.unpackAssetIdAndSubtype( packedAssetInfo )

		if packedAssetInfo[ 1 ] then
			return packedAssetInfo[ 1 ], packedAssetInfo[ 4 ]
		end

		return packedAssetInfo.id, packedAssetInfo.subtype

	end

	----------------------------------------------------------------------------

	function WFCatAssetInfoListUtils.unpackAssetInfoDetailsToTable( packedAssetInfo )

		-- For now table was already unpacked in processPackedAssetInfo.
		return packedAssetInfo

	end

end

--==============================================================================

return WFCatAssetInfoListUtils
