--[[----------------------------------------------------------------------------

WFCatBinaryUploader.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable
local SingleAssignmentDisposable = Rx.SingleAssignmentDisposable

	-- WF.oz_client.managed_catalog
local WFCatBlacklistHelper = require 'WFCatBlacklistHelper'

local WFCatContentRangeUploader = require 'WFCatContentRangeUploader'
local WFCatImportDuplicateFiles = require 'WFCatImportDuplicateFiles'
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatFileSpec = require 'WFCatFileSpec'
local WFCatPhotoImportance = require 'WFCatPhotoImportance'
local WFCatOzErrors = require 'WFCatOzErrors'
local WFCatTrashUtils = require 'WFCatTrashUtils'
local WFOzContentTypes = require 'WFOzContentTypes'
local WFCatAssetUtils = require 'WFCatAssetUtils'

	-- WF.core.diagnostic_log
local WFDiagnosticLog = require 'WFDiagnosticLog'

	-- WF.core.logger
local WFLogger = require 'WFLogger'

	-- WF.core.os
local KSDate = require 'KSDate'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.performance
local WFPerformanceReport = require 'WFPerformanceReport'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSJsonDecoder = require 'KSJsonDecoder'
local KSTableUtils = require 'KSTableUtils'
local WFAnalytics = require 'WFAnalytics'
local WFArguments = require 'WFArguments'
local WFCrypto = require 'WFCrypto'
local WFCloseSupport = require 'WFCloseSupport'
local WFRxContinuousTask = require 'WFRxContinuousTask'
local WFRxLowFidelityTimer = require 'WFRxLowFidelityTimer'
local WFPlatformInfo = require 'WFPlatformInfo'
local WFDeviceInfo = require 'WFDeviceInfo'
local WFMsgPack = require 'WFMsgPack'

	-- WF.web.http_client
local WFRxHttpClient = require 'WFRxHttpClient'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local derefOpt = assert( KSTableUtils.derefOpt )
local now = assert( KSDate.timeNowForAnimation )

local doesAssetDocumentHaveLocalBinary = assert( WFCatAssetUtils.doesAssetDocumentHaveLocalBinary )
local isValidAssetDocument = assert( WFCatIndexUtils.isValidAssetDocument )
local isNonDeletedStyleDocument = assert( WFCatIndexUtils.isNonDeletedStyleDocument )
local WFCatIndexUtils_isRevisionOnOz = assert( WFCatIndexUtils.isRevisionOnOz )
local WFCatTrashUtils_getMasterIsUploaded = assert( WFCatTrashUtils.getMasterIsUploaded )
local WFCatTrashUtils_getProxyIsUploaded = assert( WFCatTrashUtils.getProxyIsUploaded )
local WFCatTrashUtils_getXmpDevelopIsUploaded = assert( WFCatTrashUtils.getXmpDevelopIsUploaded )
local runUntilDone = WFRxRunUntilDone

local logger = WFLogger 'WFCatBinaryUploader'

--------------------------------	------------------------------------------------

local maxScheduledUploads = 20
	-- SPAM [jandrew 2016-06-09]: The maximum number of uploads to schedule at once. 
	-- The WFRxRateLimiter used in the HTTP implementation limits the number of concurrent
	-- HTTP requests to 5. However, it's good to queue up more than the max number of 
	-- concurrent requests so that we can quickly process the next batch without needing
	-- to wait for the next batch to be scheduled.
	
local uploadBlackoutWindowInSeconds = 60
	-- Number of seconds we disable the re-uploading of a binary after a successful upload.

local numberOfTriesBeforeBlacklistingUpload = 3
	-- Number of attempts to upload an original before it is blacklisted.

local failedUploadBlackoutWindowInSeconds = 300
	-- Number of seconds an original remains blacklisted after it has failed to upload n times.

local kClearFromRecentlyUploadedSeconds = 30
local kWaitForOtherUploadSeconds = 10
local kWaitForCopyFailedSeconds = 10

local kContentTypeXML = 'application/rdf+xml'
local kContentTypeDng = 'image/dng'
local kContentTypeOctetStream = 'application/octet-stream'

local uploadHttpRequestCost = WFRxHttpClient.requestCostUploadNegative

--------------------------------------------------------------------------------

local function intToBool( zeroOrOne )
	return zeroOrOne ~= 0
end

--------------------------------------------------------------------------------

local function boolToInt( isTrue )
	return isTrue and 1 or 0
end

--------------------------------------------------------------------------------

local function checkIfBlacklisted( docValue, binaryType )

	return derefOpt( docValue, '_localOnly', 'blacklistedBy', 'WFCatBinaryUploader' ) and 1 or 0
	
end

--------------------------------------------------------------------------------

local function getDatesFromDocValue( docValue )

	local importDate = derefOpt( docValue, 'importSource', 'importTimestamp' )
	
	local captureDate = derefOpt( docValue, 'captureDate' )
	
	return importDate or 1, captureDate or 1

end

--------------------------------------------------------------------------------

local function getOriginalLinkFromDocValue_IgnoringShaError( docValue )

	-- The binary uploader uses the rels/master link only to indicate whether
	-- an asset has completed upload, so return the field's contents even if
	-- there was an error in the import source status field.

	return derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )

end

--------------------------------------------------------------------------------

local function getOriginalExistsFromDocValue( docValue )
	
	return derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'exists' )

end

--------------------------------------------------------------------------------

local function getProxyExistsFromDocValue( docValue )
	
	return derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/proxy_type/2560', 'exists' ) or
		derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/proxy_type/panorama', 'exists' ) or
		derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/proxy_type/720p', 'exists' )

end

--------------------------------------------------------------------------------

local function isOriginalLinkInvalid( docValue )

	return derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'invalid' )

end

--------------------------------------------------------------------------------

local function isAssetDocumentACopy( docValue )

	return isValidAssetDocument( docValue ) and derefOpt( docValue, 'copy', 'copy_id' )

end

--------------------------------------------------------------------------------

local function isSpaceAssetWithSourceLinks( docValue )
	
	local hasSourceOriginalLink = derefOpt( docValue, '_localOnly', 'sourceOriginalLink' )
	local hasSourceProxyLink = derefOpt( docValue, '_localOnly', 'sourceProxyLink' )
	
	return isValidAssetDocument( docValue ) and docValue._spaceId 
		and ( hasSourceOriginalLink or hasSourceProxyLink )

end

--------------------------------------------------------------------------------

local function getEncodedUploadHeadersForBinaryType( binaryType, docValue )
	
	if docValue._spaceId and 
		( binaryType == 'master' or 
		binaryType == 'proxy' or
		binaryType == 'proxy_pano' or
		binaryType == 'xmp_sidecar' ) then
		-- for masters, proxies, and xmp sidecars in spaces we may need
		-- to send headers to instruct Oz to redact some of the associated 
		-- metadata to protect the contributing users privacy
		local addRedaction = function( redactionsList, additionalRedaction ) 
			if #redactionsList > 0 then	
				return redactionsList .. "," .. additionalRedaction
			else
				return additionalRedaction
			end
		end
		
		local redactions = derefOpt( docValue, '_localOnly', 'universal', 'sharedwithyou', 'contribution' ) or {}
		local redactionsList = ""

		if redactions.redactKeywords then
			redactionsList = addRedaction( redactionsList, 'keywords' )
		end
		
		if redactions.redactLocation then
			redactionsList = addRedaction( redactionsList, 'location' )
		end
		
		if redactions.redactRatingsReviews then
			redactionsList = addRedaction( redactionsList, 'ratings' )
			redactionsList = addRedaction( redactionsList, 'reviews' )
		end
		
		return #redactionsList > 0 and WFMsgPack.pack( { ['X-Redact-Meta'] = redactionsList } ) or nil
	
	else
		-- currently, for non-space assets, there are no custom headers we 
		-- may need to send on binary upload
		return nil
		
	end
	
end

-----------------------------------------------------------------------------------

local function maybeEmitMasterUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )
	
	local localPath = WFCatAssetUtils.getFullOriginalPathFromDocValue( docValue )

	local fileMaybeOnOfflineVolume = ( WFPlatformInfo.isDesktop and not localPath ) or false

	-- Applicable only for desktop
	local localVolumeGuid = WFCatAssetUtils.getOriginalVolumeGuidFromDocValue( docValue )

	-- Applicable only for desktop
	local localRelativePath

	if localVolumeGuid then
		localRelativePath = WFCatAssetUtils.getOriginalRelativePathFromDocValue( docValue )
	end

	local localOriginalSha256 = WFCatAssetUtils.getOriginalContentIdFromDocValue( docValue )

	local ozMasterUrl = getOriginalLinkFromDocValue_IgnoringShaError( docValue )

	local ozMasterCreateTemplate = derefOpt( docValue,
											 '_serverProvidedMetadata',
											 'links',
											 '/rels/master_create',
											 'href' )

	local revisionIds = derefOpt( docValue,
								  '_serverProvidedMetadata',
								  'revision_ids' )

	local latestRevisionId = revisionIds and revisionIds[ #revisionIds ]
	
	local isOzOriginalInvalid = ozMasterUrl and isOriginalLinkInvalid( docValue )
	
	local imageSha256 = derefOpt( docValue, 'importSource', 'sha256' )

	local isLocalOriginalValid = localOriginalSha256 == imageSha256
	
	-- if our local original is not a valid original for the asset record, it's not even a candidate for
	-- uploading to Oz
	if not isLocalOriginalValid then
		-- NMB-10275: no local sha256 which means that we should *only* upload the binary if there's no ozMasterUrl which 
		-- would happen if we are the only client with an original for the asset. So if either of these two exist, then the local 
		-- original with no sha256 on this machine is not a valid replacement for what's on Oz. If we could guarantee that
		-- all local original annotations have a sha256, then we could skip this check and only consider the 'isLocalOriginalValid' 
		-- value.
		if localOriginalSha256 or ozMasterUrl then			
			return
		end
	end
	
	-- NMB-5779: If we have a master file then we should upload if there is no master/proxy at all. But we 
	-- may have to wait until Oz marks it invalid so we don't upload it too soon. Import(migration) should 
	-- flag this case in the localOnly section.
	
	local isInvalid
	local waitForInvalid = derefOpt( docValue, '_localOnly', 'binaryUpload', 'waitForInvalid' )
	if waitForInvalid then
		-- does master url exist and is invalid? if it doesn't exist, then we shouldn't consider it invalid
		isInvalid = isOzOriginalInvalid
		
		-- has no master, but proxy url exist and is invalid?
		local ozProxyUrl, _, isProxyInvalid = WFCatFileSpec.getProxyLinkFromDocValue( docValue )
		isInvalid = isInvalid or ( ozProxyUrl and isProxyInvalid )
		
		-- is there no master or proxy at all? this is an edge case we'll consider invalid even if it's 
		-- technically not
		isInvalid = isInvalid or ( not ozMasterUrl and not ozProxyUrl )
		
		if not isInvalid then return end
	end
	
	-- SPAM [matthewj 2017-08-29]: re: the above 'invalid' handling - technically, if the image is marked
	-- as invalid, but we have the appropriate binary based on our local sha256 matching the server sha256 
	-- then we *should* be able to upload it
	
	local ozSha256Error = derefOpt( docValue,
								  'importSource',
								  'status',
								  'error',
								  'all',
								  'reason',
								  'sha256' ) == "does not match"
								  
	-- If there's a sha256 error that we can fix, there should be a binaryUploadError for the current ozMasterUrl
	-- If so, we can actually clear the sha256error and try to re-upload locally.
	local localSha256Ok = ozMasterUrl and derefOpt( docValue, '_localOnly', 'binaryUploadErrors', '/rels/master', ozMasterUrl )

	local blockedBySha256Error = not not ( ozSha256Error and ( not localSha256Ok ) )

	local localOriginalHasError = derefOpt( docValue, '_localOnly', 'localOriginalError' )
	
	local trialOrSubscriptionExpiredError = not not WFCatOzErrors.hasTrialOrSubscriptionExpiredError( docValue )
	
	local insufficientPermissionError = not not WFCatOzErrors.hasInsufficientPermissionError( docValue )

	local hasStorageExceededError = not not WFCatOzErrors.hasStorageExceededError( docValue )

	local copyId = derefOpt( docValue, 'copy', 'copy_id' )

	local willNotUploadMaster = KSTableUtils.derefOpt( docValue,
													   'importSource',
													   'willNotUploadMaster' )
	
	local isSpaceAsset = not not docValue._spaceId
	-- A space asset may not have a master href visible so we need to instead look at a field 'exists'
	-- under /rels/master which will be set to true once the master is uploaded even if we don't have
	-- access to redownload it
	local isSpaceAssetWithImpliedMaster = isSpaceAsset and derefOpt( docValue, '_serverProvidedMetadata',
											 'links',
											 '/rels/master',
											 'exists' ) == true
	
	local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceOriginalLink' )
	
	local sourceAssetId = derefOpt( docValue, '_localOnly', 'sourceAssetId' ) or docValue._id

	local binaryType = 'master'
	if docValue.subtype == "preset" or docValue.subtype == "profile" then
		binaryType = 'preset_master'
	elseif docValue.subtype == "camera_profile" then
		binaryType = 'camera_profile'
	end
	
	local ozHasMaster = ozMasterUrl or isSpaceAssetWithImpliedMaster
	
	local shouldUpload = false

	if WFCatTrashUtils_getMasterIsUploaded( docValue ) then

		-- If this asset was copied via a delete or restore it might not have
		-- it's master link set yet, but oz may provide a master_create link
		-- in an update because it is still copying the asset asynchronously.
		-- A subsequent update will provide the link but prevent uploading
		-- the original by referring to the original asset's link.
		shouldUpload = false

	elseif not ozHasMaster or blockedBySha256Error then
	
		-- there's no master url on Oz, or there is but it has a sha256 error (which we need to see if we can clear)
		
		shouldUpload = true
	
	elseif ozHasMaster and ozSha256Error and not blockedBySha256Error then
	
		-- there's a master url and we already know we can clear it.
	
		shouldUpload = true
	
	elseif ozHasMaster and isOzOriginalInvalid and isLocalOriginalValid then
	
		-- the master on Oz is invalid, and our local original is valid.
		
		shouldUpload = true
	
	end
	
	-- For Android client, after importing assets and before all of them get fully uploaded to OZ for some reason
	-- if the local directory gets deleted, then "localpath" of those assets will be null.
	-- in those scenarios, it has the facility to support upload from camera gallery.
	-- so perform additional checks to see whether emit shall be allowed.
	-- the asset must belongs to the same device and localUrl or localAssetId field must prsent in _localOnly
	-- localUrl or localAssetId refers to local path of the asset in the camera gallery location.

	local validAndroidPath
	if not localPath and WFPlatformInfo.isAndroid then

		local uniqueDeviceId = derefOpt( docValue, 'importSource', 'uniqueDeviceId' )
		
		if( uniqueDeviceId == WFDeviceInfo.uniqueDeviceId() ) then
		
			local localUrl = derefOpt( docValue, '_localOnly', 'localUrl' )
			local localAssetId = derefOpt( docValue, 'importSource', 'localAssetId' )
			
			if( localUrl or localAssetId ) then
				validAndroidPath = true
			end
			
		end
		
	end
	
	-- It is observed that the masterCreateUrl variable takes a while to get update from Oz.
	-- So masterCreateUrl and latestRevisionId checks here causes not to emit into index table( needsBinaryUpload ) until masterCreateUrl updated by Oz.
	-- This causes delay of 18 to 20sec in showing upload progress over cloud pannel for the mobile apps as it relies upon th pending binary upload counts
	-- Hence removing those checks so that emit shall be allowed without waiting for masterCreateUrl. Applying the same to desktop doesn't hurt.
	-- So keeping it common for all platforms.
	if ( ( localPath or validAndroidPath or ( localVolumeGuid and localRelativePath ) ) or ( copyId and not willNotUploadMaster )
		or ( isSpaceAsset and sourceLinkForCopy ) )
		and shouldUpload
		and not localOriginalHasError
		and not trialOrSubscriptionExpiredError
		and not insufficientPermissionError
		and not hasStorageExceededError
	then

		local contentType = WFCatFileSpec.getContentTypeFromDocValue( docValue )

		local fileSize = WFCatFileSpec.getOriginalSizeFromDocValue( docValue )

		local masterCreateUrl

		if ozMasterCreateTemplate and latestRevisionId then
			masterCreateUrl = ozMasterCreateTemplate:replace( '{revision_id}', latestRevisionId )
		end
		
		local uploadHeaders = getEncodedUploadHeadersForBinaryType( binaryType, docValue )
		
		local spaceId = docValue._spaceId
		
		logger:debugf( "emit %s, %s", localPath, masterCreateUrl )
		
		emit( localPath, localVolumeGuid, localRelativePath, boolToInt( fileMaybeOnOfflineVolume ), 
				masterCreateUrl, binaryType, contentType, uploadHeaders, importDate,
				captureDate, isBlacklisted, imageSha256, fileSize, boolToInt( blockedBySha256Error ), 
				priorityForUpload, sourceLinkForCopy, spaceId, sourceAssetId )

	end
		
end

--------------------------------------------------------------------------------

local function maybeEmitXmpDevelopUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )
	
	local xmpCameraRaw = derefOpt( docValue, 'develop', 'xmpCameraRaw' )
	
	if not ( type( xmpCameraRaw ) == 'table' ) then
		-- if the xmpCameraRaw is not external then there's also nothing to upload
		return
	end
	
	local isXmpDevelopInvalid = not not derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'invalid' )
	
	local xmpDevelopUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'href' )
	
	-- Depending on the space settings, the develop settings may be redacted so Oz may not include the href 
	-- in /rels/xmp/develop for us to know that it has been uploaded. Instead we'll look for exists: true to
	-- infer that the xmp_develop has been uploaded even if we can't see the link
	local isAssetInSpace = not not docValue._spaceId
	local impliedXmpDevelopUrl = isAssetInSpace and 
								derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'exists' ) and
								not derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'invalid' )
								
	if not ( isXmpDevelopInvalid or ( not xmpDevelopUrl and not impliedXmpDevelopUrl ) ) then
		-- it's not missing or invalid, there's nothing to upload
		return
	end
	
	local xmpCameraRawSha256 = xmpCameraRaw.sha256

	local localXmpDevelopInfo = WFCatFileSpec.getXmpDevelopInfoFromDocValue( docValue )

	local localSha256, localPath, localFileSize
	if type( localXmpDevelopInfo ) == 'table' then
		localSha256 = localXmpDevelopInfo.sha256 or localXmpDevelopInfo.contentId -- iOS stores under contentId
		localPath = localXmpDevelopInfo.path
		localFileSize = localXmpDevelopInfo.fileSize
	end
	
	-- construction of full path needed only for Android client and not for unit test, hence exclude for unit test
	if WFPlatformInfo.isAndroid then
		local KSPathUtils = require 'KSPathUtils'
		localPath = localPath and KSPathUtils.child( KSPathUtils.getStandardFilePath( 'carouselDocuments' ), localPath )
	end
	
	local ozXmpDevelopCreateTemplate = derefOpt( docValue,
											 '_serverProvidedMetadata',
											 'links',
											 '/rels/xmp_develop_create',
											 'href' )
	
	-- If this asset was copied via a delete or restore it might not have
	-- it's xmp/develop link set yet, but oz may provide a xmp_develop_create link
	-- in an update because it is still copying the asset asynchronously.
	-- A subsequent update will provide the link.
	local wasUploadedPreviously = WFCatTrashUtils_getXmpDevelopIsUploaded( docValue ) and not isXmpDevelopInvalid
		
	local revisionIds = derefOpt( docValue, '_serverProvidedMetadata', 'revision_ids' )

	local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceXmpDevelopLink' )

	local sourceAssetId = derefOpt( docValue, '_localOnly', 'sourceAssetId' ) or docValue._id

	local insufficientPermissionError = not not WFCatOzErrors.hasInsufficientPermissionError( docValue )

	local hasStorageExceededError = not not WFCatOzErrors.hasStorageExceededError( docValue )

	if ( ( localPath and localSha256 and localSha256 == xmpCameraRawSha256 ) or sourceLinkForCopy ) and
		xmpCameraRawSha256 and
		WFCatIndexUtils_isRevisionOnOz( docValue._revId, revisionIds  )
		and not wasUploadedPreviously
		and not insufficientPermissionError
		and not hasStorageExceededError
	then
		
		local latestRevisionId = revisionIds and revisionIds[ #revisionIds ]

		local xmpDevelopCreateUrl
		if ozXmpDevelopCreateTemplate and latestRevisionId then
			xmpDevelopCreateUrl = ozXmpDevelopCreateTemplate:replace( '{revision_id}', latestRevisionId )
		end

		local binaryType = 'xmp_develop'
		local uploadHeaders = getEncodedUploadHeadersForBinaryType( binaryType, docValue )
		
		local spaceId = docValue._spaceId
		
		logger:debugf( "emit %s, %s", localPath, xmpDevelopCreateUrl )
		
		emit( localPath, nil, nil, boolToInt( false ), xmpDevelopCreateUrl, binaryType, 
				kContentTypeXML, uploadHeaders, importDate, captureDate, isBlacklisted,
				localSha256, localFileSize, boolToInt( false ), priorityForUpload, 
				sourceLinkForCopy, spaceId, sourceAssetId )

	end

end

--------------------------------------------------------------------------------

local function maybeEmitXmpSidecarUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )

	local xmpExistsOnOz = not not ( derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp', 'href' ) or
							-- space asset with redacted xmp will have `exists: true` under /rels/xmp
							derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp', 'exists' ) )

	if not xmpExistsOnOz then
		-- Upload after initial xmp is on Oz due to Oz extraction of metadata from the master/proxy. If we
		-- do not wait, then our custom xmp may be overwritten by the extraction process. 
		return
	end
	
	local xmpLocalPath = WFCatFileSpec.getXmpSidecarPathFromDocValue( docValue )
	local xmpSha256 = nil -- not needed for xmp sidecar
	
	-- Before in app contribute, Oz did not provide a /rels/xmp_create link for clients to use to upload
	-- the sidecar files so instead we were hacking the /rels/xmp link that would get created after Oz
	-- extracted the xmp from the master/proxy.
	local xmpCreateUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp_create', 'href' )
	if xmpCreateUrl then
		local revisionIds = derefOpt( docValue, '_serverProvidedMetadata', 'revision_ids' )
		local latestRevisionId = revisionIds and revisionIds[ #revisionIds ]	
	
		xmpCreateUrl = xmpCreateUrl:replace( '{revision_id}', latestRevisionId )
	else
		xmpCreateUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp', 'href' )
	end
	
	-- WFCatTrashUtils_getXmpSidecarIsUploaded is not needed because getXmpSidecarNeedsUploadFromDocValue
	-- tracks whether the binary needs to be uploaded or not, as this value is copied
	-- by the delete and restore operations.

	local xmpNeedsUpload = WFCatFileSpec.getXmpSidecarNeedsUploadFromDocValue( docValue )
	local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceXmpLink' )
	
	local sourceAssetId = derefOpt( docValue, '_localOnly', 'sourceAssetId' ) or docValue._id
	
	local insufficientPermissionError = not not WFCatOzErrors.hasInsufficientPermissionError( docValue )
	
	local hasStorageExceededError = not not WFCatOzErrors.hasStorageExceededError( docValue )
	
	if xmpNeedsUpload and
		( xmpLocalPath or sourceLinkForCopy )
		 and not insufficientPermissionError
		 and not hasStorageExceededError
	then

		local xmpFileSize = WFCatFileSpec.getXmpSidecarFileSizeFromDocValue( docValue )
		
		local binaryType = 'xmp_sidecar'
		local uploadHeaders = getEncodedUploadHeadersForBinaryType( binaryType, docValue )
		
		local spaceId = docValue._spaceId
		
		logger:debugf( "emit %s, %s", xmpLocalPath, xmpCreateUrl )
		
		emit( xmpLocalPath, nil, nil, boolToInt( false ), xmpCreateUrl, binaryType, 
				kContentTypeXML, uploadHeaders, importDate, captureDate, isBlacklisted,
				xmpSha256, xmpFileSize, boolToInt( false ), priorityForUpload, 
				sourceLinkForCopy, spaceId, sourceAssetId )

	end

end

--------------------------------------------------------------------------------

local function maybeEmitProxyUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )
	
	local localPath = WFCatAssetUtils.getFullProxyPathFromDocValue( docValue )

	local fileMaybeOnOfflineVolume = not localPath

	local localVolumeGuid, proxySize = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )
	
	local localRelativePath = WFCatFileSpec.getProxyRelativePathFromDocValue( docValue )

	local ozProxyUrl = WFCatFileSpec.getProxyLinkFromDocValue( docValue, proxySize )

	local ozProxyCreateTemplate = derefOpt( docValue,
											 '_serverProvidedMetadata',
											 'links',
											 '/rels/proxy_create',
											 'href' )

	local revisionIds = derefOpt( docValue,
								  '_serverProvidedMetadata',
								  'revision_ids' )

	local latestRevisionId = revisionIds and revisionIds[ #revisionIds ]

	local copyId = derefOpt( docValue, 'copy', 'copy_id' )

	local willNotUploadMaster = derefOpt( docValue,
													   'importSource',
													   'willNotUploadMaster' )

	local masterPath = WFCatAssetUtils.getFullOriginalPathFromDocValue( docValue )

	local insufficientPermissionError = not not WFCatOzErrors.hasInsufficientPermissionError( docValue )

	local hasStorageExceededError = not not WFCatOzErrors.hasStorageExceededError( docValue )

	-- If this asset was copied via a delete or restore it might not have
	-- it's proxy link set yet, but oz may provide a proxy_create link
	-- in an update because it is still copying the asset asynchronously.
	-- A subsequent update will provide the link.
	local wasUploadedPreviously = WFCatTrashUtils_getProxyIsUploaded( docValue )
	
	local sourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceProxyLink' )
		
	local sourceAssetId = derefOpt( docValue, '_localOnly', 'sourceAssetId' ) or docValue._id
		
	local isSpaceAsset = not not docValue._spaceId
	local isSpaceAssetWithImpliedProxyUpload = isSpaceAsset and getProxyExistsFromDocValue( docValue )						
	
	-- Only try to upload a proxy if it doesn't have a proxy or master link.
	-- and the asset indicates it was imported from lightroom without a master.
	-- Also assume that asset has to have a proxy locally.

	if willNotUploadMaster
		and not masterPath
		and ( localPath or ( localVolumeGuid and localRelativePath ) or copyId
				or ( isSpaceAsset and sourceLinkForCopy ) )
		and not ozProxyUrl
		and not ( wasUploadedPreviously or isSpaceAssetWithImpliedProxyUpload )
		and not insufficientPermissionError
		and not hasStorageExceededError
	then

		local contentType = kContentTypeDng

		local imageSha256 = derefOpt( docValue, 'importSource', 'sha256' )

		local fileSize = WFCatFileSpec.getProxySizeFromDocValue( docValue )

		local proxyCreateUrl
		if ozProxyCreateTemplate and latestRevisionId then
			proxyCreateUrl = ozProxyCreateTemplate:replace( '{revision_id}', latestRevisionId )
			-- Make a new guid for the proxyId, Nimbus doesn't use this guid outside
			-- of the pre-built string in the proxy link (typically it's created by oz).
			local proxyId = KSIdGenerator.uuidInLowercaseHex()
			proxyCreateUrl = proxyCreateUrl:replace( '{proxy_id}', proxyId )
		end

		local binaryType = 'proxy'
		if proxySize == 'panorama' then
			binaryType = 'proxy_pano'
		end

		local uploadHeaders = getEncodedUploadHeadersForBinaryType( binaryType, docValue )
		
		local spaceId = docValue._spaceId
		
		logger:debugf( "emit %s, %s", localPath, proxyCreateUrl )
		
		emit( localPath, localVolumeGuid, localRelativePath, boolToInt( fileMaybeOnOfflineVolume ),
				proxyCreateUrl, binaryType, contentType, uploadHeaders, importDate,
				captureDate, isBlacklisted, imageSha256, fileSize, boolToInt( false ), 
				priorityForUpload, sourceLinkForCopy, spaceId, sourceAssetId )

	end

end

-------------------------------------------------------------------------------

local function canAssetBeUploaded( docValue )
	
	if( doesAssetDocumentHaveLocalBinary( docValue ) or
		isAssetDocumentACopy( docValue )  or
		isSpaceAssetWithSourceLinks( docValue ) or
		 -- Additional check for Android.
		 -- localPath may be nil.( This may happen when after meta data sync, if the asset gets deleted
		 -- in those cases, attemp will be made to look for resources at gallery)
		 -- So, for android, even though localPath is nil and if it is a valid asset document allow to
		 -- emit for further processing.
		 WFPlatformInfo.isAndroid and WFCatIndexUtils.isValidAssetDocument( docValue ) ) then
		
		return true
		
	end
	
		return false
	
end

------------------------------------------------------------------------------

local function generateIndexSchema( _, indexStoreSchema )

	check( '_', 'WFRxIndexStoreSchema' )

	indexStoreSchema:addTable {

		sql = [[CREATE TABLE needsBinaryUpload(
				localFilePath,
				localVolumeGuid,
				localRelativePath,
				fileMaybeOnOfflineVolume,
				uploadUrl,
				binaryType,
				contentType,
				uploadHeaders,
				importDate,
				captureDate,
				isBlacklisted,
				sha256,
				fileSize,
				sha256Error,
				priorityForUpload,
				sourceLinkForCopy,
				spaceId,
				sourceAssetId )]],
					-- For assets that were contributed to a shared album, sourceAssetId is
					-- the document ID of the source asset. Otherwise, it's the doc ID of
					-- the current asset. This field is only necessary to fix WF-750.
		
		docValueFilter = function( docValue )
			return docValue.type == 'asset'
		end,
		map = function( docValue, emit )

			--logger:debugf( "indexer reviewing %s", docValue )

			if canAssetBeUploaded( docValue ) then
				
				local isBlacklisted = checkIfBlacklisted( docValue )

				local importDate, captureDate = getDatesFromDocValue( docValue )
	
				local priorityForUpload = WFCatPhotoImportance.getPriorityForUpload( docValue )

				-- maybe add row for xmp/develop
				maybeEmitXmpDevelopUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload, 2 )
					-- NMB-15173 / LRMI-4195062
					-- Add an extra version param set 2 to force re-index, this is needed as as a
					-- fix was made to maybeEmitXmpDevelopUploadRecord to upload xmp_develop settings
					-- to any restored asset that originally had an external develop setting but is
					-- marked as invalid. Previously we would never upload them again.

				-- maybe add row for full xmp sidecar
				maybeEmitXmpSidecarUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )

				-- maybe add row for proxy but only if no master exists (copied Lightroom asset)
				maybeEmitProxyUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload )
				
				-- maybe add row for master
				maybeEmitMasterUploadRecord( docValue, emit, importDate, captureDate, isBlacklisted, priorityForUpload, 34 )
					-- set to 34 for fixing WF-750

			end

		end,

	}

	indexStoreSchema:addIndex( 
		[[CREATE INDEX needsBinaryUpload_isBlistedImportCapture 
			ON needsBinaryUpload( fileMaybeOnOfflineVolume, isBlacklisted, priorityForUpload DESC, importDate, captureDate )]] )

	indexStoreSchema:addIndex( 
		[[CREATE INDEX needsBinaryUpload_binaryType
			ON needsBinaryUpload( binaryType )]] )

end

--------------------------------------------------------------------------------

local function isMaster( binaryType )

	return binaryType == 'master' or binaryType == 'preset_master' or binaryType == 'camera_profile'

end

--------------------------------------------------------------------------------

local function isProxy( binaryType )
	
	return binaryType == 'proxy' or binaryType == 'proxy_pano'
	
end

--------------------------------------------------------------------------------

local function uploadFinished( self )

	if self._uploadFinished then

		-- Schedule the call to _uploadFinished to prevent stack overflow.
		WFRxScheduler.sharedScheduler():schedule( function()
			-- the value here is unimportant, but making it unique can help in debugging
			self._uploadFinished:onNext( KSDate.timeInIso8601Format() )
		end )

	end

end

--------------------------------------------------------------------------------

local function putAndTrackPerformance( self, masterUrl, binaryType, filePath, fileSize, options, docId )

	local putFile

	if isMaster( binaryType ) and fileSize >= self._chunkedUploader:chunkFileSize() then
		putFile = self._chunkedUploader:doChunkedUpload( masterUrl, filePath, options, docId )
	else
		putFile = self._catalogSpec:putBinary( masterUrl, filePath, options )
	end

	return WFPerformanceReport.wrapObservable( string.format( 'WFCatBinaryUploader: %s', binaryType ), 
		putFile, { fileSize = fileSize } )

end

--------------------------------------------------------------------------------

local function getRecentlyUploadedSha256( self, docId, binaryType )
	
	return derefOpt( self._recentlyUploaded, docId, binaryType )

end

--------------------------------------------------------------------------------

local function getUploadingAssetsWithSha256( self, sha256 )

	return sha256 and self._uploadingAssetWithSha256[ sha256 ] or 0

end

--------------------------------------------------------------------------------

local function decrementUploadingAssetWithSha256( self, docId, binaryType )

	-- Return the number of remaining assets that we are attempting to upload
	-- with the specified sha256. Only called with masters.

	if isMaster( binaryType ) then

		local sha256 = getRecentlyUploadedSha256( self, docId, binaryType )

		if self._uploadingAssetWithSha256[ sha256 ] then

			-- _uploadingAssetWithSha256 may get cleared if an error occurs.
			-- If the count goes negative it shouldn't hurt as the most timers
			-- we will queue up for a sha256 is 100 as limited by the query.
			local newCountOfAssets = self._uploadingAssetWithSha256[ sha256 ] or 0
			newCountOfAssets = newCountOfAssets - 1

			if newCountOfAssets > 0 then
				self._uploadingAssetWithSha256[ sha256 ] = newCountOfAssets
			else
				self._uploadingAssetWithSha256[ sha256 ] = nil
				newCountOfAssets = 0
			end

			return newCountOfAssets

		end

	end

	return -1

end

--------------------------------------------------------------------------------

local function addToRecentlyUploaded( self, docId, binaryType, sha256 )

	local recentlyUploadedForDocId = self._recentlyUploaded[ docId ] or {}

	if recentlyUploadedForDocId[ binaryType ] then
		logger:errorf( "addToRecentlyUploaded recentlyUploadedForDocId ALREADY SET for %s %s", binaryType, docId )
	end

	recentlyUploadedForDocId[ binaryType ] = sha256 

	self._recentlyUploaded[ docId ] = recentlyUploadedForDocId

	if isMaster( binaryType ) then
		local countOfAssets = self._uploadingAssetWithSha256[ sha256 ] or 0
		self._uploadingAssetWithSha256[ sha256 ] = countOfAssets + 1
	end

end

--------------------------------------------------------------------------------

local function clearFromRecentlyUploaded( self, reason, docId, binaryType, uploadAttempted )

	logger:infof( "clearFromRecentlyUploaded %s %s because %s", binaryType, docId, reason )

	local recentlyUploadedForDocId = self._recentlyUploaded[ docId ] or {}

	-- Note that the call to decrementUploadingAssetWithSha256 is placed inside here and
	-- controlled by a parameter because it is dependent on _recentlyUploaded
	-- In other words the ordering of when it is called must take place before
	-- _recentlyUploaded is cleared.
	if uploadAttempted then
		decrementUploadingAssetWithSha256( self, docId, binaryType )
	end

	recentlyUploadedForDocId[ binaryType ] = nil

	if next( recentlyUploadedForDocId ) then 
		self._recentlyUploaded[ docId ] = recentlyUploadedForDocId
	else
		self._recentlyUploaded[ docId ] = nil
	end

end

--------------------------------------------------------------------------------

local function delayedClearFromRecentlyUploaded( self, reason, docId, binaryType, optDelay )

	-- Because the index store takes a few seconds to update, and the doc store
	-- could have potentially just updated, we will give the index store some time
	-- to bring itself up to date.
	
	-- TODO: Is there a more deterministic way to do this? An index store API that fires
	-- when it's up to date?

	local indexStoreUpdateBuffer = optDelay or self._clearFromRecentlyUploadedSeconds

	logger:infof( "Calling clearFromRecentlyUploaded in %d seconds for doc %s type %s because %s.", indexStoreUpdateBuffer, docId, binaryType, reason )

	return self._timerAborter
		:flatMapLatest( function()
			return self._timer( indexStoreUpdateBuffer )
		end )
		:tap( function()

			clearFromRecentlyUploaded( self, reason, docId, binaryType, true )

			-- Trigger combineLatest in uploadSomeBinaries as we may need to try uploading docId again
			uploadFinished( self )

		end )
		:call( runUntilDone )
		
end

----------------------------------------------------------------------------------------------------------------

local function delayedClearFromRecentlyUploaded_NoUpload( self, reason, docId, binaryType, optDelay )

	-- Clears asset from recently uploaded without issuing a uploadFinished call
	-- unless it is the last asset related to the sha256.

	-- This is used to prevent spamming the uploader with requests to upload
	-- something new when we didn't even try to upload anything and didn't
	-- mark anything as invalid.

	-- It is used when we want to have assets wait for the first asset with
	-- the sha to upload so we can issue a copy on the subsequent assets.

	-- Decrement this immediately as we are not uploading, just delaying
	-- the next attempt as we are waiting for the first asset with the sha256
	-- to complete it's upload attempt.

	decrementUploadingAssetWithSha256( self, docId, binaryType )

	local secondsToDelay = optDelay or self._waitForOtherUploadSeconds

	logger:infof( "Calling clearFromRecentlyUploaded in %d seconds for doc %s type %s because %s.", secondsToDelay, docId, binaryType, reason )

	local sha256 = getRecentlyUploadedSha256( self, docId, binaryType )

	return self._timerAborter
		:flatMapLatest( function()

			if sha256 then
				local numAssetsWaitingForUpload = self._numAssetsWaitingForUploadWithSha256[ sha256 ] or 0
				self._numAssetsWaitingForUploadWithSha256[ sha256 ] = numAssetsWaitingForUpload + 1
			end

			return self._timer( secondsToDelay )

		end )
		:tap( function()

			clearFromRecentlyUploaded( self, reason, docId, binaryType, false )

			if sha256 then

				local numAssetsWaitingForUpload = self._numAssetsWaitingForUploadWithSha256[ sha256 ] or 0
				numAssetsWaitingForUpload = numAssetsWaitingForUpload - 1
				self._numAssetsWaitingForUploadWithSha256[ sha256 ] = numAssetsWaitingForUpload > 0 and numAssetsWaitingForUpload or nil

				if numAssetsWaitingForUpload == 0 then
					-- Trigger combineLatest in uploadSomeBinaries as we will need
					-- to upload the copies that were waiting.
					uploadFinished( self )
				end

			end

		end, function( err )
			-- clear when aborted
			self._numAssetsWaitingForUploadWithSha256[ sha256 ] = nil
		end )
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

local function uploadProgressStatusUpdate(self, operation, fileSize ) 

	check( 'WFCatBinaryUploader', '?string' )

	if not self._uploadProgress or not self._uploadProgress.progressReporter or not fileSize then
		return
	end

	if operation == "increaseBytesNeedToUpload" then

		-- Addition of asset to Active uploads
		self._uploadProgress.totalBytesNeedToUpload = self._uploadProgress.totalBytesNeedToUpload + fileSize

	elseif operation == "increaseBytesUploaded" then

		self._uploadProgress.totalBytesUploadedSoFar = self._uploadProgress.totalBytesUploadedSoFar + fileSize

	else
		return
	end
	
	if( self._uploadProgress.totalBytesNeedToUpload - self._uploadProgress.totalBytesUploadedSoFar  > 0 ) then
		
		local binaryUploadProgress = {
			totalBytesNeedToUpload = self._uploadProgress.totalBytesNeedToUpload,
			totalBytesUploadedSoFar = self._uploadProgress.totalBytesUploadedSoFar,
		}
		self._uploadProgress.progressReporter:updateProgressWithInfo( binaryUploadProgress, true )
		
	end

end

--------------------------------------------------------------------------------

local function addToActiveUploads( self, docId, binaryType, subscription, localFileSize )

	self._binariesInFlight = self._binariesInFlight + 1

	local activeUploadsForDocId = self._activeUploads[ docId ] or {}

	if activeUploadsForDocId[ binaryType ] then
		logger:errorf( "Already have subscription for uploading %s binaryType = %s", docId, binaryType )
	else
		uploadProgressStatusUpdate(self, "increaseBytesNeedToUpload", localFileSize )
	end

	activeUploadsForDocId[ binaryType ] = subscription

	self._activeUploads[ docId ] = activeUploadsForDocId

end

--------------------------------------------------------------------------------

local function removeFromActiveUploads( self, docId, binaryType, localFileSize )

	if derefOpt( self._activeUploads, docId, binaryType ) == nil then
		-- Already removed by completed or error, finally will call a second
		-- time but is needed in case the subscription is disposed before
		-- completed or error.
		return
	end

	local activeUploadsForDocId = self._activeUploads[ docId ] or {}
	
	activeUploadsForDocId[ binaryType ] = nil
	
	if next( activeUploadsForDocId ) then
		self._activeUploads[ docId ] = activeUploadsForDocId
	else
		self._activeUploads[ docId ] = nil
	end

	self._binariesInFlight = self._binariesInFlight - 1

	if self._binariesInFlight < 0 then
		logger:errorf( "_binariesInFlight went negative == %d", self._binariesInFlight )
		self._binariesInFlight = 0
	end
	if self._binariesInFlight > 0 and next( self._activeUploads ) == nil then
		logger:errorf( "_binariesInFlight == %d and _activeUploads is empty", self._binariesInFlight )
	end

end

--------------------------------------------------------------------------------

local function eventuallyClearBlackoutWindow( self, docId, binaryType, waitCount )

	-- Don't let this original be uploaded again

	waitCount = waitCount or 0

	self._timerAborter
		:flatMapLatest( function()
			return WFRxLowFidelityTimer.docValueTimer( uploadBlackoutWindowInSeconds, self._docStore, docId )
		end )
		:tap( function( docValue )

			local uploadedSha256, currentSha256
			local uploadConfirmed

			if not isValidAssetDocument( docValue ) and not isNonDeletedStyleDocument( docValue )
					and not WFCatIndexUtils.isNonDeletedDCPDocument( docValue ) then

				-- Clear out RecentlyUploaded for deleted documents.
				logger:warnf( "Asset deleted prior to upload completion for %s asset %s.", binaryType, docId )

				uploadConfirmed = true

			elseif isMaster( binaryType ) then
				
				uploadConfirmed = ( getOriginalLinkFromDocValue_IgnoringShaError( docValue ) ~= nil ) or
									( docValue._spaceId and getOriginalExistsFromDocValue( docValue ) )
				uploadedSha256 = getRecentlyUploadedSha256( self, docId, binaryType )
				currentSha256 = derefOpt( docValue, 'importSource', 'sha256' )
			
			elseif isProxy( binaryType ) then
				
				uploadConfirmed = ( WFCatFileSpec.getProxyLinkFromDocValue( docValue ) ~= nil ) or
									( docValue._spaceId and getProxyExistsFromDocValue( docValue ) )
				uploadedSha256 = getRecentlyUploadedSha256( self, docId, binaryType )
				currentSha256 = derefOpt( docValue, 'importSource', 'sha256' )
				
			elseif binaryType == 'xmp_develop' then
				
				uploadConfirmed = WFCatFileSpec.getXmpDevelopLinkFromDocValue( docValue ) ~= nil
				uploadedSha256 = getRecentlyUploadedSha256( self, docId, binaryType )
				currentSha256 = derefOpt( docValue, 'develop', 'xmpCameraRaw', 'sha256' )

			elseif binaryType == 'xmp_sidecar' then

				-- Verify _localyOnly.xmpSidecarNeedsUpload entry is removed from the docValue.
				uploadConfirmed = WFCatFileSpec.getXmpSidecarNeedsUploadFromDocValue( docValue ) == nil
				uploadedSha256 = nil
				currentSha256 = nil

			end

			if uploadedSha256 ~= currentSha256 then
			
				-- If there was a destructive edit that caused the original to change,
				-- resulting in a new sha256, we clear the asset from the list so that 
				-- we may upload the new original. It should be noted that we don't yet
				-- support destructive edits in many other parts of sync.
				
				-- This probably happens when xmp_develop is modified during upload.

				logger:warnf( "sha no longer matches for %s type %s", docId, binaryType )

				delayedClearFromRecentlyUploaded( self, "sha no longer matches", docId, binaryType )
				
			elseif uploadConfirmed then
			
				-- Once the docValue contains an Oz master href, we can safely remove
				-- the asset from the recently uploaded table since we won't try to reupload
				-- assets that have oz master links.

				delayedClearFromRecentlyUploaded( self, "upload confirmed", docId, binaryType )

			else
			
				waitCount = waitCount + 1
			
				logger:warnf( "Doc store hasn't updated with the presence of uploaded %s for %s after %d seconds",
						binaryType, docId, uploadBlackoutWindowInSeconds * waitCount )
						
				if waitCount > 6 then
					
					logger:warnf( "Assuming upload failed for asset %s. Retrying...", docId )
					
					delayedClearFromRecentlyUploaded( self, "assuming upload failed", docId, binaryType )
					
				else
				
					-- Keep waiting
					
					eventuallyClearBlackoutWindow( self, docId, binaryType, waitCount )
					
				end
			end
			
		end )
		
		:call( runUntilDone )

end

--------------------------------------------------------------------------------

local function uploadCompleted( self, docId, binaryType, uploadSha256 )

	eventuallyClearBlackoutWindow( self, docId, binaryType )
	
	local docStore = self._docStore
	
	if isMaster( binaryType ) then
		
		-- NMB-5779: We may have received the master via migration or another
		-- change which required us to wait until Oz marked the previous binaries
		-- as invalid.  If this is the case, remove the annotation that tells us 
		-- we must wait
		docStore:docValue( docId )
			:flatMap( function(docValue )
				
				local waitForInvalid = derefOpt( docValue, '_localOnly', 'binaryUpload', 'waitForInvalid' )
				
				-- NMB-7626: Previously wait for invalid was simply a boolean true. With the introduction of presets
				-- which are both small and the file is changeable, we 
				if waitForInvalid == true -- old case (< 1.4) which may exist in doc store before upgrade
				    or ( waitForInvalid == uploadSha256 ) then -- 1.4+
					return docStore:applyOverlay {
						_id = docId,
						_localOnly = {
							binaryUpload = {
								waitForInvalid = { _delete = true },
							},
						},
					}
					
				else
					return Observable.just()
				end
			end )
			:call( runUntilDone )
			
	elseif binaryType == 'xmp_sidecar' then

		-- Originals and develop settings have links set in the _serverProvidedMetadata
		-- for the binary's location on oz, whereas xmp is generated when
		-- an original is received.  So there is no way to detect the upload has
		-- completed by watching the sync feed, so update _localyOnly.xmpSidecarNeedsUpload
		-- when completed so that eventuallyClearBlackoutWindow can clear the upload
		-- from the blackout window when it's timer fires.

		docStore:applyOverlay  {
			_id = docId,
			_localOnly = {
				xmpSidecarNeedsUpload = { _delete = true },
			},
		}
		:call( runUntilDone )
		
	end

end

--------------------------------------------------------------------------------

local function addTransferStatsToDiagnosticLog( self, bucket, time, size )

	local sizeInMB = size / (1024 * 1024)

	local bucketStats = self._uploadSpeedStats.buckets[ bucket ]
	local uploadSpeed = sizeInMB / time

	if not bucketStats then

		bucketStats = {
			count = 0,
			timings = WFDiagnosticLog.makeRingBuffer { "Time (sec)" },
			sizes = WFDiagnosticLog.makeRingBuffer { "Size (MB)" },
			speed = WFDiagnosticLog.makeRingBuffer { "Upload speed (MBps)" },
		}

		self._uploadSpeedStats.buckets[ bucket ] = bucketStats

	end

	bucketStats.count = bucketStats.count + 1
	bucketStats.timings:add( time )
	bucketStats.sizes:add( sizeInMB )
	bucketStats.speed:add( uploadSpeed )

	local totalBinarySize = self._uploadSpeedStats.totalBinarySize or 0
	self._uploadSpeedStats.totalBinarySize = totalBinarySize + sizeInMB

end

--------------------------------------------------------------------------------------------------------

local function isMobileClient( self )

	-- just checking isMobile here not enough as some of desktop tcs also being run as part of mobile.
	-- in order to filter out those additional check added as shown below.
	return WFPlatformInfo.isMobile() and not not self._catalogBasePath; 

end

---------------------------------------------------------------------------------------------------------

local function WFAnalyticsTrackEventNotifySuccess( self, docId, uploadUrl, binaryType, contentType )

	WFAnalytics.trackEvent {
				['event.type'] = binaryType .. "_uploaded",
				['event.workflow'] = 'SYNC',
				['event.subcategory'] = 'binary_upload_success',
				['content.id'] = docId,
				['content.type'] = contentType,
				['content.url'] = uploadUrl,
			}

end

----------------------------------------------------------------------------------------------------------

local function WFAnalyticsTrackEventNotifyFailure( self, docId, masterPutUrl, responseData, headers, binaryType, contentType, fileSize, httpError, reason )

	local eventSubcategory

	if reason then
		eventSubcategory = reason
	elseif fileSize and fileSize >= self._chunkedUploader:chunkFileSize() then
		eventSubcategory = 'binary_upload_chunk'
	else
		eventSubcategory = 'binary_upload_data'
	end

	local subtype -- subtype, content type, filesize shall be empty for the InvalidLocalOriginal case, so do not assign default vale

	if contentType then

		if WFOzContentTypes.isVideo( contentType ) then 
			subtype = 'video'
		else
			subtype = 'image'
		end

	end

	local errorEvent = {}

	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = eventSubcategory
	errorEvent['event.url'] = masterPutUrl
	errorEvent['event.error_type'] = binaryType .. "Upload.Error"
	errorEvent['content.id'] = docId
	errorEvent['content.type'] = subtype
	errorEvent['content.mimetype'] = contentType
	errorEvent['content.size'] = fileSize 

	if not httpError then

		WFAnalytics.trackEvent( errorEvent )
		return

	end

	errorEvent['content.source'] = (subtype == 'image') and 'file' or 'data'

	local errorCode = headers and headers.status_code or 'nil'

	if not headers then

		-- Modern native HTTP implementations put "wf_err_timeout" in the error JSON on timeout
		if (string.match( responseData, "wf_err_timeout" ) or string.match( responseData, "timed out" )) then
			errorCode = "timeout"
		end

		local annotatedReason = string.format( "%s", tostring( responseData ) )
		
		errorEvent['event.subtype'] = 'OS'
		errorEvent['event.error_code'] = errorCode
		errorEvent['event.error_desc'] = annotatedReason

	else

		local ozErrorResponse = string.format( "%s",
									   tostring( responseData ) )

		local ozRequestId = headers['X-Request-Id']

		errorEvent['event.subtype'] = 'API'
		errorEvent['event.error_code'] = errorCode
		errorEvent['event.error_desc'] = ozErrorResponse
		errorEvent['event.context_guid'] = ozRequestId

	end
	
end

-----------------------------------------------------------------------------------------------------------------------

local function addHttpErrorToLog( self, docId, masterPutUrl, responseData, headers, binaryType, contentType, localSize )

	-- Special case for HTTP transit failure.

	local errorLog = self._uploadErrorLog

	if not headers then

		local annotatedReason = string.format( "%s", tostring( responseData ) )

		local logFmtStr = "Unable to communicate with Oz URL %s\n\nreason = %s"

		logger:errorf( logFmtStr, masterPutUrl, annotatedReason )

	else

		-- We were able to communicate with Oz, but Oz didn't like the response. Log this.

		if not headers then
			headers = {}
		end

		local logFmtStr = "Got status %s back from Oz\n\n"
		local logParams = { headers.status_code or 'nil' }

		if masterPutUrl then
			logFmtStr = logFmtStr .. "url we sent data to was:\n\n%s\n\n"
			logParams[ #logParams + 1 ] = masterPutUrl
		end

		logFmtStr = logFmtStr .. "HTML response body was:\n\n%s\n\n"
		logParams[ #logParams + 1 ] = responseData

		errorLog[ #errorLog + 1 ] = {

			time = KSDate.timeInIso8601FormatWithTimeZone(),

			id = docId,
			binaryType = 'master',
			error = string.format( logFmtStr, unpack( logParams ) ),

		}

		if #errorLog > 10 then
			table.remove( errorLog, 1 )
		end

		logger:errorf( logFmtStr, unpack( logParams ) )

	end

	WFAnalyticsTrackEventNotifyFailure( self, docId, masterPutUrl, responseData, headers, binaryType, contentType, localSize, true )

end

--------------------------------------------------------------------------------

local function addInvalidOriginalErrorToLog( errorLog, docId, filePath, reason )

	errorLog[ #errorLog + 1 ] = {

		time = KSDate.timeInIso8601FormatWithTimeZone(),

		id = docId,
		binaryType = 'master',
		error = reason,
		filePath = filePath,
	}

	logger:errorf( "Marking original as invalid for docId %s path %s", docId, filePath )

end

--------------------------------------------------------------------------------

local function getRequestOptionsFor( binaryType, contentType, isCopy, additionalUploadHeaders )
	
	local options = {
		dataFormat = 'filePath',
		requestPriority = WFRxHttpClient.requestPriorityMedium,
		requestCost = uploadHttpRequestCost
	}
	
	if binaryType == 'master' then
		
		local requestHeaders = {
			{
				header = 'X-Generate-Renditions',
				value = 'all',
			},
			{
				header = 'X-Read-Meta',
				value = 'xmp,location'
			},
			{
				header = 'Content-Type',
				value = contentType,
			},
			{
				header = "Accept",
				value = "application/json",
			},
		}
		
		if isCopy then
			table.insert( requestHeaders, {
				header = 'X-Async-Copy',
				value = 'true'
			} )
		end
		
		if not WFOzContentTypes.isVideo( contentType ) then
			-- don't request proxy generation for videos, but 
			-- do for image/* and raw files (application/octet-stream)
			table.insert( requestHeaders, 
				{
					header = 'X-Generate-Proxies',
					value = 'all',
				} )
		end
		
		options.requestHeaders = requestHeaders

	elseif binaryType == 'preset_master' then
		-- master for preset/profile
		
		local requestHeaders = {
			{
				header = 'Accept',
				value = 'application/json',
			},
			{
				header = 'Content-Type',
				value = contentType or kContentTypeXML,
			},
		}
		
		options.requestHeaders = requestHeaders
		
	elseif binaryType == 'camera_profile' then
		-- master for camera_profile
		local requestHeaders = {
			{
				header = 'Accept',
				value = 'application/json',
			},
			{
				header = 'Content-Type',
				value = contentType or kContentTypeOctetStream,
			},
		}
		options.requestHeaders = requestHeaders
		
	elseif binaryType == 'proxy' or binaryType == 'proxy_pano' then

		options.requestHeaders = {
				{
					header = 'Content-Type',
					value = contentType or kContentTypeDng,
				},
				{
					header = 'Accept',
					value = 'application/json',
				},
				{
					header = 'X-Generate-Renditions',
					value = 'all',
				},
				{
					header = 'X-Read-Meta',
					value = 'xmp',
				},
			}
		
		if isCopy then
			table.insert( options.requestHeaders, {
				header = 'X-Async-Copy',
				value = 'true'
			} )
		else
			local proxyType = '2560'
			if binaryType == 'proxy_pano' then
				proxyType = 'panorama'
			end
			table.insert( options.requestHeaders,
				{
					header = 'X-Proxy-Type',
					value = proxyType,
				} )
		end

	elseif binaryType == 'xmp_develop' then
		
		options.requestHeaders = {
				{
					header = 'X-Generate-Renditions',
					value = 'all',
				},
				{
					header = 'X-Generate-Proxies',
					value = 'all',
				},
				{
					header = 'Content-Type',
					value = contentType or kContentTypeXML,
				},
				{
					header = 'Accept',
					value = 'application/json',
				},
			}

	elseif binaryType == 'xmp_sidecar' then
		
		options.requestHeaders = {
				{
					header = 'Content-Type',
					value = contentType or kContentTypeXML,
				},
				{
					header = 'Accept',
					value = 'application/json',
				},
			}
			
	end
	
	for header, value in pairs( additionalUploadHeaders or {} ) do 
		local headers = options.requestHeaders or {}
		headers[ #headers + 1 ] = {
			header = header,
			value = value
		}
		options.requestHeaders = headers
	end
	
	return options

end

--------------------------------------------------------------------------------

local function recordSuccess( self, docId, startTime, uploadUrl, binaryType, localSize, response, contentType, isCopy )

	localSize = localSize or 0

	local bucket = binaryType
	if localSize >= self._chunkedUploader:chunkFileSize() then
		bucket = bucket .. "-chunked"
	end

	if isCopy then
		bucket = bucket .. "-copy"
	end

	if not string.match(bucket, "-chunked") then
		-- for chunked uploads audit happens in WFCatContentRangeUploader
		uploadProgressStatusUpdate(self, "increaseBytesUploaded", localSize )
	end

	addTransferStatsToDiagnosticLog( self, bucket, now() - startTime, localSize )
	
	local elapsedTime = now() - startTime;
	logger:infof( "Uploaded %s to %s, response = %s, time %f (%.2fKBps)",
		binaryType, uploadUrl, response,
		elapsedTime, ((localSize/elapsedTime)/1024) )

	-- This is limited to mobile and Mac App Store only because tracking successes for regular desktop 
	-- builds overwhelms highbeam. If desktop fully switches over to Dunamis, the if can be removed
	-- and we can always track the upload success.
	
	if isMobileClient( self ) or WFPlatformInfo.isMacAppStore then

		WFAnalyticsTrackEventNotifySuccess( self, docId, uploadUrl, binaryType, contentType )

	end

end

--------------------------------------------------------------------------------

local function removeSidecarFileAnnotation( self, docId )

	local binarySizes = {}
	binarySizes[ 'xmp' ] = { _delete = true }
	
	local files = {}
	files[ 'original' ] = binarySizes

	return self._docStore:applyOverlay {
			_id = docId,
			_localOnly = {
				files = files,
				xmpSidecarNeedsUpload = { _delete = true },
				xmpSidecarExists = false,
			}
		}
		
end

--------------------------------------------------------------------------------

local function applyInvalidOriginalOverlay( self, docId, filePath, binaryUrl, reason )

	reason = reason or 'InvalidLocalOriginal'

	local shouldApplyOverlay = true
	
	local volumeIsOnline = filePath and WFFileUtils.getVolumeGuidForPath( filePath )
	
	if filePath and not volumeIsOnline then
	
		-- If the file is on a volume which was disconnected
		-- in the middle of the upload, then we shouldn't apply the
		-- error overlay.
	
		shouldApplyOverlay = false

	end

	if shouldApplyOverlay then
	
		addInvalidOriginalErrorToLog( self._uploadErrorLog, docId, filePath, reason )

		WFAnalyticsTrackEventNotifyFailure( self, docId, binaryUrl, nil, nil, 'master', nil, nil, false, reason )

		-- Mark the original as invalid since the sha256 no longer matches what we imported.

		-- This can happen because of disk corruption or from an invalid copy during import.
		-- (since sha may be calculated prior to importing by the Import UI).

		-- It could also happen if the user copied a new file into the ~/Pictures/Nimbus folder
		-- manually (which they should not do).

		-- Adding localOriginalError will cause this file to no longer be considered for uploaded.

		return self._docStore:applyOverlay {
				_id = docId,
				_localOnly = {
					localOriginalError = reason,
				}
			}

	else
	
		return Observable.just( false )
	
	end

end

--------------------------------------------------------------------------------

local function resolveSha256ErrorInInputSource( self, masterUploadUrl, filePath, docId, imageSha256 )

	-- The most comment case for this to happen is when an async upload operation
	-- has a sha mismatch error.  This happens when either WFCatContentRangeUploader
	-- or WFCatContentRangeUploader complete (they get success) but oz assembles
	-- the pieces together and updates importSource.status with this error if
	-- the sha doesn't match.
 
	-- Copied assets can also get marked with the sha256error so check to see if
	-- we have a filePath before attempting the digest, if there is no filePath
	-- then just mark the asset as invalid.
	if filePath == nil then
		return applyInvalidOriginalOverlay( self, docId, filePath, masterUploadUrl )
	end

	-- Must have been from an async upload, see if we can fix it.

	-- Recalculate sha256 to verify it matches the sha256 in the importSource to
	-- verify we have the correct file to upload.

	return WFCrypto.asyncDigestFile( filePath, 'SHA256' )

	:flatMap( function( localSha256 )

		if localSha256 == imageSha256 then

			-- Update the doc so the binary uploader will try to upload this file again.

			logger:debugf( "re-uploading corrupted original for docId %s path %s", docId, filePath )

			return self._docStore:docValue( docId )
				:flatMap( function( docValue )
					
					local currentMasterUrl = derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
	
					return self._docStore:applyOverlay {
						_id = docId,
						_localOnly = { 
							-- Add a local annotation for the sha error since we're going to
							-- otherwise remove all traces that it happened.  This is necessary
							-- to prevent us from accidentally deleting the local master because
							-- we think the master is uploaded and fine (NMB-4862)
							binaryUploadErrors = {
								[ '/rels/master' ] = {
									[ currentMasterUrl ]  = "corrupt upload",
								}
							}
						},
					}
			end )
			
		else

			return applyInvalidOriginalOverlay( self, docId, filePath, masterUploadUrl )

		end

	end )

end

--------------------------------------------------------------------------------

local function disposeActiveUploads( self )

	logger:debugf( "disposeActiveUploads started" )

	-- Abort the timers we have for clearing the recently uploaded fields
	-- since we are clearing them out immediately below.  This prevents
	-- a timer from a prior upload request from being issued which can cause
	-- a file to be uploaded multiple times simultaneously.	
	self._timerAborter:onError( 'Abort' )

	-- Create a new _timerAborter that isn't in the terminal state for the
	-- next set of timers.
	self._timerAborter = Rx.ReplaySubject( 1 )
	self._timerAborter:onNext( true )

	for docId, uploadsForDocId in pairs( self._activeUploads ) do
		
		for binaryType, subscription in pairs( uploadsForDocId ) do 
		
			logger:debugf( "Aborting upload for docId %s", docId )
	
			subscription:dispose()
	
			clearFromRecentlyUploaded( self, "disposed", docId, binaryType )
			
		end
		
	end

	if self._binariesInFlight > 0 then
		logger:errorf( "_binariesInFlight == %d should be 0", self._binariesInFlight )
	end

	self._binariesInFlight = 0

	self._recentlyUploaded = {}

	self._activeUploads = {}

	self._uploadingAssetWithSha256 = {}

end

--------------------------------------------------------------------------------

local function handleUploadFailure( self, docId, binaryType, uploadUrl, responseData, headers, invalidOriginal, contentType, localSize )

	self._errorStatus:onNext( docId, responseData )

	addHttpErrorToLog( self, docId, uploadUrl, responseData, headers, binaryType, contentType, localSize )

	if invalidOriginal then
		delayedClearFromRecentlyUploaded( self, "wait for docstore update as invalidOriginal", docId, binaryType )
		return
	end
	
	if self._failedUploadBlacklistHelper:maybeBlacklist( docId ) then

		-- The asset is now blacklisted. It will be automatically removed from the
		-- blacklist after failedUploadBlackoutWindowInSeconds seconds.

		WFPerformanceReport.recordSingleValue( 'WFCatBinaryUploader:Blacklist', {
			docId = docId,
			numberOfFailedAttempts = numberOfTriesBeforeBlacklistingUpload,
			blacklistDuration = failedUploadBlackoutWindowInSeconds
		} )

		delayedClearFromRecentlyUploaded( self, "wait for docstore update as blacklisted", docId, binaryType )

	else

		-- Upload wasn't blacklisted so try to upload the original again.

		local uploadErrorDelay = 10
		delayedClearFromRecentlyUploaded( self, "upload failed, delay to let other assets upload first", docId, binaryType, uploadErrorDelay )

	end

end

--------------------------------------------------------------------------------

local function errorShaDoesNotMatch( binaryType, responseData, headers )

	if isMaster( binaryType ) and headers and headers.status_code == '400' and headers[ 'X-Result-Code' ] == '1005' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		return ok and derefOpt( response, 'errors', 'sha256', 1 ) == "does not match"

	end

	return false

end

--------------------------------------------------------------------------------

local function errorTrialOrSubscriptionExpired( binaryType, responseData, headers )

	if headers and headers.status_code == '403' and headers[ 'X-Result-Code' ] == '4300' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		return ok and derefOpt( response, 'subtype' ) == WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError

	end

	return false

end

--------------------------------------------------------------------------------

local function errorUploadMasterWhenNotAllowed( binaryType, responseData, headers )

	if isMaster( binaryType ) and headers and headers.status_code == '403' and headers[ 'X-Result-Code' ] == 'UploadMasterWhenNotAllowedError' then

		return true

	end

	return false

end

--------------------------------------------------------------------------------

local function errorInsufficientPermission( binaryType, responseData, headers )

	if headers and headers.status_code == '403' and headers[ 'X-Result-Code' ] == 'InsufficientPermissionError' then

		return true
		
	end

	return false

end

--------------------------------------------------------------------------------

local function errorPayloadMustExist( binaryType, responseData, headers )

	if not isMaster( binaryType ) then
		return false
	end

	if responseData == 'payload must exist' then
		-- Chunked uploader check file size up front as it can't be 0.
		return true
	end

	if headers and headers.status_code == '400' and headers[ 'X-Result-Code' ] == '1005' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		return ok and derefOpt( response, 'errors', 'payload', 1 ) == "must exist"

	end

	return false

end

--------------------------------------------------------------------------------

local function applyRemoveWillNotUploadMasterOverlay( self, docId )

	return self._docStore:applyOverlay {
			_id = docId,
			importSource = {
				willNotUploadMaster = { _delete = true },
			},
		}

end

--------------------------------------------------------------------------------

local function applySpecificErrorOverlay( self, docId, errorType )
	
	local overlay = WFCatOzErrors.createOverlayForOzErrorSubtype( docId, errorType )
	return self._docStore:applyOverlay( overlay )
	
end

--------------------------------------------------------------------------------

local function checkForAccountError( self, docId, binaryType, responseData, headers )

	if errorTrialOrSubscriptionExpired( binaryType, responseData, headers ) then
		
		return applySpecificErrorOverlay( self, docId, WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError )
			:flatMap( function()
				return Observable.throw( responseData, headers, true )
			end )
	
	end

	if errorInsufficientPermission( binaryType, responseData, headers ) then
		
		return applySpecificErrorOverlay( self, docId, WFCatOzErrors.Errors.kInsufficientPermissionError )
			:flatMap( function()
				return Observable.throw( responseData, headers, true )
			end )
	
	end
		
end

--------------------------------------------------------------------------------

local function errorAndMaybeNewIdIfMasterOrProxyResourceDoesNotExist( binaryType, responseData, headers )

	if isMaster( binaryType ) or binaryType == 'proxy' and headers and headers.status_code == '404' and headers[ 'X-Result-Code' ] == '1000' then

		local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )

		if ok then

			local resourceDoesNotExist = derefOpt( response, 'subtype' ) == "ResourceNotFoundError"

			return resourceDoesNotExist, response.new_id

		end

	end

	return false

end

--------------------------------------------------------------------------------

local function makeOverlayToUpdateSourceAsset( docIdToUpdate, sourceDocValue )

	-- This will either get the createFrom links, or the
	-- actual links for this new asset. We may end up failing
	-- on the createFrom links but will get another error and
	-- another new_id that we can try.

	local localOnly = {}

	localOnly.sourceOriginalLink = WFCatFileSpec.getOriginalLinkFromDocValue( sourceDocValue )
	localOnly.sourceXmpLink = WFCatFileSpec.getXmpLinkFromDocValue( sourceDocValue )
	localOnly.sourceXmpDevelopLink = WFCatFileSpec.getXmpDevelopLinkFromDocValue( sourceDocValue )
	localOnly.sourceProxyLink = WFCatFileSpec.getProxyLinkFromDocValue( sourceDocValue )

	localOnly.sourceAssetId = sourceDocValue._id

	return {
		_id = docIdToUpdate,
		_localOnly = localOnly,
	}

end

--------------------------------------------------------------------------------

local function makeOverlayToRemoveSourceAsset( docValue )

	local localOnly = {}

	localOnly.sourceOriginalLink = { _delete = true }
	localOnly.sourceXmpLink = { _delete = true }
	localOnly.sourceXmpDevelopLink = { _delete = true }
	localOnly.sourceProxyLink = { _delete = true }

	localOnly.sourceAssetId = { _delete = true }

	local hasMasterLocally = WFCatAssetUtils.hasLocalOriginal( docValue )
	local hasProxyLocally = WFCatAssetUtils.hasLocalProxy( docValue )

	local importSource
	if hasProxyLocally and not hasMasterLocally then
		importSource = {
			willNotUploadMaster = true
		}
	end

	return {
		_id = docValue._id,
		_localOnly = localOnly,
		importSource = importSource
	}

end

--------------------------------------------------------------------------------

local function applyOrRemoveNewSourceLinksOrErrorOverlayForCopy( self, docId, newSourceId, binaryType )

	-- Update source asset and links in docId with id and links retrieved from
	-- the document identified by newId.

	local docStore = self._docStore

	return docStore:docValue( docId )
	:flatMap( function( docValue )

		if newSourceId then

			-- Oz provided a new_id, so source must have been deleted.
			return Observable.just( true, newSourceId, docValue )

		else

			-- Oz didn't provide a new_id in the result headers, check to see
			-- if the source was deleted and get the _newId from the deleted doc.
			local sourceAssetId = derefOpt( docValue, '_localOnly', 'sourceAssetId' )

			return docStore:docValue( sourceAssetId )
			:map( function( existingSourceDocValue )

				local existingSourceDeleted = existingSourceDocValue._deleted

				newSourceId = existingSourceDocValue._newId

				if newSourceId then
					return existingSourceDeleted, newSourceId, docValue
				else
					return existingSourceDeleted, nil, docValue
				end

			end )

		end

	end )
	:flatMap( function( existingSourceDeleted, newSourceId, docValue )

		if newSourceId then

			return WFCatTrashUtils.getTrashOrRestoredDocFromNewId( docStore, newSourceId )
			:flatMap( function( newSourceDocValue )

				if newSourceDocValue._deleted then

					-- Remove the overlay the asset was permanently deleted.
					local updateSourceOverlay = makeOverlayToRemoveSourceAsset( docValue )
					return docStore:applyOverlay( updateSourceOverlay )
						:map( function()
							 -- Don't count as failed yet, see if we can upload a binary.
							return true
						end )

				elseif newSourceDocValue._empty then

					-- Doesn't exist yet, newId likely came from oz for asset
					-- from which we haven't received metadata.

					-- Count this as a failed attempt, it will retry with existing
					-- invalid link and get error again resulting in redoing all this
					-- in hopes that it will get an update with the links to the
					-- binary.

					return Observable.just( false )

				else

					-- Exists

					if WFCatFileSpec.getOriginalLinkFromDocValue( newSourceDocValue ) == nil
					and WFCatFileSpec.getProxyLinkFromDocValue( newSourceDocValue ) == nil then

						-- Links aren't there yet

						-- Count this as a failed attempt, it will retry with existing
						-- invalid link and get error again resulting in redoing all this
						-- in hopes that it will get an update with the links to the
						-- binary.

						return Observable.just( false )

					else

						local overlay = makeOverlayToUpdateSourceAsset( docValue._id, newSourceDocValue )
						return docStore:applyOverlay( overlay )
							:map( function()
								-- Don't count as failed as we applied a new link.
								return true
							end )

					end

				end

			end )

		else

			if existingSourceDeleted then
				-- Remove sourceLink as it no longer exists.
				local updateSourceOverlay = makeOverlayToRemoveSourceAsset( docValue )
				return docStore:applyOverlay( updateSourceOverlay )
					:map( function()
						return true -- Try again, don't count as see if we can upload a binary.
					end )
			else
				 -- Count this as a failed attempt, but retry.
				return Observable.just( false )
			end

		end

	end )

end

--------------------------------------------------------------------------------

local function uploadOneBinary( self, docId, localPath, localSize, uploadUrl, binaryType, contentType, addlUploadHeaders, sha256 )

	logger:debugf( "Uploading %s for docId %s from %s to %s", binaryType, docId, localPath, uploadUrl )

	local options = getRequestOptionsFor( binaryType, contentType, false, addlUploadHeaders )
	options.requestDescription = string.format( '%s : Uploading %s', docId, binaryType )
	
	local startTime = now()

	return putAndTrackPerformance( self, uploadUrl, binaryType, localPath, localSize, options, docId )
	:catch( function( responseData, headers )

		if errorShaDoesNotMatch( binaryType, responseData, headers ) then

			-- Recalculate sha256 to verify it matches the sha256 in the importSource to
			-- verify we have the correct file to upload.

			return WFCrypto.asyncDigestFile( localPath, 'SHA256' )
				:flatMap( function( localSha256 )

					if localSha256 == sha256 then

						return Observable.throw( responseData, headers, false )

					else

						return applyInvalidOriginalOverlay( self, docId, localPath, uploadUrl )
							:flatMap( function()
								return Observable.throw( responseData, headers, true )
							end )

					end
				end )

		end


		if errorUploadMasterWhenNotAllowed( binaryType, responseData, headers ) then

			return applyRemoveWillNotUploadMasterOverlay( self, docId )

				:flatMap( function()

					-- Throw an error so we retry the upload after the doc store
					-- has had time to push the removal of the original through
					-- the changes feed.

					return Observable.throw( responseData, headers, true )

				end )

		end

		if errorPayloadMustExist( binaryType, responseData, headers ) then

			return applyInvalidOriginalOverlay( self, docId, localPath, uploadUrl )
				:flatMap( function()
					return Observable.throw( responseData, headers, true )
				end )

		end
		
		local maybeAccountError = checkForAccountError( self, docId, binaryType, responseData, headers )
		
		if maybeAccountError then
			return maybeAccountError
		end

		return Observable.throw( responseData, headers, false )

	end )

	:tap(
		function( response )
			recordSuccess( self, docId, startTime, uploadUrl, binaryType, localSize, response, contentType )
		end,
		nil,
		function()
			uploadCompleted( self, docId, binaryType, sha256 )
		end )
		
	:catch( function( responseData, headers, invalidOriginal )

			-- On binary uploads of the xmp_develop where we request rendition generation we
			-- may get a 404 with a SourceDoesNotExist error from Oz if a master/proxy does
			-- not yet exist. In these cases, the upload succeeds, but the rendition generation
			-- does not. This failure can be treated as a success as we will request renditions
			-- again when the master/proxy is uploaded
			
			local uploadFailed = true
			
			if binaryType == 'xmp_develop' and headers.status_code == '404' then
				local ok, response = pcall( KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty, responseData )
				
				if ok and type( response ) == 'table' and response.code == "SourceDoesNotExist" then
				
					recordSuccess( self, docId, startTime, uploadUrl, binaryType, localSize, responseData, contentType )
					uploadCompleted( self, docId, binaryType, sha256 )
					
					uploadFailed = false
					
				end
				
			end

			if uploadFailed then
				handleUploadFailure( self, docId, binaryType, uploadUrl, responseData, headers, invalidOriginal, contentType, localSize )
			end
			
			return Observable.just( false )

	end )


end

--------------------------------------------------------------------------------

local function copyBinaryOnOz( self, docId, uploadUrl, binaryType, contentType, additionalUploadHeaders, sourceAssetLink )

	logger:debugf( "Copying %s for docId %s from %s to %s", binaryType, docId, sourceAssetLink, uploadUrl )

	local data = '{ "file_href": "' .. sourceAssetLink .. '" }'

	local options = getRequestOptionsFor( binaryType, 'application/json', true, additionalUploadHeaders )
	options.requestCost = WFRxHttpClient.requestCostLow
	options.dataFormat = nil

	return self._catalogSpec:putBinary( uploadUrl, data, options )
		:map( function( response, headers )
			return #data, response, headers
		end )
		:catch( function( responseData, headers )
			
			if errorUploadMasterWhenNotAllowed( binaryType, responseData, headers ) then

				return applyRemoveWillNotUploadMasterOverlay( self, docId )

					:flatMap( function()
	
						-- Throw an error so we retry the upload after the doc store
						-- has had time to push the removal of the original through
						-- the changes feed.
	
						return Observable.throw( responseData, headers, true )
	
					end )

			end
			
			local maybeAccountError = checkForAccountError( self, docId, binaryType, responseData, headers )
		
			if maybeAccountError then
				return maybeAccountError
			end
	
			return Observable.throw( responseData, headers, false )
			
		end )

end

--------------------------------------------------------------------------------

local function copyOneBinary( self, docId, binaryUrl, binaryType, sha256, contentType, addlUploadHeaders, sourceLinkForCopy )

	local startTime = now()
	
	local copyBinaryObservable = copyBinaryOnOz( self, docId, binaryUrl, binaryType, contentType, addlUploadHeaders, sourceLinkForCopy )
	
	copyBinaryObservable = copyBinaryObservable:tap( function( dataLength, response, headers )
			recordSuccess( self, docId, startTime, binaryUrl, binaryType, dataLength, response, contentType, true )
		end,
		nil, -- caller must issue call to handleUploadFailure
		function()
			uploadCompleted( self, docId, binaryType, sha256 )
		end )

	return copyBinaryObservable

end

--------------------------------------------------------------------------------

local function uploadOrCopyOneMasterOrProxy( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, addlUploadHeaders, sha256, isFirstUploadForSha256 )

	return WFCatImportDuplicateFiles.findUploadedAssetWithSha256HintNotDocId( self._catalog, sha256, docId )
		:flatMap( function( sourceAssetId, sourceHasNoMaster )

			if sourceAssetId == nil then

				-- Upload the asset.

				if isFirstUploadForSha256 then

					if localFilePath then

						return uploadOneBinary( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, addlUploadHeaders, sha256 )

					else

						-- Copied an asset that didn't have a master and source was deleted before we could upload it.

						return applyInvalidOriginalOverlay( self, docId, nil, binaryUrl, "CopyFailedSourceMissing" )
							:finally( function()
								-- delay can be shorter as we won't re-issue any http request if we
								-- didn't wait long enough, we will just end up back here.
								delayedClearFromRecentlyUploaded( self, "upload failed, wait for docstore update as CopyFailedSourceMissing", docId, binaryType, self._waitForCopyFailedSeconds )
							end )

					end

				else
				
					delayedClearFromRecentlyUploaded_NoUpload( self, "Uploading first asset with same sha256, delay so we can copy later", docId, binaryType )
					return Observable.just( true )

				end

			end

			-- Copy the asset.
			return self._docStore:docValue( sourceAssetId )
			:flatMap( function( sourceDocValue )

				-- SourceAssetLink may be nil if the source asset has an invalid original.
				local sourceAssetLink
				if isMaster( binaryType ) then
					sourceAssetLink = WFCatFileSpec.getOriginalLinkFromDocValue( sourceDocValue )
				else
					sourceAssetLink = WFCatFileSpec.getProxyLinkFromDocValue( sourceDocValue )
				end
				
				if sourceAssetLink == nil then

					if localFilePath then

						return uploadOneBinary( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, addlUploadHeaders, sha256 )

					else

						-- Copied an asset that didn't have a master that was deleted before we could upload it.
						return applyInvalidOriginalOverlay( self, docId, nil, binaryUrl, "CopyFailedSourceMissing" )
							:finally( function()
								delayedClearFromRecentlyUploaded( self, "copy failed, wait for docstore update CopyFailedSourceMissing", docId, binaryType )
							end )

					end

				else

					return copyOneBinary( self, docId, binaryUrl, binaryType, sha256, contentType, addlUploadHeaders, sourceAssetLink )
					:tapOnError( function( responseData, headers )
						handleUploadFailure( self, docId, binaryType, binaryUrl, responseData, headers, false, contentType, localFileSize )
					end )

				end

			end )

		end )

end

--------------------------------------------------------------------------------

local function maybePerformanceWrap( id, observable )

	id = id or 'unknown'
	
	return WFPerformanceReport.wrapObservable( 'WFCatBinaryUploader: ' .. id, observable )

end 

-----------------------------------------------------------------------------------------------------

local function getBinaryFullPathAndAssetOfflineStatus( self, localFileVolumeGuid, localFileRelativePath, localFilePath, binaryType )

	local binaryFullPath
	-- assetIsOffline status applicable only to Desktop. for other platforms simply return false. 
	local assetIsOffline = false

	if localFileVolumeGuid and localFileRelativePath then

		local success, filePath = pcall( 
			WFFileUtils.createFullPathFromRelativePathAndGuid,
			localFileRelativePath, localFileVolumeGuid )
			
		if success and filePath then
			binaryFullPath = localFilePath or filePath
		else
			assetIsOffline = true
		end

	elseif isMobileClient( self ) then

		local fileFullPath = require 'LrmAssetUtils'.createFullPath( self._catalogBasePath, localFilePath, binaryType )
		if fileFullPath then
			binaryFullPath = fileFullPath
		end

	end

	return binaryFullPath, assetIsOffline

end

-------------------------------------------------------------------------------------------------------

local function deleteAsset( self, docId )

	logger:debugf( "deleteAsset: delete the docId = %s", docId )
	
	return self._docStore:applyOverlay {
		_id = docId,
		_deleted = true
	}
	
end

---------------------------------------------------------------------------------

local function handleMasterNotExistsLocally( self, docId, binaryUrl, binaryType, contentType )
	
	logger:warnf( "Trying to upload %s for %s but original is missing.", binaryType, docId )
	
	return self._docStore:docValue( docId )
		:flatMap( function( docValue )
		
			local reason = 'binary_upload_missing_original'
			
			if WFPlatformInfo.isiOS then

				WFAnalyticsTrackEventNotifyFailure( self, docId, binaryUrl, nil, nil, binaryType, contentType, nil, nil, reason )
				return deleteAsset( self, docId )
				
			elseif  WFPlatformInfo.isAndroid then

				local localUrl = derefOpt( docValue, '_localOnly', 'localUrl' )
				local localAssetId = derefOpt( docValue, 'importSource', 'localAssetId' )
				local uniqueDeviceId = derefOpt( docValue, 'importSource', 'uniqueDeviceId' )

				if localAssetId and ( uniqueDeviceId == WFDeviceInfo.uniqueDeviceId() ) then
					localUrl = localAssetId
				end

				if localUrl == nil then
					return Observable.just( true );
				end

				if not WFFileUtils.exists( localUrl ) then -- marking asset as deleted as file locally was deleted
					WFAnalyticsTrackEventNotifyFailure( self, docId, binaryUrl, nil, nil, binaryType, contentType, nil, nil, reason )
					return deleteAsset( self, docId )
				end

				-- In Android user has to enable permission for storage, camera, location etc.
				-- It may so happen that these permissions may not be enabled.
				-- And also WFCrypto.digestFile() assumes it will be invoked on files that has read permission.
				-- So better check before invoking
				local isFileReadable = WFFileUtils.isFileReadable( localUrl )
				
				if not isFileReadable then
					logger:debugf( "File exist but no read permission. leave it in queue" )
					return Observable.just( true )
				end
				
				local sha256 = docValue.importSource and docValue.importSource.sha256
				sha256 = sha256 and string.upper( sha256 )

				local sha256Digest = WFCrypto.digestFile(localUrl, 'SHA256')
				local expectedSHA256 = sha256Digest and string.upper( sha256Digest )

				if expectedSHA256 and sha256 ~= expectedSHA256 then -- marking asset as deleted as file locally was modified
					WFAnalyticsTrackEventNotifyFailure( self, docId, binaryUrl, nil, nil, binaryType, contentType, nil, nil, reason )
					return deleteAsset( self, docId )
				end

				return self._docStore:applyOverlay {
					_id = docId,
					_localOnly = {
						originalFilePath = localUrl,
					}
				}
				
			end
	
		end )

end

-------------------------------------------------------------------

local function resetDetailedProgressReportCounters( self )

	if not self._uploadProgress then
		return
	end

	if( self._uploadProgress.totalBytesNeedToUpload ~= 0 and
		self._uploadProgress.totalBytesUploadedSoFar ~= 0 ) then
		
		-- reset the upload progress counters
		self._uploadProgress.totalBytesUploadedSoFar = 0
		self._uploadProgress.totalBytesNeedToUpload = 0

		local binaryUploadProgress = {
			totalBytesNeedToUpload = self._uploadProgress.totalBytesNeedToUpload,
			totalBytesUploadedSoFar = self._uploadProgress.totalBytesUploadedSoFar,
		}
		self._uploadProgress.progressReporter:updateProgressWithInfo( binaryUploadProgress, true )
		
	end

end

--------------------------------------------------------------------------

local function sendFinalDetailedProgressUpdate( self )

	if not self._uploadProgress then
		return
	end

	if( self._uploadProgress.totalBytesNeedToUpload ~= 0 and
		self._uploadProgress.totalBytesUploadedSoFar ~= 0 and
		self._uploadProgress.totalBytesNeedToUpload - self._uploadProgress.totalBytesUploadedSoFar >= 0 ) then
		
		local binaryUploadProgress = {
			totalBytesNeedToUpload = self._uploadProgress.totalBytesNeedToUpload,
			totalBytesUploadedSoFar = self._uploadProgress.totalBytesUploadedSoFar,
		}
		self._uploadProgress.progressReporter:updateProgressWithInfo( binaryUploadProgress, true )
		resetDetailedProgressReportCounters( self )
		
	end
	
end
------------------------------------------------------------------

local function uploadSomeBinaries( self )
	
	local typeOfJoin = "LEFT"
	
	if self._onlyUploadSpaceAssets then
		typeOfJoin = "INNER"
	end
	
	local needsUploadQueryStr = 
		string.format( [[SELECT needsBinaryUpload.docId, localFilePath, uploadUrl, binaryType, contentType, 
								uploadHeaders, isBlacklisted, sha256, fileSize, sha256Error, 
								sourceLinkForCopy, localVolumeGuid, localRelativePath
						FROM needsBinaryUpload
						%s JOIN space ON space.docId == needsBinaryUpload.spaceId
						WHERE( uploadUrl IS NOT NULL )
						ORDER BY fileMaybeOnOfflineVolume ASC, isBlacklisted ASC,
									priorityForUpload DESC, importDate ASC, captureDate ASC 
						LIMIT 100]], typeOfJoin )

	local needsUploadQuery =
			self._catalog:indexStore():query {
				sql = needsUploadQueryStr,
				quiesceUntilUpToDate = true,
			}

	needsUploadQuery = maybePerformanceWrap( 'query', needsUploadQuery )

	-- There's potentially for a race condition where the query might be refreshed
	-- before we've determined on this end that the upload has finished. When this happens,
	-- the below rx chain will do no work since we still think we have the maximum
	-- number of uploads in flight. This can potentialy cause sync to freeze. In
	-- order to remedy this, we use combineLatest to cause the rx chain to fire on
	-- query refresh, as well as when an upload finishes.
	
	return Observable.combineLatest(
			needsUploadQuery,
			self._uploadFinished,
			self._volumeStatusHasChanged )
			
	:flatMapLatest( function( binariesToUpload )

		return self._continuousUploadTask:throwIfNotRunning( binariesToUpload )
			-- Need to throw as uploadSomeBinaries doesn't ever complete.

	end )

	:tap( function( binariesToUpload )

--		logger:debugf( "uploadSomeBinaries raw query result = %s", binariesToUpload )
		logger:debugf( "uploadSomeBinaries raw query result = %s", #binariesToUpload )

		if #binariesToUpload == 0 then
		 	sendFinalDetailedProgressUpdate( self )
			self._continuousUploadTask:taskDidWork( false )
				-- Since we don't typically send onCompleted events we need to
				-- indicate when we don't do work.
		end

		local didWork = false

		for _, binaryToUpload in ipairs( binariesToUpload ) do	

			if not self._continuousUploadTask:isRunning() then break end

			local docId = binaryToUpload[ 1 ]
			local localFilePath = binaryToUpload[ 2 ]
			local binaryUrl = binaryToUpload[ 3 ]
			local binaryType = binaryToUpload[ 4 ]
			local contentType = binaryToUpload[ 5 ]
			local encodedUploadHeaders = binaryToUpload[ 6 ]
			local isBlacklisted = intToBool( binaryToUpload[ 7 ] )
			local sha256 = binaryToUpload[ 8 ]
			local localFileSize = binaryToUpload[ 9 ]
			local sha256Error = intToBool( binaryToUpload[ 10 ] )
			local sourceLinkForCopy = binaryToUpload[ 11 ]
			local localFileVolumeGuid = binaryToUpload[ 12 ]
			local localFileRelativePath = binaryToUpload[ 13 ]
			
			if isBlacklisted then
			
				-- Check if the asset is actually blacklisted.
				isBlacklisted = self._failedUploadBlacklistHelper:maybeResetBlacklistStatus( docId )
			
			end

			-- ensure localFilePath is a string and not bad data
			localFilePath = type( localFilePath ) == 'string' and localFilePath or nil

			local binaryFullPath, assetIsOffline = getBinaryFullPathAndAssetOfflineStatus( self, localFileVolumeGuid, localFileRelativePath, localFilePath, binaryType )

			if binaryFullPath then
				localFilePath = binaryFullPath
			end

			local binaryExistsLocally = localFilePath and WFFileUtils.exists( localFilePath )
				-- Though we could check that the original exists in the indexer, it's
				-- more efficient to check for it here since it won't run all items,
				-- but only the items before we've reached maxOriginalUploadsInFlight.
				
			if self._binariesInFlight < maxScheduledUploads and
				not getRecentlyUploadedSha256( self, docId, binaryType ) and
				not isBlacklisted
				and ( binaryExistsLocally or -- Standard Uploads
						sourceLinkForCopy or -- Copies + InAppContribute assets
						( WFPlatformInfo.isDesktop and localFilePath == nil and not assetIsOffline ) or  -- SPAM [matthewj 2019-09-25]: This case covers edge cases like testUploadCopy_sourceDeleted_noMasterLocal 
																										-- but these if cases should be restructured significantly to each handle fewer cases and this one should
																										-- be moved out of this if block. This is related to story WF-685
						( WFPlatformInfo.isMobile( )  and localFilePath == nil  and  sourceLinkForCopy) ) -- SPAM [matthewj 2019-09-27]: this allows copies/IAC through but prevents a special Android case (WF-501) from 
																										-- falling in and instead those should end up getting caught in the not ( binaryExistsLocally or sourceLinkForCopy )
																										-- case. This is related to story WF-686/WF-689 and to a lesser extent WF-685
				and getUploadingAssetsWithSha256( self, sha256 ) < maxScheduledUploads -- Prevent creating a large blacklist while uploading first asset if many copies exist, but allow parallel copies to occur.
				then
		 
				local uploadHeaders
				if encodedUploadHeaders then
					uploadHeaders = WFMsgPack.unpack( encodedUploadHeaders )
				end
				
				if localFilePath and not localFileSize then
					local fileAttributes = WFFileUtils.fileAttributes( localFilePath )
					localFileSize = fileAttributes.fileSize
				end
			
				local isFirstUploadForSha256 = getUploadingAssetsWithSha256( self, sha256 ) == 0

				addToRecentlyUploaded( self, docId, binaryType, sha256 or true )

				local subscription = SingleAssignmentDisposable()

				addToActiveUploads( self, docId, binaryType, subscription, localFileSize )

				local function removeUploadSubscription()

					-- Don't call uploadFinished from finally as it could be the result of
					-- dispose being called from disposeActiveUploads and we don't want to
					-- start other uploads in this case.
					removeFromActiveUploads( self, docId, binaryType, localFileSize )

				end

				local function uploadCompletedOrError()

					-- Call removeFromActiveUploads here (in addition to the finally) so that _binariesInFlight
					-- is reduced so the following uploadFinished call can trigger another upload as 
					-- finally is called after the completed not before it.
					removeFromActiveUploads( self, docId, binaryType, localFileSize )

					-- Trigger combineLatest above to start other uploads (in case _binariesInFlight reached maxScheduledUploads).
					uploadFinished( self )

					if self._binariesInFlight == 0 then
						self._idleStatus:onNext( true )
					end

				end

				didWork = true
				
				local action

				if sha256Error then

					local function delayedClearFromRecentlyUploadedFn()
						delayedClearFromRecentlyUploaded( self, "waiting for resolveSha256ErrorInInputSource doc update", docId, binaryType )
					end

					-- Handle asynchronous error from a prior chunked upload that is sent via catalog metadata.
					action = resolveSha256ErrorInInputSource( self, binaryUrl, localFilePath, docId, sha256 )
						:finally( delayedClearFromRecentlyUploadedFn )

				elseif sourceLinkForCopy ~= nil then

					-- Asset was copied and the original is already up on oz. Using the copy rather 
					-- than the local binary should be preferred because it will almost always be --
					-- faster than uploading the file directly.
					
					action = copyOneBinary( self, docId, binaryUrl, binaryType, sha256, contentType, uploadHeaders, sourceLinkForCopy )
					
						:catch( function( responseData, headers )

							local resourceDoesNotExist, newId = errorAndMaybeNewIdIfMasterOrProxyResourceDoesNotExist( binaryType, responseData, headers )
							
							if resourceDoesNotExist then

								return applyOrRemoveNewSourceLinksOrErrorOverlayForCopy( self, docId, newId, binaryType )
									
									:map( function( appliedOverlay )
										if appliedOverlay then
											delayedClearFromRecentlyUploaded( self, "updated or removed sourceAsset for copy", docId, binaryType )
										else
											handleUploadFailure( self, docId, binaryType, binaryUrl, responseData, headers, false, contentType, localFileSize )
										end
									end )

							end

							handleUploadFailure( self, docId, binaryType, binaryUrl, responseData, headers, false, contentType, localFileSize )
							
							return Observable.throw( responseData, headers )
							
						end )

				elseif isMaster( binaryType ) or isProxy( binaryType ) then
				
					-- proxy addresses: lightroom asset is copied while paused, then the 
					-- copy is copied (results in no sourceLinkForCopy)
					action = uploadOrCopyOneMasterOrProxy( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, uploadHeaders, sha256, isFirstUploadForSha256 )

				else

					action = uploadOneBinary( self, docId, localFilePath, localFileSize, binaryUrl, binaryType, contentType, uploadHeaders, sha256 )

				end

				subscription:setDisposable(
					action
						:finally( removeUploadSubscription )
						:subscribe( nil, uploadCompletedOrError, uploadCompletedOrError )
				)

				if self._binariesInFlight >= maxScheduledUploads then break end

			elseif not ( binaryExistsLocally or sourceLinkForCopy ) then
		 		
				if binaryType == 'xmp_sidecar' and not sourceLinkForCopy then
				
					-- Xmp sidecar files are always stored on the main volume. If the
					-- xmp sidecar file doesn't exist, then we need to delete the
					-- file annotation.

					addToRecentlyUploaded( self, docId, binaryType, sha256 or true )

					removeSidecarFileAnnotation( self, docId, binaryType )
						:finally( function()
							delayedClearFromRecentlyUploaded( self, "removing sidecar annotation", docId, binaryType )
						end )
						:call( runUntilDone )

				elseif binaryType == 'master' then
				
					if WFPlatformInfo.isMobile() then
						handleMasterNotExistsLocally( self, docId, binaryUrl, binaryType, contentType )
							:call( runUntilDone )

					elseif localFilePath then -- localFilePath can be nil doing a copy for desktop.

						logger:warnf( "Uh oh. Trying to upload %s for %s but it doesn't exist at path %s.", binaryType, docId, localFilePath )

						if isMaster( binaryType ) and not assetIsOffline then

							applyInvalidOriginalOverlay( self, docId, localFilePath, binaryUrl, 'UploadFailedMissingOriginal' )
								:finally( function()
									delayedClearFromRecentlyUploaded( self, "binary missing, marked invalid if volume not offline", docId, binaryType )
								end )
								:call( runUntilDone )
						end

					end
				
				end
				
			end

		end
		
		if didWork then
			self._continuousUploadTask:taskDidWork()
		end
		
	end )

	:catch( function( errorMessage )

		logger:debugf( "uploadSomeBinaries %s", errorMessage )

		if self._binariesInFlight > 0 then

			-- Wait until all the in-flight requests have finished.

			return self._idleStatus
				:take( 1 )
				:flatMap( function()

					self._continuousUploadTask:taskDidWork( false )
					return self._continuousUploadTask:throwIfNotStopped( errorMessage )

				end )

		else

			self._continuousUploadTask:taskDidWork( false )
			return self._continuousUploadTask:throwIfNotStopped( errorMessage )

		end

	end )

end

--============================================================================--

local WFCatBinaryUploader = {

	_catalog = false,
		-- (WFCatalog)

	_docStore = false,
		-- (WFRxDocStore)

	_catalogSpec = false,
		-- (WFCatSpec)

	_binariesInFlight = false,
		-- (number) Number of uploads that are in progress.
		
	_recentlyUploaded = false,
		-- (table) set of document IDs for which we've recently uploaded

	_activeUploads = false,
		-- (table) Table of subscriptions of each active upload.

	_uploadingAssetWithSha256 = false,
		-- (table) Table of counts of the number of assets attempted to be uploaded indexed by sha256.

	_numAssetsWaitingForUploadWithSha256 = false,
		-- (table) Table of counts of the number of assets waiting for upload to finish to be uploaded indexed by sha256.
		-- used to limit the number of times uploadSomeBinaries is called when multiple copies are waiting for the first
		-- asset to finish uploading.

	_continuousUploadTask = false,
		-- (WFRxContinuousTask) A continuous task to handle continuous uploads

	_errorStatus = false,
		-- (Subject) A Rx.Subject used to relay upload error status

	_idleStatus = false,
		-- (Subject) A Rx.Subject used to indicate idle status.

	_uploadFinished = false,
		-- (Rx BehaviorSubject) A local stream that fires when an upload completes.
	
	_volumeStatusHasChanged = false,
		-- (Rx BehaviorSubject) A local stream that fires when a volumes status changes.
	
	_failedUploadBlacklistHelper = false,
		-- (WFCatBlacklistHelper) A helper classed used to temporarily blacklist failed uploads.

	_chunkedUploader = false,
		-- (WFCatContentRangeUploader) Used to upload large files in chunks.

	_timerAborter = false,
		-- (ReplaySubject) Used to indicate the uploader is being closed by sending an error, used to clear out blackout timers.

	_uploadErrorLog = false,
		-- (table) record of recent upload errors for the diagnostic log

	_uploadSpeedStats = false,
		-- (table) record of recent upload speeds

	_clearFromRecentlyUploadedSeconds = false,
		-- (number) Delay in seconds before clearing the recently uploaded table.

	_waitForOtherUploadSeconds = false,
		-- (number) Delay in seconds before clearing the recently uploaded table waiting for another upload to finish that has the same sha256.

	_waitForCopyFailedSeconds = false,
		-- (number) Delay in seconds before clearing the recently uploaded table after marking an asset with copyFailedSourceMissing.

	_timer = false,
		-- (timer) Use by unit tests to avoid warnings generated by WFRxLowFidelityTimer.

	_catalogBasePath = false,
	-- (string) catalog base path

	_uploadProgress = false,
		-- (table) 
			-- (WFProgressReporter) progressReporter instance 
			-- (number) totalBytesNeedToUpload
			-- (number) totalBytesUploadedSoFar

	_onlyUploadSpaceAssets = false,
		-- (boolean) If true, the uploader will ignore catalog-feed assets.

	class = {
		generateIndexSchema = generateIndexSchema,
	},

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function WFCatBinaryUploader:__init( catalog, chunkedUploader, optTaskDelays, optCatSpec, optUploaderOptions )

	check( 'WFCatBinaryUploader', 'WFCatalog', '?table', '?table', '?WFCatSpec', '?table' )

	optUploaderOptions = optUploaderOptions or {}

	self:setOwned {
		_catalog = catalog:relayObject()
	}
	
	self._docStore = catalog:docStore()

	self._catalogSpec = optCatSpec or catalog:ozCatalogSpec()

	self._failedUploadBlacklistHelper = WFCatBlacklistHelper( 
												self._docStore, 
												'WFCatBinaryUploader',
												numberOfTriesBeforeBlacklistingUpload,
												failedUploadBlackoutWindowInSeconds,
												true )

	-- Applicable only for mobile clients
	if WFPlatformInfo.isMobile() and optUploaderOptions.deriveCatalogBasePath then

		local OzFilePaths = require 'OzFilePaths'
		self._catalogBasePath = not not optCatSpec and not not self._catalogSpec and OzFilePaths.getCatalogBasePath( self._catalogSpec._catalogId )
	end

	if optUploaderOptions.progressReporter then
		
		self._uploadProgress = {
			progressReporter = optUploaderOptions.progressReporter,
			totalBytesNeedToUpload = 0,
			totalBytesUploadedSoFar = 0,
		}
		
		local binaryUploadProgress = {
			totalBytesNeedToUpload = self._uploadProgress.totalBytesNeedToUpload,
			totalBytesUploadedSoFar = self._uploadProgress.totalBytesUploadedSoFar,
		}
		-- Trigger once during the startup to fire the upload progress RX chain.
		-- This will help to display pending uploads if any during restart of the app
		self._uploadProgress.progressReporter:updateProgressWithInfo( binaryUploadProgress, true )
		
	end

	self._chunkedUploader = chunkedUploader or
		WFCatContentRangeUploader( catalog:docStore(), self._catalogSpec, nil, nil, nil, self._uploadProgress )

	self._recentlyUploaded = {}

	self._activeUploads = {}

	self._binariesInFlight = 0

	self._uploadingAssetWithSha256 = {}
	self._numAssetsWaitingForUploadWithSha256 = {}

	self._onlyUploadSpaceAssets = false

	local continuousTaskDelays = optTaskDelays or {
		progressiveBackoff = true,
		errorDelay = 10,
		maxDelay = 60,
	}
	
	local continuousTask = function() 
		return uploadSomeBinaries( self ) 
	end

	self._continuousUploadTask = WFRxContinuousTask( "WFCatBinaryUploader", continuousTask, continuousTaskDelays )

	self._errorStatus = Rx.Subject()

	self._idleStatus = Rx.Subject()

	self._uploadFinished = Rx.BehaviorSubject()
	
	self._volumeStatusHasChanged = Rx.BehaviorSubject()
	
	self._timerAborter = Rx.ReplaySubject( 1 )
	self._timerAborter:onNext( true )

	self._uploadErrorLog = {}

	self._uploadSpeedStats = {
		buckets = {},
		totalSize = 0,
	}

	self._clearFromRecentlyUploadedSeconds = optUploaderOptions.clearFromRecentlyUploadedSeconds or kClearFromRecentlyUploadedSeconds
	self._waitForOtherUploadSeconds = optUploaderOptions.waitForOtherUploadSeconds or kWaitForOtherUploadSeconds
	self._waitForCopyFailedSeconds = optUploaderOptions.waitForCopyFailedSeconds or kWaitForCopyFailedSeconds

	self._timer = optUploaderOptions.timer or WFRxLowFidelityTimer.timer

	WFDiagnosticLog.addLogProvider( self )

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:close()

	check( 'WFCatBinaryUploader' )

	disposeActiveUploads( self )

	local errorStatus = self._errorStatus
	self._errorStatus = nil
	errorStatus:dispose()

	local continuousUploadTask = self._continuousUploadTask
	self._continuousUploadTask = nil
	
	if continuousUploadTask then
		continuousUploadTask:close()
	end

	local failedUploadBlacklistHelper = self._failedUploadBlacklistHelper
	self._failedUploadBlacklistHelper = nil
	
	if failedUploadBlacklistHelper then
		failedUploadBlacklistHelper:close()
	end

end

--------------------------------------------------------------------------------	

function WFCatBinaryUploader:stopContinuousUploads()

	self._continuousUploadTask:stopContinuousTask()

	disposeActiveUploads( self )

	uploadFinished( self )
		-- trigger combineLatest in uploadSomeBinaries so it can see it is stopped.

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:startContinuousUploads()

	if self._onlyUploadSpaceAssets then
		self:stopContinuousUploads()
	end

	self._onlyUploadSpaceAssets = false

	self._continuousUploadTask:startContinuousTask()

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:startContinuousUploadOfOnlySpaceAssets()

	if not self._onlyUploadSpaceAssets then
		self:stopContinuousUploads()
	end

	self._onlyUploadSpaceAssets = true

	self._continuousUploadTask:startContinuousTask()

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:errorStatus()

	return self._errorStatus

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:isDoingWork()

	return self._continuousUploadTask:isTaskDoingWork()

end

--------------------------------------------------------------------------------

--[[!

	\brief Report the status of the Nimbus Volumes
	
	\return (void)
	
--]]

function WFCatBinaryUploader:reportStatusOfVolumes( statusOfVolumes )

	check( 'WFCatBinaryUploader', 'table' )
	
	-- When a volume's status changes, namely when a volume is reconnected
	-- to Nimbus, there's a chance that we'll be able to upload binaries
	-- that were previously not accessible. 
	
	-- The query WFCatBinaryUploader uses does not filter out assets that
	-- are on offline volumes, unlike WFCatIdleDownloader, because we also
	-- support uploading "copies" that have no local file.
	
	-- Instead, we prioritize uploading the files that are known to be connected,
	-- as well as copies, and then check to see if the possibly-disconnected
	-- files can be accessed and uploaded.
	
	-- The following statement guarentees we process the disconnected -> connected
	-- files once a volume comes online. The value here is unimportant, but making
	-- it unique can help in debugging.
	
	self._volumeStatusHasChanged:onNext( KSDate.timeInIso8601Format() )
	
end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:DEBUG_getDiagnosticLogErrors()
	return logger.DEBUG_makeDiagnosticLogErrorsHtmlText()
end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:DEBUG_resetDiagnosticLogErrors()
	return logger.DEBUG_resetDiagnosticLogErrors()
end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:sendDiagnosticLogData()

	local finalStats = {}

	local activeUploads = {}
	for docId, uploadsForDocId in pairs( self._activeUploads ) do
		for binaryType, subscription in pairs( uploadsForDocId ) do
			table.insert( activeUploads, { id = docId, binaryType = binaryType } )
		end
	end

	for bucket, stats in pairs( self._uploadSpeedStats.buckets ) do
		finalStats[ bucket ] = {
			count = stats.count,
			timings = { stats.timings:visualize 'boxAndWhisker' },
			size = { stats.sizes:visualize 'boxAndWhisker' },
			speed = { stats.speed:visualize 'boxAndWhisker' },
		}
	end

	return {

		name = "Binary Uploader",
		id = 'WFCatBinaryUploader',
		weight = 510,
		parentSection = WFDiagnosticLog.Sections.SYNC,
		
		body = {
			activeUploads = activeUploads,
			binariesInFlight = self._binariesInFlight,
			recentErrors = self._uploadErrorLog,
			uploadHttpRequestCost = uploadHttpRequestCost,
			uploadSpeedStats = finalStats,
			totalBinarySize = self._uploadSpeedStats.totalBinarySize,
			chunkFileAtSize = self._chunkedUploader:chunkFileSize(),
			chunkSize = self._chunkedUploader:chunkSize(),
		},

	}

end

--------------------------------------------------------------------------------

function WFCatBinaryUploader:DEBUG_uploadCount( bucket )

	local bucketStats = self._uploadSpeedStats.buckets[ bucket ]

	return derefOpt( bucketStats, 'count' ) or 0

end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatBinaryUploader'( WFCatBinaryUploader )
