--[[----------------------------------------------------------------------------

WFCatBlacklistHelper.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.logger
local WFLogger = require 'WFLogger'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local KSData = require 'KSData'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxLowFidelityTimer = require 'WFRxLowFidelityTimer'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
KSData.registerArgumentChecker()

local runUntilDone = WFRxRunUntilDone

local logger = WFLogger 'WFCatBlacklistHelper'
--:setLevel( 'debug' )

local removeLocalIdFromBlacklist -- forward declaration


--============================================================================--

--[[!

	\class WFCatBlacklistHelper
	
	\brief This class can be used to temporarily blacklist assets. You can choose
	to ether blacklist the assets in a local table, or by annotating the localOnly\
	section of the docstore.
		
]]--

-- @mark WFCatBlacklistHelper class

local WFCatBlacklistHelper = {

	_docStore = false,
		-- (WFRxDocStore)
	
	_blacklistName = false,
		-- (string) Unique name given to this blacklist instance. Optionally
		-- used in doc store annotations.
	
	_numberOfFailuresBeforeBlacklisting = false,
		-- (number) of failures before blacklisting an Id.

	_failedBlackoutWindowInSeconds = false,
		-- (number) of seconds an Id remains blacklisted.
	
	_annotateDocStore = false,
		-- (boolean) if true, we annotate the doc store with blacklist status
		-- so we can check if an asset is blacklisted when indexing.
	
	_failureCountById = false,
		-- (table) Lookup table used for tracking failures.
		-- key: ID
		-- value: Failure Count
		
	_localBlacklist = false,
		-- (table) Lookup table for blacklisted Ids.
		-- key: ID
		-- value: true
		
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

--[[!

	\brief Creates a new WFCatBlacklistHelper instance.

	\param catalog (WFRxDocStore) A valid WFRxDocStore object.
	
	\param blacklistName (string) Unique name given to this blacklist instance.
	
	\param numberOfFailuresBeforeBlacklisting (number) of failures before blacklisting an id.
	
	\param failedBlackoutWindowInSeconds (number) of seconds an Id remains blacklisted.
	
	\param annotateDocStore (boolean) if true, we annotate the doc store with blacklist status
		so we can check if an asset is blacklisted when indexing.
	
	\return WFCatBlacklistHelper object
	
--]]

function WFCatBlacklistHelper:__init( docStore, blacklistName, numberOfFailuresBeforeBlacklisting, failedBlackoutWindowInSeconds, annotateDocStore )

	check( 'WFCatBlacklistHelper', 'WFRxDocStore', 'string', '?number', '?number', '?boolean' )

	self._docStore = docStore

	self._blacklistName = blacklistName

	self._numberOfFailuresBeforeBlacklisting = numberOfFailuresBeforeBlacklisting or 3

	self._failedBlackoutWindowInSeconds = failedBlackoutWindowInSeconds or 30

	self._annotateDocStore = annotateDocStore or false

	self._failureCountById = {}

	self._localBlacklist = {}

end

--------------------------------------------------------------------------------

--[[!

	\brief Close this istance.

	\return (void)
	
--]]

function WFCatBlacklistHelper:close()

	check( 'WFCatBlacklistHelper' )
	
	-- Clear all blacklist annotations
	
	for id, _ in pairs( self._localBlacklist ) do
		
		removeLocalIdFromBlacklist( self, id )
	
	end

end
	
--------------------------------------------------------------------------------

removeLocalIdFromBlacklist = function( self, localId, forceClear )

	self._localBlacklist[ localId ] = nil
	self._failureCountById[ localId ] = nil
	
	if self._annotateDocStore or forceClear then
	
		local blacklist = {}
		blacklist[ self._blacklistName ] = { _delete = true }

		if not WFCloseSupport.hasBeenOrWillBeClosed( self._docStore  ) then
		
			self._docStore:applyOverlay(
				{	
					_id = localId,
					_localOnly = { blacklistedBy = blacklist }
				} )
					
				:call( runUntilDone )
			
		end
			
	end
		
end

--------------------------------------------------------------------------------

local function blacklistLocalId( self, localId, blackoutTime )

	local result = Observable.just( localId )
	
	if not self._localBlacklist[ localId ] then
		
		logger:infof( "Blacklisting Id: %s", localId )

		self._localBlacklist[ localId ] = true
		
		if self._annotateDocStore then

			local blacklist = {}
			blacklist[ self._blacklistName ] = true

			result = self._docStore:applyOverlay(
				{	
					_id = localId,
					_localOnly = { blacklistedBy = blacklist }
				} )

		end
		
		-- Clear from the blacklist after a set amount of time.
			
		WFRxLowFidelityTimer.timer( self._failedBlackoutWindowInSeconds )
			:flatMapLatest( function() 
				logger:debugf( "Un-blacklisting Id: %s", localId )
				return removeLocalIdFromBlacklist( self, localId ) 
			end )
			:call( runUntilDone )
			
	else
		
		logger:infof( "Id %s is already blacklisted", localId )
	
	end

	return result:call( runUntilDone )

end

--------------------------------------------------------------------------------

function WFCatBlacklistHelper:maybeBlacklist( localId )

	check( 'WFCatBlacklistHelper', 'number|string' )

	local maybeFailureCount = self._failureCountById[ localId ]
	local newFailureCount = maybeFailureCount and maybeFailureCount + 1 or 1
	local isBlacklisted = false
	
	if newFailureCount >= self._numberOfFailuresBeforeBlacklisting then
	
		blacklistLocalId( self, localId )
		isBlacklisted = true
		
	else
	
		self._failureCountById[ localId ] = newFailureCount
		
	end
		
	return isBlacklisted

end

--------------------------------------------------------------------------------

function WFCatBlacklistHelper:addToBlacklist( localId )

	check( 'WFCatBlacklistHelper', 'number|string' )
	return blacklistLocalId( self, localId )

end

--------------------------------------------------------------------------------

function WFCatBlacklistHelper:isBlacklisted( localId )

	check( 'WFCatBlacklistHelper', 'number|string' )

	return not not self._localBlacklist[ localId ]
	
end

--------------------------------------------------------------------------------

function WFCatBlacklistHelper:maybeResetBlacklistStatus( localId )

	check( 'WFCatBlacklistHelper', 'number|string' )

	local blacklisted = true

	if not self._localBlacklist[ localId ] then
	
		-- If the asset is annotated as blacklisted but for some reason we're not tracking it,
		-- then something went wrong. This could potentially happen if the asset is blacklisted
		-- and then the app restarts. The annotation will remain in the doc store, but the
		-- _localBlacklist table will be reset. In this case, we simply remove the id from the
		-- blacklist.
		
		removeLocalIdFromBlacklist( self, localId, true ) 
	
		blacklisted = false
	
	end
	
	return blacklisted
	
end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatBlacklistHelper'( WFCatBlacklistHelper )
