--[[----------------------------------------------------------------------------

WFCatImportFileQueue.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.os
local WFFileUtils = require 'WFFileUtils'
local WFFolderWalker = require 'WFFolderWalker'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFRequestQueue = require 'WFRequestQueue'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local log = WFLogger 'WFCatImportFileQueue'
--	:setLevel( 'DEBUG' )

--============================================================================--

-- @mark WFCatImportFileQueue class

--------------------------------------------------------------------------------
--[[!
	\brief WFCatFileImportQueue maintains a list of files to be imported.
]]--

local WFCatImportFileQueue = {

	WFRequestQueue, 
		-- Base class

}

--------------------------------------------------------------------------------

function WFCatImportFileQueue:__init()

	check( 'WFCatImportFileQueue' )

end

--------------------------------------------------------------------------------

function WFCatImportFileQueue:dispose()

	check( 'WFCatImportFileQueue' )

	return self:WFRequestQueue_dispose()

end

--------------------------------------------------------------------------------

-- TODO [scouten 2015-06-10]: We will need to bring ACR or ImageCore into the
-- WF (daemon, at least) Real Soon Now. When we do, this list should go away
-- in favor of the official list from cr_sdk.

local whitelistedExtensions = KSTableUtils.mapArrayToSet {

	'TIF',		-- Canon, Kodak
	'TIFF',
	'CRW',		-- Canon
	'NEF',		-- Nikon
	'RAF',		-- Fujifilm
	'ORF',		-- Olympus
	'MRW',		-- Minolta
	'DCR',		-- Kodak
	'MOS',		-- Leaf
	'RAW',		-- Panasonic, Contax
	'PEF',		-- Pentax
	'SRF',		-- Sony
	'DNG',		-- Adobe
	'X3F',		-- Foveon
	'CR2',		-- Canon
	'CR3',		-- Canon
	'ERF',		-- Epson
	'SR2',		-- Sony
	'KDC',		-- Kodak
	'MFW',		-- Mamiya (sample files)
	'MEF',		-- Mamiya (sample camera)
	'ARW',		-- Sony
	'NRW',		-- Nikon
	'RW2',		-- Panasonic
	'RWL',		-- Leica
	'IIQ',		-- Phase One
	'3FR',		-- Hasselblad
	'FFF',		-- Hasselblad
	'SRW',		-- Samsung
	'DXO',		-- DxO
	'HEIC',		-- HEIF
	'ARQ',		-- Sony pixel shift

	'DNG',		-- Adobe
	'PSD',		-- Photoshop
	'JPG',		-- We can convert JPEG to raw so it's a raw extension!
	'JPEG',
	'JPE',		-- alternate JPEG extension: Bug 1324818.
	'PNG',
	
	'MOV',
	'M4V',
	'MP4',
	'MPE',
	'MPEG',
	'MPG4',
	'MPG',
	'AVI',
	'R3D',
	'MTS',
	'3GP',
	'3GPP',
	'M2T',
	'M2TS',

}

--------------------------------------------------------------------------------

local function addFileToQueueIfLegal( self, pathOrTable )

	local path = type( pathOrTable ) == 'table' and pathOrTable.filePath
		or pathOrTable

	local extension = path:match( "%.([^%.]+)$" )

	extension = extension and extension:upper()

	if not whitelistedExtensions[ extension ] then
		log:infof( "not queueing file because unrecognized extension: %s", path )
		return false
	end

	self:WFRequestQueue_addToQueue( pathOrTable )

	return true

end

--------------------------------------------------------------------------------
--[[!
	\brief Add a file or directory to the import queue.

	\param pathOrTable (string|table) If string then it's a path to file or
	directory to add to queue. If table then it contains a path and optional
	importTimestamp field.

	If a directory is specified, it is scanned recursively and all suitable
	matching paths are added.

	If the file extension does not match the import whitelist, the request
	will be silently ignored.
	
	\return Boolean, true if at least one item was legal to be added.

]]--

function WFCatImportFileQueue:addToQueue( pathOrTable )

	check( 'WFCatImportFileQueue', 'string|table' )

	local path = type( pathOrTable ) == 'table' and pathOrTable.filePath
		or pathOrTable

	local fileType = WFFileUtils.exists( path )

	local addedSomething = false

	if fileType == 'file' then

		addedSomething = addFileToQueueIfLegal( self, pathOrTable )

	elseif fileType == 'directory' then

		-- Naive implementation for now. May need to break this up so it doesn't
		-- hog CPU when presented with a large # of files.

		local files = {}
		local nFiles = 0

		for filePath in WFFolderWalker( path ) do
			nFiles = nFiles + 1
			files[ nFiles ] = filePath
		end

		table.sort( files )

		for _, filePath in ipairs( files ) do

			if not addedSomething then
				filePath = {
					filePath = filePath,
					importTimestamp = pathOrTable.importTimestamp
				}
			end

			if addFileToQueueIfLegal( self, filePath ) then
				addedSomething = true
			end

		end

	else

		error "no such file or directory"

	end

	return addedSomething

end

--============================================================================--

return Rx.Class 'WFCatImportFileQueue'( WFCatImportFileQueue )
