--[[----------------------------------------------------------------------------

WFCatImportTask.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'
local Disposable = Rx.Disposable
local Observable = Rx.Observable
local ReplaySubject = Rx.ReplaySubject

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua
local WFTimer = require 'WFTimer'

	-- WF.core.os
local KSDate = require 'KSDate'
local KSPathUtils = require 'KSPathUtils'
local WFFileUtils = require 'WFFileUtils'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local _ = require 'WFArguments_uuid'
local WFCrypto = require 'WFCrypto'
local WFConfiguration = require 'WFConfiguration'
local WFProgressReporter = require 'WFProgressReporter'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.oz_client.managed_catalog
local WFCatImport = require 'WFCatImport'
local WFCatImportConstants = require 'WFCatImportConstants'
local WFCatImportFileQueue = require 'WFCatImportFileQueue'
local WFCatImportReadMetadataTask = require 'WFCatImportReadMetadataTask'

	-- WF.core.performance
local WFPerformanceReport = require 'WFPerformanceReport'

local WFRatioTime = require 'WFRatioTime'

local kDontDuplicateCheck = WFConfiguration.Import.DontDuplicateCheck
local kDontGenerateSha256 = WFConfiguration.Import.DontGenerateSha256


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local logger = WFLogger 'WFCatImportTask'
--	:setLevel 'debug'

local runUntilDone = WFRxRunUntilDone

--============================================================================--

local function getVideoInfo( metadataResults )
	local info = {}
	
	-- frameRate
	if WFRatioTime.isRatioTime( metadataResults.frame_rate ) then
		info.frameRate = { metadataResults.frame_rate.numerator, metadataResults.frame_rate.denominator }
	end
	
	-- duration
	if WFRatioTime.isRatioTime( metadataResults.duration ) then
		info.duration = { metadataResults.duration.numerator, metadataResults.duration.denominator }
	end
	
	-- audioSampleRate
	info.audioSampleRate = tonumber( metadataResults.audioSampleRate )
	
	-- displayAspectRatio
	-- TODO
	
	-- videoBitrate
	-- TODO
	
	-- audioBitrate
	-- TODO
	
	-- audioSampleSize
	-- TODO
	
	-- videoCodec
	-- TODO
	
	-- audioCodec
	-- TODO
	
	-- audioChannel
	-- TODO
	
	-- fieldOrder
	info.fieldOrder = metadataResults.nativeFieldType
	
	-- storageWidth
	-- TODO
	
	-- storageHeight
	-- TODO
	
	if not next( info ) then
		return nil -- don't provide an empty table for video
	end
	
	return info
end

--------------------------------------------------------------------------------

-- @mark file processing loop

--------------------------------------------------------------------------------

local function createImportInfoFromMetadataResults( metadataResults )

	local filePath = metadataResults.filePath

	local additionalInfo = metadataResults.additionalInfo or {}

	local fileAttributes = WFFileUtils.fileAttributes( filePath )

	local fileName = KSPathUtils.leafName( filePath )
	
	local width = metadataResults.width or additionalInfo.fileWidth
	local height = metadataResults.height or additionalInfo.fileHeight

	local croppedWidth
	local croppedHeight

	if additionalInfo.croppedWidth then
		croppedWidth = additionalInfo.croppedWidth
		croppedHeight = additionalInfo.croppedHeight
	else
		croppedWidth = width
		croppedHeight = height
	end
	
	local videoInfo = metadataResults.isVideo and 
		getVideoInfo( metadataResults ) or nil
	
	return {

		xmp = metadataResults.metadata,
			-- TODO [scouten 2015-06-20]: For performance, I want to hand
			-- around a reference to the native XMP object and not the
			-- serialized XMP. This will do for now, though.

		filePath = filePath,
		fileName = fileName,
		fileSize = fileAttributes.fileSize,

		width = width,
		height = height,

		croppedWidth = croppedWidth,
		croppedHeight = croppedHeight,

		xmpCameraRaw = metadataResults.xmpCameraRaw,
		
		isVideo = metadataResults.isVideo,
		additionalInfo = videoInfo or additionalInfo,

	}

end

--------------------------------------------------------------------------------

local maxSimultaneousTasks = 5
	-- Arbitrary constant. Tune with experience.

--------------------------------------------------------------------------------

local function consumeQueue( self )

	assert( not self._queueProcessor, "Already working on consuming queue!" )

	local mrQueue = Rx.Subject()
		-- Metadata request queue. "That's 'MISTER queue' to you!"

	local fileQueue = self._fileQueue
	local progress = self._progress

	local nRequestsOutstanding = 0
	local isDone = false

	self._isBusy:onNext( true )

	local function stop()

		logger:infof( "Stopping queue via %s", ':traceback:' )

		mrQueue:onCompleted()
		progress:done()

		self._queueProcessor = false
		self._isBusy:onNext( false )

		isDone = true

	end

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	local function queueUpMoreRequests( isDelayed )

		-- This code is a bit crufty because Observable:concatMap doesn't actually
		-- simultaneously evaluate the source Observables. We use the publish
		-- assembly to achieve the desired result, which is overlapping evaluation
		-- of the metadata read operations with the original sequence re-established
		-- even if they complete out of order.

		if isDone then return end

		if self:isDisposed() then
			logger:infof( "processor disposed, ending consumeQueue loop" )
			stop()
			return
		end

		for i = 1, maxSimultaneousTasks - nRequestsOutstanding do

			local filePathOrTable = fileQueue:nextRequest()

			local filePath
			if type( filePathOrTable ) == 'table' then
				self._importTimestamp = KSDate.formatDateTimeAsUTC( filePathOrTable.importTimestamp ) .. "Z"
				filePath = filePathOrTable.filePath
			else
				filePath = filePathOrTable
			end

				-- TODO [scouten 2015-06-13]: When queue becomes persisent, need to
				-- be able to dequeue the item without removing it from the queue in
				-- the event that the app crashes or quits before fully processing
				-- this file path. For now, this is OK.

			if filePath == nil then

				-- Wait a little while to see if anything else shows up in the queue.
				-- If not, then we're done.

				if isDelayed then

					logger:infof( "nothing left in queue, ending consumeQueue loop" )
					mrQueue:onCompleted()

				else

					logger:infof( "queue empty, waiting briefly to see if anything new gets added" )

					WFTimer.afterDelayDo( .25, function()
						queueUpMoreRequests( true )
					end )

				end

				return

			end

			logger:infof( "Queued metadata read request for %s", filePath )

			local importTimestamp = self._importTimestamp
			local metadataRequest = WFCatImportReadMetadataTask.readMetadataForFileAtPath( filePath )

				:map( function( metadataResults )
					return metadataResults, importTimestamp
				end )

				:catch( function( message )

					-- We don't want to use Observable.throw here because that would cause
					-- the entire queue below to stop. Just make the system aware of this
					-- failure and continue.

					return Observable.just( {
						failure = WFCatImportConstants.kErrorCantReadMetadata,
						message = message,
						filePath = filePath
					} )

				end )

				:replay( nil, 1 )

			metadataRequest:connect() -- start it now

			mrQueue:onNext( metadataRequest )
			nRequestsOutstanding = nRequestsOutstanding + 1

		end

	end

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	-- Re-establish the original file queue sequence, then add the files to catalog.

	self._queueProcessor = mrQueue

		:concatMap( function( metadataRequest )

			-- This fires once for each metadata request we kicked off earlier,
			-- as soon as possible, but preserving the original sequence.

			local filePath

			return WFPerformanceReport.wrapObservable( 'WFCatImportTask: metadataRequest',
				metadataRequest )

				:flatMapLatest( function( metadataResults, importTimestamp )

					filePath = assert( metadataResults.filePath, "filePath missing from metadataResults" )

					if metadataResults.failure then
						return Observable.throw( metadataResults )
					end

					logger:infof( "Finished metadata read request for %s", filePath )

					return Observable.just( createImportInfoFromMetadataResults( metadataResults ) )

					:flatMapLatest( function( importInfo )

						importInfo.importTimestamp = importTimestamp

						-- Generate the sha256 here so we can do several in parallel.

						if kDontDuplicateCheck then
							importInfo.importIfDuplicate = true
						end

						if kDontGenerateSha256 then

							importInfo.sha256 = 'dont-generate'
							return Observable.just( importInfo )

						end

						return WFPerformanceReport.wrapObservable(
							'WFCatImportTask: digestFileSHA256',
							WFCrypto.asyncDigestFile( filePath, 'SHA256' ) )

						:map( function( sha256 )

							importInfo.sha256 = sha256

							return importInfo

						end )

					end )

				end )

				:catch( function( message )

					if type( message ) == 'table' then
						return Observable.just( message )
					else
						return Observable.just( {
							failure = WFCatImportConstants.kError,
							message = message,
							filePath = filePath
						} )
					end

				end )

		end )

		:map( function( importInfo )

			return Observable.just( importInfo )
				-- Delay the call to createDocumentFromImportInfo until
				-- the flatMerge subscribes to it.

			:flatMapLatest( function( importInfo )

				local function reportError( failureInfo )

					if failureInfo.message == 'matchesExistingFile' then
						logger:debugf( "%s FilePath: %s", failureInfo.message, failureInfo.filePath )
					else
						logger:errorf( "%s FilePath: %s", failureInfo.message, failureInfo.filePath )
					end

					progress:reportErrorMustDeliver( failureInfo )
					return Observable.just( "keep calm and carry on" )
				end

				if importInfo.failure then
					return reportError( importInfo )
				end

				local createDocObservable = WFCatImport.createDocumentFromImportInfo( self._catalog, self._profileCatalog, importInfo, {} )
				return WFCatImport.addDocumentsInBatches( self._catalog, createDocObservable )

				:tap( function( filePath, assetId )

					logger:infof( "imported file at path %s", filePath )
					progress:reportSuccess()

				end )

				:catch( function( failureInfo )
					return reportError( failureInfo )
				end )

			end )

		end )

		:flatMerge( 1 )
			-- Only allow 1 simultaneous call in order to serialize document
			-- creation so we can check if a duplicate is next to another other
			-- in the sequence.

	-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

	self._queueProcessor:subscribe(

		function( filePath ) -- onNext
			nRequestsOutstanding = nRequestsOutstanding - 1
			queueUpMoreRequests()
		end,

		function( err ) -- onError
			logger:errorf( "QUEUE FAILED WITH ERROR: %s", err )
				-- This should NOT happen!
			stop()
		end,

		function() -- onCompleted
    		logger:infof( "queue processing completed" )
			stop()
		end )

	-- Seed with a few overlapping requests.

	queueUpMoreRequests()

end


--============================================================================--

-- @mark WFCatImportTask class

--------------------------------------------------------------------------------
--[[!
	\brief WFCatFileImportProcessor reads files from the import file queue
		and converts them into doc store records.
]]--

local WFCatImportTask = {

	Disposable,

	_catalog = false,
		-- (WFCatalog)
		
	_profileCatalog = false,
		-- (WFCatalog) optional catalog containing the camera and lens profile data

	_fileQueue = false,
		-- (WFCatImportFileQueue)

	_isBusy = false,
		-- (Rx.ReplaySubject< boolean >) true if processing files

	_progress = false,
		-- (WFProgressReporter) used for import progress updates

	_watchingFileQueueHandle = false,
		-- (Rx.Disposable) disposal handle for file queue watcher

	_watchingFileQueueTotalSize = false,
		-- (Rx.Disposable) disposal handle for file queue totalSize watcher

	_queueProcessor = false,
		-- (Observable) strong reference to queue processor

	_importTimestamp = false,
		-- (string) timestamp when import was started

}

--------------------------------------------------------------------------------

function WFCatImportTask:__init( wfCatalog, optProfileCatalog )

	check( 'WFCatImportTask',
		   'WFCatalog' )

	self._catalog = wfCatalog
	self._profileCatalog = optProfileCatalog
	
	self._fileQueue = WFCatImportFileQueue()
	self._progress = WFProgressReporter( "Importing files" )

	self._watchingFileQueueTotalSize = self._fileQueue:totalSize()
		:subscribe( function( value )
			if type( value ) == 'number' then
				self._progress:setTotal( value )
			end
		end )

	local fileQueueHasContent = self._fileQueue:size()
		:map( function( value )
			return type( value ) == 'number' and value > 0
		end )

	self._isBusy = ReplaySubject( 1 )
	self._isBusy:onNext( false )

	local shouldStartProcessing = fileQueueHasContent
		:withLatestFrom( function( _fileQueueHasContent, _isBusy )
			return _fileQueueHasContent and not _isBusy
		end, self._isBusy )
		:distinctUntilChanged()
		:filter( function( value )
			return value
				-- Squelch transition from true to false.
		end )

	self._watchingFileQueueHandle = shouldStartProcessing
		:subscribe( function()
			logger:info( "queue has new content, starting loop" )
			consumeQueue( self )
		end )

end

--------------------------------------------------------------------------------

function WFCatImportTask:dispose()

	check( 'WFCatImportTask' )

	if self._isBusy then
		self._isBusy:onCompleted()
		self._isBusy:dispose()
		self._isBusy = nil
	end

	if self._watchingFileQueueHandle then
		self._watchingFileQueueHandle:dispose()
		self._watchingFileQueueHandle = nil
	end

	if self._watchingFileQueueTotalSize then
		self._watchingFileQueueTotalSize:dispose()
		self._watchingFileQueueTotalSize = nil
	end

	return self:Rx_Disposable_dispose()

end

--------------------------------------------------------------------------------

function WFCatImportTask:isBusy()

	check( 'WFCatImportTask' )

	return self._isBusy

end

--------------------------------------------------------------------------------

function WFCatImportTask:queueSize()

	check( 'WFCatImportTask' )

	return self._fileQueue:size()

end

--------------------------------------------------------------------------------

function WFCatImportTask:import( pathOrArrayOfPaths )

	check( 'WFCatImportTask', 'string|table' )

	if self._progress:isDone() then
		self._progress:reset()
		self._fileQueue:reset()
		self._importTimestamp = nil
	end

	local haveFilesToImport = self._fileQueue:totalSize()
		:map( function( importTotal )
			return importTotal > 0
		end )
		:take( 1 )

	if type( pathOrArrayOfPaths ) == 'string' then
		pathOrArrayOfPaths = { pathOrArrayOfPaths }
	end

	local addedSomething = false
	local importTimestamp = KSDate.timeNow()


	for _, path in ipairs( pathOrArrayOfPaths ) do

		local importInfo = {
			filePath = path,
			importTimestamp = importTimestamp
		}

		if self._fileQueue:addToQueue( importInfo ) then
			addedSomething = true
		end

	end

	haveFilesToImport
		:tap( function( haveFiles )
		
			if not ( addedSomething or haveFiles ) then
			
				-- If nothing gets added we will never call into consumeQueue
				-- and thus never call stop() which calls _progress:done().

				-- Check if something was added during this call in addition
				-- to checking the totalSize observable as it may not be updated
				-- quickly enough when adding a single file.
			
				self._progress:done()
				
			end
			
		end )
		:call( runUntilDone )

	return self._progress:status()

end

--============================================================================--

return Rx.Class 'WFCatImportTask'( WFCatImportTask )
