--[[----------------------------------------------------------------------------

WFCatMetadataPusher.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable
local ReplaySubject = Rx.ReplaySubject

	-- WF.core.logger
local WFLogger = require 'WFLogger'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxDifferenceFilterWithWindowObservable = require 'WFRxDifferenceFilterWithWindowObservable'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFAnalytics = require 'WFAnalytics'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFConfiguration = require 'WFConfiguration'
local WFDiagnosticLog = require 'WFDiagnosticLog'
local WFTableMerge = require 'WFTableMerge'
local WFRxContinuousTask = require 'WFRxContinuousTask'
local WFRxScheduler = require 'WFRxScheduler'
local WFPerformanceReport = require 'WFPerformanceReport'
local KSIdGenerator = require 'KSIdGenerator'
local WFRxLowFidelityTimer = require 'WFRxLowFidelityTimer'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.oz_client.managed_catalog
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatOzErrors = require 'WFCatOzErrors'
local WFCatTrashUtils = require 'WFCatTrashUtils'

	-- WF.oz_client.lrmobile
local OzSyncMetadataMapper

	-- WF.sqlite
local WFRxDocStore = require 'WFRxDocStore'

	-- WF.web.json'
local KSJsonEncoder = require 'KSJsonEncoder'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local derefOpt = KSTableUtils.derefOpt
local runUntilDone = WFRxRunUntilDone
local filterDifferenceWithWindow = WFRxDifferenceFilterWithWindowObservable
local timeNowForAnimation = assert( KSDate.timeNowForAnimation )

local logger = WFLogger 'WFCatMetadataPusher'
--:setLevel( 'info' )

if not WFPlatformInfo.isDesktop then
	OzSyncMetadataMapper = require 'OzSyncMetadataMapper'
end

local isTrashAssetDocument = assert( WFCatIndexUtils.isTrashAssetDocument )
local WFCatTrashUtils_createFromInfo = assert( WFCatTrashUtils.createFromInfo )
local WFCatTrashUtils_isDeleteTrash = assert( WFCatTrashUtils.isDeleteTrash )
local WFCatTrashUtils_isDirectlyDelete = assert( WFCatTrashUtils.isDirectlyDelete )
local WFCatTrashUtils_getCreateWithRevisionId = assert( WFCatTrashUtils.getCreateWithRevisionId )

local kNotEnoughUpdatesError = "Error not enough updates"

--------------------------------------------------------------------------------

local function getContinueFromLocalDocValueKey( isSpacePusher )

	local key = 'Oz-sendUpdatesContinueFrom'

	if isSpacePusher then
	
		key = 'Oz-spaceSendUpdatesContinueFrom'
	
	end
	
	return key

end

--------------------------------------------------------------------------------

local updateSendUpdatesContinueFrom = function( self )
	
	return self._catalog:docStore():localDocValue( self._continueFromLocalDocValueKey )

		:flatMap( function( docValue )
			if self._isSpacePusher and not KSTableUtils.derefOpt( docValue, 'continueFrom' ) then
				return self._catalog:docStore():localDocValue( getContinueFromLocalDocValueKey( false ) )
			else
				return Observable.just( docValue )
			end
		end )
		:map( function( docValue )
			local continueFrom = KSTableUtils.derefOpt( docValue, 'continueFrom' ) or 0

			-- If pusher is closed before we get a localDocValue
			-- _sendUpdatesContinueFrom can be nil-ed out.
			if self._sendUpdatesContinueFrom then
				self._sendUpdatesContinueFrom:onNext( continueFrom )
			end

			logger:infof( "Oz-sendUpdatesContinueFrom -> continueFrom = %s", continueFrom )

			return continueFrom

		end )

end

--------------------------------------------------------------------------------

local function combineChangedDocsWithOz( docStore, changedDocs )

	local docIds = { label = 'oz' }

	for i, docValue in ipairs( changedDocs ) do
		docIds[ i ] = docValue._localId
	end

	return docStore:bulkDocValue( docIds )
		:map( function( docValues )
			return changedDocs, docValues
		end )

end

--------------------------------------------------------------------------------

-- The following document types will not cause logger warnings if they have local
-- changes and are not in the whitelist of accepted document types for a particular
-- metadata pusher instance.

local doNotWarnDocTypes = {
	feedbackuser = true,
}

-- The following document types are supported via the catalog bulk update API on Oz

--------------------------------------------------------------------------------

local createDocTypeWhitelist = {
	asset = true,
	album = true,
	album_asset = true,
	space = true,
	space_album = true,
	comment = true,
	favorite = true,
	stack_asset = true,
	spaceinvite = true,
}

local updateDocTypeWhitelist = {
	catalog = true,
	asset = true,
	album = true,
	album_asset = true,
	space = true,
	stack_asset = true,
	spaceinvite = true,
	space_user = true,
}

local deleteDocTypeWhitelist = {
	asset = true,
	album = true,
	album_asset = true,
	space = true,
	space_album = true,
	spaceinvite = true,
	space_user = true,
	comment = true,
	favorite = true,
	stack_asset = true,
}

--------------------------------------------------------------------------------

-- The following document types are supported via the space-feed bulk update API on Oz

local spaceFeedCreateDocTypeWhitelist = {
	asset = true,
	album_asset = true,
	comment = true,
	favorite = true,
	stack_asset = true,
}

local spaceFeedUpdateDocTypeWhitelist = {
	space_user = true,
}

local spaceFeedDeleteDocTypeWhitelist = {
	asset = true,
	album_asset = true,
	comment = true,
	favorite = true,
	stack_asset = true,
}

--------------------------------------------------------------------------------

local function docUsesSpaceInCloud( doc )

	-- Note that trash assets do not use space in the cloud they will have
	-- subtype 'deleted_image' and 'deleted_video'.
	return not not (  doc and doc.type == 'asset' and
		( doc.subtype == 'image' or doc.subtype == 'video' ) )

end

--------------------------------------------------------------------------------

local album_assetCreateRemapFn = function( self, localDoc, payload, docToSend )

	docToSend.asset = localDoc.asset
	docToSend.album = localDoc.album
	
	payload.asset = nil
	payload.album = nil
	
end

local space_albumCreateRemapFn = function( self, localDoc, payload, docToSend )

	docToSend.space = localDoc.space
	docToSend.album = localDoc.album
	
	payload.space = nil
	payload.album = nil
	
end

local stack_assetCreateRemapFn = function( self, localDoc, payload, docToSend )
	
	docToSend.asset = localDoc.asset
	docToSend.stack = localDoc.stack
	
	payload.asset = nil
	payload.stack = nil
	
end

local commentCreateRemapFn = function( self, localDoc, payload, docToSend )

	docToSend.space = localDoc.space
	docToSend.asset = localDoc.asset
	
	payload.space = nil
	payload.asset = nil
	
end

local favoriteCreateRemapFn = function( self, localDoc, payload, docToSend )

	docToSend.space = localDoc.space
	docToSend.asset = localDoc.asset
	
	payload.space = nil
	payload.asset = nil
	
end

local assetCreateRemapFn = function( self, localDoc, payload, docToSend )

	if docToSend.subtype == 'preset' or docToSend.subtype == 'profile'
								or docToSend.subtype == 'camera_profile'then
		-- for presets/profiles we don't want to let duplicates in to the system
		docToSend.if_none_match = derefOpt( payload, 'importSource', 'sha256' )
	end

	local extraTrashDelta = nil

	local createFromInfo = WFCatTrashUtils_createFromInfo( localDoc )

	if createFromInfo and createFromInfo.id == localDoc._id then
		-- Can't create a docId from itself, clear createFromInfo to issue a regular create.
		createFromInfo = nil
		logger:errorf( "ERROR createFromInfo.id == localDoc._id for localDoc=\n\n%s\n\n", localDoc )
	end

	if createFromInfo then

		if isTrashAssetDocument( localDoc ) then

			-- Create trash asset by deleting.

			-- If we are creating a trash asset and the original asset existed on
			-- oz create the trash asset by sending a delete with a deleted_asset
			-- entry instead. We do this here instead of on every tombstone to
			-- better handle repeated delete restore cases.

			local createTrashFromId = createFromInfo.id
			local createTrashFromRevisionId = createFromInfo.revision_id

			if createFromInfo and localDoc.subtype == createFromInfo.subtype then

				-- Oz only allows you to create trash from a regular asset so
				-- send the restore for this even though it has been deleted
				-- locally, it will be deleted immediately by the delta below it.

				createTrashFromId = createFromInfo.lastAssetId
				createTrashFromRevisionId = createFromInfo.lastRevisionId
				local restoredSubtype = string.match( localDoc.subtype, '^deleted_(.*)' )

				extraTrashDelta = {
					id = createTrashFromId,
					revision_id = createTrashFromRevisionId,
					type = localDoc.type,
					subtype = restoredSubtype,
					action = 'create',
					payload = {
						userUpdated = localDoc.userCreated
					},
					restore_from = {
						id = createFromInfo.id
					},
				}

			end

			-- Trash assets that originally existed on oz may have had a
			-- modification applied to them either before they were trashed
			-- or if they were restored and trashed again while offline
			-- so use the revision the trash asset was created with in case
			-- another revision was applied to it.
			local createWithRevisionId = WFCatTrashUtils_getCreateWithRevisionId( localDoc )

			-- Trash assets are created via a delete if the asset they are
			-- being created for existed in the cloud. If the original asset
			-- is not in the cloud they will be created via the create request below.
			docToSend = {
				id = createTrashFromId,
				revision_id = createTrashFromRevisionId,
				type = localDoc.type,
				action = 'delete',
				deleted_asset = {
					id = localDoc._id,
					revision_id = createWithRevisionId,
					payload = {
						userCreated = localDoc.userCreated,
						purgeDate = localDoc.purgeDate,
					}
				},
			}

		else

			-- Create regular asset by restoring.

			local createFromId = createFromInfo.id
			local createFromRevisionId = createFromInfo.revision_id

			if localDoc.subtype == createFromInfo.subtype then

				-- Oz only allows you to restore from a trash asset so
				-- send the delete for the target to create a trash asset
				-- that we can restore from.

				extraTrashDelta = {
					id = createFromId,
					revision_id = createFromRevisionId,
					type = localDoc.type,
					action = 'delete',
					deleted_asset = {
						id = createFromInfo.lastAssetId,
						revision_id = createFromInfo.lastRevisionId,
						payload = {
							userCreated = createFromInfo.lastTrashUserCreated,
							purgeDate = createFromInfo.lastTrashPurgeDate
						}
					},
				}

				createFromId = createFromInfo.lastAssetId
				createFromRevisionId = createFromInfo.lastRevisionId

			end

			-- Always restore from the initial revId in case modifications
			-- have been made to the asset locally, this is done in order
			-- to have the conflict resolver see newer revisions and cause
			-- us to send a subsequent update with the modifications.
			docToSend.revision_id = WFCatTrashUtils_getCreateWithRevisionId( localDoc )

			docToSend.restore_from = {
				id = createFromId,
			}

			docToSend.payload = {
				userUpdated = localDoc.userUpdated,
			}

		end

	else

		-- Normal create path, add original if it exists, which should
		-- only occur when a trash asset is created directly on oz
		-- because the asset was imported and deleted before being synced.

		docToSend.original = localDoc.original

	end

	-- Handle freemium asset payload on mobile
	if not WFPlatformInfo.isDesktop then
		local userId = self._catalog:ozUserId()
		
		OzSyncMetadataMapper.removeEmptyTables( payload )
		OzSyncMetadataMapper.remapGlobalDocumentId( userId, payload )
	end
	
	
	-- Return docToSend, extraTrashDelta as docToSend may be changed ot use new
	-- table.  extraTrashDelta will be nil if an extra delta isn't needed

	return docToSend, extraTrashDelta

end
	
local spaceinviteCreateRemapFn = function( self, localDoc, payload, docToSend )

	docToSend.space = localDoc.space
	
	payload.space = nil
	
end
	
--------------------------------------------------------------------------------

local function replaceDeleteWithJsonNull( t )

	local didReplaceAnything = false

	for k, v in pairs( t ) do

		if type( v ) == 'table' then

			if v._delete == true then
				didReplaceAnything = true
				t[ k ] = KSJsonEncoder.json_null
			else
				didReplaceAnything = replaceDeleteWithJsonNull( v ) or didReplaceAnything
			end

		end

	end

	return didReplaceAnything

end


--------------------------------------------------------------------------------

local function deltaForCreateDoc( self, localDoc, newDocsFilteredDueToOzError )

	-- If document is deleted locally and no corresponding Oz doc,
	-- no need to send anything.

	if localDoc._deleted or not localDoc.type then
		return
	end

	if not self._createDocTypeWhitelist[ localDoc.type ] then
		logger:warnf( "Unable to upload %s document ID %s because not supported in bulk API",
					  localDoc.type, localDoc._id )
		return
	end
	
	local spaceId = localDoc._spaceId
	
	if ( self._userIsOverQuota or ( spaceId and self._spacesOverQuotaById[ spaceId ] ) ) 
			and docUsesSpaceInCloud( localDoc ) then
			
		-- We're out of storage on Oz and this new doc will use up space so we will pre-filter
		-- out the create and, instead of sending it to Oz, mark it with a StorageExceededError
		-- When storage is once again available on Oz, it should be cleared by the normal process.
		logger:debugf( "Filtering out create of %s due to pre-existing StorageExceededErrors", localDoc._id )

		local storageErrors = newDocsFilteredDueToOzError[ WFCatOzErrors.Errors.kStorageExceededError ] or {}
		storageErrors[ #storageErrors + 1 ] = localDoc._id
		newDocsFilteredDueToOzError[ WFCatOzErrors.Errors.kStorageExceededError ] = storageErrors

		return
		
	end

	if ( self._userIsExpired or ( spaceId and self._spacesExpiredById[ spaceId ] ) ) 
			and docUsesSpaceInCloud( localDoc ) then
			
		logger:debugf( "Filtering out create of %s due to trial or subscription expired", localDoc._id )

		local expiredErrors = newDocsFilteredDueToOzError[ WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError ] or {}
		expiredErrors[ #expiredErrors + 1 ] = localDoc._id
		newDocsFilteredDueToOzError[ WFCatOzErrors.Errors.kTrialOrSubscriptionExpiredError ] = expiredErrors

		return
		
	end

	local payload = {}

	for k, v in pairs( localDoc ) do
		if type( k ) == 'string' and not k:match( '^_' ) then
			payload[ k ] = v
		end
	end

	payload.original = nil
	payload.type = nil
	payload.subtype = nil

	local docToSend = {
		id = localDoc._id,
		revision_id = localDoc._revId,
		type = localDoc.type,
		subtype = localDoc.subtype,
		action = 'create',
		payload = payload,
	}

	local extraTrashDelta = nil

	local createRemapFn = self._createDocRemapFunctions[ localDoc.type ]

	if createRemapFn then
		local newDocToSend
		newDocToSend, extraTrashDelta = createRemapFn( self, localDoc, payload, docToSend )
		docToSend = newDocToSend or docToSend
	end

	return docToSend, extraTrashDelta

end

--------------------------------------------------------------------------------

local function deltaForDeleteDoc( self, localDoc, ozDoc )

	local extraTrashDelta = nil
	local subtype = localDoc.subtype
	
	local createFromInfo = WFCatTrashUtils_createFromInfo( localDoc )
	if subtype == 'deleted_image' or subtype == 'deleted_video' then
	
		if createFromInfo and WFCatTrashUtils_isDeleteTrash( localDoc ) then

			if createFromInfo.subtype ~= subtype then

				-- If we are deleting a trash asset that doesn't exist on oz but
				-- the original asset it referred to does exist on oz then
				-- send the delete to the original asset and follow up with a
				-- delete for the trash asset as well.

				extraTrashDelta = {
					id = createFromInfo.id,
					revision_id = createFromInfo.revision_id,
					type = localDoc.type,
					action = 'delete',
					deleted_asset = {
						id = localDoc._id,
						revision_id = localDoc._revId,
						payload = {
							userCreated = localDoc.userCreated, -- date doesn't matter we are deleting this immediately after it is created.
							purgeDate = localDoc.purgeDate,
						}
					},
				}
				
			elseif not ozDoc then

				-- We are deleting a trash asset that was restored, deleted, and
				-- then permanently deleted since the last sync. Just delete the
				-- trash asset we were going to create the new trash asset from.
				return {
					id = createFromInfo.id,
					revision_id = createFromInfo.revision_id,
					type = localDoc.type,
					action = 'delete',
				}

			end

		end

	elseif WFCatTrashUtils_isDirectlyDelete( localDoc ) and ( createFromInfo or ozDoc ) then
	
		-- This process only works for normal assets. The asset can be original asset or restored asset.
		-- Deletion would only be sent if the asset exists on Oz
	
		local assetId = createFromInfo and createFromInfo.id or localDoc._id
		local revisionId = createFromInfo and createFromInfo.revision_id or localDoc._revId
		local trashId = KSIdGenerator.uuidInLowercaseHex()
		local trashRevId = KSIdGenerator.uuidInLowercaseHex()
		local trashPurgeDate = KSDate.timeInIso8601Format()
		
		return 
			{
				id = trashId,
				revision_id = trashRevId,
				type = localDoc.type,
				action = 'delete',
			},
			{
				id = assetId,
				revision_id = revisionId,
				type = localDoc.type,
				action = 'delete',
				deleted_asset = {
					id = trashId,
					revision_id = trashRevId,
					payload = {
						userCreated = trashPurgeDate, -- date doesn't matter we are deleting this immediately after it is created.
						purgeDate = trashPurgeDate,
					}
				}
			}
	
	end

	-- If extraTrashDelta is set that means even though this asset doesn't
	-- exist on oz we need to also issue the delete for it as the
	-- extraTrashDelta will be creating it.

	if not ozDoc and not extraTrashDelta then
		-- If document is deleted locally and no corresponding Oz doc,
		-- no need to send anything.
		return
	end

	-- Explicitly calling out this case:

	-- if we sent a deletion to Oz,
	-- the returned doc won't include the revision_ids array. Hence the WFRxDocStore
	-- creates a new local revId which triggers this sync code again
	-- Adding this comparision here, makes this case clear how it is handled.
	-- Without it we only returned early as the localDoc no longer carries the type
	-- property which is catched by the deleteDocType whitelisting

	if ozDoc and ozDoc._deleted then
		logger:infof( "Both local and Oz docs deleted, not sending change", localDoc._id )
		return
	end

	local docType = localDoc.type or ( ozDoc and ozDoc.type )
	if self._deleteDocTypeWhitelist[ docType ] then

		return {
			id = localDoc._id,
			revision_id = localDoc._revId,
			type = docType,
			action = 'delete',
		}, extraTrashDelta

	else
		logger:infof( "Unable to upload deletion of %s due to non-whitelisted docType %s", localDoc._id, docType )
		return
	end

end

--------------------------------------------------------------------------------

local function deltaForUpdateDoc( self, localDoc, ozDoc )

	-- Local document has a new revision, calculate diffs against Oz.

	local delta = WFTableMerge.createDiffForOz( ozDoc, localDoc )

	for k in pairs( delta ) do
		if type( k ) == 'string' and k:match( '^_' ) then
			delta[ k ] = nil
		end
	end

	if not self._updateDocTypeWhitelist[ localDoc.type ] then
	
		local loggerFn = logger.warnf
	
		if doNotWarnDocTypes[ localDoc.type ] then
			loggerFn = logger.infof
		end
	
		loggerFn( logger, "Unable to update %s document ID %s because not supported in bulk API",
					  localDoc.type, localDoc._id )
		return
		
	end

	if not ozDoc._serverProvidedMetadata then
		return
	end
	local ozRevIds = derefOpt( ozDoc, '_serverProvidedMetadata', 'revision_ids' )
	
	local ozParentRevId = ozRevIds and ozRevIds[ #ozRevIds ]

	local shouldOzGenerateRenditions
	local developSection = derefOpt( delta, 'develop' )
	local xmpCameraRawSha256 = derefOpt( delta, 'develop', 'xmpCameraRaw', 'sha256' )
	if developSection and not xmpCameraRawSha256 then
		-- If there are develop changes and they're not external, then we should ask Oz to 
		-- generate the renditions. If they are external, we will ask when we upload the
		-- external settings
		shouldOzGenerateRenditions = 'all'
	end

	-- We can never update type or subtype. Drop any such changes on the floor.

	delta.original = nil
	delta.subtype = nil
	delta.type = nil
	delta.purgeDate = nil

	-- if there's nothing in the delta, we can't send it to Oz.
	if not next( delta ) then
		return
	end
	
	replaceDeleteWithJsonNull( delta )

	local docType = ozDoc.type or localDoc.type
	
	local docToSend = {
		id = localDoc._id,
		revision_id = localDoc._revId,
		action = 'update',
		type = docType,
		generate_renditions = shouldOzGenerateRenditions,
		if_match = ozParentRevId or nil,
		payload = delta,
	}

	local updateDocRemapFn = self._updateDocRemapFunctions[ localDoc.type ]
	
	if updateDocRemapFn then
		updateDocRemapFn( self, localDoc, delta, docToSend )
	end
	
	return docToSend
	
end

--------------------------------------------------------------------------------

local function convertOneDocToDelta( self, localDoc, ozDoc, newDocsFilteredDueToOzError )

	-- If document is in conflict, don't send. Wait for conflict resolution
	-- and we'll sort it out later.

	if localDoc._conflict then
		logger:infof( "Unable to upload %s document ID %s at changeCounter %s due to conflict",
					  localDoc.type, localDoc._id, localDoc._changeCounter )
		return
	end

	-- Is local document deleted?

	if localDoc._deleted then
		return deltaForDeleteDoc( self, localDoc, ozDoc )
	end
	
	local shouldSkip, reason = WFCatOzErrors.shouldSkipUploadDueToError( localDoc )
	if shouldSkip then
		logger:infof( "Unable to upload %s document with id %s due to %s",
						localDoc.type, localDoc._id, reason )
		if localDoc.type then
			self._docsFilteredDueToErrorByType[ localDoc.type ] = 
				( self._docsFilteredDueToErrorByType[ localDoc.type ] or 0 ) + 1	
		end
		
		return				
	end

	-- If no existing Oz doc, generate a create request.

	if not ozDoc then
		return deltaForCreateDoc( self, localDoc, newDocsFilteredDueToOzError )
	end

	-- If existing Oz doc and local doc are at the same revision id,
	-- omit creating the overlay

	if ozDoc and ozDoc._revId == localDoc._revId then
		return
	end

	-- Document exists both locally and on Oz. Generate update delta.

	return deltaForUpdateDoc( self, localDoc, ozDoc )

end

--------------------------------------------------------------------------------

local function convertChangedDocsAndOzDocsToDeltas( self, changedDocs, ozDocs )

	local docIdsToPosition = {}
	local newDocsFilteredDueToOzError = {}
	local result = {}
	local nResults = 0

	local ozDocsIndex = 1
	local nOzDocs = #ozDocs
	
	for i, localDoc in ipairs( changedDocs ) do

		local ozDoc

		if ozDocsIndex <= nOzDocs and ozDocs[ ozDocsIndex ]._id == localDoc._id then
			ozDoc = ozDocs[ ozDocsIndex ]
			ozDocsIndex = ozDocsIndex + 1
		end

		logger:debugf( "convertOneDocToDelta\n\nlocalDoc = %s\n\nozDoc = %s\n\n",
					   localDoc, ozDoc )

		local delta, extraTrashDelta = convertOneDocToDelta( self, localDoc, ozDoc, newDocsFilteredDueToOzError )

		if delta then
		
			if self._isSpacePusher then
			
				delta.space_id = localDoc._spaceId or nil
				assert( delta.space_id )

			end

			if delta.payload and next( delta.payload ) == nil then
				delta.payload = nil
			end
		
		end
		
		logger:infof( "delta for %s = %s\n\n", localDoc._id, delta )

		local function addDelta( delta )

			if delta then

				local docType = localDoc.type
				if docType then
					self._changedDocsToSendByType[ docType ]
						= ( self._changedDocsToSendByType[ docType ] or 0 ) + 1
				end

				nResults = nResults + 1
				result[ nResults ] = delta
				docIdsToPosition[ localDoc._localId ] = nResults

			end

		end

		-- Unfortunately oz is not able to create a trash asset
		-- from another trash asset, or restore from a regular non-trash asset.
		-- The createDoc for trashed and restored docs recognizes this and
		-- adds the extra delta needed to make the intermediate document.
		addDelta( extraTrashDelta )

		addDelta( delta )

	end

	return result, docIdsToPosition, newDocsFilteredDueToOzError

end

--------------------------------------------------------------------------------

local function makeConvertChangedDocsAndOzDocsToDeltasFn( self )
 
	return function( ... )
		local deltasToUpload, docIdsToDeltaPosition, docsFilteredDueToOzError = convertChangedDocsAndOzDocsToDeltas( self, ... )
		
		-- Mark any docs that were filtered out because we expect to hit StorageExceededError
		-- or TrialOrSubscriptionExpiredError with the appropriate error without sending it to Oz at 
		-- all. Even if this is done prematurely and Oz would have taken it, the logic that retries
		-- these should see these and send them if appropriate.
		if next( docsFilteredDueToOzError ) then
			
			-- NMB-6496: Need to look up dependent docs for these docs and mark them as well to minimize
			-- the number of errors Oz may have to deal with. The majority of these would be album_asset 
			-- records so identify those.
			
			-- sort the table of docs filtered out into some tables we can use to add dependent docs
			local allFilteredDocIds = {}
			local filteredDocIdsToErrorMap = {}
			for errorStr, docIds in pairs( docsFilteredDueToOzError ) do
				for _, docId in ipairs( docIds ) do
					allFilteredDocIds[ #allFilteredDocIds + 1 ] = docId
					filteredDocIdsToErrorMap[ docId ] = errorStr
					self._docsFilteredDueToErrorByType[ 'asset' ] = ( self._docsFilteredDueToErrorByType[ 'asset' ] or 0  ) + 1
				end
			end
			
			if not self._albumAssetsForFilteredAssetsQueryFn then
				-- in order to filter out album_asset records for assets clients should provide an appropriate query function when creating
				-- the WFCatMetadataPusher that will return the album_asset records that correspond to the asset ids given
				logger:warnf( "Assets filtered due to errors, but no _albumAssetsForFilteredAssetsQueryFn to filter out associated album_assets" )

				return WFCatOzErrors.markDocsWithOzError( self._catalog:docStore(), docsFilteredDueToOzError, 'prefiltered' )
					:takeLast( 1 )
					:map( function()
						return deltasToUpload
					end )

			else
			
				return self._albumAssetsForFilteredAssetsQueryFn( self._catalog, allFilteredDocIds )
					:take( 1 )
					:flatMap( function( albumAssetRows ) 
						
						for _, row in ipairs( albumAssetRows ) do
							local albumAssetDocId = row[ 1 ]
							local assetId = row[ 2 ]
							
							-- add them to the docsFilteredDueToOzError
							local errorStr = filteredDocIdsToErrorMap[ assetId ]
							local errors = docsFilteredDueToOzError[ errorStr ]
							errors[ #errors + 1 ] = albumAssetDocId
							
							-- check to see if they're in the deltasToUpload, if so, remove them (log this)
							local maybePosition = docIdsToDeltaPosition[ albumAssetDocId ]
							if maybePosition then
								if deltasToUpload[ maybePosition ] and deltasToUpload[ maybePosition ].action == 'create' then
									logger:warnf( "Removing create of document %s which depends on %s from changed docs to upload due to %s", 
													albumAssetDocId, assetId, errorStr )
									table.remove( deltasToUpload, docIdsToDeltaPosition[ albumAssetDocId ] )
									self._docsFilteredDueToErrorByType[ 'album_asset' ] = ( self._docsFilteredDueToErrorByType[ 'album_asset' ] or 0  ) + 1
								end
							end 
								
						end
						
						-- apply the appropriate error annotation to all of these docs
						return WFCatOzErrors.markDocsWithOzError( self._catalog:docStore(), docsFilteredDueToOzError, 'prefiltered' )
							:takeLast( 1 )
							:map( function()
								return deltasToUpload 
							end )
					end )
			end
		end
		
		return Observable.just( deltasToUpload )
	end
	
end

--------------------------------------------------------------------------------

local doSendUpdatesOnce --forward declaration

--------------------------------------------------------------------------------

local function maybeDoSendUpdatesAgain( self, maybeNewContinueFrom )
	
	assert( maybeNewContinueFrom ~= nil, 'maybeNewContinueFrom: needs to be a valid number' )
	
	-- self._moreUpdatesToProcess should either be a changeCounter value that we
	-- need to catch up to or false we have nothing more to process. If we've
	-- caught up, then we can indicate that we're done processing by setting it 
	-- to false
	if maybeNewContinueFrom and self._moreUpdatesToProcess 
		and maybeNewContinueFrom > self._moreUpdatesToProcess then 
		self._moreUpdatesToProcess = false
	end
		
	if self._moreUpdatesToProcess and not self._syncIsStopped and self._pausedBySynchronizedOperation == 0 then
		
		logger:debug( "More updates to process. Rerunning sendUpdatesOnce." )
		doSendUpdatesOnce( self )
		
	else
		
		if self._syncIsStopped then
			logger:debug( "Not running sendUpdatesOnce as sync is stopped" )
		elseif self._pausedBySynchronizedOperation > 0 then
			logger:debug( "Not running sendUpdatesOnce as sync is pausedBySynchronizedOperation" )
		else
			logger:debug( "sendUpdatesOnce has completed and is now idle." )
		end
		self._updateInProgress = false
		self._isDoingWork:onNext( false )

	end

end

--------------------------------------------------------------------------------

local function changedDocsAndTrackPerformance( docStore, continueFrom, updateId, documentBatchCount, changedDocsSource )

	return WFPerformanceReport.wrapObservable( 'WFCatMetadataPusher : getChangedDocs', 
		docStore:changedDocs( {
			fromChangeCounter = continueFrom,
			limit = math.max( 100, documentBatchCount ),
			includeDocumentsFromSource = changedDocsSource
		}, 'MetadataPusher' ),
		{ 
			fromChangeCounter = continueFrom,
			updateId = updateId
		} )
		
end

--------------------------------------------------------------------------------

local function putAndTrackPerformance( self, bulkPostUrl, json, options, nDeltasToUpload, updateId )

	return WFPerformanceReport.wrapObservable( 'WFCatMetadataPusher : sendUpdatesToOz', 
		self._catalogSpec:put( bulkPostUrl, json, options ),
		{ 
			bulkPostUrl = bulkPostUrl,
			deltasToUpload = nDeltasToUpload,
			updateId = updateId,
		} )
		
end

--------------------------------------------------------------------------------

local function retryDelayRequested( headers )
	return headers and headers.status_code == '202' and tonumber( headers[ 'Retry-After' ] )
end

--------------------------------------------------------------------------------

local function maybeSetRetryDelayRequestByOz( self, headers )

	local retryDelay = retryDelayRequested( headers )

	if retryDelay then

		if self._retryDelayRequestByOz then
			-- Don't think this can happen.
			logger:error( "Trying to add a retryDelay when one is already set, using new delay" )
		end

		self._retryDelayRequestByOz = retryDelay
		self._retryDelayRequestByOzAtTime = timeNowForAnimation()

	end

end

--------------------------------------------------------------------------------

local function maybeAddDelayRequestByOz( self )

	if self._retryDelayRequestByOz then

		local timeSinceDelayRequested = timeNowForAnimation() - self._retryDelayRequestByOzAtTime

		local retryDelay = self._retryDelayRequestByOz - timeSinceDelayRequested

		self._retryDelayRequestByOz = nil
		self._retryDelayRequestByOzAtTime = nil

		if retryDelay > 0 then
			return Observable.timer( retryDelay, WFRxScheduler.sharedScheduler() )
		end

	end

	return Observable.just(true)

end

--------------------------------------------------------------------------------

local function recordError( bulkPostUrl, rawJsonOrError, headers )

	logger:errorf( "ERROR on send changes request\n\nPUT %s\n\n%s\n\n%s\n\n",
							   bulkPostUrl, headers, rawJsonOrError )
	
	local errorEvent = {}
	
	errorEvent['event.type'] = 'error'
	errorEvent['event.workflow'] = 'SYNC'
	errorEvent['event.subcategory'] = 'metadata_pusher'
	errorEvent['event.url'] =  bulkPostUrl
	errorEvent['event.error_type'] = 'MetadataUpload.Error'
	errorEvent['event.error_code'] = headers and headers.status_code or 'Unknown'
	errorEvent['event.error_desc'] = tostring( rawJsonOrError )
	
	WFAnalytics.trackEvent( errorEvent )
end

--------------------------------------------------------------------------------
--[[!
	\brief Sends one batch of updates to Oz.

	\return Observable< number | boolean > If any new documents were sent, the
		number of documents sent. If no documents sent, returns true if there
		is new content to send (but none were generated by this batch). Returns
		false if there is nothing new to send.

	Note that this may not send \b all of the documents that need to be sent.
	It will send up to 100 of the least-recently modified documents (exact number
	subject to change). If all of the least-recently modified documents result
	in no-op deltas to send to Oz, it will send no new documents but return true,
	which is your hint that you should call this function again to try to send
	a subsequent batch.
]]--

doSendUpdatesOnce = function( self, numRequiredUpdates )

	check( 'WFCatMetadataPusher', '?number' )

	if self._sendingLatestChanges then
		return self._sendingLatestChanges
	end

	local docStore = assert( self._catalog:docStore() )

	local bulkPostUrl = self._isSpacePusher and self._catalogSpec:spaceUploadUrl() 
			or self._catalogSpec:catalogUpdatesUrl()

	local updateId = KSIdGenerator.uuidInLowercaseHex()

	local newContinueFrom, nDeltasToUpload

	-- Find out where we left off before.

	local replayedUpdatesContinueFrom = self._sendUpdatesContinueFrom:shareReplay( 1 )
	
	self._sendingLatestChanges = Observable.combineLatest(

			function( changeCounter, sendUpdatesContinueFrom )
			
				-- See comment in :countUpdatesToSend() how the changeCounter is constructed
				-- This math compensates for the one off
				
				local diff = ( changeCounter + 1 ) - sendUpdatesContinueFrom
				
				return diff
				
			end,
			
			self._catalog:docStore():changeCounter(),
			replayedUpdatesContinueFrom
		)
		
		:where( function( updateCount ) 
			return updateCount >= 0 
		end )
		
		:call( filterDifferenceWithWindow, self._documentBatchCount, 
				self._delayBeforeSendingUpdates, WFRxScheduler.sharedScheduler() )
		
		:flatMapLatest( replayedUpdatesContinueFrom )
		
		:take( 1 )
			
		-- Get this batch of updated documents from local doc store.

		:selectMany( function( continueFrom )

			return changedDocsAndTrackPerformance( docStore, continueFrom, updateId, self._documentBatchCount, self._changedDocsSource )

				:selectMany( function( changedDocs )
					
					local originalContinueFrom = changedDocs.continueFrom
					
					local docsWithNoConflictsAndNoOzLabel = {
						continueFrom = changedDocs.continueFrom 
					}
					
					for i, changedDoc in ipairs( changedDocs ) do
						
						if changedDoc._conflict then
							-- doc is in conflict, don't let change counter pass it by
							docsWithNoConflictsAndNoOzLabel.continueFrom = changedDoc._changeCounter
							break
						elseif changedDoc._labels and ( changedDoc._labels[ 'oz' ] or changedDoc._labels[ '_dontsendtooz' ] ) then
							logger:infof( "Skipping document %s which has the label 'oz' or '_dontsendtooz'", changedDoc._id )
							docsWithNoConflictsAndNoOzLabel.continueFrom = changedDoc._changeCounter + 1
						else
							-- doc is ok, change counter should be next document
							docsWithNoConflictsAndNoOzLabel[ #docsWithNoConflictsAndNoOzLabel + 1 ] = changedDoc
							docsWithNoConflictsAndNoOzLabel.continueFrom = changedDoc._changeCounter + 1
						end
						
					end

					changedDocs = docsWithNoConflictsAndNoOzLabel 
					
					-- TODO [scouten 2015-08-05]: Possible race condition?
					-- What if Oz docs gets updated in the interval between
					-- :changedDocs and the fetch of Oz docs here?

					logger:infof( "#changedDocs = %s, continueFrom = %s, originalContinueFrom = %s",
								   #changedDocs, continueFrom, originalContinueFrom )

					newContinueFrom = changedDocs.continueFrom
					
					if #changedDocs > 0 then
						return combineChangedDocsWithOz( docStore, changedDocs )
					else
						return Observable.just( changedDocs, {} )
					end

				end )

		end )

		-- Log WF-flavored documents.

		:tap( function( changedDocs, ozDocs )

			if type( changedDocs ) == 'table' and #changedDocs > 0 then
				logger:debugf( "Changed docs returned \n\n%s\n\n", changedDocs )
			else
				logger:debug( "No documents to send" )
			end

		end )

		-- Convert the local-doc / Oz-doc pairs into deltas.

		:flatMap( makeConvertChangedDocsAndOzDocsToDeltasFn( self ) )

		-- This flatMap is only used by unit tests calling _DEBUG_sendNUpdates.
		:flatMap( function( deltasToUpload )

			if numRequiredUpdates and #deltasToUpload < numRequiredUpdates then

				logger:infof( "sendUpdatesOnce numUpdates %s < numRequiredUpdates %s",
					#deltasToUpload, numRequiredUpdates )

				self._moreUpdatesToProcess = false -- don't let maybeDoSendUpdatesAgain trigger another sendUpdates

				return Observable.throw( kNotEnoughUpdatesError )

			end

			return Observable.just( deltasToUpload )

		end )

		-- Log Oz delta-flavored documents.
		:map( function( deltasToUpload )

			-- SPAM [jandrew 2017-10-11]: Even though this `map` doesn't change the
			-- data, it's important that we use `map` and not `tap` because we need
			-- _isDoingWork to fire before we start uploading metadata changes.
			-- This is because `tap` is not guaranteed to execute before the chain 
			-- completes.

			if type( deltasToUpload ) == 'table' and #deltasToUpload > 0 then
				self._isDoingWork:onNext( true )
				logger:debugf( "Will upload deltas \n\n%s\n\n", deltasToUpload )
			end
			
			return deltasToUpload

		end )

		-- Upload those documents via bulk doc API.

		:takeLast( 1 )

		:selectMany( function( deltasToUpload )

			nDeltasToUpload = #deltasToUpload

			if nDeltasToUpload > 0 then
				
				local json = KSJsonEncoder.encodeOzUploadJson( deltasToUpload )

				local options = {
					requestHeaders = {
						{ header = 'Content-Type', value = 'application/json' },
						{ header = 'Accept', value = 'application/json' },
					},
					requestMethod = 'POST',
					bypassRateLimiter = true,
				}
				
				if WFConfiguration.WFCatMetadataPusher.QuickerRetries then
					table.insert( options.requestHeaders, { header = 'X-Quicker-Retries', value = 'true' } )
				end

				logger:debugf( "Sending docs via\n\nPOST %s\n\n%s\n\n%s\n\n",
							   bulkPostUrl, options, json )

				return maybeAddDelayRequestByOz( self )
				:flatMap( function()
					return putAndTrackPerformance( self, bulkPostUrl, json, options, nDeltasToUpload, updateId )
				end )
				:tap(function()
					self._totalDocumentsSent = self._totalDocumentsSent + #deltasToUpload
					self._lastUpdateCycleWithData = KSDate.timeInIso8601FormatWithTimeZone()
					self._successfulUpdateCycles = self._successfulUpdateCycles + 1
				end)
				
			else

				return Observable.just( true )

			end

		end )

		-- Look for requested Retry-After and Log results.

		:tap( function( rawJson, headers )

			maybeSetRetryDelayRequestByOz( self, headers )

			if rawJson then
				logger:debugf( "Send docs\n\nPOST %s\n\n%s\n\n%s\n\n",
							   bulkPostUrl, headers, rawJson )
			end

		end )

		-- Record where we left off, so we don't repeat ourselves next time.

		:concat( Observable.just( true ) )
		:takeLast( 1 )

		:selectMany( function()

			logger:infof( "New continueFrom => %s", newContinueFrom )

			if newContinueFrom then

				self._sendUpdatesContinueFrom:onNext( newContinueFrom )

				return docStore:setLocalDocValue {
					_id = self._continueFromLocalDocValueKey,
					continueFrom = newContinueFrom,
				}
				
				:concat( Observable.just( true ) )
				
			else

				return Observable.just( false )

			end

		end )

		-- Log any failure.

		:tap( 
			nil, 
			function( rawJsonOrError, headers )
			
				recordError( bulkPostUrl, rawJsonOrError, headers )
				
				self._errorCount = self._errorCount + 1
				local errorDelay = 5
				local progressiveBackoff = math.min( 60, errorDelay * self._errorCount )

				if self._moreUpdatesToProcess then

					logger:debugf( "Retrying metadata push in %d seconds", errorDelay )

					WFRxLowFidelityTimer.timer( progressiveBackoff )
					:tap( function()

						-- Subtracting one from _moreUpdatesToProcess will force a re-try
						-- of the send task

						maybeDoSendUpdatesAgain( self, self._moreUpdatesToProcess - 1 )

					end )
					:call( runUntilDone )

				end
				
			end,
			function()
			
				self._errorCount = 0
				self._continuousPushTask:taskDidWork()
				maybeDoSendUpdatesAgain( self, newContinueFrom )
				
			end )

		-- Return true if new content; false if not. This is only
		-- used by unit tests.

		:map( function()

			self._sendingLatestChanges = nil
			return nDeltasToUpload or not not newContinueFrom
			
		end )
		:tapOnError( function() 
		
			self._sendingLatestChanges = nil
			
		end )
				
		-- Avoid stopping this process once started.

		:call( runUntilDone )

		return self._sendingLatestChanges

end

--------------------------------------------------------------------------------

local function sendUpdatesTask( self )

	local docStore = assert( self._catalog:docStore() )

	local changeCounter = docStore:changeCounter()
	
		-- Fire one observation when doc store's change counter climbs
		-- above where we left off.

		:flatMapLatest( function( changeCounter )

			return self._continuousPushTask:throwIfNotRunning( changeCounter )
				-- Need to throw as this function doesn't ever complete.

		end )
		
	local continueFrom = self._sendUpdatesContinueFrom:shareReplay( 1 )
	
	return Observable.combineLatest( changeCounter, continueFrom )

		:map( function( changeCounter, continueFrom )

			logger:infof( "sendUpdatesOnce got changeCounter %s, continueFrom %s", changeCounter, continueFrom )

			continueFrom = continueFrom ~= 0 and continueFrom or 1
			logger:infof( "sendUpdatesOnce waiting for changeCounter > %s", continueFrom )
		
			local shouldSendUpdates = changeCounter >= continueFrom
			
			self._moreUpdatesToProcess = changeCounter
			
			if shouldSendUpdates and self._updateInProgress then
			
				logger:info( "sendUpdatesOnce already in progress. Will reprocess once complete." )

			else
			
				self._updateInProgress = true
				
				doSendUpdatesOnce( self )
				

			end
					
		end )
		
		:catch( function( errorMessage )

			return self._continuousPushTask:throwIfNotStopped( errorMessage )

		end )

end


--============================================================================--

local WFCatMetadataPusher = {

	_catalog = false,
		-- (WFCatalog)

	_catalogSpec = false,
		-- (WFCatSpec)

	_sendUpdatesContinueFrom = false,
		-- (ReplaySubject< number >)

	_continuousPushTask = false,
		-- (WFRxContinuousTask) A continuous task to push metadata updates to Oz.

	_updateInProgress = false,
		-- (boolean) Indicates if we're in the act of pushing Metadata changes.

	_moreUpdatesToProcess = false,
		-- (boolean) Indicates if we should immediately process more updates once the
		-- current update completes.
	
	_sendingLatestChanges = false,
		-- (Rx Observable) The current observable being used to push metadata updates.

	_errorCount = false,
		-- (number) Number of repeated push metadata errors. Used in progressive backoff.
		
	_requestHeaders = false,
		-- (table) HTTP Request Headers to be used for bulk upload

	_isDoingWork = false,
		-- (Rx ReplaySubject) A Rx.Subject used to indicate wether not uploading is occuring.
	
	_retryDelayRequestByOz = false,
		-- (number) when set indicates seconds since _retryDelayRequestByOzAtTime to delay before requesting updates.

	_retryDelayRequestByOzAtTime = false,
		-- (number) the value of timeNowForAnimation clock when _retryDelayRequestByOz was last set.

	_syncIsStopped = false,
		-- (boolean) internal flag used to indicate whether sync should be stopped.

	_pausedBySynchronizedOperation = 0,
		-- (number) internal flag used to indiate sync is paused (> 0) because of a synchronized operation.

	_userIsOverQuota = false,
		-- (boolean) internal flag used to indicate whether user is over quota and we should modify how we handle
		-- asset creates on Oz
	
	_userIsExpired = false,
		-- (boolean) internal flag used to indicate whether user's trial or subscription is expired and we should modify
		-- how we handle asset creates on Oz
		
	_delayBeforeSendingUpdates = false,
		-- (number) Delay in seconds the pusher waits to collect batches before sending new updates
		
	_documentBatchCount = false,
		-- (number) Number of documents to batch before uploading to Oz
		
	_createDocRemapFunctions = false,
		-- (table) Remap functions applied during document creation on Oz
		
	_updateDocRemapFunctions = false,
		-- (table) Remap functions applied during document updates on Oz
	
	_albumAssetsForFilteredAssetsQueryFn = false,
		-- (function) A function that takes a catalog and a list of full asset ids and returns an Observable that
		-- will query for the album_assets associated with these asset ids.  The rows returned by this observable should contain
		-- the album_asset docId first and the full *asset* id from the album_asset record.  e.g.
		-- SELECT album_asset_v2.docId, album_asset_v2.assetId FROM TEMP_ids JOIN album_asset_v2 ON 
		--					TEMP_ids.docId = album_asset_v2.assetId

	_isSpacePusher = false,
		-- (boolean) if true, this pusher will only process documents with spaceIds.
		-- if false, this pusher will only process catalog-based documents.

	_continueFromLocalDocValueKey = false,
		-- (string) The local doc value key used by this pusher to track continue-from.

	_changedDocsSource = false,
		-- (string) A string indicating which documents to push to Oz: "catalog" or "spaceFeed".

	_spacesOverQuotaById = false,
		-- (table) A table used track which spaces are over quota.
		-- key: space ID
		-- value: true if the space is over quota.
		
	_spacesExpiredById = false,
		-- (table) A table used track which spaces have lapsed subscriptions.
		-- key: space ID
		-- value: true if the space owner has a lapsed subscription.
		
	-- fields storing data for the diagnostic log
	_changedDocsToSendByType = {},
		-- (table) table with counts of documents that could be sent by document type
	
	_docsFilteredDueToErrorByType = {},
		-- (table) count of uploads filtered due to preexisting error by document type
		
	_successfulUpdateCycles = 0,
		-- (number) total number of successful update cycles since app launch
		
	_totalDocumentsSent = 0,
		-- (number) total number of documents sent since app launch
		
	_lastUpdateCycleWithData = false,
		-- (string) timestamp of the last update sent to Oz that contained data

	_createDocTypeWhitelist = false,
		-- (table) A whitelist of documents that can be created on Oz.
		
	_updateDocTypeWhitelist = false,
		-- (table) A whitelist of documents that can be updated on Oz.
		
	_deleteDocTypeWhitelist = false,
		-- (table) A whitelist of documents that can be deleted from Oz.
		
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

function WFCatMetadataPusher:__init( catalog, options, optCatSpec )

	check( 'WFCatMetadataPusher', 'WFCatalog', '?table', '?WFCatSpec'  )

	self:setOwned {
		_catalog = catalog:relayObject()
	}
	
	options = options or {}
	
	self._createDocRemapFunctions = {}
	self._updateDocRemapFunctions = {}
	
	self._syncIsStopped = true
	self._pausedBySynchronizedOperation = 0

	self._catalogSpec = optCatSpec or catalog:ozCatalogSpec()

	self._sendUpdatesContinueFrom = ReplaySubject( 1 )
	
	self._delayBeforeSendingUpdates = options.delayBeforeSendingUpdates or 3 -- sec
	self._documentBatchCount = options.documentBatchCount or 1
	self._albumAssetsForFilteredAssetsQueryFn = options.albumAssetsForFilteredAssetsQueryFn
	self._isSpacePusher = not not options.isSpacePusher
	
	self._createDocTypeWhitelist = createDocTypeWhitelist
	self._updateDocTypeWhitelist = updateDocTypeWhitelist
	self._deleteDocTypeWhitelist = deleteDocTypeWhitelist
	
	if self._isSpacePusher then
	
		-- In test mode, allow the standard set of documents to be used instead of the
		-- more restrictive space-feed set. This allows us to write more thorough
		-- unit tests.
		
		if not options.testMode then
		
			self._createDocTypeWhitelist = spaceFeedCreateDocTypeWhitelist
			self._updateDocTypeWhitelist = spaceFeedUpdateDocTypeWhitelist
			self._deleteDocTypeWhitelist = spaceFeedDeleteDocTypeWhitelist
			
		end
		
		self._changedDocsSource = WFRxDocStore.changedDocsSourceSpaceFeedDocuments
		
	else
	
		self._changedDocsSource = WFRxDocStore.changedDocsSourceCatalogDocuments
		
	end
	
	self._continueFromLocalDocValueKey = getContinueFromLocalDocValueKey( self._isSpacePusher )
	
	self._spacesOverQuotaById = {}
	
	self._spacesExpiredById = {}
	
	self._updateInProgress = false
	
	self._moreUpdatesToProcess = false

	self._errorCount = 0
	
	WFDiagnosticLog.addLogProvider( self )
	
	self._requestHeaders = {
		{ header = 'Content-Type', value = 'application/json' },
		{ header = 'Accept', value = 'application/json' },
	}

	updateSendUpdatesContinueFrom( self )
		:call( runUntilDone )

	local continuousTaskDelays = options.continuousTaskDelays or {
	
		progressiveBackoff = true,
		errorDelay = 10,
		maxDelay = 60,

	}
	
	local continuousTask = function() return sendUpdatesTask( self ) end

	self._continuousPushTask = WFRxContinuousTask( "WFCatMetadataPusher_"..catalog:ozCatalogId(), continuousTask, continuousTaskDelays )

	self._isDoingWork = ReplaySubject( 1 )
	self._isDoingWork:onNext( false )
	
	self:addCreateDocRemapFunction( 'album_asset', album_assetCreateRemapFn )
	self:addCreateDocRemapFunction( 'space_album', space_albumCreateRemapFn )
	self:addCreateDocRemapFunction( 'stack_asset', stack_assetCreateRemapFn )
	self:addCreateDocRemapFunction( 'comment', commentCreateRemapFn )
	self:addCreateDocRemapFunction( 'favorite', favoriteCreateRemapFn )
	self:addCreateDocRemapFunction( 'asset', assetCreateRemapFn )
	self:addCreateDocRemapFunction( 'spaceinvite', spaceinviteCreateRemapFn )
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:close()

	check( 'WFCatMetadataPusher' )
	
	self:stopContinuousUpdates()
	
	if self._sendingLatestChanges then
		self._sendingLatestChanges:dispose()
	end
	
	if self._sendUpdatesContinueFrom then
		self._sendUpdatesContinueFrom:dispose()
		self._sendUpdatesContinueFrom = nil
	end

	local continuousPushTask = self._continuousPushTask
	self._continuousPushTask = nil
	
	if continuousPushTask then
		continuousPushTask:close()
	end

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:addCreateDocRemapFunction( docType, remapFn )

	check( 'WFCatMetadataPusher', 'string', 'function' )
	
	if self._createDocRemapFunctions[ docType ] then
		logger:warnf( 'Create Document remap function already defined for docType %s', docType )
	end
	
	self._createDocRemapFunctions[ docType ] = remapFn
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:addUpdateDocRemapFunction( docType, remapFn )

	check( 'WFCatMetadataPusher', 'string', 'function' )
	
	if self._updateDocRemapFunctions[ docType ] then
		logger:warnf( 'Update Document remap function already defined for docType %s', docType )
	end
	
	self._updateDocRemapFunctions[ docType ] = remapFn
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:addRequestHeader( name, value )

	check( 'WFCatMetadataPusher', 'string', 'string|number|boolean' )
	
	table.insert( self._requestHeaders, { header = name, value = value } )
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:removeRequestHeader( name )

	check( 'WFCatMetadataPusher', 'string' )
	
	local WFFunctionUtils = require 'WFFunctionUtils'
	
	self._requestHeaders = WFFunctionUtils.filter( function( request )
	
		if request.header == name then
			return nil
		end
		
		return request
		
	end, self._requestHeaders )
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:stopContinuousUpdates()

	check( 'WFCatMetadataPusher' )
	
	self._syncIsStopped = true
	self._continuousPushTask:stopContinuousTask()

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:sendUpdatesContinuously()

	check( 'WFCatMetadataPusher' )
	
	self._syncIsStopped = false
	if self._pausedBySynchronizedOperation == 0 then
		self._continuousPushTask:startContinuousTask()
	end

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:isDoingWork()

	check( 'WFCatMetadataPusher' )

	return self._isDoingWork

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:_DEBUG_sendUpdatesOnce()

	check( 'WFCatMetadataPusher' )

	return doSendUpdatesOnce( self )

end

--------------------------------------------------------------------------------

-- SPAM [jandrew 2019-08-29]: This function is only used by unit tests.

function WFCatMetadataPusher:_testOnly_getIsTaskRunningStream()

	check( 'WFCatMetadataPusher' )

	return self._continuousPushTask:_testOnly_getIsTaskRunningStream()

end

--------------------------------------------------------------------------------

--[[!
	\brief This function is only used in unit tests.

	Calling this function will cause the metadata pusher to continously attempt
	to push updates until it has pushed the number of updates specified by n.

	\param numRequiredUpdates (number) The number of updates to push.
	\param timeoutSeconds (number) The number of seconds at which to give up
		and issue one last request to see what updates were available.
	\param allowMoreThanN (boolean) If true, this function will not throw an
		error if more than n updates are sent.

	\return Observable - Resolves when n updates have been pushed.
]]--

function WFCatMetadataPusher:_DEBUG_sendNUpdates( numRequiredUpdates, timeoutSeconds, allowMoreThanN )

	check( 'WFCatMetadataPusher', 'number', '?number', '?boolean' )

	-- Use one hour as default timeout as, all tests will timeout sooner, all tests run in less than 10 min currently.
	timeoutSeconds = timeoutSeconds or 60*60

	logger:infof( "_DEBUG_sendNUpdates numRequiredUpdates = %s, timeOutSeconds %s, allowMoreThanN = %s",
						  numRequiredUpdates, timeoutSeconds, allowMoreThanN )

	return Observable.defer( function()

		return doSendUpdatesOnce( self, numRequiredUpdates )

		:catch( function( errorMessage )

			if errorMessage == kNotEnoughUpdatesError then
				return Observable.throw( errorMessage )
			end

			-- if we threw for some other reason send it on.
			return Observable.just( 0, errorMessage )

		end )

	end )

	:retry( 100000 ) -- can't use math.huge, retrying this many times is unlikley to succeed.

	:timeout( timeoutSeconds, WFRxScheduler.sharedScheduler() )

	:catch( function( errorMessage )

		if errorMessage == "Timeout" then

			-- Delay a little to allow tapOnError in the prior call to
			-- doSendUpdatesOnce to complete prior to re-issuing one
			-- last call to doSendUpdatesOnce.

			return Observable.timer( .1, WFRxScheduler.sharedScheduler() )
			:flatMap( function()

				-- Issue one last call to doSendUpdatesOnce without numRequiredUpdates
				-- to send whatever the pusher has to send to make it easier to
				-- debug as the HttpMock can show more detailed information as to what
				-- is happening.

				return doSendUpdatesOnce( self )

			end )

		else
			return Observable.throw( errorMessage )
		end

	end )

	:flatMap( function( count, maybeErrorMessage )

		if maybeErrorMessage then

			return Observable.throw( maybeErrorMessage )

		elseif count == 0 and count < numRequiredUpdates then

			-- If we got 0 updates and we were expecting more than 0 then show this error
			-- as no http transactions will be generated.  Otherwise send it on as the
			-- error generated by the MockedHttp will be more informative.

			return Observable.throw( string.format(
				"Too few updates were sent. Expected %d, Sent %d", numRequiredUpdates, count ) )

		elseif count > numRequiredUpdates and not allowMoreThanN then

			return Observable.throw( string.format(
				"Too many updates were sent. Expected %d, Sent %d", numRequiredUpdates, count ) )

		end

		return Observable.empty()

	end )

end

--------------------------------------------------------------------------------
--[[!
	\brief Counts the number of document updates that may need to be send to Oz.

	WARNING: This API is prone to over-reporting. Some documents may be marked
	as changed locally, but may wind up being no-ops. (For example, the change
	might have been the read-back of the change \b from Oz.)

	If updates are being sent to Oz, this observable should eventually return
	to the correct result (zero).

	\return Observable< number > number of document updates that may need to be
		sent to Oz
]]--

function WFCatMetadataPusher:countUpdatesToSend()

	check( 'WFCatMetadataPusher' )

	return Observable.combineLatest(

		function( changeCounter, sendUpdatesContinueFrom )

			logger:infof( "countUpdatesToSend changeCounter = %s, sendUpdatesContinueFrom = %s",
						  changeCounter, sendUpdatesContinueFrom )

			if type( changeCounter ) == 'number'
				and type( sendUpdatesContinueFrom ) == 'number'
			then

				if changeCounter >= sendUpdatesContinueFrom then

					-- Yes, the following math is a bit wonky. The continueFrom result
					-- from :changedDocs is inclusive (i.e. the first new revision
					-- number you are seeking). I wish now that it was the change
					-- counter of the last document considered, but fixing that
					-- is much too risky now.

					if sendUpdatesContinueFrom > 0 then
						sendUpdatesContinueFrom = sendUpdatesContinueFrom - 1
					end

					return changeCounter - sendUpdatesContinueFrom

				else

					return 0

				end

			else

				return false

			end

		end,

		self._catalog:docStore():changeCounter(),
		self._sendUpdatesContinueFrom )

		:filter( function( value )
			return value -- discard "false" placeholder that says incomplete data
		end )

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:setUserIsOverQuota( isOverQuota )

	check( 'WFCatMetadataPusher', 'boolean' )
	
	self._userIsOverQuota = isOverQuota
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:setUserIsExpired( isExpired )
	
	check( 'WFCatMetadataPusher', 'boolean' )
	
	self._userIsExpired = isExpired
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:setSpaceIsOverQuota( spaceId, isOverQuota )

	check( 'WFCatMetadataPusher', '+string', 'boolean' )
	
	self._spacesOverQuotaById[ spaceId ] = isOverQuota
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:getSpaceIsOverQuota( spaceId )

	check( 'WFCatMetadataPusher', '+string' )
	
	return self._spacesOverQuotaById[ spaceId ]
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:setSpaceIsExpired( spaceId, entitlementExpired )

	check( 'WFCatMetadataPusher', '+string', 'boolean' )
	
	self._spacesExpiredById[ spaceId ] = entitlementExpired
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:getSpaceIsExpired( spaceId )

	check( 'WFCatMetadataPusher', '+string' )
	
	return self._spacesExpiredById[ spaceId ]
	
end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:pauseReturningCounter()

	check( 'WFCatMetadataPusher' )
	
	self._pausedBySynchronizedOperation = self._pausedBySynchronizedOperation + 1

	logger:infof( "pauseReturningCounter _pausedBySynchronizedOperation = %s", self._pausedBySynchronizedOperation )

	self._continuousPushTask:stopContinuousTask()

	return self._isDoingWork
		:filter( function( isDoingWork )
			return not isDoingWork
		end )
		:flatMap( function()
			return self._catalog:docStore():localDocValue( self._continueFromLocalDocValueKey )
		end )
		:map( function( docValue )
			return KSTableUtils.derefOpt( docValue, 'continueFrom' ) or 0
		end )
		:take( 1 )

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:resume()

	check( 'WFCatMetadataPusher' )

	self._pausedBySynchronizedOperation = self._pausedBySynchronizedOperation - 1

	logger:infof( "resume _pausedBySynchronizedOperation = %s", self._pausedBySynchronizedOperation )

	if self._pausedBySynchronizedOperation < 0 then
		logger:error( "_pausedBySynchronizedOperation < 0" )
		self._pausedBySynchronizedOperation = 0
	end

	if not self._syncIsStopped and self._pausedBySynchronizedOperation == 0 then
		self._continuousPushTask:startContinuousTask()
	end

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:sendDiagnosticLogData()

	check( 'WFCatMetadataPusher' )

	return {
		
		name = 'Metadata Sync Up: ' .. self._catalogSpec:catalogId(),
		id = 'WFCatMetadataPusher-' .. self._catalogSpec:catalogId(),
		weight = 500,
		parentSection = WFDiagnosticLog.Sections.SYNC,

		body = {
			isCurrentlySyncing = not self._syncIsStopped,
			pausedBySynchronizedOperation = self._pausedBySynchronizedOperation,
			changedDocsToSendByType = self._changedDocsToSendByType,
			docsFilteredDueToErrorByType = self._docsFilteredDueToErrorByType,
			successfulUpdateCycles = self._successfulUpdateCycles,
			totalDocumentsSent = self._totalDocumentsSent,
			lastUpdateCycleWithData = self._lastUpdateCycleWithData,
			
		}
		
	}

end

--------------------------------------------------------------------------------

function WFCatMetadataPusher:DEBUG_convertOneDocToDelta( localDoc, ozDoc, newDocsFilteredDueToOzError )
	-- Only used by unit tests.
	newDocsFilteredDueToOzError = newDocsFilteredDueToOzError or {}
	return convertOneDocToDelta( self, localDoc, ozDoc, newDocsFilteredDueToOzError )
end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatMetadataPusher'( WFCatMetadataPusher )
