--[[----------------------------------------------------------------------------

WFCatSearchByCategoryText.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--
local WFCatSearchByCategoryText = {}
local WFCatSearchByTextConstants = require 'WFCatSearchByTextConstants'
local TEXT_SEARCH_CATEGORY_TABLE_VERSION = 'v0'
local KSTableUtils = require 'KSTableUtils'

local derefOpt = KSTableUtils.derefOpt
local isValidSearchableAssetDocument = assert( require 'WFCatIndexUtils'.isValidSearchableAssetDocument )

local ENABLE_CATEGORY_SEARCH = WFCatSearchByTextConstants.ENABLE_CATEGORY_SEARCH

--==============================================================================
function WFCatSearchByCategoryText.makeCategoryBasedDetails( WFCatSearchByTextFields, WFCatSearchByCategoryBasedTextFields )
	if not ENABLE_CATEGORY_SEARCH then
		return
	end
	
	for _, searchFieldSpecifications in ipairs( WFCatSearchByTextFields ) do
		local mainDocFieldPath = searchFieldSpecifications.mainDocFieldPath
		local processingDetails = searchFieldSpecifications.processingDetails
		for _, details in ipairs( processingDetails ) do
			local category = details.category
			
			local categoryObject = ( WFCatSearchByCategoryBasedTextFields[ category ])  or {}
			categoryObject.category = category
			
			local dataInfo = {}
			dataInfo.mainDocFieldPath = mainDocFieldPath
			dataInfo.additionalData =  details.additionalData
			dataInfo.textProcessMethod =  details.textProcessMethod
			dataInfo.defaultValue = details.defaultValue
		
			local existingData = categoryObject.dataProcessingDetails or {}
			table.insert( existingData, dataInfo )
			
			categoryObject.dataProcessingDetails = existingData
			WFCatSearchByCategoryBasedTextFields[ category ] = categoryObject
		end
	end
end


--------------------------------------------------------------------------------
local function getCompleteData( docValue, mainDocFieldPath, additionalDataInfo, defaultVal )
	
	local mainVal = derefOpt( docValue, unpack( mainDocFieldPath ) ) or defaultVal
	
	
	if mainVal then -- if val is not there we don't get data as of now. When needed this can be customized
		local retValue = { mainVal }
		local additionalValues = {}
		if additionalDataInfo then
			for _, additionalDataPath in ipairs( additionalDataInfo ) do
				local additionalData = derefOpt( docValue, unpack( additionalDataPath ) )
				if additionalData  then
					table.insert( additionalValues, additionalData )
				end
			end
			table.insert( retValue, additionalValues )
		end
		return retValue
	end
				
end



--------------------------------------------------------------------------------
local function addCategoryTextSearchTable( indexStoreSchema, category, processingDetails )
	-- reset oldversions of table
	--indexStoreSchema:resetTable( 'text_search' )
	indexStoreSchema:addTable {
		tableName = WFCatSearchByTextConstants.getCategoryTableName( category ),
		fields = { 'search_text' },
		isFullTextSearch = true,
		useLocalId = true,
		hasOneRowPerDocument = true,
		updateLazily = false, -- don't wait for the table to be used. Do indexing ASAP
		docValueFilter = function( docValue )
			-- can probably be slightly more selective here
			return docValue.type == 'asset'
		end,
		map = function( docValue, emit )
			-- increment this number if we want to reset the table
			TEXT_SEARCH_CATEGORY_TABLE_VERSION = 'v12'
			
			local isSearchable, searchableDoc = isValidSearchableAssetDocument( docValue )
			if isSearchable and searchableDoc then

				local searchableText = {}
				
				for _, processingDetail in ipairs( processingDetails ) do
			
					local textProcessMethod = processingDetail.textProcessMethod
					local val = getCompleteData( searchableDoc, 
									processingDetail.mainDocFieldPath, 
									processingDetail.additionalData, 
									processingDetail.defaultValue )
		 
					if val then
						if textProcessMethod then
							local processedData = textProcessMethod( val, false, category )	
							if processedData  ~= nil then
								for k,v in ipairs(processedData) do
									table.insert(searchableText, v) 
								end 
							end
						elseif val then
							searchableText[ #searchableText + 1 ] = val
						end
					end
						
				end
					
				if #searchableText > 0 then
					emit( table.concat( searchableText, ',' ) )
				end
				
			end
		end,
	}
end

--------------------------------------------------------------------------------
function WFCatSearchByCategoryText.addCategoryTextSearchTables( indexStoreSchema, WFCatSearchByCategoryBasedTextFields )
	if not ENABLE_CATEGORY_SEARCH then
		return
	end
	
	for category, categoryInfoObject in pairs( WFCatSearchByCategoryBasedTextFields ) do
		if not WFCatSearchByTextConstants.categoryBasedSearchNotNeeded[ category ]  then
			addCategoryTextSearchTable( indexStoreSchema, category, categoryInfoObject.dataProcessingDetails )
		end
	end
end	


--------------------------------------------------------------------------------
function WFCatSearchByCategoryText.addSearchToSearchSchema( searchSchema, WFCatSearchByCategoryBasedTextFields )

	if not ENABLE_CATEGORY_SEARCH then
		return
	end
	
	for category, categoryInfoObject in pairs( WFCatSearchByCategoryBasedTextFields ) do
		if not WFCatSearchByTextConstants.categoryBasedSearchNotNeeded[ category ]  then
			local tableName = WFCatSearchByTextConstants.getCategoryTableName( category )
			local sqlString = string.format( [[SELECT docid FROM %s WHERE search_text MATCH ? ORDER BY docid]] , tableName )
			local termVal =  WFCatSearchByTextConstants.getTermName( category )

			searchSchema:addSearchableData {
				term = termVal,
				dataType = 'fullTextSearch',
				sql = sqlString,
					-- unfortunately, SQLite's built-in docId for full text search differs from ours
					-- by case so, docid here is correct (even if SQLite is generally permissive about
					-- column name case in queries)
				query_params = 'text',
			}
		end
	end
	
end

return WFCatSearchByCategoryText
