--[[----------------------------------------------------------------------------

WFCatSearchByText.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--
	-- WF.core.os
local KSDate = require 'KSDate'

local KSTableUtils = require 'KSTableUtils'
local WFCatSearchQuery = require 'WFCatSearchQuery'
local WFCatSearchByTextConstants = require 'WFCatSearchByTextConstants'
local WFCatSearchByTextDataProcessors = require "WFCatSearchByTextDataProcessors"
local WFCatSearchByTextIndexedData = require "WFCatSearchByTextIndexedData"

	-- WF.core.logging
local WFLogger = require 'WFLogger'
local logger = WFLogger( 'WFSearchTables' )

local WFSearchDebugLogging = require 'WFSearchDebugLogging'

local ENABLE_CATEGORY_SEARCH = WFCatSearchByTextConstants.ENABLE_CATEGORY_SEARCH
local ENABLE_SEARCH_COMPLETIONS = WFCatSearchByTextConstants.ENABLE_SEARCH_COMPLETIONS
local ENABLE_LOCAL_TEXT_SEARCH  = WFCatSearchByTextConstants.ENABLE_LOCAL_TEXT_SEARCH

--------------------------------------------------------------------------------
-- Info : To reset the tables when new data needs to be pulled.
-- e.g adding a new WFCatSearchByTextFields or changing value from
-- any of the processor methods
-- Table versioning is  supported. But that will keep old version tables alive.
-- We have to trigger clearning of old versions if we are using that approach.
-- *TABLE_VERSION is used as an alternative
-- changing the value of this in  the table  map function will change 
-- the map funciton hash and hence re trigger table creation
local TEXT_SEARCH_TABLE_VERSION = 'v0'
local CATEGORY_TABLE_VERSION = 'v0'
local RELEVANCY_TABLE_VERSION = 'v0'

--------------------------------------------------------------------------------
local derefOpt = KSTableUtils.derefOpt

local isValidDisplayableAssetDocument = assert( require 'WFCatIndexUtils'.isValidDisplayableAssetDocument )
local noDataString = "NO DATA"

--==============================================================================

local WFCatSearchByText = {}
local WFCatSearchByTextFields = {}
local WFCatSearchByCategoryBasedTextFields = {}
local WFCatSearchByCategoryBasedTextFieldsInitialized = false

--local _internalMethods = {}
	
local categoryLacalizedMap = WFCatSearchByTextConstants.categories

--=========================Logger Debug Methods ========================================
local KSDate_timeNowForAnimation = KSDate.timeNowForAnimation

local totalTimeSpent = {}
local totalCounterOfTables = {}
local printCounters = {}

-- after search issues are analyzed, we need to switch this off.
local tableLoggingNeeded = true
local startTime = 0
local endTime = 0

local minNumberOfWritesForTimeReporting = 10
-------------------------------------------------------------------------------
--------------------------------------------------------------------------------
local function search_debugPrint( messageType, ... )
	if messageType == "e" or tableLoggingNeeded then
		WFSearchDebugLogging.search_debugPrint( "WFSearchTables", logger, messageType, ... )
	end
end

--------------------------------------------------------------------------------
-- we dont expect simultaneous write on a table. So this is stored as a single value
local function markTableWriteEnd( tableName )	
	if tableLoggingNeeded then 
	
		endTime = KSDate_timeNowForAnimation()
		local diff = endTime - startTime 
		local totalTime =  ( totalTimeSpent[ tableName ]  or 0 ) + diff
		totalTimeSpent [ tableName ] = totalTime
		
		local totalCounter = ( totalCounterOfTables[ tableName ]  or 0 ) + 1
		totalCounterOfTables[ tableName ]  = totalCounter
		
		local printCounter = ( printCounters[ tableName ]  or 0 ) + 1
		if printCounter > minNumberOfWritesForTimeReporting then
		 	search_debugPrint( "t" , "Search : WF : table write : total time spent so far in writing " , tableName, totalTime, "and count" , totalCounter  )
			printCounter = printCounter - minNumberOfWritesForTimeReporting
		end
		printCounters[ tableName ] = printCounter
	end
end

--------------------------------------------------------------------------------
-- we dont expect simultaneous write on a table. So this is stored as a single value
local function markTableWriteStart( tableName )	
	if tableLoggingNeeded then 
		startTime = KSDate_timeNowForAnimation()
		--search_debugPrint( "t" , "Search : WF : table write :  in table write of " , tableName, getFormattedTimeForSearchDebug() )
	end
end


--------------------------------------------------------------------------------
local function registerTextSearchableFields( )
	if #WFCatSearchByTextFields == 0 then
		WFCatSearchByTextIndexedData.initializeSearchIndexedFieldDetails( WFCatSearchByTextFields )
	end
end

--================================== Category search realted =====================================================
local _categorySearchMethods ={}
----------------- Category search realted  : makeCategoryBasedDetails -------------------------------------------------

function _categorySearchMethods.makeCategoryBasedDetails( )
	if WFCatSearchByCategoryBasedTextFieldsInitialized == false then
		local WFCatSearchByCategoryText = require 'WFCatSearchByCategoryText'
		WFCatSearchByCategoryText.makeCategoryBasedDetails( WFCatSearchByTextFields, WFCatSearchByCategoryBasedTextFields )
		WFCatSearchByCategoryBasedTextFieldsInitialized = true
	end
end

------------------ Category search realted  : addCategoryTextSearchTables -------------------------------------------------------------

function _categorySearchMethods.addCategoryTextSearchTables( indexStoreSchema )
	if WFCatSearchByCategoryBasedTextFieldsInitialized == false then
		registerTextSearchableFields( )
		_categorySearchMethods.makeCategoryBasedDetails( )
	end
	
	local WFCatSearchByCategoryText = require 'WFCatSearchByCategoryText'
	WFCatSearchByCategoryText.addCategoryTextSearchTables( indexStoreSchema, WFCatSearchByCategoryBasedTextFields )
end	


------------------- Category search realted  : addCategoryTablesSearchSchemas -------------------------------------------------------------
function _categorySearchMethods.addCategoryTablesSearchSchemas( searchSchema )
	if WFCatSearchByCategoryBasedTextFieldsInitialized == false then
		registerTextSearchableFields( )
		_categorySearchMethods.makeCategoryBasedDetails( )
	end
	local WFCatSearchByCategoryText = require 'WFCatSearchByCategoryText'
	WFCatSearchByCategoryText.addSearchToSearchSchema( searchSchema, WFCatSearchByCategoryBasedTextFields )
end	


--================== General data processing =====================================
-- this expects the items in order to add the def relevancy		
-------------------------------------------------------------------------------
local function makeDefaultRelevancyItems( items, defRelevance )
	local processedData = {}
	for i, data in ipairs( items or {} ) do
		processedData[ data ] = defRelevance
	end
	return processedData
end

--------------------------------------------------------------------------------
local function processAdditionalData( docValue, definitions )
	if not definitions then	
		return 
	end
	
	local additionalDataRequirement = definitions.additionalData
	if not additionalDataRequirement then
		return 
	end
	
	local additionalValues = {}
	for _,dataRequiremnentFields in ipairs( additionalDataRequirement ) do
		local val = derefOpt( docValue, unpack( dataRequiremnentFields ) ) 
		table.insert( additionalValues, val or noDataString ) -- to maintain the order
	end
	
	return additionalValues
end

--------------------------------------------------------------------------------
local function getCompleteData( mainFieldValue, docValue, mainDocFieldPath, searchFieldDefinitions )
	if  not searchFieldDefinitions then
		return
	end
	
	if mainFieldValue == nil then
		mainFieldValue = derefOpt( docValue, unpack( mainDocFieldPath ) )
	end
	
	local defaultValueForMainData = searchFieldDefinitions.defaultValue
	local val =  mainFieldValue or defaultValueForMainData
	
	
	if val then -- if val is not there we don't get data as of now. When needed this can be customized
		local retValue = { val }
		local additionalData = processAdditionalData( docValue, searchFieldDefinitions )
		if additionalData  then
			table.insert( retValue, additionalData )
		end
		return retValue
	end
				
end

--------------------------------------------------------------------------------
local function getNonRepeatingDataAndStoreTheCurrent( processedData, existingDataForThisCategory, useKey )
	if existingDataForThisCategory == nil or processedData == nil then
		return processedData
	end
	local newData = {}
	
	if useKey  then
		for key,value in pairs( processedData ) do
			-- we dont want element if they are differing only in letter case
			local valInLower = string.lower( key )
			if not existingDataForThisCategory[ valInLower ]  then
				existingDataForThisCategory[ valInLower ] = true
				newData[ key ] = value
			end
		end 
	else
		for k,value in ipairs( processedData ) do
			-- we dont want element if they are differing only in letter case
			local valInLower = string.lower( value )
			if not existingDataForThisCategory[ valInLower ] then
				existingDataForThisCategory[ valInLower ] = true
				table.insert( newData, value ) 
			end
		end 
	end
	
	
	return newData	
end

--------------------------------------------------------------------------------
local function getOrInitCategoryData( allCategoryData, categoryKey ) 
	local existingDataForThisCategory = allCategoryData[ categoryKey ]
	if existingDataForThisCategory == nil then
		existingDataForThisCategory = {}
		allCategoryData[ categoryKey ] = existingDataForThisCategory
	end
	return existingDataForThisCategory
end


--================== Table processing : text_search =====================================
--------------------------------------------------------------------------------
local function addCompleteTextTable( indexStoreSchema )
	-- reset oldversions of table
	--indexStoreSchema:resetTable( 'text_search' )
	indexStoreSchema:addTable {
		tableName = WFCatSearchByTextConstants.getTableName( 'text_search' ),
		fields = { 'search_text' },
		isFullTextSearch = true,
		useLocalId = true,
		hasOneRowPerDocument = true,
		updateLazily = false, -- don't wait for the table to be used. Do indexing ASAP
		docValueFilter = function( docValue )
			-- can probably be slightly more selective here
			return docValue.type == 'asset'
		end,
		map = function( docValue, emit )
			-- increment this number if we want to reset the table
			TEXT_SEARCH_TABLE_VERSION = 'v12'
			
			if isValidDisplayableAssetDocument( docValue ) then
			
				-- logging
				if tableLoggingNeeded then
					markTableWriteStart( 'text_search' )
				end
				
				local searchableText = {}
				local allCategoryData = {}
				
				for _, searchFieldSpecifications in ipairs( WFCatSearchByTextFields ) do
				
					local mainFields = searchFieldSpecifications.mainDocFieldPath
					local processingDetails = searchFieldSpecifications.processingDetails 
					
					for _, fieldProcessingInfo in ipairs( processingDetails or {} ) do
					
						local categoryKey = fieldProcessingInfo.category	
						local val = getCompleteData( nil, docValue,mainFields, fieldProcessingInfo )
						
						local existingDataForThisCategory = getOrInitCategoryData( allCategoryData, categoryKey ) 
						
						if val then
							local textProcessMethod = fieldProcessingInfo.textProcessMethod
							if textProcessMethod then
								local processedData = textProcessMethod( val, false, categoryKey )
								
								-- dataProcessMethods are expected to return a table
								processedData = getNonRepeatingDataAndStoreTheCurrent( processedData, existingDataForThisCategory )
								
								if processedData  ~= nil then
									for k,v in ipairs(processedData) do
										table.insert(searchableText, v) 
									end 
								end
							elseif val then
								local valInLower = string.lower( val )
								if not existingDataForThisCategory[ valInLower ] then
									existingDataForThisCategory[ valInLower ] = true
									searchableText[ #searchableText + 1 ] = val
								end
							end
						end
						
					end
					
				end
				
				if #searchableText > 0 then
					emit( table.concat( searchableText, ',' ) )
				end
				
				-- logging
				if tableLoggingNeeded then
					markTableWriteEnd( 'text_search' )
				end
			end
		end,
	}
end


--================== Table processing : relevancy =====================================
--------------------------------------------------------------------------------
local function addTextRelevanceInfoTable( indexStoreSchema )
	
	-- reset oldversions of table
	-- indexStoreSchema:resetTable( 'text_search_with_relevance' )
	indexStoreSchema:addTable {
		tableName = WFCatSearchByTextConstants.getTableName('text_search_with_relevance'),
		fields = { 'search_text', 'relevance' },
		useLocalId = true,
		updateLazily = false, -- don't wait for the table to be used. Do indexing ASAP
		docValueFilter = function( docValue )
			-- can probably be slightly more selective here
			return docValue.type == 'asset'
		end,
		map = function( docValue, emit )
			-- increment this number if we want to reset the table
			RELEVANCY_TABLE_VERSION = 'v12'
			
			if not isValidDisplayableAssetDocument( docValue ) then
				return
			end
			-- logging
			if tableLoggingNeeded then
				markTableWriteStart( 'text_search_with_relevance' )
			end
			
			local allCategoryData = {}
			for _, searchFieldSpecifications in ipairs( WFCatSearchByTextFields ) do
			
				local mainFields = searchFieldSpecifications.mainDocFieldPath
				local processingDetails = searchFieldSpecifications.processingDetails 
				
				for _, fieldProcessingInfo in ipairs( processingDetails or {} ) do
				
					local categoryKey = fieldProcessingInfo.category
					local val = getCompleteData( nil, docValue, mainFields, fieldProcessingInfo )
					local existingDataForThisCategory = getOrInitCategoryData( allCategoryData, categoryKey ) 
				
					-- val will return a table as  { mainValue, additionalVale }
					if not val then
						break -- continue
					end
					
					local defaultRel = fieldProcessingInfo.defaultRelevance
					if defaultRel == nil then
						defaultRel = 100
					end
					local relevancyProcessedData 
					
					-- if relevance process mehtod is there, it get the value directly
					-- and it should know how to process the main and additional values as needed
					local relevancyProcessMethod = fieldProcessingInfo.relevancyProcessMethod
					if relevancyProcessMethod then
						-- here we are calling the relevancy process method. 
						relevancyProcessedData = relevancyProcessMethod( val, defaultRel )
					else	
						local textProcessMethod = fieldProcessingInfo.textProcessMethod
						local   contentValue 
						-- if additional value needs processing, a text process method is must
						if textProcessMethod then
							contentValue =  textProcessMethod( val, false, categoryKey )
						else
							-- since text process method is not found, we are interested only in
							-- main data. but we want the content value in a table form
							contentValue = { WFCatSearchByTextDataProcessors.getMainData( val ) }
						end
						
						relevancyProcessedData = makeDefaultRelevancyItems( contentValue, defaultRel ) 
					end
					
					-- for the same category, if we are getting multiple similar values, we may need not store all these
					-- as search is case insestive. 
					relevancyProcessedData = getNonRepeatingDataAndStoreTheCurrent( relevancyProcessedData, existingDataForThisCategory , true )
								
					for k,v in pairs( relevancyProcessedData or  {} )  do
						emit(k,v)
					end 
				end
			end
			
			-- logging
			if tableLoggingNeeded then
				markTableWriteEnd( 'text_search_with_relevance' )
			end
		end,
	}
end

--================== Table processing : text_category (for hinting ) =====================================
local function addTextCategoryTable( indexStoreSchema )
	---- reset oldversions of table
	--indexStoreSchema:resetTable( 'text_search_with_category_info' )
	indexStoreSchema:addTable {
		tableName = WFCatSearchByTextConstants.getTableName( 'text_search_with_category_info' ),
		fields = { 'category', 'search_text' },
		useLocalId = true,
		updateLazily = false, -- don't wait for the table to be used. Do indexing ASAP
		docValueFilter = function( docValue )
			-- can probably be slightly more selective here
			return docValue.type == 'asset'
		end,
		map = function( docValue, emit )
			-- increment this number if we want to reset the table
			CATEGORY_TABLE_VERSION = 'v12'
			
			if isValidDisplayableAssetDocument( docValue ) then
				
				-- logging
				if tableLoggingNeeded then
					markTableWriteStart( 'text_search_with_category_info' )
				end
				
				local allCategoryData = {}
				
				-- for hinting of the data, we need the additional strings
				-- e.g camera is expected before every camera model so that
				-- starting with camera itself,we can show all the camera models
				local processCategoryBasedExtraStrings = true
				for _, searchFieldSpecifications in ipairs( WFCatSearchByTextFields ) do
					
					local mainFields = searchFieldSpecifications.mainDocFieldPath
					local processingDetails = searchFieldSpecifications.processingDetails 
					
					for _, fieldProcessingInfo in ipairs( processingDetails or {} ) do
					
						local searchableText = {}
						local categoryKey = fieldProcessingInfo.category
						
						local categoryLocalized = categoryLacalizedMap[ categoryKey ] 
						local val = getCompleteData( nil, docValue, mainFields, fieldProcessingInfo )
						local existingDataForThisCategory = getOrInitCategoryData( allCategoryData, categoryKey ) 
						
						if val then
							local textProcessMethod = fieldProcessingInfo.textProcessMethod
							if textProcessMethod then
								local processedData = textProcessMethod( val, processCategoryBasedExtraStrings, categoryKey )
								
								-- for the same category, if we are getting multiple similar values, we may need not store all these
								-- as search is case insestive. 
								processedData = getNonRepeatingDataAndStoreTheCurrent( processedData, existingDataForThisCategory )
								
								if processedData  ~= nil then
									for k,v in ipairs(processedData) do
										table.insert(searchableText, v) 
									end 
								end
							elseif val then
								local valInLower = string.lower( val )
								if not existingDataForThisCategory[ valInLower ] then
									existingDataForThisCategory[ valInLower ] = true
									searchableText[ #searchableText + 1 ] = val
								end
							end
							
							if #searchableText > 0 then
								emit( categoryLocalized,table.concat( searchableText, ',' ) )
							end
						end
						
					end
				end
				
				-- logging
				if tableLoggingNeeded then
					markTableWriteEnd( 'text_search_with_category_info' )
				end
			end
		end,
	}
end

--================== WFCatSearchByText : hinting All Assets ====================================
--------------------for  hinting and search tab ------------------------------------
-- This will not be called if ENABLE_SEARCH_COMPLETIONS are off on client side 
function WFCatSearchByText.allAssetsSearchStringWithCategory( wfCatalog, maybeCategory )
	local tableName = WFCatSearchByTextConstants.getTableName( 'text_search_with_category_info' )
	local sqlString = maybeCategory and 
						string.format ( [[ SELECT docId, category, search_text FROM %s WHERE category = ? ORDER BY docId ]],	tableName )
						or string.format ( [[ SELECT docId, category, search_text FROM %s ORDER BY docId ]],	tableName )

	return wfCatalog:indexStore():query( {

			calm = 60,

			sql = sqlString,

			maybeCategory 

		} )
end


--================== WFCatSearchByText : hinting Selected Assets ====================================
--------------------for  hinting and search tab ------------------------------------
-- This will not be called if ENABLE_SEARCH_COMPLETIONS are off on client side 
function WFCatSearchByText.collectSelectedAssetsSearchStrings( wfCatalog, assetIds )
	local tableName = WFCatSearchByTextConstants.getTableName( 'text_search_with_category_info' )
	local sqlString =  string.format( [[ SELECT 
						%s.docId, %s.category, %s.search_text 
						FROM TEMP_ids 
						LEFT JOIN %s 
						ON %s.docId = TEMP_ids.docId  ORDER BY %s.docId ]], tableName,tableName,tableName,tableName,tableName, tableName )
	
	return wfCatalog:indexStore():query( { 

			sql = sqlString,

			tempIds = assetIds,

			calm = WFCatSearchQuery.defaultQueryCalmRate

		})
end

--================== WFCatSearchByText : generateIndexSchema ====================================
function WFCatSearchByText.generateIndexSchema( indexStoreSchema )
	if not ENABLE_LOCAL_TEXT_SEARCH  then
		return
	end
	
	------ ------ ------ ------
	registerTextSearchableFields()
	if ENABLE_CATEGORY_SEARCH then
		_categorySearchMethods.makeCategoryBasedDetails()
	end
	
	------ ------ ------ ------
	addCompleteTextTable( indexStoreSchema )
	if ENABLE_SEARCH_COMPLETIONS then
		addTextCategoryTable( indexStoreSchema )
	end
	
	addTextRelevanceInfoTable( indexStoreSchema )
	if ENABLE_CATEGORY_SEARCH then
		_categorySearchMethods.addCategoryTextSearchTables( indexStoreSchema )
	end
	
	-- For a time after initial prerelease and before Internal Release, Nimbus used 
	-- the table name text_search_with_source instead of with_category, add this to
	-- the tables to drop so any existing data gets cleaned up
	indexStoreSchema:dropTable( 'text_search_with_source_info_v1' )
end

--================== WFCatSearchByText : addSearchToSearchSchema ====================================
function WFCatSearchByText.addSearchToSearchSchema( searchSchema )
	if not ENABLE_LOCAL_TEXT_SEARCH  then
		return
	end
	
	local tableName = WFCatSearchByTextConstants.getTableName( 'text_search' )
	local sqlString = string.format( [[SELECT docid FROM %s WHERE search_text MATCH ? ORDER BY docid]] , tableName )
	searchSchema:addSearchableData {
		term = 'text',
		dataType = 'fullTextSearch',
		sql = sqlString,
			-- unfortunately, SQLite's built-in docId for full text search differs from ours
			-- by case so, docid here is correct (even if SQLite is generally permissive about
			-- column name case in queries)
		query_params = 'text',
	}
	
	if ENABLE_CATEGORY_SEARCH then
		_categorySearchMethods.addCategoryTablesSearchSchemas( searchSchema )
	end
end

------------------------------------------------------------
function  WFCatSearchByText.isLocalTextSearchSupported()
	return ENABLE_LOCAL_TEXT_SEARCH
end

--==============================================================================

return WFCatSearchByText
