--[[----------------------------------------------------------------------------

WFCatSyncClientBase.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
___________________

Copyright 2019 Adobe

All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Adobe and its suppliers, if any. The intellectual
and technical concepts contained herein are proprietary to Adobe
and its suppliers and are protected by all applicable intellectual
property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Adobe.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFCloseSupport = require 'WFCloseSupport'
local WFArguments = require 'WFArguments'
local WFConfiguration = require 'WFConfiguration'
local WFRxRunUntilDone = require 'WFRxRunUntilDone'

	-- WF.oz_client.managed_catalog
local WFCatUser = require 'WFCatUser'
local WFCatSpec = require 'WFCatSpec'
local WFCatalog = require 'WFCatalog'
local WFCatMetadataPuller = require 'WFCatMetadataPuller'
local WFCatMetadataPusher = require 'WFCatMetadataPusher'
local WFCatSpaceFeedsManager = require 'WFCatSpaceFeedsManager'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )
local Observable = Rx.Observable
local runUntilDone = WFRxRunUntilDone

--------------------------------------------------------------------------------

local pusherDocumentBatchCount = WFConfiguration.Sync.pusherDocumentBatchCount
local pullerDocumentBatchCount = WFConfiguration.Sync.pullerDocumentBatchCount

--==============================================================================

-- @mark WFCatSyncClientBase namespace

--------------------------------------------------------------------------------

local WFCatSyncClientBase = {

	_user = false,
		-- (WFCatUser)
		
	_catSpec = false,
		-- (WFCatSpec)
		
	_catalog = false,
		-- (WFCatalog)
		
	_catalogPuller = false,
		-- (WFCatMetadataPuller)
		
	_catalogPusher = false,
		-- (WFCatMetadataPusher)
		
	_spaceFeedsManager = false,
		-- (WFCatSpaceFeedsManager)
		
	_checkOzForCatalogUpdatesSubscription = false,
		-- (Rx Disposable)
		
	WFCloseSupport.addCloseSupportMethods,
	
}

--------------------------------------------------------------------------------

function WFCatSyncClientBase:__init( options )

	check( 'WFCatSyncClientBase', 'table' )
	
	-- Required settings
	
	local serverBase = assert( options.serverBase )
	local userId = assert( options.userId )
	local userInfoCacheFolder = assert( options.userInfoCacheFolder )
	local catalogId = assert( options.catalogId )
	local rootCatalogPath = assert( options.rootCatalogPath )
	
	-- Optional settings
	
	local isSyncClientForProfileCatalog = options.isSyncClientForProfileCatalog
	local pullerTaskDelays = options.metadataPullerArgs
	local noFileAccessDb = options.noFileAccessDb
	local noCatalogSearch = options.noCatalogSearch
	local pullerPrefetch = options.pullerPretfetch
	local pusherDelayBeforeSendingUpdates = options.pusherDelayBeforeSendingUpdates
	local albumAssetsForFilteredAssetsQueryFn = options.albumAssetsForFilteredAssetsQueryFn
	
	-- Set up sync infrastructure
	
	local user = WFCatUser( serverBase, userId, userInfoCacheFolder )
	
	local catSpec = WFCatSpec( user, options.catalogDoc or catalogId )

	self._catSpec = catSpec

	local catArgs = {
		docStorePath = assert( options.docStorePath ),
		indexSpecScript = assert( options.indexSpecScript ),
		noFileAccessDb = noFileAccessDb,
		noCatalogSearch = noCatalogSearch,
	}
	
	local wfCatalog = WFCatalog( catSpec, rootCatalogPath, catArgs )
	
	-- Sync Down
	
	local pullerOptions = {
	
		prefetch = pullerPrefetch,
		documentBatchCount = pullerDocumentBatchCount,
		continuousTaskDelays = pullerTaskDelays,
		isCatalogPuller = true
		
	}
	
	local pullerDescription
	
	if isSyncClientForProfileCatalog then
		pullerDescription = "Profile Catalog Puller"
	else
	 	pullerDescription = "Catalog Puller"
	end
	
	local puller = WFCatMetadataPuller( wfCatalog, pullerDescription, pullerOptions )
	
	self:setOwned {
		_user = user,
		_catalog = wfCatalog,
		_catalogPuller = puller,
	}
		
	if not isSyncClientForProfileCatalog then
	
	 	-- Sync Up
		
		local catalogPusherOptions = {
			delayBeforeSendingUpdates = pusherDelayBeforeSendingUpdates or 2,
			documentBatchCount = pusherDocumentBatchCount or 100,
				-- Prevents the pusher from being overly aggressive at sending updates.
			albumAssetsForFilteredAssetsQueryFn = albumAssetsForFilteredAssetsQueryFn
		}
	
		local catalogPusher = WFCatMetadataPusher( wfCatalog, catalogPusherOptions )
	
		self:setOwned {
			_catalogPusher = catalogPusher,
		}
	
		-- WF-256: Reset Catalog MetadataPuller polling frequency after sending data to Oz.
	
		self._checkOzForCatalogUpdatesSubscription = self._catalogPusher:isDoingWork()
			:filter( function( isDoingWork )
				return isDoingWork
			end )
			:subscribe( function()
				self._catalogPuller:checkOzForUpdates()
			end )
	
		-- Space-based Metadata Sync
	
		local spaceFeedsManagerArgs = {
			taskDelays = pullerTaskDelays,
			pullerDocumentBatchCount = pullerDocumentBatchCount,
			pusherDocumentBatchCount = pusherDocumentBatchCount or 100,
			albumAssetsForFilteredAssetsQueryFn = albumAssetsForFilteredAssetsQueryFn,
			testMode = options.testMode -- Only used by unit tests.
		}
	
		local spaceFeedsManager = WFCatSpaceFeedsManager( wfCatalog, spaceFeedsManagerArgs )

		self:setOwned {
			_spaceFeedsManager = spaceFeedsManager
		}

		-- WF-398: Pause space-based sync during initial sync.

		self._catalogPuller:onceInitialSyncIsComplete()
			
			:map( function()
			
				self._spaceFeedsManager:reportCatalogInitialSyncIsComplete()
				
				return true
				
			end )
			
			:call( runUntilDone )
		
	end
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:close()

	check( 'WFCatSyncClientBase' )
	
	if self._checkOzForCatalogUpdatesSubscription then
	
		self._checkOzForCatalogUpdatesSubscription:dispose()
		self._checkOzForCatalogUpdatesSubscription = nil
	
	end
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:getCatalog()

	check( 'WFCatSyncClientBase' )
	
	return self._catalog
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:getCatSpec()

	check( 'WFCatSyncClientBase' )
	
	return self._catSpec
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:getCatalogPuller()

	check( 'WFCatSyncClientBase' )
	
	return self._catalogPuller
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:getCatalogPusher()

	check( 'WFCatSyncClientBase' )
	
	return self._catalogPusher
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:startSync()

	check( 'WFCatSyncClientBase' )
	
	self._catalogPuller:getUpdatesContinuously()
	
	if self._catalogPusher then
		self._catalogPusher:sendUpdatesContinuously()
	end
	
	if self._spaceFeedsManager then
		self._spaceFeedsManager:startSendingUpdates()
		self._spaceFeedsManager:startFetchingUpdates()
	end
	
	return Observable.empty()
	
end

------------------------------------------------------------------------

function WFCatSyncClientBase:stopSync()

	check( 'WFCatSyncClientBase' )
	
	self._catalogPuller:stopContinuousUpdates()
	
	if self._catalogPusher then
		self._catalogPusher:stopContinuousUpdates()
	end
	
	if self._spaceFeedsManager then
		self._spaceFeedsManager:stopSendingUpdates()
		self._spaceFeedsManager:stopFetchingUpdates()
	end
	
	return Observable.empty()
	
end

------------------------------------------------------------------------

function WFCatSyncClientBase:startSyncInResidualMode()

	check( 'WFCatSyncClientBase' )
	
	self._catalogPuller:getUpdatesContinuously()

	if self._spaceFeedsManager then
		self._spaceFeedsManager:startSendingUpdates()
		self._spaceFeedsManager:startFetchingUpdates()
	end

	return Observable.empty()
	
end

------------------------------------------------------------------------

function WFCatSyncClientBase:stopSyncInResidualMode()

	check( 'WFCatSyncClientBase' )

	self._catalogPuller:stopContinuousUpdates()

	if self._spaceFeedsManager then
		self._spaceFeedsManager:stopSendingUpdates()
		self._spaceFeedsManager:stopFetchingUpdates()
	end

	return Observable.empty()

end

------------------------------------------------------------------------
	
function WFCatSyncClientBase:startSpaceMetadataSyncInFreemiumMode()

	check( 'WFCatSyncClientBase' )

	if self._spaceFeedsManager then

		self._spaceFeedsManager:reportCatalogInitialSyncIsComplete()
			-- WF-749: In Freemium mode, the catalog does not sync. In other modes, we wait
			-- for initial sync to complete before syncing the spaces; initial catalog
			-- sync will never complete in freemium mode so we must activate this trigger
			-- manually. This allows space-feed sync to work.

		self._spaceFeedsManager:startSendingUpdates()
		self._spaceFeedsManager:startFetchingUpdates()	

	end

	return Observable.empty()

end

------------------------------------------------------------------------

function WFCatSyncClientBase:stopSpaceMetadataSyncInFreemiumMode()

	check( 'WFCatSyncClientBase' )

	if self._spaceFeedsManager then
		self._spaceFeedsManager:stopSendingUpdates()
		self._spaceFeedsManager:stopFetchingUpdates()
	end

	return Observable.empty()

end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:pausePusherReturningCounter()

	-- Used for synchronizing trash operations with the pusher.
	-- Since the space-feed assets don't end up in the trash, we
	-- only have to do this for the catalog pusher.
	
	check( 'WFCatSyncClientBase' )

	local pauseCatalogPusherReturningCounter = Observable.just( 0 )

	if self._catalogPusher then
		pauseCatalogPusherReturningCounter = self._catalogPusher:pauseReturningCounter()
	end
	
	return pauseCatalogPusherReturningCounter
	
end

------------------------------------------------------------------------

function WFCatSyncClientBase:resumePusher()

	-- Used for synchronizing trash operations with the pusher.
	-- Since the space-feed assets don't end up in the trash, we
	-- only have to do this for the catalog pusher.
	
	check( 'WFCatSyncClientBase' )

	if self._catalogPusher then
		self._catalogPusher:resume()
	end
	
	return Observable.empty()

end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:_DEBUG_startMetadataPusher()

	check( 'WFCatSyncClientBase' )
	
	if self._catalogPusher then
		self._catalogPusher:sendUpdatesContinuously()
	end
	
	if self._spaceFeedsManager then
		self._spaceFeedsManager:startSendingUpdates()
	end

	return Observable.empty()
	
end

------------------------------------------------------------------------

function WFCatSyncClientBase:_DEBUG_stopMetadataPusher()

	check( 'WFCatSyncClientBase' )
	
	if self._catalogPusher then
		self._catalogPusher:stopContinuousUpdates()
	end
	
	if self._spaceFeedsManager then
		self._spaceFeedsManager:stopSendingUpdates()
	end

	return Observable.empty()

end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:setUserIsOverQuota( ... )

	check( 'WFCatSyncClientBase' )
	
	if self._catalogPusher then
		self._catalogPusher:setUserIsOverQuota( ... )
	end
	
	return Observable.empty()
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:setUserIsExpired( ... )

	check( 'WFCatSyncClientBase' )
	
	if self._catalogPusher then
		self._catalogPusher:setUserIsExpired( ... )
	end
	
	return Observable.empty()
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:isUploadingMetadata()

	check( 'WFCatSyncClientBase' )
	
	local isPushingMetadata = Observable.just( false )
	
	if self._catalogPusher then
	
		isPushingMetadata = self._catalogPusher:isDoingWork():distinctUntilChanged()
		
	end
	
	if self._spaceFeedsManager then
	
		isPushingMetadata = Observable.combineLatest( 
				function( isPushingCatalogMetadata, isPushingSpaceFeedMetadata )
					return isPushingCatalogMetadata or isPushingSpaceFeedMetadata
				end, 
				self._catalogPusher:isDoingWork():distinctUntilChanged(),
				self._spaceFeedsManager:isUploadingMetadata() )
				
	end
	
	return isPushingMetadata
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:isDownloadingMetadata()

	check( 'WFCatSyncClientBase' )
	
	local isPullingMetadata = self._catalogPuller:isDoingWork()
	
	if self._spaceFeedsManager then
	
		isPullingMetadata = Observable.combineLatest( 
				function( isPullingCatalogMetadata, isPullingSpaceMetadata )
					return isPullingCatalogMetadata or isPullingSpaceMetadata
				end, 
				self._catalogPuller:isDoingWork(),
				self._spaceFeedsManager:isDownloadingMetadata() )
	
	end
	
	return isPullingMetadata
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:getLatestDownloadedAssetIdsStream()

	check( 'WFCatSyncClientBase' )
	
	return self._catalogPuller:getLatestDownloadedAssetIdsStream()
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:isOzCatalogDeleted()

	check( 'WFCatSyncClientBase' )
	
	return self._catalogPuller:isOzCatalogDeleted()
	
end

--------------------------------------------------------------------------------

function WFCatSyncClientBase:getRequiredUpdateStream()

	check( 'WFCatSyncClientBase' )
	
	return self._user:getRequiredUpdateStream()
	
end

--------------------------------------------------------------------------------

return Rx.Class( 'WFCatSyncClientBase', WFCatSyncClientBase )
