--[[----------------------------------------------------------------------------

WFCatSyncStatus.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.oz_client.managed_catalog
local WFCatIndexUtils = require 'WFCatIndexUtils'
local WFCatFileSpec = require 'WFCatFileSpec'
local WFCatIdleDownloader = require 'WFCatIdleDownloader'
local WFCatOzErrors = require 'WFCatOzErrors'
local WFCatTrashUtils = require 'WFCatTrashUtils'
local NmbFileManagementSettings = require 'NmbFileManagementSettings'

	-- WF.oz_client.shared.utils
local LrAssetUtils = require 'LrAssetUtils'

	-- WF.sqlite.database
local WFSQLiteUtils = require 'WFSQLiteUtils'

--------------------------------------------------------------------------------

local isAssetDocument = assert( WFCatIndexUtils.isAssetDocument )
local isValidDisplayableAssetDocument = assert( WFCatIndexUtils.isValidDisplayableAssetDocument )
local isFileBasedAssetDocument = assert( WFCatIndexUtils.isFileBasedAssetDocument )
local isVideoDocument = assert( WFCatIndexUtils.isVideoDocument )

local WFCatTrashUtils_isDeleteTrash = assert( WFCatTrashUtils.isDeleteTrash )

local derefOpt = KSTableUtils.derefOpt

local isLatestMetadataUploaded = assert( LrAssetUtils.isLatestMetadataUploaded )

--------------------------------------------------------------------------------

local WFCatSyncStatus = {}

--------------------------------------------------------------------------------

-- !!! CAUTION [jandrew 2016-12-01]: When asset sync status enums are changed,
-- we must also bump the sync status version in the _map methods to force re-indexing. 
-- IF YOU ADD OR MODIFY THE SYNC STATUS ENUMERATION, YOU MUST INCREMENT THE SYNC
-- STATUS VERSION NUMBERS.

local kAssetNeedsMetadataUpload = 'NeedsMetadataUploaded'
local kAssetNeedsOriginalUpload = 'NeedsOriginalUploaded'
local kAssetNeedsOriginalDownload = 'NeedsOriginalDownloaded'
local kAssetNeedsOriginalDownloadForOffline = 'NeedsOriginalDownloadedForOffline'
local kAssetNeedsProxyDownloadForOffline = 'NeedsProxyDownloadedForOffline'
local kAssetNeedsProxyDownload = 'NeedsProxyDownloaded'
local kAssetNeedsDevelopXmpDownload = 'NeedsDevelopXmpDownloaded'
local kAssetComplete = 'Completed'
local kAssetCompleteNoMaster = 'CompletedNoMaster'
local kAssetCompleteNoNegatives = 'CompletedNoNegatives'
local kAssetCompleteForOffline = 'CompletedForOffline'
local kAssetStorageLimitExceeded = 'StorageLimitExceeded'
local kAssetInvalidOriginal = 'InvalidLocalOriginal'
local kAssetOriginalIsOnOfflineVolume = 'OriginalIsOnOfflineVolume'
local kAssetProxyIsOnOfflineVolume = 'ProxyIsOnOfflineVolume'
local kAssetNeedsOriginalDownloadToOfflineVolume = 'NeedsOriginalDownloadToOfflineVolume'
local kAssetNeedsProxyDownloadToOfflineVolume = 'NeedsProxyDownloadToOfflineVolume'
local kAssetNeedsOriginalUploadFromOfflineVolume = 'NeedsOriginalUploadedButOnOfflineVolume'
local kAssetNeedsOriginalMigration = 'NeedsOriginalMigration'
local kAssetNeedsProxyMigration = 'NeedsProxyMigration'
local kAssetOriginalDownloadError = 'OriginalDownloadError'
local kAssetProxyDownloadError = 'ProxyDownloadError'
local kAssetDevelopXmpDownloadError = 'DevelopXmpDownloadError'
local kAssetNeedsOriginalDownloadToOutOfSpaceVolume = 'NeedsOriginalDownloadToOutOfSpaceVolume'
local kAssetNeedsProxyDownloadToOutOfSpaceVolume = 'NeedsOriginalProxyToOutOfSpaceVolume'
local kAssetSyncPaused = 'SyncPaused'
local kAssetSyncNoConnection = 'SyncNoConnection'
local kAssetTrialOrSubscriptionExpired = 'TrialOrSubscriptionExpired'
local kAssetStackComplete = 'StackComplete'
local kAssetInsufficientPermissionError = 'InsufficientPermissionError'

local kAssetSyncStatusById = {
	kAssetNeedsMetadataUpload,						-- 1
	kAssetNeedsOriginalUpload,						-- 2
	kAssetNeedsOriginalDownload,					-- 3
	kAssetNeedsOriginalDownloadForOffline,			-- 4
	kAssetNeedsProxyDownload,						-- 5
	kAssetNeedsDevelopXmpDownload,					-- 6
	kAssetComplete,									-- 7
	kAssetCompleteNoMaster,							-- 8
	kAssetStorageLimitExceeded,						-- 9
	kAssetInvalidOriginal,							-- 10
	kAssetOriginalIsOnOfflineVolume,            	-- 11
	kAssetProxyIsOnOfflineVolume,            		-- 12
	kAssetNeedsOriginalDownloadToOfflineVolume, 	-- 13
	kAssetNeedsProxyDownloadToOfflineVolume,		-- 14
	kAssetNeedsOriginalMigration,					-- 15
	kAssetNeedsProxyMigration,						-- 16
	kAssetNeedsOriginalUploadFromOfflineVolume,		-- 17
	kAssetOriginalDownloadError, 					-- 18
	kAssetProxyDownloadError,						-- 19
	kAssetDevelopXmpDownloadError,					-- 20
	kAssetNeedsOriginalDownloadToOutOfSpaceVolume, 	-- 21
	kAssetNeedsProxyDownloadToOutOfSpaceVolume, 	-- 22
	kAssetSyncPaused,								-- 23
	kAssetSyncNoConnection,							-- 24
	kAssetNeedsProxyDownloadForOffline,				-- 25
	kAssetTrialOrSubscriptionExpired,				-- 26
	kAssetStackComplete,							-- 27
	kAssetCompleteForOffline,						-- 28
	kAssetCompleteNoNegatives,						-- 29
	kAssetInsufficientPermissionError,				-- 30
}

local kAssetSyncStatus = KSTableUtils.invert( kAssetSyncStatusById )

--------------------------------------------------------------------------------

local kAssetHasLocalOriginal = 'OriginalOnDisk'
local kAssetHasLocalOriginalForOffline = 'OriginalOnDiskForOffline'
local kAssetHasLocalProxy = 'ProxyOnDisk'
local kAssetHasNothing = 'NothingOnDisk'

local kAssetLocalFileTypeById = {
	kAssetHasLocalOriginal,				-- 1
	kAssetHasLocalOriginalForOffline,	-- 2
	kAssetHasLocalProxy,				-- 3
	kAssetHasNothing,					-- 4
}

local kAssetLocalFileType = KSTableUtils.invert( kAssetLocalFileTypeById )

--------------------------------------------------------------------------------

local kAlbumSyncing = 'syncing'
local kAlbumComplete = 'synced'
local kAlbumSyncPaused = 'paused'
local kAlbumSyncNoConnection = 'offline'
local kAlbumSyncingForOfflineUse = 'syncingForOfflineUse'
local kAlbumCompleteForOfflineUse = 'syncedForOfflineUse'
local kAlbumSyncingForOfflineUseToOfflineVolume = 'syncingForOfflineUseToOfflineVolume'
local kAlbumSyncingForOfflineUseToOutOfSpaceVolume = 'syncingForOfflineUseToOutOfSpaceVolume'

local kAlbumSyncStatusById = {
	kAlbumSyncing,									-- 1
	kAlbumComplete,									-- 2
	kAlbumSyncPaused,   							-- 3
	kAlbumSyncNoConnection, 						-- 4
	kAlbumSyncingForOfflineUse, 					-- 5
	kAlbumCompleteForOfflineUse,					-- 6
	kAlbumSyncingForOfflineUseToOfflineVolume, 		-- 7
	kAlbumSyncingForOfflineUseToOutOfSpaceVolume 	-- 8
}

local kAlbumSyncStatus = KSTableUtils.invert( kAlbumSyncStatusById )

--------------------------------------------------------------------------------

local kOzHasOriginal = 'OzHasOriginal'
local kOzHasProxy = 'OzHasSmartPreview'
local kOzHasNothing = 'OzHasNothing'

--------------------------------------------------------------------------------

local function docValueHasLocalOriginal( docValue )

	local localOriginalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
	local localOriginalRelativePath = WFCatFileSpec.getOriginalRelativePathFromDocValue( docValue )
	return localOriginalVolumeGuid and localOriginalRelativePath

end

--------------------------------------------------------------------------------

local function docValueHasLocalProxy( docValue )

	local localProxyVolumeGuid = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )
	local localProxyRelativePath = WFCatFileSpec.getProxyRelativePathFromDocValue( docValue )
	return localProxyVolumeGuid and localProxyRelativePath

end

--------------------------------------------------------------------------------

local function assetNeeded( docValue )

	local ozHasURIForOriginal = WFCatFileSpec.getOriginalLinkFromDocValue( docValue ) ~= nil
	local haveLocalPathForOriginal = docValueHasLocalOriginal( docValue )
	local needsOriginalDownload = not haveLocalPathForOriginal and ozHasURIForOriginal 

	local ozHasURIForProxy = WFCatFileSpec.getProxyLinkFromDocValue( docValue ) ~= nil
	local haveLocalPathForProxy = docValueHasLocalProxy( docValue )
	local needsProxyDownload = not haveLocalPathForOriginal and not haveLocalPathForProxy and ozHasURIForProxy
	
	local haveLocalPathForXmp = WFCatFileSpec.getXmpDevelopInfoFromDocValue( docValue ) ~= nil
	local ozHasURIForXmp = WFCatFileSpec.getXmpDevelopLinkFromDocValue( docValue ) ~= nil
	local needsDevelopXmpDownload = not haveLocalPathForXmp and ozHasURIForXmp 
	
	local isMarkedForOffline = WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue )

	local sufficientRenditionFoundOnOz = WFCatFileSpec.checkOzForNonThumbnailRendition( docValue )

	local waitingForProxyOnOz = derefOpt( docValue, '_serverProvidedMetadata', 'startedWaitingForProxyTimestamp' )

	local keepAllOriginals = NmbFileManagementSettings.keepAllOriginals()
	local keepAllProxies = NmbFileManagementSettings.keepAllProxies()
	
	-- SPAM [jandrew 2017-09-12]: We pass `nil` for two of the parameters because they should be
	-- ignored when calculating what we NEED to sync as opposed to what we SHOULD try to sync;
	-- the difference between "need" and "should" in this context is that an asset that needs
	-- syncing still needs to be synced even if there's no disk space available. On the other hand,
	-- an asset that needs to be synced SHOULD NOT be synced if there's no disk space available.

	local binaryType, isBinaryRequired = WFCatIdleDownloader.whatToDownload(
		keepAllOriginals,
		keepAllProxies,
		needsOriginalDownload,
		haveLocalPathForOriginal,
		needsProxyDownload,
		haveLocalPathForProxy,
		needsDevelopXmpDownload,
		isMarkedForOffline,
		nil, -- originalVolumeOfflineOrFull.
		nil, -- proxyVolumeOfflineOrFull
		sufficientRenditionFoundOnOz,
		waitingForProxyOnOz )

	-- Sync status is not affected by purely idle downloads. It's only affected by idle
	-- downloads that are required. 
	
	if not isBinaryRequired then 
		binaryType = nil
	end

	return binaryType

end 

--------------------------------------------------------------------------------

local function isStack( docValue )
	
	return not not ( docValue.subtype == 'stack' )

end 

--------------------------------------------------------------------------------

local function doesAssetNeedOriginalUploaded( docValue )
	
	local ozMasterUrl, isOzOriginalInvalid = WFCatFileSpec.getOriginalLinkFromDocValue( docValue )
	local hasLocalOriginal = docValueHasLocalOriginal( docValue )
	local hasOriginalSourceLinkForCopy = derefOpt( docValue, '_localOnly', 'sourceOriginalLink' )
	
	local imageSha256 = derefOpt( docValue, 'importSource', 'sha256' )
	local localOriginalSha256 = hasLocalOriginal and WFCatFileSpec.getLocalOriginalSha256( docValue )
	local isLocalOriginalValid = localOriginalSha256 == imageSha256
	
	local isSpaceAsset = not not docValue._spaceId
	local isSpaceAssetWithImpliedMaster = isSpaceAsset and derefOpt( docValue, '_serverProvidedMetadata',
											 'links',
											 '/rels/master',
											 'exists' ) == true
	
	local ozHasMaster = ozMasterUrl or isSpaceAssetWithImpliedMaster
	
	local needsUpload = false
	
	if ( hasLocalOriginal or hasOriginalSourceLinkForCopy ) and not ozHasMaster then
	
		needsUpload = true
		
	elseif hasLocalOriginal and ozMasterUrl and isOzOriginalInvalid and isLocalOriginalValid then
	
		needsUpload = true
	
	end
	
	return needsUpload
	
end

--------------------------------------------------------------------------------

local function isMarkedForOfflineUse( docValue )
	
	return WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue )

end

--------------------------------------------------------------------------------

local function doesAssetNeedOriginalDownloaded( docValue )
	
	return assetNeeded( docValue ) == 'original'

end

--------------------------------------------------------------------------------

local function doesAssetNeedProxyDownloaded( docValue )

	return assetNeeded( docValue ) == 'proxy'
	
end

--------------------------------------------------------------------------------

local function doesAssetNeedDevelopXmpDownloaded( docValue )
	
	return assetNeeded( docValue ) == 'xmp_develop'
	
end

--------------------------------------------------------------------------------

local function hasInvalidOriginalError( docValue )
	return KSTableUtils.derefOpt( docValue, '_localOnly', 'localOriginalError' )
end

--------------------------------------------------------------------------------

local function hasOriginalDownloadError( docValue )

	local originalHref = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/master', 'href' )
	
	local permanentOriginalDownloadErrorForCurrentHref = KSTableUtils.derefOpt( docValue, '_localOnly', 'binaryDownloadErrors', '/rels/master', originalHref )
	
	return not not permanentOriginalDownloadErrorForCurrentHref
	
end

--------------------------------------------------------------------------------

local function hasProxyDownloadError( docValue )
	
	local proxyRel
	if isVideoDocument( docValue ) then
		proxyRel = '/rels/rendition_type/720p'
	else
		proxyRel = '/rels/proxy_type/2560'
	end
	
	local proxyHref = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links', proxyRel, 'href' )
	
	local permanentProxylDownloadErrorForCurrentHref = KSTableUtils.derefOpt( docValue, '_localOnly', 'binaryDownloadErrors', proxyRel, proxyHref )
	
	return not not permanentProxylDownloadErrorForCurrentHref
end

--------------------------------------------------------------------------------

local function hasXmpDevelopDownloadError( docValue )

	local xmpDevelopHref = KSTableUtils.derefOpt( docValue, '_serverProvidedMetadata', 'links', '/rels/xmp/develop', 'href' )
	
	local permanentXmpDevelopDownloadErrorForCurrentHref = KSTableUtils.derefOpt( docValue, '_localOnly', 'binaryDownloadErrors', 'rels/xmp/develop', xmpDevelopHref )
	
	return not not permanentXmpDevelopDownloadErrorForCurrentHref
	
end

--------------------------------------------------------------------------------

-- !!! CAUTION [jandrew 2016-12-01]: When the following function, or functions within,
-- are changed and result in different outputs, we must bump the sync status version
-- numbers found in the _map methods to force re-indexing. 

function WFCatSyncStatus.getAssetStatus( docValue )

	local assetSyncStatusId

	if docValue._deleted then

		local links = KSTableUtils.derefOpt( docValue,
									 '_serverProvidedMetadata',
									 'links' )

		if not links then
			-- This gets executed if you delete an asset that hasn't been
			-- uploaded yet, it is called by the _map function in asset_status_table().
			return nil
		end

		assetSyncStatusId = kAssetNeedsMetadataUpload

	elseif isStack( docValue ) then
		
		if not isLatestMetadataUploaded( docValue ) then
			assetSyncStatusId = kAssetNeedsMetadataUpload
		else
			assetSyncStatusId = kAssetStackComplete
		end
	
	else
	
		if not isLatestMetadataUploaded( docValue ) then

			if WFCatOzErrors.hasStorageExceededError( docValue ) then
				assetSyncStatusId = kAssetStorageLimitExceeded
			elseif WFCatOzErrors.hasTrialOrSubscriptionExpiredError( docValue ) then
				assetSyncStatusId = kAssetTrialOrSubscriptionExpired
			elseif WFCatOzErrors.hasInsufficientPermissionError( docValue ) then
				assetSyncStatusId = kAssetInsufficientPermissionError
			else
				assetSyncStatusId = kAssetNeedsMetadataUpload
			end

		elseif doesAssetNeedOriginalUploaded( docValue ) then

			if hasInvalidOriginalError( docValue ) then
				assetSyncStatusId = kAssetInvalidOriginal
			elseif WFCatOzErrors.hasStorageExceededError( docValue ) then
				assetSyncStatusId = kAssetStorageLimitExceeded
			elseif WFCatOzErrors.hasTrialOrSubscriptionExpiredError( docValue ) then
				assetSyncStatusId = kAssetTrialOrSubscriptionExpired
			elseif WFCatOzErrors.hasInsufficientPermissionError( docValue ) then
				assetSyncStatusId = kAssetInsufficientPermissionError
			else
				assetSyncStatusId = kAssetNeedsOriginalUpload
			end

		elseif doesAssetNeedOriginalDownloaded( docValue ) then
			
			if isMarkedForOfflineUse( docValue ) or
				NmbFileManagementSettings.keepAllOriginals() then
				assetSyncStatusId = kAssetNeedsOriginalDownloadForOffline
			else
				assetSyncStatusId = kAssetNeedsOriginalDownload
			end

		elseif doesAssetNeedProxyDownloaded( docValue ) then
			
			if isMarkedForOfflineUse( docValue ) or
				NmbFileManagementSettings.keepAllProxies() then
				assetSyncStatusId = kAssetNeedsProxyDownloadForOffline
			else
				assetSyncStatusId = kAssetNeedsProxyDownload
			end
			
		elseif doesAssetNeedDevelopXmpDownloaded( docValue ) then
			
			assetSyncStatusId = kAssetNeedsDevelopXmpDownload
		
		else
			
			if hasOriginalDownloadError( docValue ) then
				
				assetSyncStatusId = kAssetOriginalDownloadError
				
			elseif hasProxyDownloadError( docValue ) then
			
				assetSyncStatusId = kAssetProxyDownloadError
			
			elseif hasXmpDevelopDownloadError( docValue ) then
			
				assetSyncStatusId = kAssetDevelopXmpDownloadError
				
			elseif WFCatFileSpec.getOriginalLinkFromDocValue( docValue ) and 
				( isMarkedForOfflineUse( docValue ) or
				NmbFileManagementSettings.keepAllOriginals() ) then
				
				assetSyncStatusId = kAssetCompleteForOffline
				
			elseif WFCatFileSpec.getOriginalLinkFromDocValue( docValue ) then
				
				assetSyncStatusId = kAssetComplete
			
			elseif WFCatFileSpec.getProxyLinkFromDocValue( docValue ) then
			
				assetSyncStatusId = kAssetCompleteNoMaster
			
			else
			
				assetSyncStatusId = kAssetCompleteNoNegatives
			
			end

		end

	end

	return kAssetSyncStatus[ assetSyncStatusId ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )

	local allVolumeAgents = wfCatalog:getAllVolumeAgents()

	local volumeAgentStatusStreams = {}
	
	local originalsPathInfo = wfCatalog:fileSpec():originalsPathInfo()
	local volumeGuidForOriginals = originalsPathInfo.volumeGuid
	
	local proxiesPathInfo = wfCatalog:fileSpec():proxiesPathInfo()
	local volumeGuidForProxies = proxiesPathInfo.volumeGuid
	
	-- Build up a table of volume agent status streams.

	for _, agent in pairs( allVolumeAgents ) do
		table.insert( volumeAgentStatusStreams, agent:getVolumeStatusStream() )
	end

	table.insert( volumeAgentStatusStreams, Rx.Observable.just( nil ) )
		-- Add a nil observable in case there are no volume streams. This can
		-- happen in unit tests.
		
	return Rx.Observable.combineLatest( unpack( volumeAgentStatusStreams ) )
	
		:map( function( ... )
		
			local knownVolumes = {}
			local onlineVolumes = {}
			local outOfSpaceVolumes = {}
			
			for _, volumeStatus in ipairs( arg ) do

				local volumeGuid = volumeStatus.volumeGuid
				local isOnline = volumeStatus.isOnline
				local outOfSpace = volumeStatus.isOutOfSpace

				knownVolumes[ volumeGuid ] = true
				
				onlineVolumes[ volumeGuid ] = isOnline
				
				outOfSpaceVolumes[ volumeGuid ] = outOfSpace
				
			end
			
			return volumeGuidForOriginals, volumeGuidForProxies,
					onlineVolumes, outOfSpaceVolumes, knownVolumes
		
		end )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedAssetStatusAsStr( wfCatalog, syncStatusAndVolumeInfos )

	--[[ 
	
		True asset sync status is made up 3 separate data points.
		
		1. The current state of the asset as generated by "getAssetStatus". This
			is the same status that is stored in sync status index store table.
	
		2. The current state of the volumes where the assets are stored. If an asset
		 	needs to be downloaded, but the volume we need to download it to is
			offline, that's a new unique state.
		
		3. The current global state of sync; is sync paused and/or offline.

	]]--
	
	local mappedVolumeStatuses = WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )
	
	local volumeAgents = wfCatalog:getAllVolumeAgents()
	local volumeAgent = volumeAgents[ next( volumeAgents ) ]
	local syncPausedStream = volumeAgent and volumeAgent:isSyncPausedStream() or Rx.Observable.just( false )
	local internetConnectionStream = volumeAgent and volumeAgent:hasInternetConnectionStream() or Rx.Observable.just( true )
	
	-- syncStatusAndVolumeInfos is an Observable that produces an array of rows containing
	-- 1: sync status constant mapping to WFCatSyncStatus.getAssetStatus
	-- 2: originalVolumeGuid
	-- 3: proxyVolumeGuid
	-- 4: markedAsOffline
	
	return Rx.Observable.combineLatest(
			syncStatusAndVolumeInfos,
			mappedVolumeStatuses )
		:map( function( syncStatusAndVolumeInfoRows, volumeGuidForOriginals, volumeGuidForProxies, 
				onlineVolumes, outOfSpaceVolumes, knownVolumes )
			
			local combinedSyncStatus
			
			for _, syncStatusAndVolumeInfoRow in ipairs( syncStatusAndVolumeInfoRows ) do
			
				local syncStatus = syncStatusAndVolumeInfoRow[ 1 ]
				local volumeGuidUsedByOriginal = syncStatusAndVolumeInfoRow[ 2 ]
				local volumeGuidUsedByProxy = syncStatusAndVolumeInfoRow[ 3 ]
				
				local isOriginalOnKnownVolume = volumeGuidUsedByOriginal and knownVolumes[ volumeGuidUsedByOriginal ]
				local isProxyOnKnownVolume = volumeGuidUsedByProxy and knownVolumes[ volumeGuidUsedByProxy ]
				
				local isOriginalOnOnlineVolume = volumeGuidUsedByOriginal and onlineVolumes[ volumeGuidUsedByOriginal ] 
				local isProxyOnOnlineVolume = volumeGuidUsedByProxy and onlineVolumes[ volumeGuidUsedByProxy ]
				
				local isVolumeForOriginalsOnline = onlineVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOnline = onlineVolumes[ volumeGuidForProxies ]
			
				local isVolumeForOriginalsOutOfSpace = outOfSpaceVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOutOfSpace = outOfSpaceVolumes[ volumeGuidForProxies ]
				
				-- Calculate the sync status.
				
				local currentDocSyncStatus = WFCatSyncStatus.calculateSyncStatusWithVolumeStatus( 
						syncStatus, volumeGuidUsedByOriginal, volumeGuidUsedByProxy,
						isOriginalOnKnownVolume, isProxyOnKnownVolume,
						isVolumeForOriginalsOnline, isVolumeForProxiesOnline,
						isVolumeForOriginalsOutOfSpace, isVolumeForProxiesOutOfSpace,
						isOriginalOnOnlineVolume, isProxyOnOnlineVolume )
				
				if not combinedSyncStatus then
					combinedSyncStatus = currentDocSyncStatus
				elseif combinedSyncStatus ~= currentDocSyncStatus then
					combinedSyncStatus = nil
					break
				end
					
			end
					
			return combinedSyncStatus
					
		end )
		
		:flatMap( function( syncStatus )
		
			return Rx.Observable.combineLatest( 
				Rx.Observable.just( syncStatus ),
				syncPausedStream,
				internetConnectionStream )
				
		end)
		
		:map( function( syncStatus, isSyncPaused, isOnline )
		
			return WFCatSyncStatus.calculateSyncStatusWithGlobalStatus( syncStatus, isSyncPaused, not isOnline )
		
		end )
		
		:map( WFCatSyncStatus.translateAssetSyncStatus )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedOzStatusFromIdsAsStr( wfCatalog, assetIds )
	
	return wfCatalog:indexStore():query {
			sql = [[SELECT ozHasOriginal, ozHasProxy FROM localStatusAndOzStatus JOIN TEMP_ids ON localStatusAndOzStatus.docId = TEMP_ids.docId]],
			tempIds = assetIds,
			calm = 2
		}:map( function( ozHasRows )
			
			local combinedOzStatusStr
			
			local previousOzHasOriginal
			local previousOzHasProxy
			
			-- returns kOzHasOriginal if all assets have original on Oz, kOzHasProxy if all assets have only a proxy on Oz
			-- kOzHasNothing if all assets have nothing on Oz and nil if the assets are mixed
			for _, ozHas in ipairs( ozHasRows ) do
			
				local ozHasOriginal = ozHas[ 1 ] == 1
				local ozHasProxy = ozHas[ 2 ] == 1
			
				if not combinedOzStatusStr then
				
					if ozHasOriginal then
					
						combinedOzStatusStr = kOzHasOriginal
					
					elseif ozHasProxy then
					
						combinedOzStatusStr = kOzHasProxy
					
					else
					
						combinedOzStatusStr = kOzHasNothing
					
					end
					
				elseif previousOzHasOriginal ~= ozHasOriginal
					or previousOzHasProxy ~= ozHasProxy then
				
					combinedOzStatusStr = nil
					break
					
				end
				
				previousOzHasProxy = ozHasProxy
				
				previousOzHasOriginal = ozHasOriginal
				
			end
		
			return combinedOzStatusStr
		
		end )

end

--------------------------------------------------------------------------------

local function localFileType( hasOriginal, isOriginalOnOfflineVolume, hasOriginalError, isMarkedForOffline, hasProxy, isProxyOnOfflineVolume )

	local assetLocalFileTypeId = kAssetHasNothing
	
	if hasOriginal and not isOriginalOnOfflineVolume and not hasOriginalError then
		if isMarkedForOffline then
			assetLocalFileTypeId = kAssetHasLocalOriginalForOffline
		else
			assetLocalFileTypeId = kAssetHasLocalOriginal
		end
	elseif hasProxy and not isProxyOnOfflineVolume then
		assetLocalFileTypeId = kAssetHasLocalProxy
	end
	
	return kAssetLocalFileType[ assetLocalFileTypeId ]
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getLocalFileTypeForAsset( docValue, isOriginalOnOfflineVolume, isProxyOnOfflineVolume )
	
	return localFileType( 
		docValueHasLocalOriginal( docValue ),
		isOriginalOnOfflineVolume,
		hasInvalidOriginalError( docValue ),
		WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue ),
		docValueHasLocalProxy( docValue ),
		isProxyOnOfflineVolume )
		
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedLocalFileTypeAsStr( wfCatalog, syncStatusAndVolumeInfos )
	-- syncStatusAndVolumeInfos is an Observable that produces an array of rows containing
	-- 1: sync status constant mapping to WFCatSyncStatus.getAssetStatus
	-- 2: originalVolumeGuid
	-- 3: proxyVolumeGuid
	-- 4: markedAsOffline
	local mappedVolumeStatuses = WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )
	
	local volumeAgents = wfCatalog:getAllVolumeAgents()
	local volumeAgent = volumeAgents[ next( volumeAgents ) ]
	
	return Rx.Observable.combineLatest(
		syncStatusAndVolumeInfos,
		mappedVolumeStatuses )

		:map( function( syncStatusAndVolumeInfoRows, volumeGuidForOriginals, 
				volumeGuidForProxies, onlineVolumes )
					
			local combinedLocalFileType
			
			for _, syncStatusAndVolumeInfoRow in ipairs( syncStatusAndVolumeInfoRows ) do
					
				local volumeGuidUsedByOriginal = syncStatusAndVolumeInfoRow[ 2 ]
				local volumeGuidUsedByProxy = syncStatusAndVolumeInfoRow[ 3 ]
				local isMarkedForOffline = syncStatusAndVolumeInfoRow[ 4 ] == 1
			
				local isOriginalOnOfflineVolume = volumeGuidUsedByOriginal and not onlineVolumes[ volumeGuidUsedByOriginal ] 
				local isProxyOnOfflineVolume = volumeGuidUsedByProxy and not onlineVolumes[ volumeGuidUsedByProxy ]
				
				local docLocalFileType = localFileType(
					not not volumeGuidUsedByOriginal,
					isOriginalOnOfflineVolume,
					nil, -- need to fix this has InvalidOriginalError shouldn't be empty
					isMarkedForOffline,
					not not volumeGuidUsedByProxy,
					isProxyOnOfflineVolume )
					
				if not combinedLocalFileType then
				
					combinedLocalFileType = docLocalFileType
					
				elseif combinedLocalFileType ~= docLocalFileType then
				
					combinedLocalFileType = nil
					break
				
				end
				
			end
					
			return combinedLocalFileType
					
		end )

		:map( WFCatSyncStatus.translateAssetSyncLocalFileType )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getCombinedMarkedForOfflineUse( wfCatalog, syncStatusAndVolumeInfos )
	-- syncStatusAndVolumeInfos is an Observable that produces an array of rows containing
	-- 1: sync status constant mapping to WFCatSyncStatus.getAssetStatus
	-- 2: originalVolumeGuid
	-- 3: proxyVolumeGuid
	-- 4: markedAsOffline
	return syncStatusAndVolumeInfos
		:map( function( isMarkedForOfflineRows )
			
			-- returns true if all assets are marked for offline, false if no assets are marked for offline
			-- and nil if the assets are mixed
			local combinedMarkedForOfflineUse

			for _, row in ipairs( isMarkedForOfflineRows ) do

				local markedForOfflineUse = row[ 4 ] == 1

				if combinedMarkedForOfflineUse then
				
					combinedMarkedForOfflineUse = markedForOfflineUse
					
				elseif combinedMarkedForOfflineUse ~= markedForOfflineUse then
				
					combinedMarkedForOfflineUse = nil
					break
				
				end

			end

			return combinedMarkedForOfflineUse

		end )

end

--------------------------------------------------------------------------------

local function asset_status_table()

	local _map = function( docValue, emit )

		if docValue._deleted and not WFCatTrashUtils_isDeleteTrash( docValue ) then

			-- Don't count deleted assets unless they are delete trash as Cloud
			-- Trash generates another asset for each deleted or restored asset.
			return

		end

		if isAssetDocument( docValue ) then

			local status = WFCatSyncStatus.getAssetStatus( docValue )
			
			if status then

				local originalVolumeGuid = WFCatFileSpec.getOriginalVolumeGuidFromDocValue( docValue )
				local proxyVolumeGuid = WFCatFileSpec.getProxyVolumeGuidFromDocValue( docValue )

				local assetId = docValue._id

				emit( assetId, status, originalVolumeGuid, proxyVolumeGuid )
				
			end
			
		end
		
	end

	-- SPAM [jandrew 2016-05-31]: 
	-- The contents of the asset sync status table are invalidated when switching 
	-- between file management strategies. However, the index store
	-- will automatically refresh any table whose mapping function changes between
	-- app runs. Here, we take advantage of this by providing different mapping functions
	-- based on the file management strategy being used. The only difference
	-- between the methods is the no-op third parameter to the _map method: it's enough
	-- to trigger the rebuild when the user changes modes.

	-- The 4th parameter (also a no-op parameter) is a table version value that is used
	-- to cause reindexing when the implementatin of _map or calls underneath change
	-- the status value.

	-- Note: The last two parameters are what make each mapping function unique.
	-- THEY SHOULD ALWAYS BE RAW NUMBERS, AND NOT VARIABLES.

	local defaultMap = function( docValue, emit )
		_map( docValue, emit, 1, 25 )
	end
	
	local keepAllOriginalsMap = function( docValue, emit )
		_map( docValue, emit, 2, 25 )
	end

	local keepAllProxiesMap = function( docValue, emit )
		_map( docValue, emit, 3, 25 )
	end
 
 	local keepEverythingMap = function( docValue, emit )
		_map( docValue, emit, 4, 25 )
	end
 
	local mapFn 
	
	if NmbFileManagementSettings.keepAllProxies() and NmbFileManagementSettings.keepAllOriginals() then
		mapFn = keepEverythingMap
	elseif NmbFileManagementSettings.keepAllProxies() then
		mapFn = keepAllProxiesMap
	elseif NmbFileManagementSettings.keepAllOriginals() then
		mapFn = keepAllOriginalsMap
	else
		mapFn = defaultMap
	end
	
	return {
		
		sql = [[CREATE TABLE asset_sync_status_ui( assetId, status INTEGER, volumeGuidUsedByOriginal, volumeGuidUsedByProxy )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		map = mapFn,
		docValueFilter = function( docValue )
			return isAssetDocument( docValue )
		end,
	}
	
end

--------------------------------------------------------------------------------

local function asset_status_by_assetId_index( )
	return [[CREATE INDEX asset_sync_status_ui_by_assetId_status ON asset_sync_status_ui( assetId, status )]]
end

--------------------------------------------------------------------------------

local function asset_status_by_status_index( )
	return [[CREATE INDEX asset_sync_status_ui_by_status ON asset_sync_status_ui( status, assetId )]]
end

--------------------------------------------------------------------------------

local function asset_local_file_type_table( )

	return {

		sql = [[CREATE TABLE asset_sync_local_file_type_ui( assetId, localFileType INTEGER )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		docValueFilter = isFileBasedAssetDocument,
		map = function( docValue, emit )
		
			if isValidDisplayableAssetDocument( docValue ) and 
				isFileBasedAssetDocument( docValue ) then
			
				local TABLE_VERSION
				TABLE_VERSION = TABLE_VERSION or 2  -- a dummy change to force the reindexing

				local localFileType = WFCatSyncStatus.getLocalFileTypeForAsset( docValue )

				-- SPAM [jandrew 2018-09-05]: The assets table does not emit for assets
				-- that do not contain a capture date. In order to maintain consistency
				-- between the tables, I've added that here as well.

				local captureDate = derefOpt( docValue, 'captureDate' )

				if localFileType and captureDate then

					local assetId = docValue._id	
					emit( assetId, localFileType )

				end

			end
		end,

	}

end

--------------------------------------------------------------------------------

local function asset_local_file_type_by_assetId_index( )
	return [[CREATE INDEX asset_sync_local_file_type_ui_by_assetId_status ON asset_sync_local_file_type_ui( assetId, localFileType )]]
end

--------------------------------------------------------------------------------

local function asset_local_file_type_by_localFileType_index( )
	return [[CREATE INDEX asset_sync_local_file_type_ui_by_status ON asset_sync_local_file_type_ui( localFileType )]]
end

--------------------------------------------------------------------------------

local function album_status_table( )

	return {

		sql = [[CREATE TABLE album_sync_status_ui( albumId, status INTEGER )]],
		useLocalId = true,
		hasOneRowPerDocument = true,
		docValueFilter = function( docValue )
			return docValue.type == 'album'
		end,
		map = function( docValue, emit )
			if docValue and docValue.type == 'album' then
				
				local TABLE_VERSION
				TABLE_VERSION = TABLE_VERSION or 2  -- a dummy change to force the reindexing

				local albumId = docValue._id

				local status = kAlbumSyncStatus[ kAlbumComplete ]
				
				if docValue._deleted and not derefOpt( docValue, '_serverProvidedMetadata', 'links' )  then
					-- this deleted album is either not yet up on Oz or has been on Oz and was deleted
					-- so we can avoid emitting for this album
					return
				end
				
				local isMarkedForOffline = WFCatFileSpec.getMarkedForOfflineUseFromDocValue( docValue )

				if isLatestMetadataUploaded( docValue ) then
				
					if isMarkedForOffline then
				
						status = kAlbumSyncStatus[ kAlbumCompleteForOfflineUse ]
					
					else
					
						status = kAlbumSyncStatus[ kAlbumComplete ]
						
					end
					
				else
				
					if isMarkedForOffline then
				
						status = kAlbumSyncStatus[ kAlbumSyncingForOfflineUse ]
					
					else
						
						status = kAlbumSyncStatus[ kAlbumSyncing ]
					
					end
					
				end

				emit( albumId, status )

			end
		end,	
	}

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.generateIndexSchema( indexStoreSchema )
	
	indexStoreSchema:addTable( asset_status_table() )
	
	indexStoreSchema:addIndex( asset_status_by_assetId_index() )
	
	indexStoreSchema:addIndex( asset_status_by_status_index() )
	
	indexStoreSchema:addTable( asset_local_file_type_table() )
	
	indexStoreSchema:addIndex( asset_local_file_type_by_assetId_index() )
	
	indexStoreSchema:addIndex( asset_local_file_type_by_localFileType_index() )
	
	indexStoreSchema:addTable( album_status_table() )
	
end

--------------------------------------------------------------------------------

local function toBool( int )

	return int and int > 0

end

--------------------------------------------------------------------------------

local function getSyncingStatusIds( allSyncStatuses, isStatusForAlbums )
	
	-- Certain statuses (kAssetComplete, kAssetCompleteNoMaster, kAssetInvalidOriginal, etc.) 
	-- would indicate complete asset. All other assets are considered 'syncing'.
	
	local nonSyncingStates = {
		kAssetComplete,
		kAssetCompleteNoMaster,
		kAssetCompleteNoNegatives,
		kAssetCompleteForOffline,
		kAssetInvalidOriginal,
		kAssetOriginalDownloadError,
		kAssetProxyDownloadError,
		kAssetDevelopXmpDownloadError,
		kAssetStackComplete,
		kAssetInsufficientPermissionError
	}

	local keepAllOriginals = NmbFileManagementSettings.keepAllOriginals()

	if isStatusForAlbums and not keepAllOriginals then
		table.insert( nonSyncingStates, kAssetNeedsOriginalDownloadForOffline )
		table.insert( nonSyncingStates, kAssetNeedsProxyDownloadForOffline )
	end

	local nonSyncingStateSet = KSTableUtils.mapArrayToSet( nonSyncingStates )

	-- Copy status ids into a table that can be unpacked.
	
	local syncingStatusIds = {}

	for state, statusId in pairs( allSyncStatuses ) do

		if not nonSyncingStateSet[ state ] then

			table.insert( syncingStatusIds, statusId )

		end

	end

	return syncingStatusIds

end

--------------------------------------------------------------------------------

local function getErrorStatusIds( allSyncStatuses )
	
	local errorSyncStates = {
		kAssetInvalidOriginal,
		kAssetOriginalDownloadError,
		kAssetProxyDownloadError,
		kAssetDevelopXmpDownloadError,
	}
	
	local errorStatusIds = {}
	
	for _, errorState in ipairs( errorSyncStates ) do
		
		if allSyncStatuses[ errorState ] then
			table.insert( errorStatusIds, allSyncStatuses[ errorState ] )
		end
		
	end
	
	return errorStatusIds
	
end

--------------------------------------------------------------------------------

local function basicAlbumStatusStream( wfCatalog )
	
	return wfCatalog:indexStore():query {
			sql = [[SELECT albumId, status
						FROM album_sync_status_ui ]],
			calm = 0.25
		}
		:map( function( albumStatuses )
			
			local statusForAlbum = {}
			
			for _, albumStatus in ipairs( albumStatuses ) do
			
				local albumId = albumStatus[ 1 ]
				local status = albumStatus[ 2 ]
				
				statusForAlbum[ albumId ] = kAlbumSyncStatusById[ status ]
					
			end
			
			return statusForAlbum
			
		end )
end

--------------------------------------------------------------------------------

local function getSyncingAlbumsQuerySQL( numberOfSyncingStatuses )
	
		return string.format( 
				[[SELECT DISTINCT albumId
					FROM album_asset_v2
					WHERE assetId IN
						(
							SELECT assetId 
								FROM asset_sync_status_ui
								WHERE status IN ( %s )
						) ]], WFSQLiteUtils.createVariableArgList( numberOfSyncingStatuses ) )
						
	end
	
--------------------------------------------------------------------------------

function WFCatSyncStatus.getAllAlbumsFinishedUploading( wfCatalog )
	
	local albumSyncStatus = basicAlbumStatusStream( wfCatalog )
	
	local pendingAssetUploadStatuses = {
		-- these statuses indicate we're still syncing the 
		-- base asset data UP to Oz and therefore cannot
		-- send the data to the third party service yet
		kAssetNeedsMetadataUpload,	
		kAssetNeedsOriginalUpload,
	}
	
	local assetsPendingUpload = getSyncingAlbumsQuerySQL( #pendingAssetUploadStatuses )
	
	local albumsPendingAssetUpload = wfCatalog:indexStore():query {
			sql = assetsPendingUpload,
			calm = 3,
			unpack( pendingAssetUploadStatuses )
		}
		:map( function( syncingAlbums )
		
			local statusForAlbum = {}
			
			for _, syncingAlbum in ipairs( syncingAlbums ) do
				statusForAlbum[ syncingAlbum[ 1 ] ] = kAlbumSyncing
			end
			
			return statusForAlbum
			
		end )
		
	return Rx.Observable.combineLatest( albumSyncStatus, albumsPendingAssetUpload )
		:map( function( albums, albumsWithAssetsPendingUpload )
			local completed = {}
			for albumId, status in pairs( albums ) do
				if ( status == kAlbumComplete or
					status == kAlbumCompleteForOfflineUse ) and 
					not albumsWithAssetsPendingUpload[ albumId ] then
					completed[ albumId ] = true
				end
			end
			return completed
		end )
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getSyncStatusForAlbums( wfCatalog, optIgnoreAlbumSyncStatus )
	
	local albumStatusesStream
	
	local keepAllOriginals = NmbFileManagementSettings.keepAllOriginals()
	
	if optIgnoreAlbumSyncStatus then 
	
		-- Don't consider the sync status of the actual album document when determining 
		-- the status to show for the album. Essentially this means that we only consider
		-- the sync status of the underlying asset documents since this also ignores the 
		-- album_asset documents.
		albumStatusesStream = wfCatalog:indexStore():query {
			sql = [[SELECT albumId, isMarkedForOffline FROM albums ]],
			calm = 0.25
		}
		:map( function( albums )
		
			local statusForAlbum = {}
			
			for _, album in ipairs( albums ) do
			
				local albumId = album[ 1 ]
				local isMarkedForOfflineUse = toBool( album[ 2 ] )
				
				statusForAlbum[ albumId ] = isMarkedForOfflineUse and kAlbumCompleteForOfflineUse or kAlbumComplete

			end
			
			return statusForAlbum
			
		end )
		
	else 
	
		albumStatusesStream = basicAlbumStatusStream( wfCatalog )
	
	end 
	
	local syncingStatusIds = getSyncingStatusIds( kAssetSyncStatus, true )
	
	local query = getSyncingAlbumsQuerySQL( #syncingStatusIds )
	
	local syncingAlbumsFromAssets = wfCatalog:indexStore():query {
			sql = query,
			calm = 3,
			unpack( syncingStatusIds )
		}
		
		:map( function( syncingAlbums )
		
			local statusForAlbum = {}
			
			for _, syncingAlbum in ipairs( syncingAlbums ) do
				statusForAlbum[ syncingAlbum[ 1 ] ] = kAlbumSyncing
			end
			
			return statusForAlbum
			
		end )
	
	local syncingAlbumsMarkedForOfflineUseFromAssets = Rx.Observable.just( {} )
	
	if not keepAllOriginals then
	
		syncingAlbumsMarkedForOfflineUseFromAssets = wfCatalog:indexStore():query {
				sql = [[ SELECT DISTINCT albumId
							FROM album_asset_v2
							INNER JOIN albums USING( albumId )
							WHERE albums.isMarkedForOffline = 1
							AND assetId IN
							(
								SELECT assetId 
									FROM asset_sync_status_ui
									WHERE status IN ( ?, ? ) 		  									
								UNION
								SELECT assetId 
									FROM assets 
									WHERE isMarkedForOffline = 0
							) ]],
				calm = 3,
				kAssetSyncStatus[ kAssetNeedsOriginalDownloadForOffline ],
				kAssetSyncStatus[ kAssetNeedsProxyDownloadForOffline ]
			}
			
			:map( function( syncingAlbums )
			
				local statusForAlbum = {}
				
				for _, syncingAlbum in ipairs( syncingAlbums ) do
					statusForAlbum[ syncingAlbum[ 1 ] ] = kAlbumSyncingForOfflineUse
				end
				
				return statusForAlbum
				
			end )
			
	end
	
	local sharedWithYouAlbumIdsStream = wfCatalog:indexStore():query {
				sql = [[ SELECT DISTINCT albumId
							FROM albums
							WHERE spaceId IS NOT NULL]],
				calm = 3,
			}
			
			:map( function( rows )
			
				local sharedWithYouAlbums = {}
				
				for _, row in ipairs( rows ) do
					sharedWithYouAlbums[ row[ 1 ] ] = true
				end
				
				return sharedWithYouAlbums
				
			end )
	
	local volumeAgents = wfCatalog:getAllVolumeAgents()
	local volumeAgent = volumeAgents[ next( volumeAgents ) ]
	local syncPausedStream = volumeAgent and volumeAgent:isSyncPausedStream() or Rx.Observable.just( false )
	local internetConnectionStream = volumeAgent and volumeAgent:hasInternetConnectionStream() or Rx.Observable.just( true )
	local mappedVolumeStatuses = WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )
	
	return Rx.Observable.combineLatest( albumStatusesStream, syncingAlbumsFromAssets, 
		syncingAlbumsMarkedForOfflineUseFromAssets, syncPausedStream, 
		internetConnectionStream, sharedWithYouAlbumIdsStream, mappedVolumeStatuses )

		:calm( 0.5, WFRxScheduler.sharedScheduler() )
		
		:map( function( albumStatusMap, albumStatusFromAssetsMap, albumMarkedForOfflineStatusFromAssetsMap,
			isSyncPaused, isSyncOnline, sharedWithYouAlbumIds, volumeGuidForOriginals, _, onlineVolumes, 
			outOfSpaceVolumes )

			local combinedAlbumStatuses = {}
			
			local isVolumeForOriginalsOnline = onlineVolumes[ volumeGuidForOriginals ]
			local isVolumeForOriginalsOutOfSpace = outOfSpaceVolumes[ volumeGuidForOriginals ]
			
			for albumId, albumStatus in pairs( albumStatusMap ) do
			
				local combinedStatus
				
				local albumStatusFromAssets = albumMarkedForOfflineStatusFromAssetsMap[ albumId ] 
						or albumStatusFromAssetsMap[ albumId ]
				
				if not sharedWithYouAlbumIds[ albumId ]
					and ( albumStatus == kAlbumSyncingForOfflineUse 
					or albumStatusFromAssets == kAlbumSyncingForOfflineUse
					or ( albumStatus == kAlbumCompleteForOfflineUse and albumStatusFromAssets == kAlbumSyncing )
					or ( keepAllOriginals and ( albumStatus == kAlbumSyncing or albumStatusFromAssets == kAlbumSyncing ) ) ) then
				
					if isSyncPaused then
					
						combinedStatus = kAlbumSyncPaused
						
					elseif not isSyncOnline then
					
						combinedStatus = kAlbumSyncNoConnection
					
					elseif not isVolumeForOriginalsOnline then
	
						combinedStatus = kAlbumSyncingForOfflineUseToOfflineVolume
					
					elseif isVolumeForOriginalsOutOfSpace then
				
						combinedStatus = kAlbumSyncingForOfflineUseToOutOfSpaceVolume
					
					else
					
						combinedStatus = kAlbumSyncingForOfflineUse
					
					end
				
				elseif albumStatus == kAlbumSyncing or albumStatusFromAssets == kAlbumSyncing then
				
					if isSyncPaused then
					
						combinedStatus = kAlbumSyncPaused
						
					elseif not isSyncOnline then
					
						combinedStatus = kAlbumSyncNoConnection
						
					else
					
						combinedStatus = kAlbumSyncing
						
					end
					
				elseif not sharedWithYouAlbumIds[ albumId ]
					and ( ( albumStatus == kAlbumCompleteForOfflineUse and albumStatusFromAssets == kAlbumComplete )
					or ( keepAllOriginals and ( albumStatus == kAlbumComplete or albumStatusFromAssets == kAlbumComplete ) ) ) then
					
					combinedStatus = kAlbumCompleteForOfflineUse
					
				elseif albumStatus == kAlbumComplete or albumStatusFromAssets == kAlbumComplete then
				
					combinedStatus = kAlbumComplete
				
				else
				
					combinedStatus = albumStatus
				
				end
				
				combinedAlbumStatuses[ albumId ] = combinedStatus
				
			end
			
			return combinedAlbumStatuses
			
		end )
	
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getSyncStatusOfParentCollectionOfAlbums( wfCatalog, albumIds )
	
	return wfCatalog:indexStore():query {
			sql = [[SELECT album_sync_status_ui.albumId, album_sync_status_ui.status
						FROM album_sync_status_ui
						INNER JOIN albums ON albums.parentId == album_sync_status_ui.albumId
						INNER JOIN TEMP_ids ON TEMP_ids.docId = albums.docId ]],
			calm = 2,
			tempIds = albumIds
		}
		
		:map( function( albumStatuses )
			
			local syncedAlbumsById = {}
			
			for _, albumStatus in ipairs( albumStatuses ) do
			
				local albumId = albumStatus[ 1 ]
				local statusId = albumStatus[ 2 ]
				
				syncedAlbumsById[ albumId ] = WFCatSyncStatus.isAlbumCompleted( statusId )
					
			end
			
			return syncedAlbumsById
			
		end )
	
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getSyncStatusOfAlbumsOfAlbumAssets( wfCatalog, albumAssetIds )
	
	return wfCatalog:indexStore():query {
			sql = [[SELECT album_sync_status_ui.albumId, album_sync_status_ui.status
						FROM album_sync_status_ui
						INNER JOIN album_asset_v2 ON album_asset_v2.albumId == album_sync_status_ui.albumId
						INNER JOIN TEMP_ids ON TEMP_ids.docId = album_asset_v2.docId ]],
			calm = 2,
			tempIds = albumAssetIds
		}
		
		:map( function( albumStatuses )
			
			local syncedAlbumsById = {}
			
			for _, albumStatus in ipairs( albumStatuses ) do
			
				local albumId = albumStatus[ 1 ]
				local statusId = albumStatus[ 2 ]
				
				syncedAlbumsById[ albumId ] = WFCatSyncStatus.isAlbumCompleted( statusId )
					
			end
			
			return syncedAlbumsById
			
		end )
	
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getSyncStatusOfAssetsOfAlbumAssets( wfCatalog, albumAssetIds )

	return wfCatalog:indexStore():query {
			sql = [[SELECT asset_sync_status_ui.assetId, asset_sync_status_ui.status
						FROM asset_sync_status_ui
						INNER JOIN album_asset_v2 ON album_asset_v2.assetId == asset_sync_status_ui.assetId
						INNER JOIN TEMP_ids ON TEMP_ids.docId = album_asset_v2.docId ]],
			calm = 2,
			tempIds = albumAssetIds
		}
		
		:map( function( assetStatuses )
			
			local syncedAssetsById = {}
			
			for _, assetStatus in ipairs( assetStatuses ) do
			
				local assetId = assetStatus[ 1 ]
				local statusId = assetStatus[ 2 ]
				
				syncedAssetsById[ assetId ] = WFCatSyncStatus.isAssetCompleted( statusId )
					
			end
			
			return syncedAssetsById
			
		end )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.syncPendingForAnyAsset( wfCatalog )

	local syncingStatusIds = getSyncingStatusIds( kAssetSyncStatus )

	local isSyncPendingSQL = string.format( 
			[[SELECT status 
					FROM asset_sync_status_ui 
					WHERE status IN ( %s )
					LIMIT 1]], WFSQLiteUtils.createVariableArgList( #syncingStatusIds ) )

	return wfCatalog:indexStore():query {
			sql = isSyncPendingSQL,
			unpack( syncingStatusIds )
		}
		:map( function( results )
			return KSTableUtils.derefOpt( results, 1, 1 ) ~= nil
		end )
		:distinctUntilChanged()

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getLocalIdsWithSyncPendingOrError( wfCatalog )
	
	local syncingStatusIds = getSyncingStatusIds( kAssetSyncStatus )
	
	local errorStatusIds = getErrorStatusIds( kAssetSyncStatus )
	
	local syncingOrErrorStatusIds = KSTableUtils.appendArray( syncingStatusIds, errorStatusIds )
	
	local assetsWithSyncPendingOrErrorSQL = string.format(
		[[SELECT docId 
			FROM asset_sync_status_ui 
			WHERE status IN ( %s )]],
		WFSQLiteUtils.createVariableArgList( #syncingOrErrorStatusIds ) )
	
	return wfCatalog:indexStore():query {
			sql = assetsWithSyncPendingOrErrorSQL,
			unpack( syncingOrErrorStatusIds )
		}
		:map( function( rows )
			local docIds = {}
			for _, row in ipairs( rows ) do
				table.insert( docIds, row[ 1 ] )
			end
			return docIds
		end )
end

--------------------------------------------------------------------------------

-- Allows for identification and count of assets pending restoration, which is
-- needed for 'pancake' messaging purposes when a restore happens while the user
-- is out of Oz storage or has an expired subscription.

function WFCatSyncStatus.getNumAssetsWithRestorationPending( wfCatalog )
	
	local assetsWithRestorationPendingSQL = 
		[[SELECT COUNT(1), MIN(purgeDate) FROM assetsPendingRestore]]
	
	return wfCatalog:indexStore():query {
			sql = assetsWithRestorationPendingSQL,
			calm = 2,
		}
		:map( function( rows )
			local row = rows[ 1 ]
			return { numAssets = row[ 1 ], firstPurgeDate = row[ 2 ] }
		end )
end

--------------------------------------------------------------------------------
function WFCatSyncStatus.getDocIdsAndStatusWithSyncPendingOrError( wfCatalog )

	local syncingStatusIds = getSyncingStatusIds( kAssetSyncStatus )
	
	local errorStatusIds = getErrorStatusIds( kAssetSyncStatus )
	
	local syncingOrErrorStatusIds = KSTableUtils.appendArray( syncingStatusIds, errorStatusIds )
	
	local assetsWithSyncPendingOrErrorSQL = string.format(
		[[SELECT docId, assetId, status
			FROM asset_sync_status_ui 
			WHERE status IN ( %s )]],
		WFSQLiteUtils.createVariableArgList( #syncingOrErrorStatusIds ) )
		
	return wfCatalog:indexStore():query {
			sql = assetsWithSyncPendingOrErrorSQL,
			unpack( syncingOrErrorStatusIds )
		}
		:map( function( rows )
			local results = {}
			for _, row in ipairs( rows ) do
				table.insert( results, { docId = row[ 1 ], assetId = row[ 2 ], status = row[ 3 ] } )
			end
			return results
		end )
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.persistentSyncErrorForAnyAsset( wfCatalog )
	
	local errorStatusIds = getErrorStatusIds( kAssetSyncStatus )
	
	return wfCatalog:indexStore():query {
			sql = string.format( 
				[[SELECT status FROM asset_sync_status_ui WHERE status IN ( %s ) LIMIT 1]],
				WFSQLiteUtils.createVariableArgList( #errorStatusIds ) ),
			unpack( errorStatusIds )
		}
		:map( function( results )
			return KSTableUtils.derefOpt( results, 1, 1 ) ~= nil
		end )
		:distinctUntilChanged()

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getSyncStatusForAssets( wfCatalog )
	
	return wfCatalog:indexStore():query {
		sql = [[SELECT status, count( status ) FROM asset_sync_status_ui GROUP BY status]],
		}
		:map( function( rows )
			local statusCounts = {}
			
			for _, row in ipairs( rows ) do
				local statusId = row[ 1 ]
				local statusCount = row[ 2 ]
				
				if statusId and statusCount then
					local status = kAssetSyncStatusById[ statusId ]
					if status then
						statusCounts[ status ] = tonumber( statusCount ) or 0
					end
				end
			end
			
			for _, status in pairs( kAssetSyncStatusById ) do
				if not statusCounts[ status ] then 
					statusCounts[ status ] = 0
				end
			end
			
			return statusCounts
		end )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.isDoingInitialSync( wfCatalog )

	return wfCatalog:docStore():ongoingLocalDocValue( 'Oz-isDoingInitialSync' )
	
		:map( function( docValue )
		
			local isDoingInitialSync = true
			
			if docValue.isDoingInitialSync ~= nil then
			 
				isDoingInitialSync = docValue.isDoingInitialSync
				
			end
			
			return isDoingInitialSync
			
		end )
		
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.translateAssetSyncStatus( status )
	
	return kAssetSyncStatusById[ status ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.translateAssetSyncLocalFileType( localFileType )
	
	return kAssetLocalFileTypeById[ localFileType ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.completeNoErrorSyncStatus( )
	
	return require 'WFCatSearchBySyncStatus'.completeNoErrorSyncStatus( )

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.isAssetCompleted( statusOrId )

	if type( statusOrId ) ~= 'string' then
		statusOrId = kAssetSyncStatusById[ statusOrId ]
	end

	return statusOrId == kAssetComplete
		or statusOrId == kAssetCompleteNoMaster
		or statusOrId == kAssetCompleteNoNegatives
		or statusOrId == kAssetCompleteForOffline
		or statusOrId == kAssetInvalidOriginal
		or statusOrId == kAssetOriginalDownloadError
		or statusOrId == kAssetProxyDownloadError
		or statusOrId == kAssetDevelopXmpDownloadError
		or statusOrId == kAssetStackComplete
		or statusOrId == kAssetInsufficientPermissionError
		
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.isAlbumCompleted( statusOrId )

	if type( statusOrId ) ~= 'string' then
		statusOrId = kAlbumSyncStatusById[ statusOrId ]
	end

	return statusOrId == kAlbumComplete
		or statusOrId == kAlbumCompleteForOfflineUse

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.calculateSyncStatusWithVolumeStatus( 
		statusOrId, hasLocalOriginal, hasLocalProxy, 
		isOriginalOnKnownVolume, isProxyOnKnownVolume,
		isVolumeForOriginalsOnline, isVolumeForProxiesOnline,
		isVolumeForOriginalsOutOfSpace, isVolumeForProxiesOutOfSpace,
		isOriginalOnOnlineVolume, isProxyOnOnlineVolume )

	if type( statusOrId ) ~= 'string' then
		statusOrId = kAssetSyncStatusById[ statusOrId ]
	end

	if statusOrId == kAssetNeedsOriginalDownload
		or statusOrId == kAssetNeedsOriginalDownloadForOffline then
		
		if not isVolumeForOriginalsOnline then
		
			-- Original needs to be downloaded to an offline volume.
			statusOrId = kAssetNeedsOriginalDownloadToOfflineVolume
			
		elseif isVolumeForOriginalsOutOfSpace then
		
			-- Original needs to be downloaded to an out-of-space volume.
			statusOrId = kAssetNeedsOriginalDownloadToOutOfSpaceVolume
		
		end
		
	elseif statusOrId == kAssetNeedsProxyDownload then
	
		if not isVolumeForProxiesOnline then
		
			-- Proxy needs to be downloaded to an offline volume.
			statusOrId = kAssetNeedsProxyDownloadToOfflineVolume
			
		elseif isVolumeForProxiesOutOfSpace then
		
			-- Proxy needs to be downloaded to an out-of-space volume.
			statusOrId = kAssetNeedsProxyDownloadToOutOfSpaceVolume
		
		end
	
	elseif statusOrId == kAssetNeedsOriginalUpload then

		if hasLocalOriginal and not isOriginalOnKnownVolume then
		
			-- Original needs to be migrated to the original volume.
			statusOrId = kAssetNeedsOriginalMigration
		
		elseif hasLocalOriginal and not isOriginalOnOnlineVolume then
		
			-- Original needs to be uploaded, but is on an offline volume.
			statusOrId = kAssetNeedsOriginalUploadFromOfflineVolume
			
		end
		
	elseif statusOrId == kAssetComplete
		or statusOrId == kAssetCompleteNoMaster
		or statusOrId == kAssetCompleteNoNegatives
		or statusOrId == kAssetCompleteForOffline
		or statusOrId == kAssetInvalidOriginal 
		or statusOrId == kAssetOriginalDownloadError
		or statusOrId == kAssetProxyDownloadError
		or statusOrId == kAssetDevelopXmpDownloadError
		or statusOrId == kAssetStackComplete
		or statusOrId == kAssetInsufficientPermissionError then
		
		if hasLocalProxy and not isProxyOnKnownVolume then
		
			-- Proxy needs to be migrated to the proxy volume.
			statusOrId = kAssetNeedsProxyMigration
			
		elseif hasLocalProxy and not isProxyOnOnlineVolume then
		
			-- Proxy is downloaded, but on an offline volume.
			statusOrId = kAssetProxyIsOnOfflineVolume
			
		end
		
		if hasLocalOriginal and not isOriginalOnKnownVolume then
		
			-- Original needs to be migrated to the original volume.
			statusOrId = kAssetNeedsOriginalMigration
		
		elseif hasLocalOriginal and not isOriginalOnOnlineVolume then
		
			-- Original is downloaded, but on an offline volume.
			statusOrId = kAssetOriginalIsOnOfflineVolume
			
		end
		
	end

	return kAssetSyncStatus[ statusOrId ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.calculateSyncStatusWithGlobalStatus( 
		statusOrId, isSyncPaused, isOffline )

	if type( statusOrId ) ~= 'string' then
		statusOrId = kAssetSyncStatusById[ statusOrId ]
	end

	if isSyncPaused or isOffline then
	
		if statusOrId == kAssetNeedsMetadataUpload
			or statusOrId == kAssetNeedsOriginalUpload
			or statusOrId == kAssetNeedsOriginalDownload
			or statusOrId == kAssetNeedsOriginalDownloadForOffline
			or statusOrId == kAssetNeedsProxyDownload
			or statusOrId == kAssetNeedsProxyDownloadForOffline
			or statusOrId == kAssetNeedsDevelopXmpDownload
			or statusOrId == kAssetNeedsOriginalDownloadToOfflineVolume
			or statusOrId == kAssetNeedsProxyDownloadToOfflineVolume
			or statusOrId == kAssetNeedsOriginalUploadFromOfflineVolume
			or statusOrId == kAssetNeedsOriginalDownloadToOutOfSpaceVolume
			or statusOrId == kAssetNeedsProxyDownloadToOutOfSpaceVolume then
			
			statusOrId = isOffline and kAssetSyncNoConnection or kAssetSyncPaused
			
		end
		
	end

	return kAssetSyncStatus[ statusOrId ]

end

--------------------------------------------------------------------------------

function WFCatSyncStatus.offlineVolumesRequiringSyncing( wfCatalog )

	local syncStatusQuery = wfCatalog:indexStore():query {
		sql = [[
			SELECT DISTINCT status, volumeGuidUsedByOriginal, volumeGuidUsedByProxy
			FROM asset_sync_status_ui
			WHERE status IN ( ?, ?, ?, ?, ? )
		]],
		kAssetSyncStatus[ kAssetNeedsOriginalDownload ],
		kAssetSyncStatus[ kAssetNeedsOriginalDownloadForOffline ],
		kAssetSyncStatus[ kAssetNeedsProxyDownload ],
		kAssetSyncStatus[ kAssetNeedsProxyDownloadForOffline ],
		kAssetSyncStatus[ kAssetNeedsOriginalUpload ],
		calm = 1,
	}
	
	-- Combine the asset sync status query with the volume info status stream. We
	-- use this information to map the asset sync status to a new status based
	-- on the connection status of the volumes.
	
	local mappedVolumeStatuses = WFCatSyncStatus.getVolumeInfoForSyncStatus( wfCatalog )
	
	local volumesRequiringSyncingStream 
		= Rx.Observable.combineLatest( syncStatusQuery, mappedVolumeStatuses )
		
			:map( function( rows, volumeGuidForOriginals, volumeGuidForProxies, 
					onlineVolumes, outOfSpaceVolumes, knownVolumes )
		
				local volumesRequiringSync = {}
			
				local isVolumeForOriginalsOnline = onlineVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOnline = onlineVolumes[ volumeGuidForProxies ]
			
				local isVolumeForOriginalsOutOfSpace = outOfSpaceVolumes[ volumeGuidForOriginals ]
				local isVolumeForProxiesOutOfSpace = outOfSpaceVolumes[ volumeGuidForProxies ]
			
				for _, row in ipairs( rows ) do
				
					local syncStatusId = row[ 1 ]
					local volumeGuidUsedByOriginal = row[ 2 ]
					local volumeGuidUsedByProxy = row[ 3 ]
					
					local isOriginalOnKnownVolume = knownVolumes[ volumeGuidUsedByOriginal ]
					local isProxyOnKnownVolume = knownVolumes[ volumeGuidUsedByProxy ]
					
					local isOriginalOnOnlineVolume = onlineVolumes[ volumeGuidUsedByOriginal ] 
					local isProxyOnOnlineVolume = onlineVolumes[ volumeGuidUsedByProxy ]
					
					-- Calculate the new sync status.
					
					syncStatusId = WFCatSyncStatus.calculateSyncStatusWithVolumeStatus( 
							syncStatusId, volumeGuidUsedByOriginal, volumeGuidUsedByProxy,
							isOriginalOnKnownVolume, isProxyOnKnownVolume,
							isVolumeForOriginalsOnline, isVolumeForProxiesOnline,
							isVolumeForOriginalsOutOfSpace, isVolumeForProxiesOutOfSpace,
							isOriginalOnOnlineVolume, isProxyOnOnlineVolume )
					
					local syncStatus = kAssetSyncStatusById[ syncStatusId ]
					
					if syncStatus == kAssetNeedsOriginalUploadFromOfflineVolume then
					
						volumesRequiringSync[ volumeGuidUsedByOriginal ] = true
					
					elseif syncStatus == kAssetNeedsOriginalDownloadToOfflineVolume then
						
						volumesRequiringSync[ volumeGuidForOriginals ] = true
						
					elseif syncStatus == kAssetNeedsProxyDownloadToOfflineVolume then
					
						volumesRequiringSync[ volumeGuidForProxies ] = true
					
					end
					
				end
			
				return volumesRequiringSync
			
			end )
			
	return volumesRequiringSyncingStream	
		
end

--------------------------------------------------------------------------------

function WFCatSyncStatus.getAssetSyncStatusList()

	return kAssetSyncStatus

end

--------------------------------------------------------------------------------

return WFCatSyncStatus
