
--[[----------------------------------------------------------------------------

WFCatSynchronizer.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.oz_client.managed_catalog
local WFCatBinaryUploader = require 'WFCatBinaryUploader'
local WFCatConflictResolver = require 'WFCatConflictResolver'
local WFCatIdleDownloader = require 'WFCatIdleDownloader'
local WFCatSyncClient = require 'WFCatSyncClient'
local WFCatStorage = require 'WFCatStorage'

local WFCatSyncAutoDeleter = require 'WFCatSyncAutoDeleter'
local WFCatStaleBinaryRemover = require 'WFCatStaleBinaryRemover'
local WFCatStyleSynchronizer = require 'WFCatStyleSynchronizer'
local WFCatPersonBinaryDownloader = require 'WFCatPersonBinaryDownloader'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFConfiguration = require 'WFConfiguration'
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local KSPrefs = require 'KSPrefs'

	-- Lightroom CC
local NmbConnectors = require 'NmbConnectors'
local NmbErrorHandler = require 'NmbErrorHandler'
local NmbOfflineAssetsTask = require 'NmbOfflineAssetsTask'
local NmbFileManagementSettings = require 'NmbFileManagementSettings'

--------------------------------------------------------------------------------
	
local debug_disableSync = WFConfiguration.Sync.disable or false

local check = assert( WFArguments.check )

--==============================================================================

local WFCatSynchronizerMethods = {

	_wfCatalog = false,
		-- (WFCatalog)

	_pixelProviderProcessor = false,
		-- (PixelProviderProcessor)

	_metadataSyncClient = false,
		-- (WFCatSyncClient)
		
	_ozStorageMonitor = false,
		-- (WFCatStorage)

	_conflictResolver = false,
		-- (WFCatConflictResolver)
		
	_autoDeleter = false,
		-- (WFCatSyncAutoDeleter)
		
	_profileSyncClient = false,
		-- (WFCatSyncClient)

	_idleDownloader = false,
		-- (WFCatIdleDownloader)

	_staleBinaryRemover = false,
		-- (WFCatStaleBinaryRemover)

	_binaryUploader = false,
		-- (WFCatBinaryUploader)
	
	_styleSynchronizer = false,
		-- (WFCatStyleSynchronizer)
		
	_personThumbnailDownloader = false,
		-- (WFCatPersonBinaryDownloader)
		
	_nmbOfflineAssetsTask = false,
		-- (NmbOfflineAssetsTask)
		
	_errorHandler = false,
		-- (WFCatErrorHandler)
		
	_isSyncRunning = false,
		-- (boolean)
		
	_residualMode = false,
		-- (boolean)
		
	_freemiumMode = false,
		-- (boolean)

	_restrictedCountryMode = false,
		-- (boolean)

	_catalogValidationStream = false,
		-- (Observable<boolean> denoting Oz/local catalog validity for syncing)
		
	_idleDownloaderDelaySubscription = false,
		-- (Rx Timer Subscription)
		
	_isSyncRunningStream = false,
		-- (BehaviorSubject)
	
	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------

local function buildDownloadBinaryStream( pixelProviderProcessor, idlePixStreamIds )

	local pixStreamIdsLut = KSTableUtils.invert( idlePixStreamIds )

	return pixelProviderProcessor
			:pixStreamIdsPixelPathStream( pixStreamIdsLut )
			:tap( 

				function( pixStreamId, localAssetId )

					pixelProviderProcessor
						:runUntilDone( 
							"removeRequest", 
								pixStreamId, localAssetId )

				end )

end

--------------------------------------------------------------------------------

local function buildDownloadBinaryFunc( pixelProviderProcessor, idlePixStreamIds )

	return function( localAssetId, binaryType, binarySize, isMarkedForOffline, hasLocalOriginal )

		local pixStreamId = idlePixStreamIds[ binaryType ]

		pixelProviderProcessor
			:runUntilDone( 
				"requestBinaryWhenIdle", 
					pixStreamId, localAssetId, binaryType, binarySize,
					isMarkedForOffline, hasLocalOriginal )
		
	end

end

--------------------------------------------------------------------------------

local function buildBatchCancelDownloadsFunc( pixelProviderProcessor, idlePixStreamIds )

	return function( binaryType, localAssetIds )

		local pixStreamId = idlePixStreamIds[ binaryType ]

		pixelProviderProcessor
			:runUntilDone( 
				"batchRemoveRequests", pixStreamId, localAssetIds )
		
	end

end

--------------------------------------------------------------------------------

local function validateLocalAndOzCatalogMatch( self )
	
	self._wfCatalog:docStore():localDocValue( 'Oz-nextUpdateFeedLink' )
		:map( function( maybeNextFeedLinkTbl )
			local isValid = true
			local ozCatalogId = self._wfCatalog:ozCatalogId()
			
			if ozCatalogId and type( maybeNextFeedLinkTbl ) == 'table'  then
				if maybeNextFeedLinkTbl.ozCatalogId then
					isValid = maybeNextFeedLinkTbl.ozCatalogId == ozCatalogId
				elseif maybeNextFeedLinkTbl.link then
					isValid = not not string.find( maybeNextFeedLinkTbl.link, ozCatalogId )
				end
			end
			
			self._catalogValidationStream:onNext( isValid )
			
			return isValid
		end )
		:call( WFRxRunUntilDone )

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:__init( wfCatalog, profileCatalog, pixelProviderProcessor, 
	workerThreadAgrs, syncEnabled, residualMode, freemiumMode )

	check( 'WFCatSynchronizer', 'WFCatalog', '?WFCatalog', 'PixelProviderProcessor', 'table', '?boolean', '?boolean' )

	self._wfCatalog = wfCatalog
	self._pixelProviderProcessor = pixelProviderProcessor
	
	self._conflictResolver = WFCatConflictResolver( wfCatalog:docStore(), wfCatalog:indexStore() )
	self._ozStorageMonitor = WFCatStorage( wfCatalog )
	
	self._metadataSyncClient = WFCatSyncClient( wfCatalog )
	
	self._styleSynchronizer = WFCatStyleSynchronizer( wfCatalog, profileCatalog, workerThreadAgrs )
	
	self._personThumbnailDownloader = WFCatPersonBinaryDownloader(wfCatalog)
	
	self._catalogValidationStream = Rx.ReplaySubject( 1 )
	validateLocalAndOzCatalogMatch( self )

	-- monitor quota
	self._ozStorageMonitor:isOverStorage()
		:distinctUntilChanged()
		:map( function( isOverStorage )
			self._metadataSyncClient:setUserIsOverQuota( isOverStorage )
		end )
		:call( WFRxRunUntilDone )
	
	-- monitor expiration
	wfCatalog:ozUser():isSubscriptionExpired()
		:distinctUntilChanged()
		:map( function( isExpired )
			self._metadataSyncClient:setUserIsExpired( isExpired )
		end )
		:call( WFRxRunUntilDone )
	
	self._autoDeleter = WFCatSyncAutoDeleter( wfCatalog )
	
	if profileCatalog then					  
		self._profileSyncClient = WFCatSyncClient( profileCatalog, {
																	metadataPullerArgs = {
																	 	progressiveBackoff = false,
																		noWorkDelay = 60 * 60 * 24, -- 1 day
																		errorDelay = 60 * 60, -- 1 hour
																		maxDelay = 60 * 60 * 24, -- 1 day 
																	},
																	isSyncClientForProfileCatalog = true
																} )
		self._profileSyncClient:addCloseAction( function()
			profileCatalog:close()
		end )
	end
	
	local idlePixStreamBaseId = string.format( "%s:%s", 'WFCatSynchronizer', KSIdGenerator.uuidInLowercaseHex() )
	
	local idlePixStreamIds = {}
	idlePixStreamIds[ 'original' ] = string.format( "%s:%s", idlePixStreamBaseId, 'original' )
	idlePixStreamIds[ 'proxy' ] = string.format( "%s:%s", idlePixStreamBaseId, 'proxy' )
	idlePixStreamIds[ 'xmp_develop' ] = string.format( "%s:%s", idlePixStreamBaseId, 'xmp_develop' )
	
	local idleDownloadStream = buildDownloadBinaryStream( pixelProviderProcessor, idlePixStreamIds )
	local idleDownloadFunc = buildDownloadBinaryFunc( pixelProviderProcessor, idlePixStreamIds )
	local batchCancelDownloadsFunc = buildBatchCancelDownloadsFunc( pixelProviderProcessor, idlePixStreamIds ) 
	
	self._idleDownloader = WFCatIdleDownloader( 
			wfCatalog, idleDownloadFunc, idleDownloadStream, batchCancelDownloadsFunc,
			NmbFileManagementSettings.keepAllOriginals(),
			NmbFileManagementSettings.keepAllProxies() )

	self._binaryUploader = WFCatBinaryUploader( wfCatalog )

	self._staleBinaryRemover = WFCatStaleBinaryRemover( wfCatalog, self._idleDownloader )

	self._nmbOfflineAssetsTask = NmbOfflineAssetsTask( wfCatalog )

	self._errorHandler = NmbErrorHandler( wfCatalog )
	
	self._residualMode = residualMode
	
	self._freemiumMode = freemiumMode
	
	self._isSyncRunningStream = Rx.BehaviorSubject( syncEnabled )
	
	-- check for the latest connectors available 
	Observable.timer( 10, WFRxScheduler.sharedScheduler() )
		:flatMap( function()
			return NmbConnectors.updateAvailableConnectorsCache( wfCatalog )
		end )
		:call( WFRxRunUntilDone )
	
	if syncEnabled then
		
		-- delay the start of the idle downloader a bit so that we're not overaggressively pulling binaries at the 
		-- start of sync when other things are likely more important to get through the sync process
		-- The purge process is setup to delay for 15 seconds on startup; it's good to have the idle downloader
		-- delay slightly longer than purge so purge can prevent the idle downloader from running if needed.
		
		self:startSyncing( 20 )
		
	end

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:close()

	check( 'WFCatSynchronizer' )

	self:stopSyncing()
	
	if self._profileSyncClient then 
		self._profileSyncClient:close()
	end
	
	if self._catalogValidationStream then
		self._catalogValidationStream:dispose()
	end
	
	if self._isSyncRunningStream then
		self._isSyncRunningStream:dispose()
	end
	
	self._autoDeleter:close()
	
	self._metadataSyncClient:close()
	self._ozStorageMonitor:close()
	self._idleDownloader:close()
	self._styleSynchronizer:close()
	self._personThumbnailDownloader:close()
	self._binaryUploader:close()
	self._errorHandler:close()
	self._staleBinaryRemover:close()
	self._conflictResolver:close()
	self._nmbOfflineAssetsTask:close()
	
end
		
--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:startSyncing( maybeIdleDownloaderDelay )
	
	if WFCloseSupport.hasBeenOrWillBeClosed( self._wfCatalog ) then
		-- ignore start request if catalog is closed/closing
		return
	end
	
	local ozUserId = self._wfCatalog:ozUserId()
	
	if not ozUserId:match( '^0+$' ) and not debug_disableSync and not self._restrictedCountryMode then
		
		if self._isSyncRunning then
			return
		end
		
		self._catalogValidationStream
			:take( 1 )
			:map( function( isValid )
			
				if not isValid then	
					-- catalog is not valid, don't allow syncing
					self:stopSyncing()
			
				else
				
					-- catalog is fine, start up sync
					self._isSyncRunning = true
					self._isSyncRunningStream:onNext( true )
								
					if self._profileSyncClient then
						self._profileSyncClient:startSync()
					end
					
					if self._residualMode then
						self._metadataSyncClient:startSyncInResidualMode()
					elseif self._freemiumMode then
						self._metadataSyncClient:startSpaceMetadataSyncInFreemiumMode()
					else
						self._metadataSyncClient:startSync()
					end
					
					self._idleDownloaderDelaySubscription =
						Observable.timer( maybeIdleDownloaderDelay or 0.1, WFRxScheduler.sharedScheduler() )
							:map( function()
								self._idleDownloader:start()
							end )
							:subscribe()
							
					self._styleSynchronizer:start()
			
					if self._freemiumMode or self._residualMode then
						self._binaryUploader:startContinuousUploadOfOnlySpaceAssets()
					else
						self._binaryUploader:startContinuousUploads()
					end
	
				end
			end )
			:call( WFRxRunUntilDone )
	end

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:stopSyncing()

	if self._profileSyncClient then 
		self._profileSyncClient:stopSync()
	end
	
	if self._residualMode then
		self._metadataSyncClient:stopSyncInResidualMode()
	elseif self._freemiumMode then
		self._metadataSyncClient:stopSpaceMetadataSyncInFreemiumMode()
	else
		self._metadataSyncClient:stopSync()
	end
	
	if self._idleDownloaderDelaySubscription then
		self._idleDownloaderDelaySubscription:dispose()
		self._idleDownloaderDelaySubscription = nil
	end
	
	self._idleDownloader:stop()
	
	self._styleSynchronizer:stop()
	
	self._binaryUploader:stopContinuousUploads()
	
	self._isSyncRunning = false
	self._isSyncRunningStream:onNext( false )
	
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:enableResidualMode()
	
	self._freemiumMode = false
	self._residualMode = true

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:disableResidualMode()
	
	self._residualMode = false

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:enableFreemiumMode()
	
	self._residualMode = false
	self._freemiumMode = true

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:disableFreemiumMode()
	
	self._freemiumMode = false

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:enableRestrictedCountryMode()
	
	self._restrictedCountryMode = true

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:catalogValidationStream()

	return self._catalogValidationStream

end
--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isSyncRunningNow()
	
	return self._isSyncRunning

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isSyncRunningStream()
	
	return self._isSyncRunningStream

end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:shouldSyncBeRunning()
	
	return not KSPrefs.syncPaused
		-- KSPrefs.syncPaused always contains the correct value for whether or
		-- not sync should be enabled.
		
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isUploadingBinaries()
	return self._binaryUploader:isDoingWork()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isDownloadingRequiredOriginalsOrProxies()
	return self._idleDownloader:isDownloadingRequiredBinary()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isUploadingMetadata()
	return self._metadataSyncClient:isUploadingMetadata()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isDownloadingMetadata()
	return self._metadataSyncClient:isDownloadingMetadata()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isDownloadingStyles()
	return self._styleSynchronizer:isDownloadingStyles()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getLatestDownloadedAssetIdsStream()
	return self._metadataSyncClient:getLatestDownloadedAssetIdsStream()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:isOzCatalogDeleted()
	return self._metadataSyncClient:isOzCatalogDeleted()
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getIdleDownloader()
	return self._idleDownloader
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getBinaryUploader()
	return self._binaryUploader
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getStyleSynchronizer()
	return self._styleSynchronizer
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getPersonThumbnailDownloader()
	return self._personThumbnailDownloader
end

--------------------------------------------------------------------------------

function WFCatSynchronizerMethods:getRequiredUpdateStream()

	-- Listen on different threads, all of which make requests to Oz, for a required update error.

	return Rx.Observable.combineLatest( function( maybeRequiresUpdate1, maybeRequiresUpdate2, maybeRequiresUpdate3 )
			return maybeRequiresUpdate1 or maybeRequiresUpdate2 or maybeRequiresUpdate3
		end,
		self._metadataSyncClient:getRequiredUpdateStream(),
		self._profileSyncClient:getRequiredUpdateStream(),
		self._wfCatalog:ozUser():getRequiredUpdateStream() )
		
end

--------------------------------------------------------------------------------

return Rx.Class( 'WFCatSynchronizer', WFCatSynchronizerMethods )
