--[[----------------------------------------------------------------------------

WFCatTrashTestSession.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2019 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Adobe and its suppliers, if any. The intellectual
and technical concepts contained herein are proprietary to Adobe
and its suppliers and are protected by all applicable intellectual
property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Adobe.

----------------------------------------------------------------------------]]--

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.rxlua
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.os
local WFPlatformInfo = require "WFPlatformInfo"

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFRxBatchedBulkApplyOverlay = require 'WFRxBatchedBulkApplyOverlay'

	-- WF.core.table
local WFTableArchiving = require 'WFTableArchiving'

	-- WF.oz_client.managed_catalog
local WFCatBinaryUploader = require 'WFCatBinaryUploader'
local WFCatConflictResolver = require 'WFCatConflictResolver'
local WFCatErrorHandler = require 'WFCatErrorHandler'
local WFCatMetadataPuller = require 'WFCatMetadataPuller'
local WFCatMetadataPusher = require 'WFCatMetadataPusher'
local WFCatMockCatalog = require 'WFCatMockCatalog'
local WFCatSpecMock = require 'WFCatSpecMock'
local WFCatSyncStatus = require 'WFCatSyncStatus'
local LrmCatSyncStatus = require 'LrmCatSyncStatus'
local WFCatTestSession = require 'WFCatTestSession'

	-- WF.web.http_client
local WFRxHttpClientMock = require 'WFRxHttpClientMock'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'


local logger = require 'WFLogger'( 'WFCatTrashTestSession' )

local doRequestResponseLogging = require 'WFConfiguration'.trashTest.enableRequestResponseLogging

if doRequestResponseLogging then
	logger:enable( 'logfile' )
end


local kOptIndexSpec
local kAlbumAssetQuery
if WFPlatformInfo.isDesktop then
	kAlbumAssetQuery = [[SELECT album_asset_v2.docId, album_asset_v2.assetId FROM TEMP_ids JOIN album_asset_v2 ON
							TEMP_ids.docId = album_asset_v2.assetId]]
else
	kOptIndexSpec = 'LrmCatalogIndexDefinitions.lua'
	kAlbumAssetQuery = [[SELECT album_asset.docId, album_asset.assetId FROM TEMP_ids JOIN album_asset ON
							TEMP_ids.docId = album_asset.assetId]]
end

--------------------------------------------------------------------------------

local check = WFArguments.check

local stripped = '<<stripped>>'

--==============================================================================

local function stripUpdatedSinceTime( url )

	local relativeUrl = url:replace( 'https://photos-stage.adobe.io', '')

	local updatedSinceTime = relativeUrl:match( "updated_since=(.*)" )
	if updatedSinceTime then

		relativeUrl = relativeUrl:replace( updatedSinceTime, stripped )

	end

	return relativeUrl

end

--------------------------------------------------------------------------------

local function maybeAddQuickerRetriesHeader( options )

	for _, header in ipairs( options.requestHeaders or {} ) do
		if header.header and header.header:lower() == 'x-quicker-retries' then
			return
		end
	end

	options.requestHeaders = options.requestHeaders or {}

	options.requestHeaders[ #options.requestHeaders + 1 ] = {
		header = 'X-Quicker-Retries',
		value = 'true',
	}

end

----------------------------------------------------------------------------

local function makeWFCatSpecSpyWithHttpMockTracking( rxTest, catSpec, options )

	local defaultTransform = options.defaultTransform
	local injectQuickerRetries = options.injectQuickerRetries

	local skipResponseDataCheck = false

	local function compareData( url, optArgs, traceback, mockData, ozData, transformFn )

		if skipResponseDataCheck then
			return
		end
		local mockDataAsTable = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( mockData )
		local ozDataAsTable = KSJsonDecoder.decodeWebSafeJson_stripNull_ignoreEmpty( ozData )

		-- WFHttpClientImpCurl.cpp's OnResponseReceived_L() is mapping empty responses to nil
		-- whereas WFHttpClientTask.m pushResponseDataOrError does not.
		-- so if we get a nil response and are expecting an empty table set it to an empty table
		if ozDataAsTable == nil and mockDataAsTable and next(mockDataAsTable) == nil then
			ozDataAsTable = {}
		end

		-- Add same url and optArgs to make it easier to debug failures
		if mockDataAsTable then
			mockDataAsTable._DEBUG_url = url
			mockDataAsTable._DEBUG_optArgs = optArgs
			mockDataAsTable._DEBUG_traceback = traceback
		end

		if ozDataAsTable then
			ozDataAsTable._DEBUG_url = url
			ozDataAsTable._DEBUG_optArgs = optArgs
			ozDataAsTable._DEBUG_traceback = traceback
		end

		if transformFn then
			mockDataAsTable, ozDataAsTable = transformFn( mockDataAsTable, ozDataAsTable, true )
		else
			mockDataAsTable, ozDataAsTable = defaultTransform( mockDataAsTable, ozDataAsTable, true )
		end

		-- Compare the mock data to the real oz data to make sure the mocks
		-- are correct, and that we are getting what is expected from oz.
		rxTest:expectEqual( mockDataAsTable, ozDataAsTable )

	end

	local function compareResultHeaders( url, optArgs, traceback, mockResultHeaders, ozResultHeaders )

		-- For now just compare status_code instead of entire result header.

		-- Sample header:
		--	["Access-Control-Allow-Credentials"] = "true",
		--	["Access-Control-Allow-Methods"] = "GET,HEAD,PUT,DELETE,POST,OPTIONS",
		--	["Access-Control-Max-Age"] = "60",
		--	Connection = "keep-alive",
		--	["Content-Length"] = "154",
		--	["Content-Type"] = "application/json",
		--	Date = "Mon, 20 Aug 2018 17:27:49 GMT",
		--	Server = "APIP",
		--	Status = "415 Unsupported Media Type",
		--	["X-Account-Id"] = "2e28b0705701e8fe6f7cba1c79c891e7",
		--	["X-Client-Id"] = "WFMacTest",
		--	["X-Client-Identifier"] = "Lightroom",
		--	["X-Gw-App-DevMode"] = "true",
		--	["X-Request-Id"] = "qfRKFJB89dmnqv1jSZprBB4dbDi1b0Hh",
		--	["X-Result-Code"] = "1003",
		--	["X-Result-Subtype"] = "InvalidContentTypeError",
		--	["X-Traffic-Type"] = "customer",
		--	retryCount = 2,
		--	status_code = "415",
		--	wf_time = 1.1892045750283,

		if ozResultHeaders.status_code ~= mockResultHeaders.status_code then

			local mockResultsTable = {
				_DEBUG_url = url,
				_DEBUG_optArgs = optArgs,
				_DEBUG_traceback = traceback,
				resultHeaders = mockResultHeaders
			}

			local ozResultsTable = {
				_DEBUG_url = url,
				_DEBUG_optArgs = optArgs,
				_DEBUG_traceback = traceback,
				resultHeaders = ozResultHeaders
			}

			rxTest:expectEqual( mockResultsTable, ozResultsTable )

		end

	end

	local function put( self, url, data, optArgs )

		if injectQuickerRetries then
			-- Add here to be compatible with unit test request headers
			-- that specify this in case we remove the always add below.
			maybeAddQuickerRetriesHeader( optArgs )
		end

		return self._httpMock.put( stripUpdatedSinceTime( url ), data, optArgs )
		:flatMap( function( mockData, mockResultHeaders )

			local requestNum = self._httpMock.requestCount()

			-- Oz says graceful shutdown should allow use of quicker retries on all
			-- http traffic. see https://www.pivotaltracker.com/story/show/164908026
			-- turn the option on for all tests. Add here so http mock doesn't need
			-- to specify X-Quicker-Retries for all the requests.
			maybeAddQuickerRetriesHeader( optArgs )

			return catSpec:put( url, data, optArgs )
			:map( function( ozData, ozResultHeaders )
				local traceback = self._httpMock.requestTraceback( requestNum )
				local transformFn = self._httpMock.requestResponseTransformFn( requestNum )
				compareData( url, optArgs, traceback, mockData, ozData, transformFn )
				compareResultHeaders( url, optArgs, traceback, mockResultHeaders, ozResultHeaders )
				
				if doRequestResponseLogging then
					local verb = optArgs.requestMethod or 'PUT'
					logger:infof( "\n\n######################## %s ########################\n\n%s\n\ngot response\n\n%s\n\n",
						verb,
						WFTableArchiving.tableToString( KSJsonDecoder.decodeWebSafeJson( data ) ),
						WFTableArchiving.tableToString( KSJsonDecoder.decodeWebSafeJson( ozData ) ) )
				end

				return ozData, ozResultHeaders
			end )
-- TODO [kcarlson 2018-06-27]: handle errors

		end )
	end

	local function get( self, url, optArgs )
	
		if injectQuickerRetries then
			-- Add here to be compatible with unit test request headers
			-- that specify this in case we remove the always add below.
			maybeAddQuickerRetriesHeader( optArgs )
		end

		local originalOptArgs = KSTableUtils.deepCopy( optArgs )
		local dontCompareToMockResponse = optArgs.dontCompareToMockResponse
		optArgs.dontCompareToMockResponse = nil

		return self._httpMock.get( stripUpdatedSinceTime( url ), originalOptArgs )
		:flatMap( function( mockData, mockResultHeaders )

			local requestNum = self._httpMock.requestCount()

			-- Oz says graceful shutdown should allow use of quicker retries on all
			-- http traffic. see https://www.pivotaltracker.com/story/show/164908026
			-- turn the option on for all tests. Add here so http mock doesn't need
			-- to specify X-Quicker-Retries for all the requests.
			maybeAddQuickerRetriesHeader( optArgs )

			return catSpec:get( url, optArgs )
			:map( function( ozData, ozResultHeaders )

				if not dontCompareToMockResponse then
					local traceback = self._httpMock.requestTraceback( requestNum )
					local transformFn = self._httpMock.requestResponseTransformFn( requestNum )
					compareData( url, optArgs, traceback, mockData, ozData, transformFn )
					compareResultHeaders( url, optArgs, traceback, mockResultHeaders, ozResultHeaders )
				end

				if doRequestResponseLogging then
					logger:infof( "\n\n######################## GET ########################\n\n%s\n\ngot response:\n\n%s\n\n",
						url, WFTableArchiving.tableToString( KSJsonDecoder.decodeWebSafeJson( ozData ) ) )
				end
				
				return ozData, ozResultHeaders
			end )
-- TODO [kcarlson 2018-06-27]: handle errors

		end )
	end

	local function putBinary( self, url, data, optArgs )

		local copyOfOptArgs = KSTableUtils.deepCopy( optArgs )

		-- Issue the mocked call first since it will return results immediately
		-- and more likely in order.
		return self._httpMock.put( stripUpdatedSinceTime( url ), data, copyOfOptArgs )
		:catch( function( mockError, mockResultHeaders )
			return Observable.just( mockError, mockResultHeaders )
		end )
		:flatMap( function( mockData, mockResultHeaders )

			-- Get requestNum here in case two simultaneous putBinary calls
			-- return results out the order they were issued in. As we can
			-- issue an xmp_develop and master request simultaneously.
			local requestNum = self._httpMock.requestCount()

			return catSpec:putBinary( url, data, optArgs )
			:map( function( ozData, ozResultHeaders )
				local traceback = self._httpMock.requestTraceback( requestNum )
				compareResultHeaders( url, optArgs, traceback, mockResultHeaders, ozResultHeaders )
				return ozData, ozResultHeaders
			end )
			:catch( function( ozErrorJson, ozResultHeaders )
				local traceback = self._httpMock.requestTraceback( requestNum )
				compareResultHeaders( url, optArgs, traceback, mockResultHeaders, ozResultHeaders )
				return Observable.throw( ozErrorJson, ozResultHeaders )
			end )

		end )
	end

	local function getBinary( self, url, optArgs )
		return self._httpMock.get( stripUpdatedSinceTime( url ), optArgs )
		:flatMap( function( mockData, mockResultHeaders )

			local requestNum = self._httpMock.requestCount()

			return catSpec:getBinary( url, optArgs )
			:map( function( ozData, ozResultHeaders )
				local traceback = self._httpMock.requestTraceback( requestNum )
				compareResultHeaders( url, optArgs, traceback, mockResultHeaders, ozResultHeaders )
				return ozData, ozResultHeaders
			end )
-- TODO [kcarlson 2018-06-27]: handle errors

		end )
	end

	local function expectHttpRequest( self, httpRequest )

		check( 'WFCatSpec' )

		self._httpMock.addHttpRequest( httpRequest )

	end

	local WFCatSpecSpy = {
		_httpMock = WFRxHttpClientMock( rxTest, {} ),

		getBinary = getBinary,
		putBinary = putBinary,
		put = put,
		get = get,
		catalogHref = function( self, ...) return catSpec:catalogHref( ... ) end,
		catalogUpdatesUrl = function( self, ...) return catSpec:catalogUpdatesUrl( ... ) end,
		expectHttpRequest = expectHttpRequest,
		requestCount = function( self ) return self._httpMock.requestCount() end,
	}

	return Rx.Class( 'WFCatSpec', WFCatSpecSpy )()

end

----------------------------------------------------------------------------

local function bulkDelete( catalog, docIds  )

	-- very simple bulkDelete that does not delete album_assets automatically

	local overlays = {}

	for i, docId in ipairs( docIds ) do
		overlays[ #overlays + 1 ] = {
			_id = docId,
			_deleted = true,
			_localOnly = { deletedLocally = true },
		}
	end

	return WFRxBatchedBulkApplyOverlay( "bulkDeleteTest", catalog:docStore(), overlays )

end

----------------------------------------------------------------------------

local function deleteDoc( catalog, docId )
	return bulkDelete( catalog, { docId } )
end

----------------------------------------------------------------------------

local continuousTaskDelays = {
	progressiveBackoff = false,
	noWorkDelay = 1,
	maxNoWorkDelay = 1,
	errorDelay = 1,
	maxDelay = 1,
}

----------------------------------------------------------------------------

local function albumAssetsForFilteredAssetsQueryFn( catalog, assetIds )

	-- This function is used to find album_assets related to assets that
	-- are blocked from upload because of an ozError.

	return catalog:indexStore():queryOnce {
		sql = kAlbumAssetQuery,
		tempIds = assetIds,
	}

end

----------------------------------------------------------------------------

local function setupOzTestSession( rxTest, options )

	local catalogTestSession = WFCatTestSession( rxTest, 'NmbCatalog', kOptIndexSpec )

	local catalog, catalogSpec

	rxTest:expectSingleValue( catalogTestSession:catalog()
		:map( function( _catalog )
			check( 'WFCatalog' )
			catalog = _catalog
			catalogSpec = _catalog:ozCatalogSpec()
			return _catalog
		end )
	, ':ignore:' )

	rxTest:expectSingleValue( catalog:ozUser():getAccountDocFromOz(), ':ignore:' )

	local catalogSpecSpy = makeWFCatSpecSpyWithHttpMockTracking( rxTest, catalogSpec, options )

	local conflictResolver = WFCatConflictResolver( catalog:docStore(), catalog:indexStore() )
	rxTest:addCloseTarget( conflictResolver )

	local pusher = WFCatMetadataPusher( catalog, {
			delayBeforeSendingUpdates = .01,
			continuousTaskDelays = continuousTaskDelays,
			albumAssetsForFilteredAssetsQueryFn = albumAssetsForFilteredAssetsQueryFn,
		},
		catalogSpecSpy )

	rxTest:addCloseTarget( pusher )

	local pullerOptions = {
		continuousTaskDelays = continuousTaskDelays,
		catSpec = catalogSpecSpy,
		isCatalogPuller = true
	}

	local puller = WFCatMetadataPuller( catalog, "Test Puller", pullerOptions )
	rxTest:addCloseTarget( puller )

	local uploader = WFCatBinaryUploader( catalog, nil, continuousTaskDelays, catalogSpecSpy )

	rxTest:finally( function()

		uploader:stopContinuousUploads()
		uploader:close()

		catalog:indexStore():close()

	end )

	return catalogTestSession, catalog, catalogSpecSpy, pusher, puller, uploader, conflictResolver

end

--------------------------------------------------------------------------------

local function setupMockedCatalog( rxTest, catalogId, options )

	options = options or {}

	local userId = '87f526b44f9e41eca6cb5a550f84ceb6'

	local catalogSpec = WFCatSpecMock( catalogId, userId, rxTest, {} )
	local catalog = WFCatMockCatalog.createWFCatalogForTestWithSpec( rxTest, catalogSpec, options.catalogClassType, kOptIndexSpec )

	local conflictResolver = WFCatConflictResolver( catalog:docStore(), catalog:indexStore() )
	rxTest:addCloseTarget( conflictResolver )

	local pusher = WFCatMetadataPusher( catalog, {
			delayBeforeSendingUpdates = .01,
			continuousTaskDelays = continuousTaskDelays,
			albumAssetsForFilteredAssetsQueryFn = albumAssetsForFilteredAssetsQueryFn,
		} )

	rxTest:addCloseTarget( pusher )

	local pullerOptions = {
		continuousTaskDelays = continuousTaskDelays,
		isCatalogPuller = true
	}

	local puller = WFCatMetadataPuller( catalog, "Test Puller", pullerOptions )
	rxTest:addCloseTarget( puller )

	local uploader = WFCatBinaryUploader( catalog, nil, continuousTaskDelays, nil, {
			clearFromRecentlyUploadedSeconds = 1,
			timer = function( seconds )
				return Observable.timer( seconds, WFRxScheduler.sharedScheduler() )
			end,
		} )

	rxTest:finally( function()

		uploader:stopContinuousUploads()
		uploader:close()

		catalog:indexStore():close()

	end )

	return catalog, catalogSpec, pusher, puller, uploader, conflictResolver

end

--------------------------------------------------------------------------------

local function setupTestSession( rxTest, options )

	local testSession, catalog, catalogSpec, pusher, puller, uploader, conflictResolver

	if options.mock then
		local mockedCatalogId = '38d6f02c0177438aa3552161cc3f427f'
		catalog, catalogSpec, pusher, puller, uploader, conflictResolver = setupMockedCatalog( rxTest, mockedCatalogId, options )
		testSession = true
	else
		testSession, catalog, catalogSpec, pusher, puller, uploader, conflictResolver = setupOzTestSession( rxTest, options )
		assert( testSession )
	end

	local catSyncStatus = WFPlatformInfo.isDesktop and WFCatSyncStatus or LrmCatSyncStatus

	local errorHandler = WFCatErrorHandler( catalog,
		{
			deleteAssets = bulkDelete,
			deleteIndividualAlbum = deleteDoc,
			getSyncStatusOfAssetsOfAlbumAssets = catSyncStatus.getSyncStatusOfAssetsOfAlbumAssets,
			getSyncStatusOfAlbumsOfAlbumAssets = catSyncStatus.getSyncStatusOfAlbumsOfAlbumAssets,
			getSyncStatusOfParentCollectionOfAlbums = catSyncStatus.getSyncStatusOfParentCollectionOfAlbums,
		},
		{ debugMode = true } )

	rxTest:addCloseTarget( errorHandler )

	local pusherControl = {
		pauseReturningCounter = function()
			return pusher:pauseReturningCounter()
		end,
		resume = function()
			return pusher:resume()
		end,
	}

	return testSession, catalog, catalogSpec, pusher, puller, uploader, pusherControl, errorHandler, conflictResolver

end

--------------------------------------------------------------------------------

return {
	setupTestSession = setupTestSession,
	stripUpdatedSinceTime = stripUpdatedSinceTime,
}
