--[[----------------------------------------------------------------------------

WFCatalogBase.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local _ = require 'WFArguments_uuid'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.oz_client.managed_catalog
local WFCatFileSpec = require 'WFCatFileSpec'
 
	-- WF.sqlite.doc_store
local WFRxDocStore = require 'WFRxDocStore'

	-- WF.sqlite.index_store
local WFRxIndexStore = require 'WFRxIndexStore'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )


--============================================================================--

-- @mark WFCatalogBase class

local WFCatalogBase = {

	_wfCatSpec = false,
		-- (WFCatSpec) specification for the catalog user and catalog ID
	
	_wfCatFileSpec = false,
		-- (WFCatFileSpec) specification for where the catalog files should live on local storage

	_docStore = false,
		-- (WFRxDocStore)

	_indexStore = false,
		-- (WFRxIndexStore)

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Create the base object for a WFCatalog.

	\param ozCatalogSpec (WFCatSpec) specification for the catalog user and catalog ID
	\param rootCatalogPath (string) root directory under which to store catalog related files
	\param optArgs (table) optional args for doc store/index store creation
		- docStorePath	   -- path to doc store
		- indexSpecScript  	-- name of script to use for index store
]]--

function WFCatalogBase:__init( ozCatalogSpec, rootCatalogPath, optArgs )

	check( '_', 'WFCatSpec', 'string' )

	local optDocStorePath, optIndexSpecScript
	
	if optArgs then
		optDocStorePath = optArgs.docStorePath
		optIndexSpecScript = optArgs.indexSpecScript
	end

	local wfCatFileSpec = WFCatFileSpec( rootCatalogPath, ozCatalogSpec:catalogId(), optDocStorePath )
	self._wfCatFileSpec = wfCatFileSpec
		-- WFCatFileSpec is not a close target
		
	local docStore = WFRxDocStore( wfCatFileSpec:docStorePath() )

	local indexStore = WFRxIndexStore( docStore,
									   optIndexSpecScript or 'WFCatIndexDefinitions.lua' )
	
	self._wfCatSpec = ozCatalogSpec
		-- WFCatSpec is not a close target.

	self:setOwned {
		_docStore = docStore,
		_indexStore = indexStore,
	}

	WFCloseSupport.mustClose( self )

end

--------------------------------------------------------------------------------

function WFCatalogBase:__tostring()

	return string.format( "WFCatalog( %s )", tostring( self._wfCatSpec ) )

end

--------------------------------------------------------------------------------

function WFCatalogBase:ozCatalogSpec()

	return self._wfCatSpec

end

--------------------------------------------------------------------------------

function WFCatalogBase:fileSpec()

	return self._wfCatFileSpec

end

--------------------------------------------------------------------------------

function WFCatalogBase:ozCatalogId()

	return self._wfCatSpec:catalogId()

end

--------------------------------------------------------------------------------

function WFCatalogBase:ozUser()

	return self._wfCatSpec:wfOzUser()

end

--------------------------------------------------------------------------------

function WFCatalogBase:ozUserId()

	return self._wfCatSpec:wfOzUser():userId()

end

--------------------------------------------------------------------------------

function WFCatalogBase:docStore()

	return self._docStore

end

--------------------------------------------------------------------------------

function WFCatalogBase:indexStore()

	return self._indexStore

end

--------------------------------------------------------------------------------

return Rx.Class 'WFCatalog'( WFCatalogBase )
