--[[----------------------------------------------------------------------------

WFCrContext.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.rxlua
local WFRxNativeObservable = require 'WFRxNativeObservable'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.acr
local WFCrContextImp = require 'WFCrContextImp'

--------------------------------------------------------------------------------

assert( WFPlatformInfo.hasACR, "ACR integration not enabled" )

local check = assert( WFArguments.check )

--------------------------------------------------------------------------------

local logger = require "WFLogger"( "WFCrContext" )

--============================================================================--

-- @mark WFCrContext class

--[[!
	\brief WFCrContext provides Lua access to ACR's cr_context object.

	Use this for access to common ACR operations. Extend as necessary for
	additional ACR operations that are not yet supported.
]]--

local WFCrContext = {

	_imp = false,
		-- (WFCrContextImp)

	_filePath = false,
		-- (string) path to the CR negative we're reading

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new WFCrContext object for a single still image file.

	\param filePath (string) path to still image file

	Construction of the WFCrContext object should be very fast. The file is not
	actually opened until some later operation needs it.
]]--

local WFCrContextImp_make = WFCrContextImp.make

--------------------------------------------------------------------------------

function WFCrContext:__init( filePath )

	check( 'WFCrContext', '+string' )

	self._imp = WFCrContextImp_make( filePath )

	self._filePath = filePath

end


--------------------------------------------------------------------------------

function WFCrContext:__tostring()

	return string.format( "WFCrContext( %s )", self._filePath )

end

--------------------------------------------------------------------------------
--[[!
	\brief Read metadata from this file.

	\return Observable< table > resolves to table containing the following members:
		- \c metadata (string) XMP metadata from file
		- \c isRawFile (boolean) true if the file is a non-rendered file
		- \c additionalInfo (table) other metadata that doesn't fit in XMP
			- embeddedXmp (boolean) true if file contains XMP inside the file
			- fileHeight (number) height of file in pixels
			- fileWidth (number) weidth of file in pixels
			- incrementalWhiteBalance (boolean) true if the file's white balance settings are incremental (ie. for output-referred files) or absolute (ie. for scene-referred files)
			- monochrome (boolean) true if the file is a monochrome file
			- ExifDateTimeValue (table) capture date/time (all values are numbers)
				- year
				- month
				- day
				- hour
				- minute
				- second
]]--

local WFCloseSupport_hasBeenOrWillBeClosed =
		WFCloseSupport.hasBeenOrWillBeClosed

local function logAndWrapInErrorObservable( errStr, ... )

	local errorString = select( '#', ... ) == 0 and 
							errStr or string.format( errStr, ... )

	logger:error( errorString )

	return Rx.Observable.throw( errorString )

end
--------------------------------------------------------------------------------

local function wrapInReturnErrorObservableIfClosed( f )

	return function( self, ... )

		if WFCloseSupport_hasBeenOrWillBeClosed( self ) then

			return logAndWrapInErrorObservable( "Can't use closed %s", self )

		end

		return f( self, ... )

	end

end

--------------------------------------------------------------------------------

local WFCrContextImp_readMetadata = WFCrContextImp.readMetadata

--------------------------------------------------------------------------------

function WFCrContext:readMetadata()

	return WFRxNativeObservable( 
			WFCrContextImp_readMetadata, 
				self._imp )

end

--------------------------------------------------------------------------------

WFCrContext.readMetadata = wrapInReturnErrorObservableIfClosed(
								WFCrContext.readMetadata )

--------------------------------------------------------------------------------
--[[!
	\brief Create a JPEG rendition of this file.

	\param settings (string) XMP settings to apply to the source file
	\param longSide (number) maximum # pixels on the long side
	\param quality (number) quality (0 <= quality <= 1)

	\return Observable< KSData > blob containing the requested JPEG rendition
]]--

local WFCrContextImp_render = WFCrContextImp.render

--------------------------------------------------------------------------------

function WFCrContext:render( settings, userOrientation, longSide, quality )

	if type( settings ) ~= "string" or settings == "" then

		return logAndWrapInErrorObservable( 
					"Invalid settings passed to %s:render()", self )

	end

	if type( userOrientation ) ~= "number" then

		return logAndWrapInErrorObservable( 
					"Invalid userOrientation passed to %s:render()", self )

	end

	if type( longSide ) ~= "number" then

		return logAndWrapInErrorObservable( 
					"Invalid longSide passed to %s:render()", self )

	end

	if type( quality ) ~= "number" then

		return logAndWrapInErrorObservable( 
					"Invalid quality passed to %s:render()", self )

	end

	return WFRxNativeObservable( 
			WFCrContextImp_render, 
				self._imp, settings, userOrientation, longSide, quality )

end

--------------------------------------------------------------------------------

WFCrContext.render = wrapInReturnErrorObservableIfClosed( WFCrContext.render )

--------------------------------------------------------------------------------
--[[!
	\brief Render this file and encode it in the requested format.

	\param settings (string) XMP settings to apply to the source file
	\param outputParams (table)
	\param formatParams (table)

	\return Observable< KSData > blob containing the requested rendition
]]--

local WFCrContextImp_renderFinalImageAndEncode = 
		WFCrContextImp.renderFinalImageAndEncode

--------------------------------------------------------------------------------

function WFCrContext:renderFinalImageAndEncode( settings, outputParams, formatParams )

	if type( settings ) ~= "string" or settings == "" then

		return logAndWrapInErrorObservable( 
					"Invalid settings passed to %s:renderFinalImageAndEncode()", self )

	end

	if type( outputParams ) ~= "table" then

		return logAndWrapInErrorObservable( 
					"Invalid outputParams passed to %s:renderFinalImageAndEncode()", self )

	end

	if type( formatParams ) ~= "table" then

		return logAndWrapInErrorObservable( 
					"Invalid formatParams passed to %s:renderFinalImageAndEncode()", self )

	end

	return WFRxNativeObservable( 
			WFCrContextImp_renderFinalImageAndEncode, 
				self._imp, settings, outputParams, formatParams )

end

--------------------------------------------------------------------------------

WFCrContext.renderFinalImageAndEncode = wrapInReturnErrorObservableIfClosed(
										WFCrContext.renderFinalImageAndEncode )

--------------------------------------------------------------------------------
--[[!
	\brief Update the metadata of the file. If the file format doesn't
		support adding metedata, metadata will be added to a sidecar.

	\param settings (string) XMP settings to apply to the source file
	
	\return table containing the following info
		isEmbedded (bool) whether the metedata is embedded or added as 
			sidecar.
		imageBlob Observable< KSData > blob containing the updated image
			Will be available only if isEmbedded is true
		sidecarBlob Observable< KSData > blob containing the sidecar data
			Will be available only if isEmbedded is false
]]--

local WFCrContextImp_updateMetadata = WFCrContextImp.updateMetadata

--------------------------------------------------------------------------------

function WFCrContext:updateMetadata( settings )

	if type( settings ) ~= "string" or settings == "" then

		return logAndWrapInErrorObservable( 
					"Invalid settings passed to %s:updateMetadata()", self )

	end

	return WFRxNativeObservable( 
				WFCrContextImp_updateMetadata, 
					self._imp, settings )

end

--------------------------------------------------------------------------------

WFCrContext.updateMetadata = wrapInReturnErrorObservableIfClosed(
									WFCrContext.updateMetadata )

--------------------------------------------------------------------------------
--[[!
	\brief Merges the metadata provided with the metadata of the file.
		Final metadata will be the union of both, except for the
		develop settings. Develop settings from the input xmp will
		always override the one from the file.

	\param settings (string) XMP settings to merge
	\param xmpSidecar (string or nil) XMP from sidecar to merge

	\return Observable< table > resolves to table containing the following members:
		- \c metadata (string) XMP metadata from file
]]--

local WFCrContextImp_mergeMetadata = WFCrContextImp.mergeMetadata

--------------------------------------------------------------------------------

function WFCrContext:mergeMetadata( settings, xmpSidecarData )

	if type( settings ) ~= "string" then

		return logAndWrapInErrorObservable(
					"Invalid settings passed to %s:mergeMetadata()", self )

	end

	return WFRxNativeObservable(
			WFCrContextImp_mergeMetadata,
				self._imp, settings, xmpSidecarData )

end

--------------------------------------------------------------------------------

WFCrContext.mergeMetadata = wrapInReturnErrorObservableIfClosed(
								WFCrContext.mergeMetadata )

--------------------------------------------------------------------------------

function WFCrContext:close()

	check( 'WFCrContext' )

	if self._imp then
		self._imp:close()
	end

	self._imp = nil

end

--------------------------------------------------------------------------------

return Rx.Class 'WFCrContext'( WFCrContext )
