--[[----------------------------------------------------------------------------

WFCrypto.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2019 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local WFArguments = require 'WFArguments'
local KSData = require 'KSData'

local WFRxNativeObservable = require "WFRxNativeObservable"
local WFCryptoImp = require  "WFCryptoImp"

--==============================================================================

KSData.registerArgumentChecker()

local WFCrypto = {}

--------------------------------------------------------------------------------

--[[!
	\class WFCrypto
	\brief This namespace provides static WFDigestProvider instances
		for these digest algorithms:
		<ul>
		<li>WFCrypto.MD4 (only available in Lightroom)</li>
		<li>WFCrypto.MD5</li>
		<li>WFCrypto.SHA1</li>
		<li>WFCrypto.SHA256</li>
		<li>WFCrypto.SHA384</li>
		<li>WFCrypto.SHA512</li>
		<li>WFCrypto.HMAC</li></ul>
	These instances are static properties of the namespace. They are NOT
	functions, but due to limitations of our documentation engine, they
	are presented here as functions.

	You can use these instances to encode data in a single operation.
	You can also use these singleton instances to create your own instances
	of WFCrypto to perform streaming digest operations.  Streaming data in
	smaller chunks reduces the memory overhead of the digest operation.
]]
	
--[[!
	\brief (WFDigestProvider) The shared singleton instance for MD4.
	THIS IS NOT A FUNCTION, it is a static property of the namespace.
]]
function WFCrypto.MD4()
	-- DOXYGEN STUB
end
	
--[[!
	\brief (WFDigestProvider) The shared singleton instance for MD5.
	THIS IS NOT A FUNCTION, it is a static property of the namespace.
]]
function WFCrypto.MD5()
	-- DOXYGEN STUB
end
	
--[[!
	\brief (WFDigestProvider) The shared singleton instance for SHA1.
	THIS IS NOT A FUNCTION, it is a static property of the namespace.
]]
function WFCrypto.SHA1()
	-- DOXYGEN STUB
end
	
--[[!
	\brief (WFDigestProvider) The shared singleton instance for SHA256.
	THIS IS NOT A FUNCTION, it is a static property of the namespace.
]]
function WFCrypto.SHA256()
	-- DOXYGEN STUB
end
	
--[[!
	\brief (WFDigestProvider) The shared singleton instance for SHA384.
	THIS IS NOT A FUNCTION, it is a static property of the namespace.
]]
function WFCrypto.SHA384()
	-- DOXYGEN STUB
end
	
--[[!
	\brief (WFDigestProvider) The shared singleton instance for SHA512.
	THIS IS NOT A FUNCTION, it is a static property of the namespace.
]]
function WFCrypto.SHA512()
	-- DOXYGEN STUB
end
	
--[[!
	\brief (WFDigestProvider) The shared singleton instance for HMAC.
	THIS IS NOT A FUNCTION, it is a static property of the namespace.
]]
function WFCrypto.HMAC()
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

WFArguments.registerChecker( 'WFDigestAlgorithm', {
	MD4 = true,
	MD5 = true,
	SHA1 = true,
	SHA256 = true,
	SHA384 = true,
	SHA512 = true,
	HMAC = true,
} )

--------------------------------------------------------------------------------

--[[!
	\brief Digest a file by processing it in 4kB chunks.  This
	limits the overall memory consumed by the process.
	
	\param filePath (string) the file path to digest
	\param algorithm (string) the digest algorithm to use
	
	\return The binary digest data as a string, and as a hex string.
]]
function WFCrypto.digestFile( filePath, algorithm )

	WFArguments.check( "+string", "WFDigestAlgorithm" )
	
	local digester = WFCrypto[ algorithm ].init()
	local f = io.open( filePath, "rb" )
	while true do
		local bytes = f:read( 1024 * 1024 * 4 )
		if not bytes then break end
		digester:update( bytes )
	end
	f:close()
	return digester:digest()
	
end

--------------------------------------------------------------------------------

local WFDigestProvider = {}
-- @mark WFDigestProvider namespace

--[[!
	\class WFDigestProvider
	\brief Provides the ability to encode a string or KSBlob in a supported
	digest protocol, directly using the \c digest() function, or by
	streaming data through a WFDigest instance created by the
	WFDigestProvider's \c init() function.
	
	Shared singleton instances for each supported digest algorithm
	are available as static properties of the WFDigest namespace.
--]]

--------------------------------------------------------------------------------

--[[!
	\brief Encodes data according to the digest algorithm,
	returning both the binary digest and a hex-string version of the digested data.
		
	\param stringOrBlob The data to be encoded, as a string or KSData object.
	\param ... For the HMAC protocol ONLY, there are two additional parameters:
	<ul><li>(string) The name of a secondary digest algorithm (such as "SHA1")</li>
		<li> (string or KSData) A secret.</li></ul>
	\return The binary digest data as a string, and as a hex string.
--]]

function WFDigestProvider.digest( --[[string or blob]] stringOrBlob, ... )
	WFArguments.check( 'string|KSData' )
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

--[[!
	\brief Creates a WFDigest instance that you can use for a streaming digest operation
		using the algorithm supported by the singleton WFDigestProvider instance.
	
	\param ... For the HMAC protocol ONLY, there are two additional parameters:
	<ul><li>(string) The name of a secondary digest algorithm (such as "SHA1")</li>
		<li> (string or KSData) A secret.</li></ul>

	\return (WFDigest) The new digest object
--]]

function WFDigestProvider.init( ... )
	-- DOXYGEN STUB
	
	-- here to eliminate lint warning
	return WFDigestProvider
end

--------------------------------------------------------------------------------

-- @mark WFCrypto namespace

--[[!
	\brief Allows additional input to the ongoing streaming digest
		operation.
	
	\param input (string or KSData) The additional input.
--]]

function WFCrypto:update( --[[string or KSData]] input )
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

--[[!
	\brief Computes the digest for the accumulated bytes in a
		streaming digest operation.
 
 	As a side-effect this WFDigest instance is reset and is ready
	to compute another hash.
	
	\return The binary digest data as a string, and as a hex string.
--]]
function WFCrypto:digest( )
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

local function checkDigestAlgorithms( ... )

	local function CheckIsWFDigestAlgorithm( arg )
		WFArguments.check( "WFDigestAlgorithm" )
	end

	local digestCount = select( '#', ... )
	for i = 1, digestCount do
		CheckIsWFDigestAlgorithm( select( i, ... ) )
	end

end

--------------------------------------------------------------------------------

--[[!
	\brief Calculates digests for the specified file on another thread.
	This observable will schedule the digests when the subscription happens.
	\param filePath (string) The absolute path to the file to digest.
	\param ... < strings > The names of the digests to calculate.
	\return Observable that sends a single value that contains the digest values
		as a tuple of strings.
]]--

function WFCrypto.asyncDigestFile( filePath, ... )

	WFArguments.check( "+string" )

	checkDigestAlgorithms( ... )

	return WFRxNativeObservable( WFCrypto.asyncDigestFileImp, filePath, ... )
		:unpack()

end

--------------------------------------------------------------------------------

function WFCrypto.asyncDigestData( data, ... )
	
	checkDigestAlgorithms( ... )
		
	return WFRxNativeObservable( WFCrypto.asyncDigestDataImp, data, ... )
		:unpack()
		
end

--------------------------------------------------------------------------------

--[[!
	\brief Asynchronously generate an RSA key pair.
	The key length is fixed at 4096 bits.
		
	\return Observable<string, string> An observable that yields the private and public key in PEM format.
]]
function WFCrypto.asyncGenerateKeyPair()
	return WFRxNativeObservable( WFCryptoImp.asyncGenerateKeyPairImp )
		:unpack()
end

--------------------------------------------------------------------------------

--[[!
	\brief Generate an RSA key pair.
	The key length is fixed at 4096 bits.
		
	\return private and public key as strings in PEM format.
]]
function WFCrypto.generateKeyPair()
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

--[[!
	\brief Generate a Certificate Signing Request.
	For peer-to-peer, the commonName should be the catalog hash, and the altName should be 
	the device name.
	
	
	\param commonName (string) to embed into the Certificate's subject
	\param altName (string) to embed into the Certificate's subject alternative name
	\param privateKey (string) the private key - PEM encoded - for the certificate
		
	\return The CSR as a string in PEM format.
]]
function WFCrypto.generateCSR( commonName, altName, privateKey )
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

--[[!
	\brief This class represents a X509 certificate.

	\param pem Certificate in PEM format
	
	It has the following readonly properties:
	<ul>
		<li>WFCertificate.commonName</li>
		<li>WFCertificate.altName</li>
		<li>WFCertificate.notBefore (in ISO-8601 format)</li>
		<li>WFCertificate.notAfter (in ISO-8601 format)</li>
	</ul>
]]

function WFCrypto.WFCertificate( pem )
	-- DOXYGEN STUB
end

WFCrypto.WFCertificate = {}

--------------------------------------------------------------------------------

--[[!
	\brief Verify a certificate	.
	
	\param caStore a WFCAStore containing the root certificates to verify against.
		
	\return true or false.
]]

function WFCrypto.WFCertificate:verify( caStore )
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

--[[!
	\brief Verify a certificate	and check it is currently valid.
	
	\param caStore a WFCAStore containing the root certificates to verify against.
		
	\return true or false.
]]

function WFCrypto.WFCertificate:verifyAndValidNow( caStore )
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

--[[!
	\brief Verify a certificate, check it is currently valid and the catalogHash matches.
	
	\param caStore a WFCAStore containing the root certificates to verify against.
	\param catalogHash The hash of the catalogId for verification.
		
	\return true or false.
]]
function WFCrypto.WFCertificate:verifyAndValidNowForCatalogHash( caStore, catalogHash )
	-- DOXYGEN STUB
end

--[[!
	\brief This class represents a store of root certificates.
]]

function WFCrypto.WFCAStore()
	-- DOXYGEN STUB
end

WFCrypto.WFCAStore = {
}

--[[!
	\brief Add a certificate to the CAStore
	
	\param certificate WFCertificate to add.
	\param 
]]
function WFCrypto.WFCAStore:add( certificate )
	-- DOXYGEN STUB
end

--------------------------------------------------------------------------------

-- Most of WFCrypto is implemented natively -- the previous value of the local WFCrypto
-- was mostly for documentation purposes.

for k, v in pairs( WFCryptoImp ) do

	if type( v ) == 'table' then

		-- special handling for digest implementations
		local init = v.init
		
		if init then
		
			v.digest = function( input, ... )
				local ctx = init( ... )
				assert( ctx, "Invalid params" )
				ctx:update( input )
				return ctx:digest( ctx )
			end

		end

	end

	WFCrypto[ k ] = v
end

--------------------------------------------------------------------------------

return WFCrypto