--[[----------------------------------------------------------------------------

WFFileUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2014 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- WF.core.os
local KSPathUtils = require 'KSPathUtils'
local WFFileUtilsImp = require 'WFFileUtilsImp'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.utils
local WFArguments = require 'WFArguments'


--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

--============================================================================--
--[[!
	\class WFFileUtils

	\brief This namespace provides common file system operations.
]]--

-- @mark WFFileUtils namespace

local WFFileUtils = {}

--------------------------------------------------------------------------------

-- @mark basic file operations

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether a folder or file exists.

	\param path (string) The path to confirm.
	
	\return (string or boolean) "directory" or "file" if the path references an
		existing folder or file, otherwise (boolean) false.
]]--

function WFFileUtils.exists( path )

	check( '+string' )

	return WFFileUtilsImp.exists( path )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports whether a folder is empty.

	\param path (string) The path to the folder.
	
	\return (boolean) True if the folder is empty, othewise false.
	
	NOTE: On Mac/iOS, hidden files are ignored. A folder will be reported as empty
	if it only contains hidden files such as .DS_STORE
	
]]--

function WFFileUtils.empty( path )

	check( '+string' )

	return WFFileUtilsImp.empty( path )

end

--------------------------------------------------------------------------------
--[[!
	\brief Copies a file.

	\param src (string) The absolute path of the source file.
	\param dest (string) The absolute path of the destination.
	
	If source and destination strings are identical, this function is a no-op.

	\warning (April 2014): Unlike the old KSFileUtils.copy, this function
		throws a Lua error if unable to copy.
	
	\return dest (string) returns the second argument; useful for chaining
]]--

function WFFileUtils.copy( src, dest )

	check( '+string', '+string' )
	
	if src == dest then
		return src
	end
	
	WFFileUtilsImp.copy( src, dest )
	
	return dest

end

--------------------------------------------------------------------------------
--[[!
	\brief Moves a file.
	
	\param src (string) The absolute path of the source file.
	\param dest (string) The absolute path of the destination.
	
	If source and destination strings are identical, this function is a no-op.

	\warning (April 2014): Unlike the old KSFileUtils.move, this function
		throws a Lua error if unable to move.
]]--

function WFFileUtils.move( src, dest )

	check( '+string', '+string' )
	
	if src == dest then
		return src
	end
	
	return WFFileUtilsImp.move( src, dest )

end

--------------------------------------------------------------------------------
--[[!
	\brief Immediately deletes a file, or a folder with all of its contents.
	
	\warning Be careful with this. In most cases calling \c moveToTrash would be better.

	\param path (string) The absolute path of a file or folder.
	\param optOnlyDeleteDirIfEmpty ( boolean ) If true, the folder will be
	only be deleted if it's empty.
	
	\warning (April 2014): Unlike the old KSFileUtils.delete, this function
		throws a Lua error if unable to delete.
]]--

function WFFileUtils.delete( path, optOnlyDeleteDirIfEmpty )
	
	check( '+string', '?boolean' )
	
	if WFFileUtilsImp.exists( path ) == 'directory' and optOnlyDeleteDirIfEmpty then
	
		if WFFileUtilsImp.empty( path ) then
			return WFFileUtilsImp.delete( path )
		else
			return --noOp
		end
		
	end
	
	return WFFileUtilsImp.delete( path )

end

--------------------------------------------------------------------------------
--[[!
	\brief Moves a file to the Trash or Recycle folder for its volume.
	
	\param filePath (string) The absolute path of the file.
	
	\warning In iOS, there is no such thing as a Trash folder, so this function
		performs an immediate deletion.
]]--

function WFFileUtils.moveToTrash( filePath )
	
	check( '+string' )

	-- On Win10, we can do a "soft-delete" which sends the file to Recycle Bin.

	if WFPlatformInfo.isWindows then
		return WFFileUtilsImp.delete( filePath, false )
			-- false indicates soft-delete
	end

	-- Some platforms don't have a trash API. If that's the case here,
	-- just do a delete.

	local moveItemToTrash = WFFileUtilsImp.moveItemToTrash
	if not moveItemToTrash then
		return WFFileUtils.delete( filePath )
	end

	-- If file doesn't exist, this is a harmless no-op.

	if not WFFileUtils.exists( filePath ) then
		return
	end

	moveItemToTrash( filePath )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new folder. This function will create any intermediate
		folders along the way that might be missing. (In other words, you do not
		have to call mkdir for any parent folders.)

	\param path (string) The absolute path of the new folder.

	\warning (April 2014): Unlike the old KSFileUtils.mkdir, this function
		throws a Lua error if unable to create the folder.
]]--

function WFFileUtils.mkdir( path )

	check( '+string' )
	
	return WFFileUtilsImp.mkdir( path )

end

--------------------------------------------------------------------------------
--[[!
	\brief Perform an operation on a temporary file and move the temporary into
		a location outside the application's sandbox if successful.
	
	This function is mostly meaningful on Windows RT. Otherwise, it simply calls
	the function with the desired final path, skipping the temporary file generation.
	
	\param path (string) The absolute path of the intended final file.
	\param func (function) The function to call. It takes a single argument, which
		is the temporary (sandbox-approved) path for the file.
	
	\return whatever results or error were returned by \c func
]]--

function WFFileUtils.doForAuthorizedPath( path, func )

	check( '+string|table', 'function' )

    if type( path ) == 'table' then

        local collisionBehavior = path.collisionBehavior
        path = KSPathUtils.child( path.folder, path.filename )

        if collisionBehavior == 'GenerateUniqueName' then
            path = WFFileUtils.ensureFileNameIsUnique( path )
        elseif collisionBehavior == 'FailIfExists' then
            assert( not WFFileUtils.exists( path ),
                    "File already exists and collisionBehavior is FailIfExists" )
        end

    end

    return func( path )

end

--------------------------------------------------------------------------------

function WFFileUtils.linkFile( sourcePath, destinationPath )

	check( '+string', '+string' )

	if WFFileUtilsImp.linkFile then
		return WFFileUtilsImp.linkFile( sourcePath, destinationPath )
	end

end

--------------------------------------------------------------------------------

-- @mark unique file name operations

--------------------------------------------------------------------------------
--[[!
	\brief Adapts a file name so that it will be unique within its containing folder.

	\param path (string) The absolute path to a file which may or may not exist.

	\return (string) If no file exists at this location, returns the same path
		as on input. If the file exists, the path is modified so that it is unique.
		If unable to construct a suitable path, will return nil.
	
	\warning There is a slight risk of a race condition with this API in that
		someone else could create a file with the same name in between this
		call and the operation that creates the file.
]]--

function WFFileUtils.ensureFileNameIsUnique( folderPath )

	check( '+string' )
	
	return WFFileUtilsImp.ensureFileNameIsUnique( folderPath )

end

--------------------------------------------------------------------------------
--[[!
	\brief Copies a file from one path to another. If a file already exists at the
		destination path, chooses a new name in the destination folder so as not to
		overwrite the existing file.

	If source and destination strings are identical, this function is a no-op.

	\param src (string) The absolute path to the file to copy.
	\param dest (string) The absolute path to the desired destination path.
	\param tryHardLink (boolean) Mac & iOS only: Optional flag which if true will
		cause this routine to try a hard link before making a copy.

	\return (string) The complete path to the new file.

	Throws an error if unable to complete this operation.
]]--

function WFFileUtils.copyToUniquelyNamedFile( src, dest, tryHardLink )

	check( '+string', '+string' )
	
	if src == dest then
		return src
	end
	
	return WFFileUtilsImp.copyToUniquelyNamedFile( src, dest, tryHardLink )

end

--------------------------------------------------------------------------------
--[[!
	\brief Moves a file from one path to another. If a file already exists at the
		destination path, chooses a new name at the destination so as not to
		overwrite the existing file.
	
	If source and destination strings are identical, this function is a no-op.

	\param source (string) The absolute path to the file to move.
	\param dest (string) The absolute path to the desired destination path.
	
	\return (string) The complete path to the new file.

	Throws an error if unable to complete this operation.
]]--

function WFFileUtils.moveToUniquelyNamedFile( src, dest )

	check( '+string', '+string' )
	
	if src == dest then
		return src
	end
	
	return WFFileUtilsImp.moveToUniquelyNamedFile( src, dest )

end

--------------------------------------------------------------------------------

-- @mark file attributes

--------------------------------------------------------------------------------
--[[!
	\brief Reports the size, creation data, and modification date of a file.

	\param path (string) The absolute path of the file.
 
	\return (table) A table with the following entries:
		<ul>
			<li>fileSize: the number of bytes </li>
			<li>fileCreationDate: the number of seconds since Jan 1, 2000.</li>
			<li>fileCreationDateTimezoneOffset</li>
			<li>fileModificationDate: the number of seconds since Jan 1, 2000.</li>
			<li>fileModificationDateTimezoneOffset</li>
		</ul>

	If the file does not exist or the OS calls to retrieve these values fails,
	returns an empty table.
]]--

function WFFileUtils.fileAttributes( path )

	check( '+string' )
	
	return WFFileUtilsImp.fileAttributes( path )

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the file-system date attributes of a file.

	\param path (string) The absolute path of the file.
	\param attributes (table) A table that can have entries for
		<b>fileCreationDate</b> and <b>fileModificationDate</b>.

	Throws an error if unable to complete this operation.
]]--

function WFFileUtils.setFileAttributes( path, attributes )

	check( '+string', 'table' )
	
	return WFFileUtilsImp.setFileAttributes( path, attributes )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports the space available and other statistics about the file system
		where a particular file is stored.

	\param path (string) The absolute path of the file.
 
	\return (table) A table with the following entries:
		<ul>
			<li>fileSystemSize: the total number of bytes of storage on
				this file system</li>
			<li>fileSystemFreeSize: the number of bytes on this file system
				that are currently unused</li>
			<li>fileSystemNumber: a system-dependent ID number that uniquely
				identifies this file system</li>
		</ul>

	Throws an error if unable to complete this operation.
]]--

function WFFileUtils.fileSystemAttributes( path )

	check( '+string' )
	
	return WFFileUtilsImp.fileSystemAttributes( path )

end

--------------------------------------------------------------------------------
--[[!
	\brief Reports how much space is available at a given path.
	
	This is a convenience wrapper for \c fileSystemAttributes.
	
	\param path (string) The absolute path to report on.
	
	\return (number) the number of bytes on this file system that are currently unused
	\return (number) the total number of bytes of storage on this file system
	
	Throws an error if unable to complete this operation.
]]--

function WFFileUtils.freeSpaceAtPath( path )

	check( '+string' )

	local attrs = WFFileUtils.fileSystemAttributes( path )

	local fileSystemFreeSize = attrs and attrs.fileSystemFreeSize
	local fileSystemSize = attrs and attrs.fileSystemSize

	return fileSystemFreeSize, fileSystemSize

end

--------------------------------------------------------------------------------
--[[!
	\brief Sets the "do not back up" attribute on a file.
	
	Only meaningful on iOS; no-op on all other platforms.

	\param path (string) The absolute path of the file.
--]]

function WFFileUtils.setDoNotBackUpFlag( path )

	check( '+string' )

	if WFFileUtilsImp.setDoNotBackUpFlag then
		return WFFileUtilsImp.setDoNotBackUpFlag( path )
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Removes the "do not back up" attribute on a file.

	Only meaningful on iOS; no-op on all other platforms.

	\param path (string) The absolute path of the file.
--]]

function WFFileUtils.removeDoNotBackUpFlag( path )

	check( '+string' )

	if WFFileUtilsImp.removeDoNotBackUpFlag then
		return WFFileUtilsImp.removeDoNotBackUpFlag( path )
	end

end


--------------------------------------------------------------------------------

-- @mark Alias resolution

--------------------------------------------------------------------------------
--[[!
	\brief Resolve any aliases included in the file path.
	
	Only meaningful on Mac OS X; no-op on all other platforms.

	\param path (string) The absolute path of the file.
	
	\return (string) The fully-resolved absolute path of the file.
		On platforms other than Mac OS X, simply returns the input path.
--]]

function WFFileUtils.resolveAliases( path )

	check( '+string' )

	if WFFileUtilsImp.resolveAliases then
		return WFFileUtilsImp.resolveAliases( path )
	else
		return path
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief Calculates the size of all files at a given path. If the path is a file, it
		returns the file size. If the path is a directory, it should return the size of 
		all files underneath that directory.
		
	If the path does not exist, it will return 0
	
	\param path (string) The path to calculate the size for.
	
	\return (number) The size of all files at the specified path
--]]

function WFFileUtils.calculatePathSize( path )
	
	check( '+string' )
	
	return WFFileUtilsImp.calculatePathSize( path )

end

--------------------------------------------------------------------------------

local function _macVolume( path )
	
	local _, _, name = string.find( path, "^(/Volumes/[^/]+)" )

	return name

end

--------------------------------------------------------------------------------

local function _winDrive( path )
	
	local _, _, name = string.find( path, "^([a-zA-Z]:)" )

	return name and string.upper( name )

end

--------------------------------------------------------------------------------

local function _uncPath( path )
	
	local _, _, name = string.find( path, "^(\\\\[^\\]+\\)" )

	return name

end

--------------------------------------------------------------------------------
--[[!
	\brief Gets the volume for a specified path.
	
	\param path (string) The path to extract the volume from.
	
	\return (string) The volume for the specified path.
--]]

function WFFileUtils.volumeForPath( path )

	check( '+string' )

	local volume
	
	volume = _macVolume( path )
	volume = volume or _winDrive( path )
	volume = volume or _uncPath( path )

	volume = volume or "/"
	
	return volume

end

--------------------------------------------------------------------------------
--[[!
	\brief Figures out if a specified path is on an offline volume.
	
	\param path (string) The path to test.
	
	\return (boolean) True if the volume is offline, false otherwise.
--]]

function WFFileUtils.isPathOnOfflineVolume( path )

	if not path then
		return false -- no path given, not an offline volume
	end

	local volumePath = WFFileUtils.volumeForPath( path )

	if not volumePath then
		return false -- malformed path, not an offline volume
	end

	return WFFileUtils.exists( volumePath ) ~= "directory"

end

--------------------------------------------------------------------------------
--[[!
	\brief MAC ONLY: Figures out the name of the volume represented by '/'
	
	\return (string) The name of the volume.
--]]

function WFFileUtils.getRootVolumeName()

	if WFFileUtilsImp.getRootVolumeName then
		return WFFileUtilsImp.getRootVolumeName()
	end
	
	return '/'

end

--------------------------------------------------------------------------------
--[[!
	\brief Get the GUID of the volume in the specified path.
	
	\param path (string) A file path.
	
	\return 
		(string|boolean) GUID of the volume. 
			Returns false if it cannot be obtained.
		(?string) Mac Only: Error message, if we can't get the volume guid.
		(?string) Mac Only: Disk info, if we can't get the volume info.
--]]

function WFFileUtils.getVolumeGuidForPath( path )

	check( '+string' )

	local mountPoint = WFFileUtils.volumeForPath( path )

	-- On Windows, we need to make sure the mount point ends in '\'. 
	-- volumeForPath strips this leaving something like "C:"
	
	if mountPoint and 
		mountPoint ~= '/' and 
		WFPlatformInfo.isWindows then
	
		mountPoint = mountPoint .. '\\'
		
	elseif WFPlatformInfo.isMac then
	
		if mountPoint == '/' then
	
			mountPoint = WFFileUtils.getRootVolumeName()
		
		end
	
	end

	if mountPoint and WFFileUtilsImp.getVolumeGuidForMountPoint then
		return WFFileUtilsImp.getVolumeGuidForMountPoint( mountPoint )
	end
	
	return mountPoint

end

--------------------------------------------------------------------------------
--[[!
	\brief Get the volume using the specified GUID.
	
	\param guid (string) The volume GUID.
	
	\return (string|boolean) The mount point of the volume. Returns false if it cannot be obtained.
		Example on Windows: C:\ or F:\
		Example on Mac: /Volumes/Foo or /
--]]

function WFFileUtils.getVolumeFromGuid( guid )
	
	check( '+string' )

	local mountPoint, isNetworkDrive
	
	if WFFileUtilsImp.getVolumeMountPointFromGuid then
		mountPoint, isNetworkDrive = WFFileUtilsImp.getVolumeMountPointFromGuid( guid )
	else
		mountPoint = guid
	end
	
	if mountPoint then
	
		if WFPlatformInfo.isMac then
	
			if mountPoint == WFFileUtils.getRootVolumeName() then
				return '/'
			elseif not isNetworkDrive then
				return '/Volumes/' .. mountPoint
			end
	
		elseif WFPlatformInfo.isWindows and isNetworkDrive then
		
			return WFFileUtils.volumeForPath( mountPoint )
		
		end
	
	end
	
	return mountPoint
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Get the volume name using the specified GUID.
	
	\param guid (string) The volume GUID.
	
	\return (string|boolean) The name of the volume. Returns false if it cannot be obtained.
		Example on Windows: "OSDisk" or "THUMB"
		Example on Mac: "Foo" or "Macintosh HD" 
--]]

function WFFileUtils.getVolumeNameFromGuid( guid )
	
	check( '+string' )

	local volumeName
	
	if WFFileUtilsImp.getVolumeNameFromGuid then
		volumeName = WFFileUtilsImp.getVolumeNameFromGuid( guid )
	elseif WFFileUtilsImp.getVolumeMountPointFromGuid then
		volumeName = WFFileUtilsImp.getVolumeMountPointFromGuid( guid )
	else
		volumeName = guid
	end
	
	return volumeName

end

--------------------------------------------------------------------------------
--[[!
	\brief Given the volume guid, check to see if the volume is online.
	
	\param guid (string) The volume GUID.
	
	\return (boolean) Returns true if the volume is online, false otherwise.
--]]

function WFFileUtils.isVolumeGuidOnline( guid )

	check( '+string' )

	local volumeName = WFFileUtils.getVolumeFromGuid( guid )
	
	return volumeName and WFFileUtils.exists( volumeName ) == "directory" or false

end

--------------------------------------------------------------------------------
--[[!
	\brief Remove the volume prefix from the beginning of an absolute path.
	
	\param path (string) A file path.
	
	\return (string) The file path without the volume prefix. For example:
	
	Example Mac: /Volumes/foo/bar/baz.jpg -> /bar/baz.jpg
	Example Mac: /git/foo/bar -> git/foo/bar
	Example Windows: c:\git\foo\bar -> git\foo\bar
	
--]]


function WFFileUtils.removeVolumeFromPath( path )

	check( 'string' )

	local volume = WFFileUtils.volumeForPath( path )
	return volume and string.sub( path, string.len( volume ) + 1, -1 )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Get the full path of a file given the path relative to the volume identifier,
	and the guid of the volume.
	
	\param relativePath (string) The path of the file, sans volume identifier.
	\param guid (string) The volume GUID.
	
	\return (string) The absolute path to the file.
--]]

function WFFileUtils.createFullPathFromRelativePathAndGuid( relativePath, guid )

	check( 'string', '+string' )

	local volumePath = WFFileUtils.getVolumeFromGuid( guid )
	
	return volumePath and KSPathUtils.child( volumePath, relativePath )

end

--------------------------------------------------------------------------------
--[[!
	\brief MAC ONLY: Check to see if a path points to a mounted DMG.
	
	\param path (string) A file path.
	
	\return (boolean) True if the path points to a mounted DMG. False otherwise.
	
--]]

function WFFileUtils.isMountedDMG( path )

	check( 'string' )

	if WFFileUtilsImp.isMountedDMG then
		local volume = WFFileUtils.volumeForPath( path )
		return volume and WFFileUtilsImp.isMountedDMG( volume )
	end
	
	return false

end

--------------------------------------------------------------------------------
--[[!
	\brief MAC ONLY: Check to see if a volume is read only.
	
	\param guid (string) The volume GUID.
	
	\return (boolean) True if the volume is read-only. This is the case for NTFS
		formatted drives.
	
--]]

function WFFileUtils.isVolumeReadOnly( guid )

	check( 'string' )

	if WFFileUtilsImp.isVolumeReadOnly then
		return WFFileUtilsImp.isVolumeReadOnly( guid )
	end
	
	return false

end

--------------------------------------------------------------------------------
--[[!
	\brief Check to see if a directory is read only.
	
	\param path (string) The path to the directory.
	
	\return (boolean) True if the directory is read-only. False otherwise.
	
--]]

function WFFileUtils.isDirectoryReadOnly( path )

	check( 'string' )

	local isReadOnly = false 

	local guid = WFFileUtils.getVolumeGuidForPath( path )

	if guid and WFFileUtilsImp.isVolumeReadOnly then
		isReadOnly = WFFileUtilsImp.isVolumeReadOnly( guid )
	end
	
	if not isReadOnly and WFFileUtilsImp.isDirectoryReadOnly then
		isReadOnly = WFFileUtilsImp.isDirectoryReadOnly( path )
	end
	
	return isReadOnly

end

--------------------------------------------------------------------------------
--[[!
	\brief WIN ONLY: Check to see if a path points to a CDROM
	
	\param path (string) A file path.
	
	\return (boolean) True if the path points to a CDROM. False otherwise.
	
--]]

function WFFileUtils.isCDROM( path )

	check( 'string' )

	if WFFileUtilsImp.isCDROM then
		local volume = WFFileUtils.volumeForPath( path )
		return volume and WFFileUtilsImp.isCDROM( volume )
	end
	
	return false

end

--------------------------------------------------------------------------------

function WFFileUtils.resolveAliasesForPath( path )

	check( 'string' )

	if WFFileUtilsImp.resolveAliasesForPath then
		path = WFFileUtilsImp.resolveAliasesForPath( path )
	end
	
	return path

end

--------------------------------------------------------------------------------
--[[!
	\brief Get the contents of a file. This works with Unicode paths
	and files, which is not the case for io.open.
	
	\param path (string) A file path.
	
	\return (string) The file contents.
--]]

function WFFileUtils.getFileContents( path )

	check( 'string' )

	local contents = nil

	if WFFileUtilsImp.getFileContents then
		contents = WFFileUtilsImp.getFileContents( path )
	end
	
	return contents

end

--------------------------------------------------------------------------------
--[[!
	\brief Write contents to a file. This works with unicode file
		paths and content.
	
	\param path (string) A file path.
	\param contents (string) The contents to write to the file.
	
	\return (void)
--]]

function WFFileUtils.saveToFile( path, contents )

	check( 'string', 'string' )

	if WFFileUtilsImp.saveToFile then
		WFFileUtilsImp.saveToFile( path, contents )
	end

end

--------------------------------------------------------------------------------
--[[!
	\brief check to see if the file is readable.	
	\param path (string) A file path.
	
	\return (boolean) True if the file is readable. False otherwise.
--]]

function WFFileUtils.isFileReadable( filePath )
	
	check( 'string' )
	
	local fh = io.open( filePath, "rb" )
	if( fh ) then
		fh:close()
	end
	
	return fh ~= nil
	
end

------------------------------------------------------------------------------------

return WFFileUtils
