--[[----------------------------------------------------------------------------

WFMockHttpRequest.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local Rx = require 'Rx'

local WFMockHttpRequest = {
	
	_method = false,
	_url = false,
	_urlTransformationFn = false,
	_expectedRequestData = false,
	_requestValidationFn = false,
	_actualRequestDataTransformationFn = false,
	_requestDataTransformationFn = false,
	_expectedRequestHeaders = false,
	_requestOptions = false,
	_responseData = false,
	_responseCode = false,
	_responseHeaders = false,
	_responseTransformationFn = false,
	_traceback = false,

}
	
function WFMockHttpRequest:__init()
	self._traceback = debug.traceback( "", 4 )
end

function WFMockHttpRequest:get( url )
	self._url = url
	self._method = 'GET'
	return self
end

function WFMockHttpRequest:post( url )
	self._url = url
	self._method = 'POST'
	return self
end

function WFMockHttpRequest:put( url )
	self._url = url
	self._method = 'PUT'
	return self
end

function WFMockHttpRequest:expectRequestData( data )
	self._expectedRequestData = data
	return self
end

function WFMockHttpRequest:expectRequestDataAfterTransformation( data, transformFn )
	-- SPAM [matthewj 2018-04-27]: the transformation function is run on the data 
	-- coming in from the actual put/post call. The output of this will then be compared
	-- to the expected request data
	self._expectedRequestData = data
	self._actualRequestDataTransformationFn = transformFn
	return self
end

function WFMockHttpRequest:requestUrlTransformation( urlTransformFn )
	self._urlTransformationFn = urlTransformFn
	return self
end

function WFMockHttpRequest:requestDataTransformation( transformFn )
	-- transformFn can transform both expected and actual data
	self._requestDataTransformationFn = transformFn
	return self
end

function WFMockHttpRequest:validateRequestData( validationFn )
	self._requestValidationFn = validationFn
	return self
end

function WFMockHttpRequest:withRequestHeaders( requestHeaders )
	self._expectedRequestHeaders = requestHeaders
	return self
end

function WFMockHttpRequest:withRequestOptions( options )
	self._requestOptions = options
	return self
end

function WFMockHttpRequest:respond( code, data )
	self._responseCode = code
	self._responseData = data
	return self
end

function WFMockHttpRequest:withResponseHeaders( responseHeaders )
	self._responseHeaders = responseHeaders
	return self
end

function WFMockHttpRequest:withResponseTransformationFn( transformFn )
	self._responseTransformationFn = transformFn
	return self
end

function WFMockHttpRequest:build()
	
	local request = {}
	if self._url then
		request.url = self._url
	end
	
	if self._urlTransformationFn then
		request.urlFilter = self._urlTransformationFn
	end
	
	if self._expectedRequestData then
		request.data = self._expectedRequestData
	end
	
	if self._requestValidationFn then
		request.validationFn = self._requestValidationFn
	end
	
	if self._actualRequestDataTransformationFn then
		request.dataFilter = self._actualRequestDataTransformationFn	
	end

	if self._requestDataTransformationFn then
		request.dataTransformation = self._requestDataTransformationFn
	end

	if self._requestOptions then
		request.options = self._requestOptions
	end
	
	if self._expectedRequestHeaders then
		request.options = request.options or {}
		request.options.requestHeaders = self._expectedRequestHeaders
	end
	
	if self._responseData then
		request.response = self._responseData
	end
	
	if self._responseHeaders then
		request.headers = self._responseHeaders
	end

	if self._responseCode then
		
		request.headers = request.headers or {}
		request.headers.status_code = self._responseCode or '200'
		if self._responseCode >= '400' then
			local fullResponseHeaders = self._responseHeaders or {}
			fullResponseHeaders.status_code = self._responseCode
			request.fail = {
				self._responseData or {},
				fullResponseHeaders,
			}
		end
	end

	if self._responseTransformationFn then
		request.responseTransformationFn = self._responseTransformationFn
	end

	request.traceback = self._traceback

	return request
end

return Rx.Class 'WFMockHttpRequest'( WFMockHttpRequest )