--[[----------------------------------------------------------------------------

WFNativeVideoContext.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2018 Adobe
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law. Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.rxlua
local WFRxNativeObservable = require 'WFRxNativeObservable'

	-- WF.core.utils
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.NativeVideo
local WFNativeVideo = require 'WFNativeVideo'
local WFNativeVideoContextImp = require 'WFNativeVideoContextImp'


--------------------------------------------------------------------------------

assert( WFPlatformInfo.useNativeVideo, "NativeVideo integration not enabled" )

local check = assert( WFArguments.check )

	-- WF.core.logging
-- local WFLogger = require 'WFLogger'
-- local logger = WFLogger 'WFNativeVideo'
	--:setLevel( 'DEBUG' )

--------------------------------------------------------------------------------

local _frameTimeForRendering = 0
--------------------------------------------------------------------------------

local function assertIfClosed( self )

	if WFCloseSupport.hasBeenOrWillBeClosed( self ) then
		error( "WFNativeVideoContext can not be used after it has been closed", 3 )
	end

end

--------------------------------------------------------------------------------

local function waitForClipInit( self )

	return WFNativeVideo.waitForSession()
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFNativeVideoContextImp.initClip, self._imp )
		end)
		:map( function()
			return true
		end)

end

--------------------------------------------------------------------------------

local function combineInfoAndMetadata( info, xmpResult )

	if info and xmpResult and xmpResult.success then
		info.metadata = xmpResult.xmp
	end
	
	return info

end

--------------------------------------------------------------------------------

local function getDimensionsToRenderFromLongSide( longSide, actualWidth, actualHeight )

	local actualLongSide = math.max( actualWidth, actualHeight )
	local widthToRender, heightToRender
	if actualLongSide == actualWidth then
		widthToRender = math.min( longSide, actualLongSide )
		heightToRender = math.floor( ( actualHeight * widthToRender ) / actualWidth )
	else
		heightToRender = math.min( longSide, actualLongSide )
		widthToRender = math.floor( ( actualWidth * heightToRender ) / actualHeight )
	end

	return widthToRender, heightToRender
end

--============================================================================--

-- @mark WFNativeVideoContext class

--[[!
	\brief WFNativeVideoContext provides single entry point to all
	native video operations. 

	NativeVideo initialization and usage of session and clip,
	caching of clip etc., are handled by this class, so that
	clients don't need to worry about all the intricacies. 
]]--

local WFNativeVideoContext = {

	_imp = false,
		-- ( WFNativeVideoContextImp )

	_filePath = false,
		-- (string) original file path

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Creates a new WFNativeVideoContext object for a single video file.

	\param filePath (string) path to video file

	
]]--

function WFNativeVideoContext:__init( filePath )

	check( 'WFNativeVideoContext', '+string' )

	self._imp = WFNativeVideoContextImp.make( filePath )
	
	self._filePath = filePath

end


--------------------------------------------------------------------------------

function WFNativeVideoContext:__tostring()

	return string.format( "WFNativeVideoContext( %s )", self._filePath )

end

--------------------------------------------------------------------------------
--[[!
	\brief Read metadata from this file.

]]--

function WFNativeVideoContext:readMetadata()

	check( 'WFNativeVideoContext' )

	assertIfClosed( self )
	
	return waitForClipInit( self )
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFNativeVideoContextImp.getInfo, self._imp )
				:combineLatest( combineInfoAndMetadata, 
					WFRxNativeObservable( WFNativeVideoContextImp.getXMP, self._imp ) )
		end )
end

--------------------------------------------------------------------------------
--[[!
	\brief Get the frame corresponding to the frame time

]]--

function WFNativeVideoContext:getFrame( frameTime, width, height, quality )

	check( 'WFNativeVideoContext', 'number', 'number', 'number', 'number' )

	assertIfClosed( self )
	
	return waitForClipInit( self )
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFNativeVideoContextImp.getFrame, self._imp,
				frameTime, width, height, quality )
		end)

end

--------------------------------------------------------------------------------
--[[!
	\brief Generate a preview jpeg for the video

]]--

function WFNativeVideoContext:render( longSide, quality )

	check( 'WFNativeVideoContext', 'number', 'number' )

	assertIfClosed( self )
	
	return waitForClipInit( self )
		:flatMapLatest( function( )
			return WFRxNativeObservable( WFNativeVideoContextImp.getInfo, self._imp )
				:map( function( info )
					if info then
						return info.width, info.height
					else
						return nil
					end
				end)
				:flatMapLatest( function( width, height )

					if type( width ) ~= "number" or type( height ) ~= "number" then

						return Rx.Observable.throw( "WFNativeVideoContextImp.getInfo didn't return valid dimensions" )

					end

					local widthToRender, heightToRender = 
						getDimensionsToRenderFromLongSide( longSide, width, height )

					return WFRxNativeObservable( WFNativeVideoContextImp.getFrame, self._imp,
							_frameTimeForRendering, widthToRender, heightToRender, quality )
				end)

		end)

end

--------------------------------------------------------------------------------
--[[!
	\brief Closes the NativeVideo context and the
	associated clip. 

]]--

function WFNativeVideoContext:close()

	check( 'WFNativeVideoContext' )

	if self._imp then
		self._imp:close()
	end

	self._imp = nil

end

--------------------------------------------------------------------------------

return Rx.Class 'WFNativeVideoContext'( WFNativeVideoContext )
