--[[----------------------------------------------------------------------------

WFOzDocUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2013 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remains the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and are protected by all applicable
intellectual property laws, including trade secret and copyright laws.
Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local WFArguments = require 'WFArguments'
local KSTableUtils = require 'KSTableUtils'

local WFCatUrlUtils = require 'WFCatUrlUtils'
local WFCatFileSpec = require 'WFCatFileSpec'

local WFRxThrottledBatchedBulkDocValue = require 'WFRxThrottledBatchedBulkDocValue'

--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local resolveLinksToAbsoluteUrls = assert( WFCatUrlUtils.resolveLinksToAbsoluteUrls )

--============================================================================--

local WFOzDocUtils = {}

--------------------------------------------------------------------------------

function WFOzDocUtils.escapeUnderscoreKeyword( keyword )

	-- Ensure we don't build any keys that start with reserved
	-- character _ -- if a keyword starts with that, we replace
	-- it with the UTF "FULLWIDTH LOW LINE" U+FF3F (UTF-8 EF BC BF)

	if keyword:sub( 1, 1 ) == "_" then
		keyword = "\239\188\191" .. keyword:sub( 2 )
	end

	return keyword

end

--------------------------------------------------------------------------------

local function stripPayloadLinkNoise( noisyTable )

	if noisyTable then
		noisyTable.links = nil
		return noisyTable
	end

end

--------------------------------------------------------------------------------

function WFOzDocUtils.remapBasicOzDocument( ozDocument, baseUrl )
	
	-- Rearrange document to fit WF conventions instead of Oz conventions.
	-- This is a destructive transformation, but that's OK because the original
	-- document is no longer used after this call.

	-- Invert the relationship between payload and server-provided metadata.

	local result = ozDocument.payload or {}

	result._serverProvidedMetadata = ozDocument
	ozDocument.payload = nil

	result._id = ozDocument.id
	ozDocument.id = nil

	result.type = ozDocument.type
	ozDocument.type = nil

	result.subtype = ozDocument.subtype
	ozDocument.subtype = nil
	
	result.original = ozDocument.original
	ozDocument.original = nil

	if result.type == 'album_asset' then

		result.album = stripPayloadLinkNoise( ozDocument.album )
		ozDocument.album = nil

		result.asset = stripPayloadLinkNoise( ozDocument.asset )
		ozDocument.asset = nil

	end
	
	if result.type == 'comment' then
	
		result.space = stripPayloadLinkNoise( ozDocument.space )
		ozDocument.space = nil
		
		result.asset = stripPayloadLinkNoise( ozDocument.asset )
		ozDocument.asset = nil
		
	end
	
	if result.type == 'favorite' then
	
		result.space = stripPayloadLinkNoise( ozDocument.space )
		ozDocument.space = nil
		
		result.asset = stripPayloadLinkNoise( ozDocument.asset )
		ozDocument.asset = nil
		
	end
	
	if result.type == 'space_album' then
	
		result.space = stripPayloadLinkNoise( ozDocument.space )
		ozDocument.space = nil
		
		result.album = stripPayloadLinkNoise( ozDocument.album )
		ozDocument.album = nil
		
	end
	
	if result.type == 'stack_asset' then
	
		result.stack = stripPayloadLinkNoise( ozDocument.stack )
		ozDocument.stack = nil

		result.asset = stripPayloadLinkNoise( ozDocument.asset )
		ozDocument.asset = nil

	end

	if result.type == 'spaceinvite' then
	
		result.space = stripPayloadLinkNoise( ozDocument.space )
		ozDocument.space = nil

		result.user = stripPayloadLinkNoise( ozDocument.user )
		ozDocument.user = nil

		result.space_user = stripPayloadLinkNoise( ozDocument.space_user )
		ozDocument.space_user = nil

	end

	if result.type == 'space_user' then
	
		result.space = stripPayloadLinkNoise( ozDocument.space )
		ozDocument.space = nil

		result.user = stripPayloadLinkNoise( ozDocument.user )
		ozDocument.user = nil

	end

	local spm = result._serverProvidedMetadata
	
	-- Oz might send us relative links to the current base url of this request
	-- so we respect the current base url and create absolute url's in the format
	-- /v2/catalogs/123...
	
	if baseUrl and spm and spm.links then
	
		resolveLinksToAbsoluteUrls( baseUrl, spm.links )
		
	end
	
	-- error docs may have additional links in a separate section. if it's present
	-- convert these too
	if baseUrl and spm and spm.source and spm.source.links then
		
		resolveLinksToAbsoluteUrls( baseUrl, spm.source.links )
	
	end
	
	if baseUrl and spm and spm.matching_asset then
		if spm.matching_asset.href then
			spm.matching_asset.href = WFCatUrlUtils.resolveToAbsoluteUrl( baseUrl, spm.matching_asset.href )
		end
	end
	
	-- add the baseUrl to the server provided metadata so that we can fix up 
	-- any unresolved links at a future point, if necessary
	spm.base = baseUrl
	
	-- Docstore requires the revision id list to be non-empty,
	-- if present.
	result._revIds = ozDocument.revision_ids and #ozDocument.revision_ids > 0 
		and ozDocument.revision_ids
		or nil
	
	return result
	
end

--------------------------------------------------------------------------------

function WFOzDocUtils.getRevisionId( docValue )
	
	local revIds = docValue.revIds
	
	if type( revIds ) ~= 'table' or #revIds == 0 then 
		return nil
	else
		return revIds[ #revIds ] 
	end

end

--------------------------------------------------------------------------------

function WFOzDocUtils.isAssetDocFullySyncedWithOz( docValue )
	
	check( 'table' )
	
	local labels = docValue._labels or {}
	
	local hasBinaryOnOz = WFCatFileSpec.getOriginalLinkFromDocValue( docValue )
			or WFCatFileSpec.getProxyLinkFromDocValue( docValue )
	
	return not not ( labels[ 'oz' ] and hasBinaryOnOz )

end

--------------------------------------------------------------------------------

-- The following function resolves to `true` if all of the assets
-- are fully synced with Oz, and there are no pending local changes
-- that have yet to sync.

function WFOzDocUtils.areAssetsFullySyncedWithOz( wfCatalog, assetIds )
	
	check( 'WFCatalog', 'table' )
	
	return WFRxThrottledBatchedBulkDocValue( "WFOzDocUtils_areAssetsFullySyncedWithOz", wfCatalog:docStore(), assetIds, 100, 10 )
			
			:scan( function( docValues, batchOfDocValues )
	
				docValues = KSTableUtils.appendArray( docValues, batchOfDocValues )
				return docValues
				
			end, {} )
			
			:takeLast( 1 )
			
			:map( function( docValues )
			
				for _, docValue in ipairs( docValues ) do
				
					if not WFOzDocUtils.isAssetDocFullySyncedWithOz( docValue ) then
					
						return false
						
					end
				
				end
				
				return true
			
			end )

end

--------------------------------------------------------------------------------

return WFOzDocUtils
