--[[----------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2016 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

----------------------------------------------------------------------------]]--

local Rx = require 'Rx'

local WFArguments = require 'WFArguments'

	-- WF.core.os
local WFPlatformInfo = require 'WFPlatformInfo'
local WFURI = require "WFURI"

	-- WF.core.logging
local WFLogger = require 'WFLogger'

local KSStringUtils = require "KSStringUtils"
local KSTableUtils = require "KSTableUtils"
local KSJsonEncoder = require "KSJsonEncoder"

--------------------------------------------------------------------------------

local logger = WFLogger 'WFOzUserDeviceManagement'

--============================================================================--

--[[
This namespace contains functions to manage the management of iOS push service device tokens
]]--

-- @mark WFOzUserDeviceManagement namespace

local WFOzUserDeviceManagement = {}

--------------------------------------------------------------------------------

function WFOzUserDeviceManagement.deleteDeviceToken( wfUser, compareTokenUrl )

	local appId = WFPlatformInfo.getAppIdentifier()
	local apsEnv = 'production'
	
	if appId:find( '-develop' .. '$' ) then
		apsEnv = 'development'
	end

	local o = wfUser:userInfoCache():docValue( wfUser:userId() )

	return o:flatMap( function ( docValue )
					
		local deleteTokenUrl = KSTableUtils.derefOpt( docValue, '_localOnly', 'device_token' )

		if not deleteTokenUrl then
			-- No token location
			return Rx.Observable.just( wfUser )
		end

		if compareTokenUrl and compareTokenUrl == deleteTokenUrl then
			return Rx.Observable.just( wfUser )
		end
		
		local queryString = WFURI.buildQueryString {
			app_id = appId,
			aps_env = apsEnv,
		}
		
		if queryString ~= "" and string.sub( deleteTokenUrl, -1, -1 ) ~= '?' then
			deleteTokenUrl = deleteTokenUrl .. "?"
		end
		
		if queryString then
			deleteTokenUrl = deleteTokenUrl .. queryString
		end
				
		local deleteArgs = {
			requestMethod = 'DELETE',
			returnHttpErrorsAsData = true,
			requestHeaders = {
				{ header = 'Content-Type', value = 'application/json' },
			},
		}
		
		return wfUser:get( deleteTokenUrl, deleteArgs )
			:flatMap( function( message, headers )
			
				local status_code = KSTableUtils.derefOpt( headers, 'status_code' )
				-- If the status_code returned by Oz is '204' or '404' we interpret it as a success.
				if not ( status_code == "204" or status_code == "404" ) then
													
					logger:errorf( "deleteDeviceToken: message = %s, headers = %s",
								   message or '(nil)', headers or '(nil)' )
					
					local responseString = 'deleteDeviceToken: error deleting token from Oz status_code: ' .. (status_code or '(nil)')
					return Rx.Observable.throw( responseString )

				else

					-- Delete the token location locally.

					return wfUser:userInfoCache():applyOverlay
						{
							type = 'user',
							_id = wfUser:userId(),
							_localOnly = {
								device_token = { _delete = true },
							}
						}
						:flatMapLatest( function()
							return Rx.Observable.just( wfUser )
						end )
					
				end
			end )	
			
		end )
end

--------------------------------------------------------------------------------

WFOzUserDeviceManagement.registerDeviceToken = function( wfUser, token )

	WFArguments.check( 'table', 'KSData' )
	
	local appId = WFPlatformInfo.getAppIdentifier()
	local apsEnv = 'production'
	
	if appId:find( '-develop' .. '$' ) then
		apsEnv = 'development'
	end
	
	local postArgs = {
		requestMethod = 'POST',
		requestHeaders = {
			{ header = 'Content-Type', value = 'application/json' },
		},
	}

	-- Convert the token to hex
	
	local hexToken = KSStringUtils.convertToHex( token )
	
	local postParams = KSJsonEncoder.encodeOzUploadJson
		{
			device_token = hexToken,
			app_id = appId,
			aps_env = apsEnv,
		}
				
	return wfUser
		
		:take( 1 )
		
		:where( function( user )
			return user ~= nil
		end )
		
		:flatMapLatest( function( user )
		
			local requestUrl = string.format( '/v2/accounts/%s/devices',
									  user:userId() )
									  
			return user:put( requestUrl, postParams, postArgs )
		
		end )
		
		:flatMapLatest( function( data, headers )
			
			if headers.status_code ~= "201" or not headers.Location then

				local responseString = tostring( data )
				return Rx.Observable.throw( responseString )

			else
				-- Delete any old token first, if it's NOT the same as the
				-- one we just registered. This prevents us deleting the token
				-- in the case that we try to register the same token twice.
				
				return wfUser
					:take( 1 )
		
					:where( function( user )
						return user ~= nil
					end )
					
					:flatMapLatest( function( user )
				
						return WFOzUserDeviceManagement.deleteDeviceToken( user, headers.Location )
					
							:flatMapLatest( function()
						
								-- Then save the location of new token
								
								return user:userInfoCache():applyOverlay
									{
										type = 'user',
										_id = user:userId(),
										_localOnly = {
											device_token = headers.Location,
										},
									}
									
							end )
							
					end )
					
			end

		end )



end

--------------------------------------------------------------------------------

return WFOzUserDeviceManagement
