--[[----------------------------------------------------------------------------

WFRtmfpPartnerSession.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'

	-- WF.core
local WFConfiguration = require 'WFConfiguration'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

	-- WF.core.lua
local KSData = require 'KSData'

	-- WF.core.rxlua
local WFRxRunUntilDone = require 'WFRxRunUntilDone'
local WFRxScheduler = require 'WFRxScheduler'

	-- WF.core.utils
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'

	-- WF.rtmfp.imp
local WFRtmfpErrorCodes = require 'WFRtmfpErrorCodes'

	-- WF.web.json
local KSJsonDecoder = require 'KSJsonDecoder'
local KSJsonEncoder = require 'KSJsonEncoder'

local Observable = Rx.Observable
local ReplaySubject = Rx.ReplaySubject

--==============================================================================

local rtmfpPartnerSessionLog =  WFLogger 'WFRtmfpPartnerSession'
--rtmfpPartnerSessionLog:enable( 'print' )

local check = assert( WFArguments.check )
KSData.registerArgumentChecker()

local DISABLE_SESSION_TIMEOUT = WFConfiguration.RTMFP.DisableSessionTimeout
local BIG_TIMEOUT = 9999999
local runUntilDone = WFRxRunUntilDone

--------------------------------------------------------------------------------

-- FIXME [jandrew 1/22/2015]: This should be a real secret.
local theSecret = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

--------------------------------------------------------------------------------

-- Handshake Requests/Responses

local hiCommand = 'HI'
local helloCommand = 'HELLO'
local howdyCommand = 'HOWDY'
local goodbyeCommand = 'GOODBYE'
local pingCommand = 'PING'

local _commandMetadata = { _sessionPartnerCommand = true, no_echo = true }

local commandMetadata = function() return KSTableUtils.shallowcopy( _commandMetadata ) end

--------------------------------------------------------------------------------

local hiMessage = KSJsonEncoder.encodeOzUploadJson { command = hiCommand, secret = theSecret }

local goodbyeMessage = KSJsonEncoder.encodeOzUploadJson { command = goodbyeCommand, secret = theSecret } 

local pingMessage = KSJsonEncoder.encodeOzUploadJson { command = pingCommand } 

local oneSecond = 1000
local oneMinute = 60 * oneSecond
local oneHour = 60 * oneMinute

local partnerSessionModifiers = {

	maxRTXTime = oneSecond * 5,
	maxIdleTime = oneMinute,
	nearCloseTime = oneSecond * 5,
	sessionKeepaliveTime = DISABLE_SESSION_TIMEOUT and BIG_TIMEOUT or oneHour,
	
}

--============================================================================--

--[[!
	\class WFRtmfpPartnerSession
	\brief This namespace provides a layer on top of WF RTMFP for communicating
	with another WF RTMFP node. WFRtmfpPartnerSession objects have a one-to-one
	relationship with other WFRtmfpPartnerSession objects.
]]--

-- @mark WFRtmfpPartnerSession namespace

local WFRtmfpPartnerSession = {

	_rtmfpClient = false,
		-- ( WFRtmfpClient ) RTMFP object
		
	_commandFlowStream = false,
		--- ( ReplaySubject< WFRtmfpFlow > ) 
		
	_highPriorityFlowStream = false,
		-- ( ReplaySubject< WFRtmfpFlow > ) 
		
	_bigDataFlowStream = false,
		-- ( ReplaySubject< WFRtmfpFlow > ) 

	_smallDataFlowStream = false,
		-- ( ReplaySubject< WFRtmfpFlow > ) 

	_initializedSession = false,
		-- ( boolean ) True if we initialized session opening. False if a partner did.
		
	_initSubscription = false,
		-- ( Rx observer ) Sets _sessionNeedsReinitialization and _sessionIsReinitializing status.

	_sessionNeedsReinitialization = false,
		-- ( boolean ) Indicates if the session needs to be reinitialized
	
	_sessionIsReinitializing = false,
		-- ( boolean ) Indicates if the session is in the process of reinitializing
		
	_incomingStream = false,
		-- ( ReplaySubject< what, ... > ) A stream for receiving data from a session partner

	_incomingStreamSubscription = false,
		-- ( Rx Disposable ) A subscription to the _incomingStream to handle special messages.

	_partnerAddress = false,
		-- ( string ) The network address of the partner
		
	_partnerPort = false,
		-- ( int ) The network port of the partner

	_sessionInit = false,
		-- ( BehaviorSubject ) Resolves once the session has opened or reinitialized.

	_isShutdown = false,
		-- (boolean) Set when shut down

	_flowStatusSubscriptions = false,
		-- (table ) A table of flow status subscriptions.
		-- key: Rx disposable
		-- value: true

	_pingPartnerSubscription = false,
		-- ( Rx Disposable ) A subscription to the ping-partner heartbeat.
		
	_disposeOnClose = false,
		-- ( table ) A table of disposables to dispose when closed.
		-- key: Rx disposable
		-- value: true

	WFCloseSupport.addCloseSupportMethods,
	
}

--------------------------------------------------------------------------------

local function isGoodbyeMessage( rtmfpRequest )

	local metadata = rtmfpRequest:getMetadata()
	
	if metadata._sessionPartnerCommand then
	
		local data = KSJsonDecoder.decodeWebSafeJson( rtmfpRequest:getData() )
				
		if data and data.command == goodbyeCommand then

			return true
	
		end

	end

	return false

end

--------------------------------------------------------------------------------

function WFRtmfpPartnerSession:__init( rtmfpClient )
	
	check( 'WFRtmfpPartnerSession', 'WFRtmfpClient' )
	
	rtmfpPartnerSessionLog:debugf( "__init( rtmfpClient ID = %s )", rtmfpClient:getId() )

	self._rtmfpClient = rtmfpClient

	self._flowStatusSubscriptions = setmetatable( {}, { __mode = 'k' } )

	self._commandFlowStream = ReplaySubject( 1 )
	self._highPriorityFlowStream = ReplaySubject( 1 )
	self._bigDataFlowStream = ReplaySubject( 1 )
	self._smallDataFlowStream = ReplaySubject( 1 )

	-- Set up event stream for managing incoming message
	
	self._incomingStream = ReplaySubject( 1 )

	local disposeFn = function()
	
		if self._incomingStreamSubscription then
			self._incomingStreamSubscription:dispose()
			self._incomingStreamSubscription = nil
		end
	
	end

	self._incomingStreamSubscription = self._incomingStream:subscribe(
		function( rtmfpRequest)
			if not self._initializedSession then
				if isGoodbyeMessage( rtmfpRequest ) then
					-- The RTMFP Partner has ended the session
					if not DISABLE_SESSION_TIMEOUT then
						self:close()
					end
				end

			end
		end,
		disposeFn,
		disposeFn )

end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve the Command Flow.
	
	\return An Observable that fires once with the WF RTMFP Flow.
	
--]]

function WFRtmfpPartnerSession:getCommandFlowOnce()

	check( 'WFRtmfpPartnerSession' )
	
	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.throw( "The WFRtmfpPartnerSession has already been shutdown." )
	end
	
	return self._commandFlowStream
	
		:filter( function( maybeFlow )
			return maybeFlow
		end )
	
		:take( 1 )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve the Small Data Flow.
	
	\return An Observable that fires once with the WF RTMFP Flow.
	
--]]

function WFRtmfpPartnerSession:getDefaultDataFlowOnce()

	check( 'WFRtmfpPartnerSession' )
	
	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.throw( "The WFRtmfpPartnerSession has already been shutdown." )
	end
	
	return self._smallDataFlowStream
	
		:filter( function( maybeFlow )
			return maybeFlow
		end )
		
		:take( 1 )
			
end

--------------------------------------------------------------------------------

--[[!

	\brief Ping the other partner.
	
	\return void
	
--]]

function WFRtmfpPartnerSession:pingPartner()

	check( 'WFRtmfpPartnerSession' )
	
	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.throw( "The WFRtmfpPartnerSession has already been shutdown." )
	end
	
	return self._smallDataFlowStream
	
		:filter( function( maybeFlow )
			return maybeFlow
		end )
		
		:take( 1 )
	
		:flatMap( function( flow )
			return flow:send( pingMessage, commandMetadata() )
		end )
		
end

--------------------------------------------------------------------------------

--[[!

	\brief Setup a "heart beat" monitor between the RTMFP partners
	
	\param maybeOnErrorCallbackFn (optional function) This function is called on
		every failed beat.
		
	\param maybeOnSuccessCallbackFn (optional function) This function is called on
		every successful beat.
		
	\return void
	
--]]

function WFRtmfpPartnerSession:setupHeartBeatMonitor( maybeOnErrorCallbackFn, maybeOnSuccessCallbackFn )

	check( 'WFRtmfpPartnerSession', '?function', '?function' )

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return
	end

	if self._pingPartnerSubscription then
		self._pingPartnerSubscription:dispose()
		self._pingPartnerSubscription = nil
	end

	local noopFn = function() --[[noop]] end
	
	maybeOnSuccessCallbackFn = maybeOnSuccessCallbackFn or noopFn

	-- SPAM [jandrew 2017-06-26]: The following is a blind attempt to fix the
	-- 'daemon keeps crashing' error. It appears that when a computer
	-- goes to sleep, the connection between the daemon and Nimbus process
	-- is for some reason lost, which forces the daemon to restart. Hopefully
	-- this explicit pinging every second will prevent that.

	self._pingPartnerSubscription = Observable.timer( 2, 2, WFRxScheduler.sharedScheduler() ) 
	
		:flatMap( function()
			return self:pingPartner():concat( Observable.just( true ) )
		end )
		
		:tap( maybeOnSuccessCallbackFn, noopFn, maybeOnSuccessCallbackFn )
		
		:catch( function( ... )
		
			if maybeOnErrorCallbackFn then
				maybeOnErrorCallbackFn()
			end
			
			return Observable.empty()
			
		end )
		
		:subscribe()
		
end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve flow by priority
	
	\param (string) name of the desired flow (e.g. 'command', 'high', 'normal', 'low', etc.)
	
	\return An Observable that fires once with WF RTMFP Flow object with the given 
		priority or nil if none exists for the given priority
	
--]]

function WFRtmfpPartnerSession:maybeGetFlowOnceByPriorityName( priority )
	
	check( 'WFRtmfpPartnerSession' )
	
	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.throw( "The WFRtmfpPartnerSession has already been shutdown." )
	end
	
	local flowStream
	
	if priority == 'high' then
		flowStream = self._highPriorityFlowStream
	elseif priority == 'low' then
		flowStream = self._bigDataFlowStream
	elseif priority == 'normal' then
		flowStream = self._smallDataFlowStream
	elseif priority == 'command' then
		flowStream = self._commandFlowStream
	end
	
	return flowStream
	
		:filter( function( maybeFlow )
			return maybeFlow
		end )
		
		:take( 1 )
	
end

--------------------------------------------------------------------------------

function WFRtmfpPartnerSession:reinitializeFlows()

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return
	end

	rtmfpPartnerSessionLog:warn( "The RTMFP partner session needs to be reinitialized." )

	self._sessionNeedsReinitialization = true

	if self._initializedSession then

		if not self._sessionIsReinitializing then
	   	
			self:_closeFlows():call( runUntilDone )
		
		end
	
	else
	
		-- If the session needs reinitialization, wait a set period of time
		-- for whomever initiated the session to reinitiate. If we hear
		-- nothing, assume the worst and close.
	
		-- Why 15 seconds? The old wait time was 0 seconds which didn't work 
		-- and caused the hiccup NMB-6031. 15 seconds is sort of arbitrary, 
		-- but it's more than enough time for the UI process to attempt to 
		-- reinitialize a session ( the UI process will detect and restart the
		-- daemon process in, at most, 10s ) but not too much time such that 
		-- the daemon appears zombied.

		local maxWaitTime = 15
	
		Observable.timer( maxWaitTime, WFRxScheduler.sharedScheduler() )
		
			:tap( function()
			
				if self._sessionNeedsReinitialization then
					rtmfpPartnerSessionLog:warn( "The RTMFP partner session is CLOSING because it has not been reinitialized." )
					self:close()
				else
					rtmfpPartnerSessionLog:warn( "The RTMFP partner session is NOT closing because it has been reinitialized." )
				end
				
			end )
			
			:call( runUntilDone )
	
	end
	
end
		
--------------------------------------------------------------------------------

function WFRtmfpPartnerSession:subscribeToFlowStatus( flowStatus )

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return
	end

	local disposable 
	
	local disposeFn = function()
		
		disposable:dispose()
		disposable = nil
		
	end
	
	if flowStatus then
	
		disposable = flowStatus:subscribe(
			function( status )
				if status == false then
					self:reinitializeFlows()
				end
			end,
			function( err )
				self:reinitializeFlows()
				disposeFn()
			end,
			function ()
				disposeFn()
			end )
			
		self._flowStatusSubscriptions[ disposable ] = true
	
	end
	
end

--------------------------------------------------------------------------------

local function cleanupFlowStatusSubscriptions( self )

	for disposable, _ in pairs( self._flowStatusSubscriptions ) do

		disposable:dispose()
		disposable = nil
		
	end

end

--------------------------------------------------------------------------------

--[[!

	\brief Initiates a new session with an RTMFP partner session by opening three
	 Flow objects and performing the appropriate "HELLO"/"HOWDY" handshake.

	\param addr (string) The URL/IP of the RTMFP peer to connect to.
	\param port (integer) The port of the RTMFP peer to connect to.
	\param secret (optional string) The Partner Session secret. This must match the
		secret of the RTMFP peer.
	\param verifyResponseFunction (optional function) An optional function used to verify
		additional information within the response of the RTMFP partner. In the event
		the verify function succeeds, it should return true. If it fails, it should return
		false along with an error message. Example
		
		verifyResponseFunction( data )
		
			if data.version == myVersion
				return true
			else
				return false, "Incompatible version number"
			end
		
		end
		
	\return An observable that completes upon success, or resolves to an error state.
	
--]]

function WFRtmfpPartnerSession:openSessionWithPartner( addr, port, secret, verifyResponseFunction )

	check( 'WFRtmfpPartnerSession', 'string', 'number', '?string', '?function' )
	
	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.throw( "The WFRtmfpPartnerSession has already been shutdown." )
	end
	
	cleanupFlowStatusSubscriptions( self )
	
	self._partnerAddress = addr
	self._partnerPort = port
	self._initializedSession = true
	
	secret = secret or theSecret
	verifyResponseFunction = verifyResponseFunction or function() return true end
	
	rtmfpPartnerSessionLog:debugf( "openSessionWithPartnerP( addr = %s, port = %s )", addr, port )
	
	-- Create the command flow and register it with the RTMFP Client object.
	-- This allows the client to know when to forward along incoming messages to
	-- this partner's event stream.
	
	local commandFlow = self._rtmfpClient:openFlow( addr, port, 'high' )
	self._commandFlowStream:onNext( commandFlow )
	
	self:subscribeToFlowStatus( commandFlow:flowOpenStatus() )
	self._rtmfpClient:linkPartnerFlowToSessionStream( commandFlow:getGuid(), self._incomingStream )
	
	-- Create the high priority flow
	
	local highPriorityFlow = self._rtmfpClient:openFlow( addr, port, 'high' )
	self._highPriorityFlowStream:onNext( highPriorityFlow )
	
	self:subscribeToFlowStatus( highPriorityFlow:flowOpenStatus() )
	self._rtmfpClient:linkPartnerFlowToSessionStream( highPriorityFlow:getGuid(), self._incomingStream )
	
	-- Create the Big Data Flow
	
	local bigDataFlow = self._rtmfpClient:openFlow( addr, port, 'low' )
	self._bigDataFlowStream:onNext( bigDataFlow )
	
	self:subscribeToFlowStatus( bigDataFlow:flowOpenStatus() )
	self._rtmfpClient:linkPartnerFlowToSessionStream( bigDataFlow:getGuid(), self._incomingStream )
	
	-- Create the Small Data Flow
	
	local smallDataFlow = self._rtmfpClient:openFlow( addr, port, 'normal' )
	self._smallDataFlowStream:onNext( smallDataFlow )
	
	self:subscribeToFlowStatus( smallDataFlow:flowOpenStatus() )
	self._rtmfpClient:linkPartnerFlowToSessionStream( smallDataFlow:getGuid(), self._incomingStream )
	
	-- Send a large "Hello" message to the peer on the command flow. This includes
	-- the Flow Guids, as well as a secret. The Flow Guids are included so that the
	-- session partner can open return flows using the same Guids.
	
	local helloMessage = KSJsonEncoder.encodeOzUploadJson {
	
			command = helloCommand,
			secret = secret,
			commandFlowGuid = commandFlow:getGuid(),
			bigDataFlowGuid = bigDataFlow:getGuid(),
			smallDataFlowGuid = smallDataFlow:getGuid(),
			highPriorityFlowGuid = highPriorityFlow:getGuid(),
			
		}
		
	-- Send a "Hi" message to the peer on both the big data and small data
	-- Flows. In RTMFP, Flows do not open until data is sent across them. By
	-- sending data across the flows now, it allows the WF RTMFP partner session
	-- to easily establish return flows.
		
	return bigDataFlow:send( hiMessage, commandMetadata() )
		
		:concat( smallDataFlow:send( hiMessage, commandMetadata() ) )
		
		:concat( highPriorityFlow:send( hiMessage, commandMetadata() ) ) 
		
		:concat( commandFlow:requestData( helloMessage, commandMetadata() ) )
		
		:map( function( json )
			
			local data = KSJsonDecoder.decodeWebSafeJson( tostring( json ) )
			
			-- Verify the response type and the secret
			
			if data and data.command == howdyCommand then
			
				if data.secret == theSecret then
			
					local success, maybeErrorMessage = verifyResponseFunction( data.other or {} )
					
					if not success then
						return false, maybeErrorMessage or "Unable to verify the RTMFP peer's response."
					end
			
				else
				
					return false, "The RTMFP peer's secret did not match."
				
				end
				
			else
			
				return false, "Unable to establish a session."
			
			end
			
			return true
			
		end )
			
		:flatMapLatest( function( success, maybeErrorMessage )
		
			if not success then
			
				return self:_closeFlows()
				
					:flatMapLatest( Observable.throw( maybeErrorMessage ) )
			
			end
	
			return commandFlow:modifySession( partnerSessionModifiers )
	
		end )
		
		:tapOnCompleted( function()
		
			self._sessionNeedsReinitialization = false
		
		end )
		
		:replay( nil, 1 )
		
		:refCount()

end

--------------------------------------------------------------------------------

--[[!

	\brief Reinitiates a session with the RTMFP partner if the session was closed.

	\return An observable that completes upon success, or resolves to an error state.

--]]

function WFRtmfpPartnerSession:reopenSessionWithPartnerIfNeeded( maybeInitCallback, maybeFinalizationCallback )

	check( 'WFRtmfpPartnerSession', '?function', '?function' )

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.throw( "The WFRtmfpPartnerSession has already been shutdown." )
	end

	-- If the session is already reinitializing, it's a no-op
	
	if self._sessionIsReinitializing then
		return self._sessionInit
	end
	
	if self._sessionNeedsReinitialization then
	
		rtmfpPartnerSessionLog:warn( "Now reopening the RTMFP partner session." )
	
		if self._initSubscription then
			self._initSubscription:dispose()
			self._initSubscription = nil
		end
	
		self._sessionIsReinitializing = true
		
		-- Close the existing flows.
	
		self._sessionInit = self:_closeFlows()
		
			:flatMapLatest( function()
			
				if maybeInitCallback then
					maybeInitCallback()
				end
				
				return Observable.just( true )
				
			end )
		
			-- Attempt to initiate a new session with the Partner
		
			:flatMapLatest( function()
				return self:openSessionWithPartner( self._partnerAddress, self._partnerPort )
			end)
	
			:retryWhen( function( failures )
				return failures
					:map( function( errorStr, errorCode )
					
						rtmfpPartnerSessionLog:warnf( "Failed to reopen partner session: %s, %d", errorStr, errorCode )
					
						return errorCode ~= WFRtmfpErrorCodes.INSTANCE_CLOSED
						
					end )
			end )
		
			:replay( nil, 1 )
		
			:refCount()
		
		self._initSubscription = self._sessionInit:subscribe(
			nil,
			nil,
			function()
			
				if maybeFinalizationCallback then
					maybeFinalizationCallback()
				end
			
				self._sessionNeedsReinitialization = false
				self._sessionIsReinitializing = false
				
			end )
		
		return self._sessionInit
		
	else
		return Observable.empty()
	end
			
end

--------------------------------------------------------------------------------

--[[!

	\brief Completes the opening of a session with an RTMFP partner. This method is
	called in response to openSessionToPeerP.

	\param rtmfpRequest (WFRtmfpRequest) The WFRtmfpRequest object associated with the
	'HELLO' message created by the partner.
	\param cmdInfo (table) Contains information about the session.
	\param otherData (optional table) Contains additional information to be sent to the partner.
	
	\return An observable that completes upon success, or resolves to an error state.
	
--]]

function WFRtmfpPartnerSession:openReturnSessionWithPartner( rtmfpRequest, cmdInfo, otherData )

	check( 'WFRtmfpPartnerSession', 'WFRtmfpRequest', 'table' )

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.throw( "The WFRtmfpPartnerSession has already been shutdown." )
	end

	cleanupFlowStatusSubscriptions( self )

	self._initializedSession = false

	-- Create the command flow and register it with the RTMFP Client object.
	-- This allows the client to know when to forward along incoming messages to
	-- this partner session's event stream.
	
	local commandFlow = self._rtmfpClient:openReturnFlow( cmdInfo.commandFlowGuid, 'high' )
	self._commandFlowStream:onNext( commandFlow )
	
	self:subscribeToFlowStatus( commandFlow:flowOpenStatus() )
	self._rtmfpClient:linkPartnerFlowToSessionStream( commandFlow:getGuid(), self._incomingStream )
	
	-- Verify the type of command and the secret
	
	if cmdInfo and cmdInfo.command == helloCommand and cmdInfo.secret == theSecret then
		
		-- Create the high priority flow
	
		local highPriorityFlow = self._rtmfpClient:openReturnFlow( cmdInfo.highPriorityFlowGuid, 'high' )
		self._highPriorityFlowStream:onNext( highPriorityFlow )
		
		self:subscribeToFlowStatus( highPriorityFlow:flowOpenStatus() )
		self._rtmfpClient:linkPartnerFlowToSessionStream( highPriorityFlow:getGuid(), self._incomingStream )
	
		-- Create the Big Data Flow
		
		local bigDataFlow = self._rtmfpClient:openReturnFlow( cmdInfo.bigDataFlowGuid, 'low' )
		self._bigDataFlowStream:onNext( bigDataFlow )
		
		self:subscribeToFlowStatus( bigDataFlow:flowOpenStatus() )
		self._rtmfpClient:linkPartnerFlowToSessionStream( bigDataFlow:getGuid(), self._incomingStream )
		
		-- Create the Small Data Flow
		
		local smallDataFlow = self._rtmfpClient:openReturnFlow( cmdInfo.smallDataFlowGuid, 'normal' )
		self._smallDataFlowStream:onNext( smallDataFlow )
		
		self:subscribeToFlowStatus( smallDataFlow:flowOpenStatus() )
		self._rtmfpClient:linkPartnerFlowToSessionStream( smallDataFlow:getGuid(), self._incomingStream )
	
		-- Send a "Howdy" message to the peer. This lets the peer know we now have a complete
		-- WF RTMFP partner session opened.
		
		local howdyMessage = KSJsonEncoder.encodeOzUploadJson {
		
				command = howdyCommand,
				secret = theSecret,
				other = otherData or {},
				
			}
			
		-- Send a "Hi" message to establish the big and small data flows. Send a "HOWDY"
		-- message as a response across the command Flow to finalize the WF RTMFP partner
		-- session.
			
		return bigDataFlow:send( hiMessage, commandMetadata() )
		
			:concat( smallDataFlow:send( hiMessage, commandMetadata() ) )
			
			:concat( highPriorityFlow:send( hiMessage, commandMetadata() ) )
			
			:concat( rtmfpRequest:respond( commandFlow, howdyMessage, commandMetadata() ) )
			
			:concat( commandFlow:modifySession( partnerSessionModifiers ) )
			
			:tapOnCompleted( function()
		
				self._sessionNeedsReinitialization = false
			
			end )
			
			:catch( function( ... )
		
				return self:_closeFlows()
					
					:flatMapLatest( Observable.throw( ... ) )
		
			end )
			
			:replay( nil, 1 )
		
			:refCount()

	else
	
		return rtmfpRequest:respond( commandFlow, goodbyeMessage, commandMetadata() )
		
			:concat( self:_closeFlows() )
			
			:flatMapLatest( Observable.throw( "Unable to establish a session." ) )
			
			:replay( nil, 1 )
		
			:refCount()
			
	end

end

--------------------------------------------------------------------------------

--[[!


	\brief An internal class method used to close all open WF RTMFP partner session 
	Flow objects.

	\return <code>Observable</code> An observable that completes after the Flows
	have been properly closed.
	
--]]

function WFRtmfpPartnerSession:_closeFlows()

	check( 'WFRtmfpPartnerSession' )
	
	self._sessionNeedsReinitialization = true
	
	local commandFlowCloseAction = self._commandFlowStream
		:take( 1 )
		:flatMap( function( commandFlow )
			
			local closeFlowObservable = Observable.just( true )
			
			if commandFlow then
			
				self._rtmfpClient:unlinkPartnerFlowToSessionStream( commandFlow:getGuid() )
				
				closeFlowObservable = commandFlow:send( goodbyeMessage, commandMetadata() )
					:catch( Observable.empty() )
					:concat( commandFlow:shutdown() )
					
			end
			
			return closeFlowObservable
			
		end )
	
	local function cleanupFlow( flow )
	
		local closeFlowObservable = Observable.just( true )
			
		if flow then
		
			self._rtmfpClient:unlinkPartnerFlowToSessionStream( flow:getGuid() )
			closeFlowObservable = flow:shutdown()
				
		end
		
		return closeFlowObservable
	
	end
	
	local highPriorityFlowCloseAction = self._highPriorityFlowStream
		:take( 1 )
		:flatMap( cleanupFlow )

	local bigDataFlowCloseAction = self._bigDataFlowStream
		:take( 1 )
		:flatMap( cleanupFlow )
		
	local smallDataFlowCloseAction = self._smallDataFlowStream
		:take( 1 )
		:flatMap( cleanupFlow )

	return Observable.just( true )
	
		:concat( commandFlowCloseAction )
		
		:concat( highPriorityFlowCloseAction )
		
		:concat( bigDataFlowCloseAction )
		
		:concat( smallDataFlowCloseAction )
	
		:catch( Observable.just( true ) )
		
		:map( function( ... )
		
			-- Shut down all the existing flow streams and create new flow streams.
			
			local function createNewFlowStream( flowStream )
				flowStream:onCompleted()
				return ReplaySubject( 1 )
			end
			
			self._commandFlowStream = createNewFlowStream( self._commandFlowStream )
			self._highPriorityFlowStream = createNewFlowStream( self._highPriorityFlowStream )
			self._bigDataFlowStream = createNewFlowStream( self._bigDataFlowStream )
			self._smallDataFlowStream = createNewFlowStream( self._smallDataFlowStream )

			return ...
		
		end )
	
		:takeLast( 1 )
	
end

--------------------------------------------------------------------------------

--[[!

	\brief Closes this WF RTMFP partner session and all of it's Flow objects.

	\return <code>Observable</code> An observable that completes after the session
	has been properly closed.
	
--]]

function WFRtmfpPartnerSession:shutdown()

	check( 'WFRtmfpPartnerSession' )

	if self._isShutdown then
		-- Slight race condition here, if two callers call shutdown
		-- in quick succession, the second may complete before the 
		-- first.  However, as I think about how we manage lifetime
		-- and do a managed close of this object, the likelihood of two
		-- shutdown calls in a row seems small, and the problem with
		-- trying to support that via :replay() and :refCount() is that
		-- our call to _closeFlows() could end up not being disposed
		-- (which could cause resource leaks) or if the second call
		-- happens much later after the first shutdown() has disposed
		-- it's subscription then it would still cause problems if the
		-- second subscription caused refCount() to re-subscribe.
		return Observable.empty()
	end

	self._isShutdown = true

	rtmfpPartnerSessionLog:debugf( "Closing WFRtmfpPartnerSession" )
	
	local pingPartnerSubscription = self._pingPartnerSubscription
	self._pingPartnerSubscription = nil
	
	if pingPartnerSubscription then
		pingPartnerSubscription:dispose()
	end
	
	return self:_closeFlows()
	
		:catch( Observable.just( 1 ) )
	
		:flatMapLatest( function()
			
			rtmfpPartnerSessionLog:debugf( "Cleaning up WFRtmfpPartnerSession" )
			
			local incomingStream = self._incomingStream
			self._incomingStream = nil
		
			if incomingStream then
				incomingStream:onCompleted()
				incomingStream:dispose()
			end
	
			local commandFlowStream = self._commandFlowStream
			self._commandFlowStream = nil
			
			if commandFlowStream then
				commandFlowStream:dispose()
			end
			
			local highPriorityFlowStream = self._highPriorityFlowStream
			self._highPriorityFlowStream = nil
			
			if highPriorityFlowStream then
				highPriorityFlowStream:dispose()
			end
			
			local bigDataFlowStream = self._bigDataFlowStream
			self._bigDataFlowStream = nil
			
			if bigDataFlowStream then
				bigDataFlowStream:dispose()
			end
			
			local smallDataFlowStream = self._smallDataFlowStream
			self._smallDataFlowStream = nil
			
			if smallDataFlowStream then
				smallDataFlowStream:dispose()
			end

			self._rtmfpClient = nil

			return Observable.empty()

		end )

end

--------------------------------------------------------------------------------

--[[!

	\brief Add a disposable to be called during close.

	\param disposable (Rx.Disposable) A disposable to dispose if closed.

	\return nothing.

--]]

function WFRtmfpPartnerSession:addCloseDisposable( disposable )

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return
	end

	if not self._disposeOnClose then

		self._disposeOnClose = {}

		self:addCloseAction( function()

			for disposable, _ in pairs( self._disposeOnClose ) do
				disposable:dispose()
			end

			self._disposeOnClose = {}

		end )

	end

	self._disposeOnClose[ disposable ] = true

end

--------------------------------------------------------------------------------

--[[!

	\brief Remove a disposable to be called during close.

	\param disposable (Rx.Disposable) The disposable to remove.

	\return nothing.

--]]

function WFRtmfpPartnerSession:removeCloseDisposable( disposable )

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return
	end

	if self._disposeOnClose then
		self._disposeOnClose[ disposable ] = nil
	end

end

--------------------------------------------------------------------------------

--[[!

	\brief Close this object.

--]]

function WFRtmfpPartnerSession:close()

	return self:shutdown()
		:call( runUntilDone )
		
end

--------------------------------------------------------------------------------

--[[!

	\brief Retrieve a stream for incoming messages from a connected
	partner session.

	\return <code>ReplaySubject<RTMFP requests></code> Incoming messages
	
--]]

function WFRtmfpPartnerSession:listen()
	
	check( 'WFRtmfpPartnerSession' )

	if self._isShutdown then
		rtmfpPartnerSessionLog:warnf( "The WFRtmfpPartnerSession has already been shutdown." )
		return Observable.empty()
	end

	if not self._incomingStream then
		return Observable.empty()
	end

	return self._incomingStream

		:filter( function( rtmfpRequest )
		
			local requestMetadata = rtmfpRequest:getMetadata()
			return not requestMetadata._sessionPartnerCommand
			
		end )

end

--------------------------------------------------------------------------------

WFRtmfpPartnerSession = Rx.Class( 'WFRtmfpPartnerSession', WFRtmfpPartnerSession )

--==============================================================================

return WFRtmfpPartnerSession
