--[[----------------------------------------------------------------------------

WFRxHttpClientMock.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

--------------------------------------------------------------------------------

This mock is intended as a replacement for the WFRxHttpClient namespace.
It can be used when you want to simulate certain request / response patterns
that are difficult to emulate otherwise.

------------------------------------------------------------------------------]]

	-- RxLua
local Rx = require 'Rx'
local Observable = Rx.Observable

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.reactive
local WFRxScheduler = require "WFRxScheduler"

	-- WF.core.utils
local WFArguments = require 'WFArguments'

local WFMockHttpRequest = require 'WFMockHttpRequest'
	
--------------------------------------------------------------------------------

local check = assert( WFArguments.check )

local scheduler = WFRxScheduler.sharedScheduler()

--============================================================================--

return function( rxTest, expectedRequestResponsePairs )

	check( 'WFRxTestCaseRunner', 'table' )

	local requestCount = 0
	local requestErrors = 0
	local lastRequestTime = KSDate.timeNowForAnimation()
	local prevRequestTime = lastRequestTime
	local requestsProcessed = Rx.ReplaySubject( 1 )

	rxTest:finally( function()

		if requestCount ~= #expectedRequestResponsePairs then
			error( string.format( "Expected test to make %d calls, only received %d",
								  #expectedRequestResponsePairs, requestCount ) )
		end

		if requestErrors ~= 0 then
			-- Handle case where error was trapped by a pcall.
			error( "Mock Http error encountered in test" )
		end

	end )

	local function getPutImp( url, data, options, sequencing )

		prevRequestTime = lastRequestTime
		lastRequestTime = KSDate.timeNowForAnimation()
		requestCount = requestCount + 1

		rxTest:log( string.format( "Request #%s url %s", requestCount, url ) )

		local requestInfo = expectedRequestResponsePairs[ requestCount ]
						or { URL = "NO FURTHER REQUESTS" }

		if data or requestInfo.data then

			data = tostring( data )

			if requestInfo.dataFilter then
				data = requestInfo.dataFilter( data )
			end

			data = data:gsub( "client_id=[^&]+", "client_id=(mumble)" )
			data = data:gsub( "client_secret=[^&]+", "client_secret=(secret)" )

		end

		if options then
		
			if options.requestHeaders then
				for _, header in ipairs( options.requestHeaders ) do
					if header.header == "Accept-Language" then
						header.value = "(language)"
					end
				end
			end
			
			-- Remove some options that are specific to the rate limiter,
			-- and not the actual HTTP request.
			
			options.requestDescription = nil
			
		end
		
		if requestInfo.dataTransformation then
			data, requestInfo.data = requestInfo.dataTransformation( data, requestInfo.data )
		end
		
		if requestInfo.validationFn then
			if requestInfo.validationFn( data ) then
				requestInfo.data = data
			else
				requestInfo.data = "The actual request data failed validation" 
			end
		end

		if requestInfo.urlFilter then
			url = requestInfo.urlFilter( url )
		end
		
		if requestInfo.options and requestInfo.options.requestHeaders then
			table.sort( requestInfo.options.requestHeaders, function( a, b ) return a.header < b.header end )
		end
		
		if options and options.requestHeaders then
			table.sort( options.requestHeaders, function( a, b ) return a.header < b.header end )
		end
		
		local expectedRequestInfo = {
			URL = requestInfo.url,
			_count = requestCount,
			_traceback = requestInfo.traceback,
			options = requestInfo.options,
			data = requestInfo.data,
		}

		local actualRequestInfo = {
			URL = url,
			_count = requestCount,
			_traceback = requestInfo.traceback,
			-- don't compare options unless the mock request declared 
			-- some to validate, because not all tests need to be pedantic
			options = requestInfo.options and options or nil, 
			data = data,
		}
	
		if expectedRequestInfo.data == ':ignore:' then
			expectedRequestInfo.data = nil
			actualRequestInfo.data = nil
		end
		
		requestErrors = requestErrors + 1
		rxTest:expectEqual( expectedRequestInfo, actualRequestInfo )
		requestErrors = requestErrors - 1

		local fail = requestInfo.fail
		local failFn = ( options and options.returnHttpErrorsAsData ) and Observable.just or Observable.throw
		if type( fail ) == 'table' then
			return failFn( unpack( fail, 1, #fail ) )
		elseif fail then
			return failFn( fail )
		end

		local response = requestInfo.response

		if requestInfo.responseFilter then
			response = requestInfo.responseFilter( response )
		end

		-- Schedule requestsProcessed's onNext to occur at the next idle time
		-- after getOrPut's result is sent.
		scheduler:schedule( requestCount, function( _, processedCount )
			requestsProcessed:onNext( processedCount )
		end )

		local result = Observable.just( response, requestInfo.headers )
		if requestInfo.responseDelayBySeconds then
			result = result:delay( requestInfo.responseDelayBySeconds, scheduler )
		end

		return result
	end

	return {

		get = function( url, options, sequencing )
			return getPutImp( url, nil, options, sequencing )
		end,

		put = function( url, data, options )
			return getPutImp( url, data, options )
		end,

		requestCount = function()
			return requestCount
		end,

		requestsProcessed = function()
			return requestsProcessed
		end,

		requestTraceback = function( optRequestNum )
			local requestNum = optRequestNum or requestCount
			return expectedRequestResponsePairs[ requestNum ].traceback
		end,

		requestResponseTransformFn = function( optRequestNum )
			local requestNum = optRequestNum or requestCount
			return expectedRequestResponsePairs[ requestNum ].responseTransformationFn
		end,

		timeBetweenRequests = function()
			return lastRequestTime - prevRequestTime
		end,
		
		addHttpRequest = function( httpRequest )
			if WFMockHttpRequest:isClassOf( httpRequest ) then
				httpRequest = httpRequest:build()	
			end
			table.insert( expectedRequestResponsePairs, httpRequest )
		end,
	}

end
