--[[----------------------------------------------------------------------------

WFRxHttpServerUtils.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2019 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

local Rx = require "Rx"

--------------------------------------------------------------------------------

local WFRxHttpServerUtils = {}

--------------------------------------------------------------------------------

--[[ \brief Helper for starting an HTTP Server request.

	\param handlers A list of pattern, methodList, handler

		When the request is started, the handler list is searched for a matching pattern. This is done 
		using string.match, so if you need to match an entire string, it is recommended to use a restrictive 
		pattern for it. For example, if you only want to match '/', use the pattern '^/$.
		
	    When the pattern matches and the methodList contains the requested method, the handler function 
		will be called with these parameters:
        - path (string): The path of the request uri (that matches pathOrPattern)
        - info: a table with:
                - \c query (string): The query string (if present)
                - \c remoteAddr (string): The address of the requester
                - \c method (string): the HTTP Method
                - \c matches (array): if pathOrPattern contains capture expressions, the matches will be in this array
                - \c postData (string): The data passed into a POST request (if present)

        Upon return, the function must return an Observable. If the request cannot be served, an error
        should be returned by setting an explicit status code, for example:
		return Rx.Observable.just{ statusCode = 404, data = 'not found' }

        The Observable will be subjected to a :take( 1). Longer running observables are not supported, because they
        could hog the server for a long time.

	\param catalog The HTTP Server's catalog reference
	\param uri The request uri
	\param query The request's query string
	\param remoteAddr The client's address
	\param method GET or POST 
	\param postData the data for the POST or nil

	\returns an Observable(table), where the table contains either:
		- statusCode and data
		- contentType and data
		- contentType and filePath
]]--

function WFRxHttpServerUtils.startRequest( handlers, catalog, uri, query, remoteAddr, method, postData )

	local obs = Rx.Observable.just{ statusCode = 404, data = 'not found' }

	for _, handlerInfo in ipairs( handlers ) do

		local path, methods, handler = unpack( handlerInfo )

		local matches = { string.match( uri, path ) }

		if next( matches ) then

			local methodSupported = false

			for _, m in ipairs( methods ) do
				if m == method then
					methodSupported = true
					break
				end
			end

			if methodSupported then

				local info = {
					query = query,
					remoteAddr = remoteAddr,
					method = method,
					matches = matches,
					postData = postData
				}

				local success
				success, obs = pcall( handler, catalog, uri, info )

				if not success then
					obs = Rx.Observable.throw( 500, obs )
				end

				break

			end

		end
	end

	return obs:take( 1 )

end

return WFRxHttpServerUtils