--[[----------------------------------------------------------------------------

WFRxIndexStoreQueryServer.lua

--------------------------------------------------------------------------------

ADOBE CONFIDENTIAL
------------------
Copyright 2015 Adobe Systems Incorporated
All Rights Reserved.

NOTICE: All information contained herein is, and remain the property of
Adobe Systems Incorporated and its suppliers, if any. The intellectual and
technical concepts contained herein are proprietary to Adobe Systems
Incorporated and its suppliers and may be covered by U.S. and Foreign
Patents, patents in process, and are protected by trade secret or copyright
law.  Dissemination of this information or reproduction of this material is
strictly forbidden unless prior written permission is obtained from
Adobe Systems Incorporated.

------------------------------------------------------------------------------]]

	-- Rx
local Rx = require 'Rx'

	-- WF.core.os
local KSDate = require 'KSDate'

	-- WF.core.rxlua
local WFRxScheduler = require 'WFRxScheduler'
local WFRxControlledObservableWithBuffer = require 'WFRxControlledObservableWithBuffer'
local WFRxDisposeOnUnsubscribeObservable = require 'WFRxDisposeOnUnsubscribeObservable'

	-- WF.core.string
local KSStringUtils = require 'KSStringUtils'

	-- WF.core.utils
local KSIdGenerator = require 'KSIdGenerator'
local KSTableUtils = require 'KSTableUtils'
local WFArguments = require 'WFArguments'
local WFCloseSupport = require 'WFCloseSupport'
local WFPlatformInfo = require 'WFPlatformInfo'

	-- WF.core.logging
local WFLogger = require 'WFLogger'

--------------------------------------------------------------------------------

local Observable = Rx.Observable

local controlledWithBuffer = WFRxControlledObservableWithBuffer
local disposeOnUnsubscribe = WFRxDisposeOnUnsubscribeObservable

local check = assert( WFArguments.check, "Cannot find WFArguments.check" )
local kSemiColonByte = 59
local string_format = string.format

local logger = WFLogger 'WFRxIndexStoreQueryServer'
--	:setLevel 'debug'

--------------------------------------------------------------------------------

local function throttledMap( observable, timeRatio, mapFn )

	local getTimeNow = KSDate.timeNowForAnimation

	local lastDelayCount = 1

	local noSoonerSb = Rx.Subject()

	local noSooner = noSoonerSb
		:replay( nil, 1 )
	noSooner:connect()

	noSoonerSb:onNext( getTimeNow() )

	noSooner = noSooner:map( function( nextTime )
		local now = getTimeNow()
		local delay = nextTime - now
		if delay > 0 then
			return Rx.Observable.timer( delay, WFRxScheduler.sharedScheduler() )
		else
			return Rx.Observable.just( true )
		end
	end )
	:switch()
	:map( function()
		lastDelayCount = lastDelayCount + 1
		return lastDelayCount
	end )

	local count = 1
	local lastProcessed = 1

	local updateCount = observable:map( function()
		count = count + 1
		return count
	end )

	local rateLimited = noSooner
		:combineLatest( function( delayCount, count, ... )

			if count <= lastProcessed then
				return Rx.Observable.never()
			end

			if lastDelayCount == delayCount then
				return Rx.Observable.just( ... )
			end
			return Rx.Observable.never()

		end, updateCount, observable )
		:switch()

	local function timeIt( start, ... )
		lastProcessed = count

		local noSooner = getTimeNow()
		noSooner = noSooner + ( noSooner - start ) * timeRatio
		noSoonerSb:onNext( noSooner )

		return ...
	end

	return rateLimited:map( function( ... )

		local start = getTimeNow()
		return timeIt( start, mapFn( ... ) )

	end )

end


--============================================================================--
--[[!
	\brief A \c WFRxIndexStoreQueryServer is used by the index store implementation
		to keep the index store tables up to date.

	It should only be accessed by \c WFRxIndexStoreImp.
]]--

local queryServerMethods = {

	_mapServer = false,
		-- (WFRxIndexStoreMapServer) map server for this index store

	_tableNames = false,
		-- (table) names of the index tables declared in this schema
	
	_manualTableNames = false,
		-- (table) names of the manual tables (no map function) declared in this schema

	_nColumnsPerManualTable = false,
		-- (table) how many columns are in each manual table
		
	_needsUpdateFromManualTable = false,
		-- (table) Array of Observable per manual table to trigger query execution
		
	_queryRequests = false,
		-- (table) current query Observable requests
			-- key: (Observable< table >) stream returned to caller
			-- value: (table) arguments to :query, plus
				-- relevantTables: (table) set of table names to update
				-- nUpdatesSent: (number) how many times have we broadcasted results?
				
	_controllableQueryRequests = false,
		-- (table) current controllable query Observable requests
			-- key: (string) queryId
			-- value: (WFRxControlledObservableWithBuffer) to request query results
				
	_nQuerysCreated = 0,
		-- (number) how many query requests have ever been created?

	_virtualTables = false,
		-- (table) names and Observable< boolean > that updates whenever
		-- the virtual table has more data.  These are poked in by the imp.

	WFCloseSupport.addCloseSupportMethods,

}

--------------------------------------------------------------------------------
--[[!
	\brief Create a query server to support an index store implementation.

	\param mapServer (\c WFRxIndexStoreMapServer) map server providing the index tables
	\param tableNames (table) set of table names declared in this schema
]]--

function queryServerMethods:__init( mapServer, indexStoreSchema )

	check( 'WFRxIndexStoreQueryServer', 'WFRxIndexStoreMapServer', 'WFRxIndexStoreSchema' )

	self._virtualTables = {}
	self._needsUpdateFromManualTable = {}
	self._mapServer = mapServer
	self._mapServer:addCloseTarget( self )

	self._tableNames = indexStoreSchema:tableNames()
	self._manualTableNames = indexStoreSchema:manualTableNames()
	self._nColumnsPerManualTable = indexStoreSchema:nColumnsPerManualTable()
	
	self._queryRequests = {}
	self._controllableQueryRequests = setmetatable( {}, { __mode = 'kv' } )
	
	for manualTableName in pairs( self._manualTableNames ) do
			self._needsUpdateFromManualTable[ manualTableName ] = Rx.ReplaySubject( 1 )
			
			self._needsUpdateFromManualTable[ manualTableName ]
				:onNext( { 
					tableHasNewIndexData = true,
					tableUpToDateCounter = math.huge
				} )
				
	end
	
end

--------------------------------------------------------------------------------

local function identifyRelevantTables( tableNames, sql )

	-- Identify all of our index tables that might be relevant for this query.
	-- Naive implementation: Simply search for each table name in query string.

	local result = {}

	for tableName in pairs( tableNames ) do

		-- Naive search, too: Since SQL table names can't contain Lua pattern
		-- special characters, just use the table name as search pattern.

		if sql:find( "[^a-zA-Z]" .. tableName .. "[^a-zA-Z]" ) then
			result[ tableName ] = true
		end

	end

	return result

end
 
--------------------------------------------------------------------------------
--[[!
	\brief Create an observable that tracks the results of a SQL query on the index store.

	\param args (table) arguments for query
		- \c sql: (string) SQL query to be performed
		- \c threadName: (optional, string) name of thread requesting the query
		- \c traceback: (optional, string) traceback of call site in requesting thread
		- \c calm: (optional, number) if present, calm query result updates to no more
			than one every _n_ seconds
		- \c nArgs: (optional, number) number of query arguments (helps with \c nil
			values in the middle of the list)
		- 1 .. n: (optional) query arguments to be substituted for ? placeholders in query

	\return Observable< table > query results (array of arrays syntax)

	\b WARNING: Use this method sparingly. Creating many query Observables will adversely
		affect the overall performance of the application.

	\b WARNING: This Observable may, at times, return information that is based on
		incomplete index table results. It will eventually catch up and update
		results if needed. (Out of date information is especially likely for the
		first value returned.)
]]--

function queryServerMethods:query( args )

	check( 'WFRxIndexStoreQueryServer', 'table' )

	-- Validate args.

	local sql = args.sql

	if string.byte( sql, -1 ) ~= kSemiColonByte then
		sql = sql .. ';'
	end

	local encoder = args.encoder
	if encoder ~= nil then assert( type( encoder ) == 'string' ) end

	local storeAsTemporaryIdTable = args.storeAsTemporaryIdTable
	if storeAsTemporaryIdTable then
		assert( encoder == nil )
		encoder = '< temp table >' .. tostring( KSIdGenerator.localCounter() )
	end
	local quiesceUntilUpToDate = args.quiesceUntilUpToDate
	if quiesceUntilUpToDate ~= nil then
		assert( type( quiesceUntilUpToDate ) == 'boolean', 'quiesceUntilUpToDate must be true or false' )
	end

	-- Identify relevant tables.

	local relevantTablesSet = identifyRelevantTables( self._tableNames, sql )

	local nTables = 0
	local relevantTables = {}

	for tableName in pairs( relevantTablesSet ) do
		nTables = nTables + 1
		relevantTables[ nTables ] = tableName
	end

	-- Housekeeping ...

	local mapServer = self._mapServer
	self._nQuerysCreated = self._nQuerysCreated + 1

	args.nUpdatesSent = 0
	args.relevantTables = relevantTables

	-- Do we have new data from any of the tables?

	local tablesUpdated = {}

	for i, relevantTable in ipairs( relevantTables ) do
		tablesUpdated[ i ] = mapServer:keepTableUpToDate( relevantTable )
	end
	
	for manualTableName in pairs( identifyRelevantTables( self._manualTableNames, sql ) ) do
		tablesUpdated[ #tablesUpdated + 1 ] = self._needsUpdateFromManualTable[ manualTableName ]
	end

	for tableName in pairs( identifyRelevantTables( self._virtualTables, sql ) ) do
		local observable = self._virtualTables[ tableName ]
		tablesUpdated[ #tablesUpdated + 1 ] = 
			observable
				:map(
					function( updateInfo )

						if type( updateInfo ) == "table" then

							return updateInfo

						end

						return {

							tableUpToDateCounter = tonumber( updateInfo ) or 0,

							-- currently virtual tables always return 0, so 
							-- they are behind and should never cause a new
							-- query to fire. That may be incorrect behavior,
							-- but this code preserves it for now.
							tableHasNewIndexData = false,

						}

					end )

	end

	nTables = #tablesUpdated

	if nTables == 0 then
		error( "WFRxIndexStoreQueryServer:query: did not find any relevant tables in query " .. sql, 2 )
	end

	-- We use the laterE event to start the clock on calming even though
	-- we don't use this first event to trigger a re-query. This prevents us
	-- from re-running the query (and publishing results) any sooner than
	-- (interval) seconds after the query Observable is created.
	-- If you disable this logic, expect the test named testQuery_outOfDate_join
	-- to fail with a "too soon" error.

	local anyTableUpdated

	if quiesceUntilUpToDate then

		-- keep track of the revCounter for each table
		anyTableUpdated = Observable.combineLatest(
				self._mapServer:currentChangeCounter(),
				unpack( tablesUpdated, 1, nTables )
			 )

	else
		anyTableUpdated = Observable.combineLatest( unpack( tablesUpdated, 1, nTables ) )
	end

	local needsToDeliverFirstResult = true
	
	-- Fetch query results.

	local nQueryArgs = args.nArgs or #args
	local queryResultsNow = function()
	
		needsToDeliverFirstResult = false
		args.nUpdatesSent = args.nUpdatesSent + 1
		
		return mapServer:query( sql, encoder, args.tempIds, unpack( args, 1, nQueryArgs ) )
		
	end

	-- Raw query event stream. Fires whenever new data is available in any of the
	-- index tables, modulo the calming factor.

	-- The calming factor now scales with the performance of the query;
	-- as results get big, we calm more and more.

	if quiesceUntilUpToDate then

			-- We consider ourselves "up to date" as soon as all the tables are
			-- current with the doc store rev counter as of the time the query
			-- started.
			-- If additional documents have arrived in the time it took us
			-- to bring the index store up to date, then we might not include those
			-- new documents in this query

		anyTableUpdated = anyTableUpdated:filter( function( docStoreRevCounter, ... )

			if not docStoreRevCounter then
				return
			end

			local n = select( '#', ... )

			for i = 1, n do

				local tableUpdateInfo = select( i, ... )

				local tableUpToDateCounter = 
						tonumber( tableUpdateInfo.tableUpToDateCounter ) or 0

				if tableUpToDateCounter <= docStoreRevCounter then
					return false
				end

			end

			return true

		end )
	end
	
	anyTableUpdated = anyTableUpdated:where( function( ... )
		
		if needsToDeliverFirstResult then
			return true
		end
		
		local startIndex = 1
		if quiesceUntilUpToDate then
			startIndex = 2
		end
		
		local n = select( '#', ... )
		
		for i = startIndex, n do
			
			local tableUpdateInfo = select( i, ... )
			
			if tableUpdateInfo.tableHasNewIndexData then

				return true

			end

		end
		
		return false
		
	end )
	
	if not args.queryOnce then
	
		-- But we don't want to update too often in the non-calming case. Note that we start with the timer
		-- running so that we are calm from the get go, implemented via the :debounce() operator.
		-- In the case of a given calming factor we emit the first result immediately and start calming
		-- afterwards
		
		-- In order to speed up the WF unit tests, we remove all calming from index store queries.
		
		if args.calm and not WFPlatformInfo.isUnitTest() then 
			anyTableUpdated = anyTableUpdated:calm( args.calm, WFRxScheduler.sharedScheduler() )
		else
			anyTableUpdated = anyTableUpdated:debounce( .125, WFRxScheduler.sharedScheduler() )
		end
	
	end
	
	local rawQuery = throttledMap( anyTableUpdated, 4, queryResultsNow )

	if storeAsTemporaryIdTable then
	
		local lastTempTable
		
		-- in error, completed, or disposed, clean up the temp table.
		local cleanUpTempTables = function()
			if lastTempTable and not WFCloseSupport.hasBeenOrWillBeClosed( self ) then

					-- but only if db hasn't already been closed.  Upon close, temp tables
					-- are cleaned up automatically by SQlite :)

				local tempTableName = lastTempTable
				lastTempTable = nil

				local dropSql = string_format( 'DROP TABLE IF EXISTS %s', tempTableName )

				self._mapServer._db:exec( dropSql )

			end
		end
		
		rawQuery = rawQuery:replay( nil, 1 )

		rawQuery:connect():add( Rx.Disposable( cleanUpTempTables ) )
		
		rawQuery = rawQuery:refCount()
		:tap( function( tableName )
			if lastTempTable and lastTempTable ~= tableName then
				cleanUpTempTables()
			end
			lastTempTable = tableName
		end, cleanUpTempTables, cleanUpTempTables )

	end

	local previousResult, previousRowCount
		-- Squelch events if the actual query results are unchanged.
	local resultQuery = rawQuery:filter( function( newValue, rowCount )

		if previousRowCount == rowCount
					and KSTableUtils.deepEquals( previousResult, newValue ) then
			return false
		else
			previousResult = newValue
			previousRowCount = rowCount
			return true
		end

	end )
	
	local bufferSize = 1
	local initialRequestCount = math.huge
	
	if args.requestCount then
		initialRequestCount = args.requestCount
	end
	
	local controlledObservable = resultQuery:call( controlledWithBuffer, bufferSize, initialRequestCount )
	
	local queryId = args.queryId
	
	if args.requestCount and not queryId then
		logger:warnf( 'Default requestCount for Query modified, but no queryId is set' )
	end
	
	local unsubscribeObservable = controlledObservable:call( disposeOnUnsubscribe )
	
	if queryId then
	
		self._controllableQueryRequests[ queryId ] = controlledObservable
		
		unsubscribeObservable:addDisposable( Rx.Disposable.create( function()
		
			self._controllableQueryRequests[ queryId ] = nil
			
		end ) )
		
	end
	
	args._controlledObservable = controlledObservable

	unsubscribeObservable:addDisposable( Rx.Disposable.create( function()
	
		args._controlledObservable = nil
		self._queryRequests[ unsubscribeObservable ] = nil
		
	end ) )
	
	-- Remember the query.

	self._queryRequests[ unsubscribeObservable ] = args
	
	return unsubscribeObservable

end

--------------------------------------------------------------------------------

function queryServerMethods:bulkInsertManualTable( tableName, rows )

	check( 'WFRxIndexStoreQueryServer', 'string', 'table' )

	if #rows == 0 then
		return Observable.throw( "Empty value" )
	end
	
	if not self._manualTableNames[ tableName ] then
		error( "WFRxIndexStoreQueryServer:bulkInsertManualTable: did not find the manual table name" )
	end

	local mapServer = self._mapServer
	local db = mapServer._db
	local exec = db.exec
	
	local nColumns = self._nColumnsPerManualTable[ tableName ]
	local queryPlaceholders
				
	if nColumns > 1 then
		queryPlaceholders = string.format( '?%s', string.rep( ', ?', nColumns - 1 ) )
	else
		queryPlaceholders = '?'
	end
	
	local sql = string.format( 'INSERT INTO %s VALUES (%s)', tableName, queryPlaceholders )
	
	local result = db:withTransactionDo( function( db )

		local result = {}

		for i, row in ipairs( rows ) do
			
			local res = exec( db, sql, unpack( row ) )
			
			result[ i ] = res

		end
		
		return result

	end, db )
	
	self._needsUpdateFromManualTable[ tableName ]
		:onNext(
			{ 
				tableHasNewIndexData = true,
				tableUpToDateCounter = math.huge
			}
		)

	return Observable.just( result )

end

--------------------------------------------------------------------------------

function queryServerMethods:clearManualTable( tableName )

	check( 'WFRxIndexStoreQueryServer', 'string' )

	if not self._manualTableNames[ tableName ] then
		error( "WFRxIndexStoreQueryServer:clearManualTable: did not find the manual table name" )
	end

	local db = self._mapServer._db
	local exec = db.exec
	
	exec( db, string.format( 'DELETE FROM %s', tableName ) )
	
	self._needsUpdateFromManualTable[ tableName ]
		:onNext(
			{ 
				tableHasNewIndexData = true,
				tableUpToDateCounter = math.huge
			}
		)
	
	return Observable.just( true )
	
end

--------------------------------------------------------------------------------
--[[!
	\brief Return summary statistics about query Observables that are currently
		running.

	\return (table) array of outstanding query Observables. Each entry contains:
		- \c sql: (string) SQL query
		- \c args: (table, array of strings) query arguments, if any
		- \c relevantTables: (table, set of strings) which tables are used for this table?
		- \c nUpdatesSent: (number) how many times has this query sent updates?
		- \c calm: (optional, number) calming interval, if applied
		- \c threadName: (optional, string) which thread requested this Observable? (if provided)
		- \c traceback: (optional, string) traceback in requesting thread (if provided)
]]--

function queryServerMethods:queryStats()

	-- Send information about query requests that are outstanding.

	local filteredQueryRequests = {}
	local nQueries = 0

	for resultE, queryInfo in pairs( self._queryRequests ) do

		if not WFCloseSupport.hasBeenOrWillBeClosed( resultE ) then

			nQueries = nQueries + 1

			local traceback = queryInfo.traceback

			if traceback then
				traceback = KSStringUtils.split( queryInfo.traceback, '\n' )
				table.remove( traceback, 1 )
			end

			filteredQueryRequests[ nQueries ] = {
				sql = queryInfo.sql,
				relevantTables = queryInfo.relevantTables,
				nUpdatesSent = queryInfo.nUpdatesSent,
				calm = queryInfo.calm,
				threadName = queryInfo.threadName,
				traceback = traceback,
			}

			if queryInfo[ 1 ] then

				local args = {}
				for i = 1, #queryInfo do
					args[ i ] = queryInfo[ i ]
				end

				filteredQueryRequests[ nQueries ].args = args

			end

		end

	end

	table.sort( filteredQueryRequests, function( a, b )
		return a.sql < b.sql
	end )

	filteredQueryRequests.nQuerysCreated = self._nQuerysCreated

	return filteredQueryRequests

end

--------------------------------------------------------------------------------

function queryServerMethods:requestFromQuery( queryId, requestCount )

	check( 'WFRxIndexStoreQueryServer', 'string', 'number' )
	
	local foundQuery = false
	
	local controlledQuery = self._controllableQueryRequests[ queryId ]
	
	if controlledQuery and not WFCloseSupport.hasBeenOrWillBeClosed( controlledQuery ) then
	
		foundQuery = true
			
		controlledQuery:request( requestCount )
			
	end
	
	if not foundQuery then
	
		logger:warnf( 'No active query to request from was found for queryId: %s', queryId )
		
	end
	
	return Observable.just( foundQuery )
	
end

--------------------------------------------------------------------------------

return Rx.Class 'WFRxIndexStoreQueryServer'( queryServerMethods )
